#ifndef _IOI_H_
#define _IOI_H_


/*
 * IOimg.h
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    I/O functions for frames.
 *
 */

#include "nrctypes.h"

/*
 * Defines
 */


/*
 * Structs and typedefs
 */


/*
 * Function prototypes
 */

void ioiRotateLongTermMem(struct frame **longTermMem,
			  int *numFramesInLongTermMem,
			  int longTermMemSize);

/*
 *
 * ioiPutFrameYUV:
 *
 * Parameters:
 *      image     I:  Frame structure
 *      fileName  I:  Name of the output file
 *      n         I:  Frame number
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Writes Y, u, and v components of the given frame structure 
 *      to files fileName.n.Y, fileName.n.U, and fileName.n.V. Assumes
 *      4:2:0 format
 *
 * Returns:
 *      Nothing.
 */

void ioiPutFrameYUV(struct frame *image, char *fileName, int32 n,
                    int32 picSizeX, int32 picSizeY);


/*
 *
 * ioiClearSequence:
 *
 * Parameters:
 *      fileName  I:  Name of the output file
 *
 * Function:
 *      Clears the Y, u, and v files of the given sequence name.
 *
 * Returns:
 *      Nothing.
 */

void ioiClearSequence(char *fileName);


/*
 *
 * ioiCalcErrYUV:
 *
 * Parameters:
 *      firstFrame    I:  First frame
 *      secondFrame   I:  Second frame
 *      diffFrame     O:  Difference frame
 *      scale         I:  Scaling parameter
 *      pelsY         I:  Number of pixels in luminance frame
 *      psnrY         O:  PSNR value of the Y frame
 *      psnrU         O:  PSNR value of the u frame
 *      psnrV         O:  PSNR value of the v frame
 *      sqErrY        O:  Squared error of the Y frames
 *      sqErrU        O:  Squared error of the u frames
 *      sqErrV        O:  Squared error of the v frames
 *
 * Function:
 *      Calculates differences between Y, u, and v components of the
 *      given frames. Computes also PSNR values (peak value assumed to 
 *      be 255) and sums of squared errors for the components. Assumes
 *      4:2:0 format. Pixel values in difference images are scaled and
 *      offset of 128 is added:
 *      
 *        diffFRame = 128 + scale*(firstFrame - secondFrame)
 *
 * Returns:
 *      Nothing.
 */

void ioiCalcErrYUV(struct frame *firstFrame, struct frame *secondFrame, 
                   struct frame *diffFrame, int32 scale, int32 pelsY, 
                   float64 *psnrY, float64 *psnrU, float64 *psnrV,
                   int32 *sqErrY, int32 *sqErrU, int32 *sqErrV);


/*
 *
 * ioiAllocFrame:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Allocates frame structure.
 *
 * Returns:
 *      Nothing.
 */

void ioiAllocFrame(struct frame *fr, int32 piSizeX, int32 picSizeY);


/*
 *
 * ioiCopyFrame:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      destFr    I:  Destination frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Copies a frame structure into another.
 *
 * Returns:
 *      Nothing.
 */

void ioiCopyFrame(struct frame *srcFr, struct frame *destFr,
                  int32 picSizeX, int32 picSizeY);

/*
 *
 * ioiInitFrame:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Initializes a picture with some default constant value.
 *
 * Returns:
 *      Nothing.
 */

void ioiInitFrame(struct frame *srcFr, int32 picSizeX, int32 picSizeY);


/*
 *
 * ioiExchangeFrames:
 *
 * Parameters:
 *      srcFr     I:  Source frame structure
 *      destFr    I:  Destination frame structure
 *
 * Function:
 *      Exchanges the pointer addresses of two frame structures.
 *
 * Returns:
 *      Nothing.
 */

void ioiExchangeFrames(struct frame *srcFr, struct frame *destFr);


/*
 *
 * ioiFreeFrame:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *
 * Function:
 *      Frees frame structure.
 *
 * Returns:
 *      Nothing.
 */

void ioiFreeFrame(struct frame *fr);


/*
 *
 * ioiReadFrameYUV:
 *
 * Parameters:
 *      fr        I:  Frame structure
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *
 * Function:
 *      Reads Y, u, and v components of frame from files. Assumes
 *      4:2:0 format.
 *
 * Returns:
 *      Nothing.
 */

void ioiReadFrameYUV(struct frame *fr, int32 nr, int32 nc, FILE *fileY, 
                     FILE *fileU, FILE *fileV);


/*
 *
 * ioiSkipFrames:
 *
 * Parameters:
 *      n         I:  Number of frames to skip
 *      picSizeX  I:  Width of the luminance frame
 *      picSizeY  I:  Height of the luminance frame
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *
 * Function:
 *      Skips given number of frames. Assumes
 *      4:2:0 format.
 *
 * Returns:
 *      Nothing.
 */

void ioiSkipFrames(int32 n, int32 nr, int32 nc, FILE *fileY, FILE *fileU,
                   FILE *fileV);


/*
 *
 * ioiOpenFile:
 *
 * Parameters:
 *      fileY     I:  File pointer for Y component
 *      fileU     I:  File pointer for u component
 *      fileV     I:  File pointer for v component
 *      seqName   I:  Sequence name
 *
 * Function:
 *      Opens sequence for reading. Y, u, and v components are
 *      assumed to be in files seqName.y, seqName.u, and seqName.v.
 *
 * Returns:
 *      Nothing.
 */

void ioiOpenFile(FILE **fileY, FILE **fileU, FILE **fileV, char *seqName);

/*
 *
 * ioiPutImage:
 *
 * Parameters:
 *      fileName  I:  Name of the output file
 *      image     I:  Byte image
 *      picSizeX  I:  Width of the frame
 *      picSizeY  I:  Height of the frame
 *
 * Function:
 *      Writes image to file.
 *
 * Returns:
 *      Nothing.
 */

void ioiPutImage(char *fileName, u_char *image, int32 picSizeX, 
                        int32 picSizeY);
#endif
