#ifndef _IPD_H_
#define _IPD_H_

/*
 * IntraPredDecoder.h
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    High-level functions for decoding INTRA macroblocks and blocks.
 *
 */


#include "nrctypes.h"
#include "mainFlags.h"
#include "decDemux.h"
#define LAST                    2
#define RUN                     64
#define LEVEL                   256
#define MODE_INTRA_PRED_VSCAN   6
#define MODE_INTRA_PRED_HSCAN   7
#define MODE_INTRA_FRAME        8
#define BLK_MODE_INTER          1
#define BLK_MODE_INTRA          0

#define IPD_NOT_CLIP_BLOCK      0
#define IPD_CLIP_BLOCK          1

#define IPD_DC_PRED             128

/*
 * Function implementations
 */

void decDecodeIntraMccRegion(struct frame *recoFr, /* Out: Reconstruction frame      */
                             int16 *slicePredMap,  /* In:  Slice label buffer        */
                             int currSliceID,      /* In:  Label of current slice    */
                             int sliceXmin,        /* In:  Xmin  of current slice    */
                             int sliceXmax,        /* In:  Xmax  of current slice    */
                             int sliceYmin,        /* In:  Ymin  of current slice    */
                             int sliceYmax,        /* In:  Ymax  of current slice    */
                             object *currBlockOrMBstruct, /* In: Ptr to MB/B object  */
                             int regionSize,       /* In:  Region size (16 or 8)     */
                             int picSizeX,         /* In:  Width of the image        */
                             int picSizeY,         /* In:  Height of the image       */
                             int mode,             /* In:  Intra in I or P frames    */
                             int qp,               /* In:  Quant. param. for luma    */
                             int *bitsPErrLuma,    /* Out: Bits for pred err in luma */
                             int *bitsPErrChroma,  /* Out: Bits for pred err chroma  */
                             int *bitsIPredLuma,   /* Out: Bits for luma intra pred  */
                             int *bitsIPredChroma, /* Out: Bits for chroma intra pred*/
                             vdxInstance_t
                               *instance,          /* In:  Current instance          */
                             int *BitError);       /* Out: For error handling        */


/*
 *
 * ipdDecodeIFramePred:
 *
 * Parameters:
 *      recFrame    O:  Reconstructed frame
 *      picSizeX    I:  Width of the frame
 *      picSizeY    I:  Height of the frame
 *      qp          I:  Quantization parameter for luminance
 *      cumBitsSliceHeaders I:  Accumulates paddding and emul prev bits
 *      bitsDctY    O:  Number of bits for DCT coeffs of Y frame
 *      bitsDctU    O:  Number of bits for DCT coeffs of u frame
 *      BitsDctV    O:  Number of bits for DCT coeffs of u frame
 *      BitsOvrY    O:  Number of bits for Y overhead
 *      BitsOvrUV   O:  Number of bits for u,v overhead
 *
 * Function:
 *      Decodes INTRA frame encoded using predictive coding.
 *   
 * Returns:
 *      Total number of bits needed to code the frame
 */

int32 ipdDecodeIFramePred(struct frame *recFrame, int32 picSizeX, 
                          int32 picSizeY, int32 qp, int32 *bitsDctY, 
                          int32 *bitsDctU, int32 *bitsDctV, 
                          int32 *bitsOvrY, int32 *bitsOvrUV,
                          int32 *cumBitsSliceHeaders,
                          vdxInstance_t *instance, int *BitError);


/*
 *
 * ipdInvQuantInvDct:
 *
 * Parameters:
 *      data    B:  DCT coefficients and pixel values 
 *      qp      I:  Quantization parameter
 *      mode    I:  Used mode (INTRA or INTER)
 *      doClip  I:  Flag for clipping 
 *
 * Function:
 *      Performs inverse quantization and inverse DCT. Mode specifies 
 *      how coefficients are quantized. If doClip is set, pixel values
 *      are clipped between 0 and 255.
 *   
 * Returns:
 *      Nothing.
 */

void ipdInvQuantInvDct(int data[8][8], int32 qp, int32 mode, int32 doClip);


/*
 *
 * ipdInvQuantCoeffs:
 *
 * Parameters:
 *      coeffs  O:  Resulting DCT coefficients 
 *      data    I:  Quantized DCT coefficients
 *      qp      I:  Quantization parameter
 *      mode    I:  Used mode (INTRA or INTER)
 *
 * Function:
 *      Performs inverse quantization. Mode specifies how the DC 
 *      coefficients are treated.
 *   
 * Returns:
 *      Nothing.
 */

void ipdInvQuantCoeffs(float64 coeffs[8][8], int data[8][8], int32 qp,
                       int32 mode);


/*
 *
 * ipdInitVLCTable:
 *
 * Parameters:
 *      Nothing.
 *
 * Function:
 *      Initializes the VLC table for DCT run-level pairs.
 *   
 * Returns:
 *      Nothing.
 */

void ipdInitVLCTable();


/*
 *
 * ipdCountCoefBits:
 *
 * Parameters:
 *      last    I:  Flag for last coefficient
 *      run     I:  Run length
 *      level   I:  Quantized level
 *
 * Function:
 *      Calculates the bits for one Run-Level-Last
 *      event.
 *   
 * Returns:
 *      Number of bits.
 */

int32 ipdCountCoefBits(int32 last, int32 run, int32 level);


/*
 *
 * ipdInvDct:
 *
 * Parameters:
 *      coeff   I:  Block of DCT coefficients
 *      d       I:  Block of pixel values
 *
 * Function:
 *      Descans zigzag-scanned coefficients and does
 *      inverse DCT on 64 coefficients.
 *
 * Returns:
 *      Nothing.
 */

void ipdInvDct(float64 coeff[8][8], float64 d[8][8]);

/*
 *
 * ipdCalcCoeffs:
 *
 * Parameters:
 *      coeffs  I:  DCT coefficients
 *
 * Function:
 *      Calculates the number of non-zero DCT coefficients.
 *
 * Returns:
 *      Number of coefficients. 
 */

int32 ipdCalcCoeffs(int32 coeffs[8][8]);


/***************************************************************************
 * Function    : Decode_I_Frame()
 *
 * Purpose     : Decodes the frame & slice layer bitstream for Intra
 *               coded pictures. Intra prediction is used followed by
 *               MCC prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : void
 *
 *               Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

int32 Decode_I_Frame(struct frame *recoFr, /* Out: Reconstruction frame          */
                     int32 picSizeX,         /* In:  Width of the image            */
                     int32 picSizeY,         /* In:  Height of the image           */
                     int32 frameNumber,      /* In:  Picture frame number          */
                     int32 qp,               /* In:  Quant. param. for luma        */
                     int32 *bitsDctY,        /* Out: Picture amount of coeff bits Y*/
                     int32 *bitsDctU,        /* Out: Picture amount of coeff bits U*/
                     int32 *bitsDctV,        /* Out: Picture amount of coeff bits V*/
                     int32 *bitsOvrY,        /* Out: Picture amount of ovh bits Y  */
                     int32 *bitsOvrUV,       /* Out: Picture amount of ovh bits UV */
                     int32 *cumBitsSliceHeaders, /* Out: Accumulates bits for header */
                     int16 *slicePredMap,    /* In:  Slice label buffer            */
                     int sliceMode,          /* In:  Flag to signal slice mode     */
                     vdxInstance_t
                       *instance,            /* In:  Current instance          */
                     int *BitError);         /* Out: For error handling        */

/*
 *
 * InvQuantCoeffs:
 *
 * Parameters:
 *      coeffs  O:  Resulting DCT coefficients 
 *      data    I:  Quantized DCT coefficients
 *      qp      I:  Quantization parameter
 *      mode    I:  Used mode (INTRA or INTER)
 *
 * Function:
 *      Performs inverse quantization. Mode specifies how the DC 
 *      coefficients are treated.
 *   
 * Returns:
 *      Nothing.
 */

void ipdInvQuantCoeffs2(float64 coeffs[8][8], int32 data[8][8], int32 qp, 
                        int32 mode);

#endif
