/*
 * biblin.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for bit buffer handling.
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "nccglob.h"
#include "Biblin.h"
#include "nrctypes.h"
#include "debug.h"

#ifdef BIB_VDEBUF
#include "vdebuf.h"
#include "sync.h"
#ifdef BIB_H223_BUF
#include "h223buf.h"
#endif
#endif

#ifndef BIB_REDIRECT_ALLOCATION
#include <stdlib.h>
#define bibMalloc malloc
#define bibCalloc calloc
#define bibRealloc realloc
#define bibDealloc free
#else
#include BIB_ALLOC_H
#define bibMalloc BIB_MALLOC
#define bibCalloc BIB_CALLOC
#define bibRealloc BIB_REALLOC
#define bibDealloc BIB_DEALLOC
#endif

#define BIB_OK NCC_OK
#define BIB_EOS (BIB_OK + 1)


/*
 * Local function prototypes
 */

static u_int32 bibFetchBits(int numberOfBits, bibBuffer_t *buffer, int get,
   int flush, int *numberOfBitsGot, int16 *errorCode);

#ifdef BIB_PSC_INSERT
static void bibPreprocess( bibBuffer_t *buf );
#endif

#ifdef BIB_VDEBUF
static u_int32 bibShowBitsSlave(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int16 *errorCode);
static void bibNewBuffer(bibBuffer_t *buffer, int fFlush, int16 *errorCode);
#endif


/*
 *
 * bibCreate
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    srcBuffer                  buffer containing the data
 *    srcBufferLength            the length of the buffer
 *    errorCode                  error code
 *
 * Function:
 *    This function creates a bit buffer from a buffer given as a parameter.
 *
 * Returns:
 *    The bibCreate function returns a pointer to a bibBuffer_t
 *    structure. If the function is unsuccessful NULL is returned.
 *
 * Error codes:
 *    ERR_BIB_STRUCT_ALLOC       if the bit buffer structure could not be
 *                               allocated
 */

bibBuffer_t *bibCreate(void *srcBuffer, unsigned srcBufferLength, int16 *errorCode)
{
   bibBuffer_t *bibBuffer = NULL;

   bibBuffer = (bibBuffer_t *) bibMalloc(sizeof(bibBuffer_t));
   if (bibBuffer == NULL) {
      *errorCode = ERR_BIB_STRUCT_ALLOC;
      deb("bibCreate: ERROR - bibMalloc failed.\n");
      return NULL;
   }

   bibBuffer->baseAddr = (u_char *) srcBuffer;
   bibBuffer->size = srcBufferLength;
   bibBuffer->getIndex = 0;
   bibBuffer->bitIndex = 7;
   bibBuffer->bitsLeft = (u_int32) srcBufferLength << 3;
   bibBuffer->numBytesRead = 0;

   #ifdef BIB_VDEBUF
      if (dlstOpen(&bibBuffer->bufList) < 0) {
         bibDealloc(bibBuffer);
         return NULL;
      }
      bibBuffer->numBitsBeforeCurrListItem = 0;
   #endif

   #ifdef BIB_PSC_INSERT
   bibBuffer->numEmulPreventBits = 0;
      bibPreprocess(bibBuffer);
   #endif

   return bibBuffer;
}


/*
 *
 * bibDelete
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    buffer                     a pointer to a bit buffer structure
 *    errorCode                  error code
 *
 * Function:
 *    The bibDelete function frees the memory allocated for the buffer.
 *
 * Returns:
 *    Nothing
 *
 * Error codes:
 *    None
 */

void bibDelete(bibBuffer_t **buffer, int16 *errorCode)
{
   if (*buffer) {
      #ifdef BIB_VDEBUF
      {
         bibListItem_t *currItem;

         if (dlstHead(&(*buffer)->bufList, &currItem) < 0) {
            *errorCode = ERR_BIB_BUFLIST;
            return;
         }

         do {
            if (dlstRemove(&(*buffer)->bufList, &currItem) < 0) {
               *errorCode = ERR_BIB_BUFLIST;
               return;
            }

            if (currItem) {
               #ifdef BIB_H223_BUF
                  if (vdeBufReturnDataBuffer(
                     ((h223Buffer_t *) currItem->buf)->retId, 
                     currItem->buf) < 0) {
                     *errorCode = ERR_BIB_BUFLIST;
                     return;
                  }
               #endif

               bibDealloc(currItem);
            }
         } while (currItem);
            
         dlstClose(&(*buffer)->bufList);
      }
      #endif

      bibDealloc((void *) (*buffer)->baseAddr);
      bibDealloc((void *) (*buffer));

      *buffer = NULL;
   }
}



/*
 *
 * bibFlushBits
 * bibGetBits
 * bibShowBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    numberOfBits               the number of bits wanted (1..32)
 *    buffer                     a pointer to a bit buffer structure
 *    numberOfBitsGot            the number of bits actually got
 *    syncStartIndex             the number of the first bit belonging to 
 *                               a synchronization code (1 .. numberOfBits).
 *                               Bits are numbered starting from 1 given to
 *                               the most significant bit of the returned code.
 *                               Thus, syncStartIndex - 1 most significant bits
 *                               of the returned code are valid coding
 *                               parameters (do not belong to a sync code).
 *                               0 is returned if the returned code does not
 *                               contain any part of a synchronization code.
 *    errorCode                  error code
 *
 * Function:
 *    The bibFlushBits function removes the next numberOfBits bits from
 *    the buffer.
 *
 *    The bibGetBits function gets the next numberOfBits bits from the buffer.
 *    The returned bits are removed.
 *
 *    The bibShowBits function gets the next numberOfBits bits from the
 *    buffer. The returned bits are not removed from the buffer.
 *
 * Returns:
 *   bibGetBits and bibShowBits:
 *       the next numberOfBits bits / the rest of the buffer if the end of
 *       the input file has been reached
 *
 * Error codes:
 *    ERR_BIB_NOT_ENOUGH_DATA    if one tries to get more bits from the buffer
 *                               than there is left
 *
 *    if EXTRA_CHECKS is defined:
 *    ERR_COM_NULL_POINTER       if pointer == NULL
 *    ERR_BIB_NUM_BITS           numberOfBits is illegal
 */

#ifdef BIB_NO_EOB
void bibFlushBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode)

#else
void bibFlushBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int16 *errorCode)
#endif

{

   #ifdef BIB_VDEBUF
      if (buffer->baseAddr == NULL) {
         bibNewBuffer(buffer, 1, errorCode);
         if (*errorCode)
            return;
      }

      bibFetchBits(numberOfBits, buffer, 0, 1, numberOfBitsGot, errorCode);

      if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
         int bitCount = *numberOfBitsGot;

         *errorCode = 0;

         bibNewBuffer(buffer, 1, errorCode);
         if (*errorCode)
            return;

         bibFlushBits(numberOfBits - bitCount, buffer, numberOfBitsGot, errorCode);

         *numberOfBitsGot += bitCount;
      }

   #elif defined(BIB_NO_EOB)

      *syncStartIndex = NO_BIT_ERROR;

      bibFetchBits(numberOfBits, buffer, 0, 1, numberOfBitsGot, errorCode);

      if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
         *errorCode = 0;
         *syncStartIndex = *numberOfBitsGot + 1;
         *numberOfBitsGot = numberOfBits;
      }

   #else
      bibFetchBits(numberOfBits, buffer, 0, 1, numberOfBitsGot, errorCode);
   #endif
}


/*
 *
 * bibGetBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * See bibFlushBits.
 *
 */

#ifdef BIB_NO_EOB
u_int32 bibGetBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode)

#else
u_int32 bibGetBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int16 *errorCode)
#endif

{
   #ifdef BIB_VDEBUF
      u_int32 retValue;

      if (buffer->baseAddr == NULL) {
         bibNewBuffer(buffer, 1, errorCode);
         if (*errorCode)
            return 0;
      }

      retValue = bibFetchBits(numberOfBits, buffer, 1, 1, numberOfBitsGot, 
         errorCode);

      if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
         int bitCount = *numberOfBitsGot;
         u_int32 retStore;

         *errorCode = 0;

         bibNewBuffer(buffer, 1, errorCode);
         if (*errorCode)
            return retValue;

         retStore = (retValue <<= (numberOfBits - *numberOfBitsGot));

         retValue = bibGetBits(numberOfBits - bitCount, buffer, numberOfBitsGot, 
            errorCode);

         retValue |= retStore;

         *numberOfBitsGot += bitCount;
      }
      return retValue;

   #elif defined(BIB_NO_EOB)
      u_int32 retValue;

      *syncStartIndex = NO_BIT_ERROR;

      retValue = bibFetchBits(numberOfBits, buffer, 1, 1, numberOfBitsGot, 
			      errorCode);

      if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
         *errorCode = 0;
         retValue <<= (numberOfBits - *numberOfBitsGot);
         *syncStartIndex = *numberOfBitsGot + 1;
         *numberOfBitsGot= numberOfBits;
      }

      return retValue;

   #else
      return bibFetchBits(numberOfBits, buffer, 1, 1, numberOfBitsGot, errorCode);
   #endif
}


/*
 *
 * bibNumberOfFlushedBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    buffer                     a pointer to a bit buffer structure
 *
 * Function:
 *    The bibNumberOfFlushedBytes returns the number of bits which
 *    are got (bibGetBits) or flushed (bibFlushBits) from the buffer since
 *    the buffer was created
 *
 * Returns:
 *    See above.
 *
 * Error codes:
 *    None.
 */

u_int32 bibNumberOfFlushedBits(bibBuffer_t *buffer)
{
#ifdef EXTRA_CHECKS
   if (!buffer)
      return 0;
#endif

   return buffer->numBytesRead * 8 + 7 - buffer->bitIndex;
}


/*
 *
 * bibNumberOfFlushedBytes
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    buffer                     a pointer to a bit buffer structure
 *
 * Function:
 *    The bibNumberOfFlushedBytes returns the number of whole bytes which
 *    are got (bibGetBits) or flushed (bibFlushBits) from the buffer since
 *    bibCreate was called.
 *
 * Returns:
 *    See above.
 *
 * Error codes:
 *    None.
 */

u_int32 bibNumberOfFlushedBytes(bibBuffer_t *buffer)
{
#ifdef EXTRA_CHECKS
   if (!buffer)
      return 0;
#endif

   return buffer->numBytesRead;
}


/*
 *
 * bibNumberOfRewBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    buffer                     a pointer to a bit buffer structure
 *
 * Function:
 *    The bibNumberOfRewBits returns the number of bits which can be
 *    successfully rewinded.
 *
 * Returns:
 *    See above.
 *
 * Error codes:
 *    None.
 */

u_int32 bibNumberOfRewBits(bibBuffer_t *buffer)
{
#ifdef EXTRA_CHECKS
   if (!buffer)
      return 0;
#endif

   #ifdef BIB_VDEBUF
      return buffer->numBitsBeforeCurrListItem +
         buffer->getIndex * 8 + 7 - buffer->bitIndex;
   #else
      return buffer->numBytesRead * 8 + 7 - buffer->bitIndex;
   #endif
}


/*
 *
 * bibRewindBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    numberOfBits               the number of bits to rewind
 *    buffer                     a pointer to a bit buffer structure
 *    errorCode                  error code
 *
 * Function:
 *    This function rewinds the pointers to the buffer
 *    so that already flushed or got bits can be read
 *    again.
 *
 * Returns:
 *    Nothing.
 *
 * Error codes:
 *    ERR_BIB_CANNOT_REWIND      if numberOfBits is larger than which is
 *                               possible to get (see also bibNumberOfRewBits).
 *
 *    If EXTRA_CHECKS is defined:
 *    ERR_COM_NULL_POINTER       if buffer == NULL
 */

void bibRewindBits(u_int32 numberOfBits, bibBuffer_t *buffer, int16 *errorCode)
{
#ifdef EXTRA_CHECKS
   if (!buffer) {
      *errorCode = ERR_COM_NULL_POINTER;
      deb("bibRewindBits: ERROR - buffer == NULL\n");
      return;
   }
#endif

   #ifdef BIB_VDEBUF
   {
      bibListItem_t *item;
      u_int32 numBitsInCurrBuf;

      if (bibNumberOfRewBits(buffer) < numberOfBits) {
         *errorCode = ERR_BIB_CANNOT_REWIND;
         return;
      }

      numBitsInCurrBuf = buffer->getIndex * 8 + 7 - buffer->bitIndex;

      while (numberOfBits > numBitsInCurrBuf) {
         numberOfBits -= numBitsInCurrBuf;
         buffer->numBytesRead -= buffer->getIndex;

         if (dlstPrev(&buffer->bufList, &item) < 0) {
            *errorCode = ERR_BIB_BUFLIST;
            return;
         }
         
         #ifdef BIB_H223_BUF
         {
            h223Buffer_t *h223Buf = (h223Buffer_t *) item->buf;
            buffer->baseAddr = h223Buf->dataptr;
            buffer->size = h223Buf->datalen;
            buffer->getIndex = h223Buf->datalen;
            buffer->bitIndex = 7;
            buffer->bitsLeft = 0;
            numBitsInCurrBuf = h223Buf->datalen * 8;
            buffer->numBitsBeforeCurrListItem -= numBitsInCurrBuf;
         }
         #endif
      }
   }
   #endif

   /* All bits to rewind are in the latest byte */
   if (numberOfBits <= (u_int32) (7 - buffer->bitIndex))
      buffer->bitIndex += numberOfBits;

   /* Bits to rewind are within several bytes */
   else {
      u_int32
         numBitsWithoutFirstByte = numberOfBits - (7 - buffer->bitIndex),
         numWholeBytes = numBitsWithoutFirstByte / 8,
         numBitsInLastByte = numBitsWithoutFirstByte - numWholeBytes * 8;

      if (numBitsInLastByte) {
         if (buffer->getIndex >= numWholeBytes + 1) {
            buffer->getIndex -= numWholeBytes + 1;
            buffer->bitIndex = numBitsInLastByte - 1;
            buffer->numBytesRead -= numWholeBytes + 1;
         }
         else {
            *errorCode = ERR_BIB_CANNOT_REWIND;
            deb("bibRewindBits: ERROR - cannot rewind.\n");
         }
      }
      else {
         if (buffer->getIndex >= numWholeBytes) {
            buffer->getIndex -= numWholeBytes;
            buffer->bitIndex = 7;
            buffer->numBytesRead -= numWholeBytes;
         }
         else {
            *errorCode = ERR_BIB_CANNOT_REWIND;
            deb("bibRewindBits: ERROR - cannot rewind.\n");
         }
      }
   }
   buffer->bitsLeft += numberOfBits;
}


/*
 *
 * bibShowBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * See bibFlushBits.
 *
 */

#ifdef BIB_NO_EOB
u_int32 bibShowBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int *syncStartIndex, int16 *errorCode)

#else
u_int32 bibShowBits(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int16 *errorCode)
#endif

{
   #ifdef BIB_VDEBUF
      int16 ownError = 0;
      u_int32 retValue;

      if (buffer->baseAddr == NULL) {
         bibNewBuffer(buffer, 1, errorCode);
         if (*errorCode)
            return 0;
      }

      retValue = bibShowBitsSlave(numberOfBits, buffer, numberOfBitsGot, 
         &ownError);
      if (ownError && ownError != ERR_BIB_NOT_ENOUGH_DATA)
         return retValue;

      bibRewindBits(*numberOfBitsGot, buffer, errorCode);

      if (ownError)
         *errorCode = ownError;

      return retValue;

   #elif defined(BIB_NO_EOB)
      u_int32 retValue;

      *syncStartIndex = NO_BIT_ERROR;
      
      retValue = bibFetchBits(numberOfBits, buffer, 1, 0, numberOfBitsGot, 
         errorCode);

      if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
         *errorCode = 0;
         retValue <<= (numberOfBits - *numberOfBitsGot);
         *syncStartIndex = *numberOfBitsGot + 1;
         *numberOfBitsGot= numberOfBits;
      }

      return retValue;

   #else
      return bibFetchBits(numberOfBits, buffer, 1, 0, numberOfBitsGot, errorCode);
   #endif
}


/*
 * Local functions
 */

/*
 *
 * bibFetchBits
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    See bibFlushBits for the parameters except for
 *    get          if == 0 the bit values are ignored
 *    flush        if != 0 the bits got are removed from the buffer
 *
 * Function:
 *    The bibFetchBits function is used as a subroutine of bibGetBits,
 *    bibFlushBits and bibShowBits.
 *
 * Returns:
 *    See bibFlushBits.
 *
 * Error codes:
 *    See bibFlushBits.
 */

static u_int32 bibFetchBits(int numberOfBits, bibBuffer_t *buffer, int get,
   int flush, int *numberOfBitsGot, int16 *errorCode)
{
   static const u_char
      msbMask[8] = {1, 3, 7, 15, 31, 63, 127, 255},
      lsbMask[9] = {255, 254, 252, 248, 240, 224, 192, 128, 0};
   u_char
      *baseAddr = buffer->baseAddr;
   int
      bitIndex = buffer->bitIndex;
   unsigned
      startIndex,       /* the index of the first byte to read */
      startMask,        /* the mask for the first byte */
      endIndex,         /* the index of the last byte to read */
      endMask,          /* the mask for the last byte */
      endShift,         /* the number of shifts after masking the last byte */
      endNumberOfBits,  /* the number of bits in the last byte */
      newBitIndex,      /* bitIndex after getting the bits */
      newGetIndex,      /* getIndex after getting the bits */
      numBytesFlushed,  /* the number of bytes flushed from the buffer */
      size = buffer->size,
      getIndex = buffer->getIndex;
   u_int32
      returnValue = 0;

#ifdef EXTRA_CHECKS
   if (!buffer) {
      *errorCode = ERR_COM_NULL_POINTER;
      deb("bibFetchBits: ERROR - buffer is NULL.\n");
      return 0;
   }

   /* Check that number of bits is legal, i.e. from 1 to 32. */
   if (numberOfBits <= 0 || numberOfBits > 32) {
      *errorCode = ERR_BIB_NUM_BITS;
      deb("bibFetchBits: ERROR - numberOfBits is illegal.\n");
      return 0;
   }
#endif

   /* Check if enough bits are available. */
   if (buffer->bitsLeft < (u_int32) numberOfBits) {
      *errorCode = ERR_BIB_NOT_ENOUGH_DATA;
      #ifndef BIB_NO_EOB
         deb("bibFetchBits: ERROR - not enough data.\n");
      #endif
      numberOfBits = (int) buffer->bitsLeft;
   }
   *numberOfBitsGot = numberOfBits;

   startIndex = getIndex;
   startMask = msbMask[bitIndex];

   if (bitIndex + 1 >= numberOfBits) {
      /* The bits are within one byte. */
      endShift = bitIndex - numberOfBits + 1;
      endMask = lsbMask[endShift];
      returnValue = (baseAddr[startIndex] & startMask & endMask) >>
         endShift;
      if (endShift > 0) {
         newBitIndex = bitIndex - numberOfBits;
         newGetIndex = getIndex;
      }
      else {
         newBitIndex = 7;
         newGetIndex = getIndex + 1;
      }
   }

   else {
      /* The bits are in multiple bytes. */
      numberOfBits -= bitIndex + 1;
      endNumberOfBits = numberOfBits & 7;

      newBitIndex = 7 - endNumberOfBits;
      newGetIndex = getIndex + (numberOfBits >> 3) + 1;

      if (get) {
         /* Count the return value. */
         endIndex = newGetIndex;
         endShift = 8 - (numberOfBits & 7);
         endMask = lsbMask[endShift];

         returnValue = baseAddr[startIndex] & startMask;
         startIndex++;

         while (startIndex != endIndex) {
            returnValue <<= 8;
            returnValue += baseAddr[startIndex];
            startIndex++;
         }

         returnValue <<= endNumberOfBits;
         returnValue += (baseAddr[endIndex] & endMask) >> endShift;
      }
   }

   if (flush) {
      /* Update indexes. */
      buffer->getIndex = newGetIndex;
      buffer->bitIndex = newBitIndex;

      buffer->bitsLeft -= *numberOfBitsGot;

      numBytesFlushed = newGetIndex - getIndex;

      buffer->numBytesRead += numBytesFlushed;
   }

   return returnValue;
}


#ifdef BIB_PSC_INSERT
/*
 *
 * bibPreprocess
 *    Copyright (c) 1996 Nokia Research Center, Tampere, FINLAND
 *
 * Parameters:
 *    buf  I: Pointer to a bit buffer structure.
 *
 * Function:
 *    Preprocesses the given buffer searching for the patterns
 *    00000000 00000000 010001, 00000000 00000000 10, or
 *    00000000 00000000 011110.
 *    After finding the first pattern, it removes the last '1' in order
 *    to do inverse PSC emulation prevention.
 *    After finding the second or third pattern, it removes the last '0'
 *    in order to do inverse SSC or EOS emulation prevention.
 * 
 * Returns:
 *    Nothing.
 *
 * Error codes:
 */

static void bibPreprocess(bibBuffer_t *buf)
{
  int               
    bits, sh, foundPrevBit;
  unsigned char     
    *w, *p = w = (unsigned char*)buf->baseAddr,
    bitMask1, bitMask2;
  unsigned
    i;

  sh = 8;
  bits = 0;

  /*
   * Scan through the bit buffer byte by byte.
   */
  for (i = 0; i < buf->size - 1; i++)
  {
    /*
     * Check if there are two consecutive bytes of zeros.
     */
    if (*p == 0 && i < buf->size - 2 && p[1] == 0)
    {
      /* 
       * Check the third byte in a row for start code emulation.
       */
      unsigned char c = p[2];

      foundPrevBit = 0;

      /*
       * SSC emulation prevention.
       */
      if ((c & 0xc0) == 0x80) 
      {
        foundPrevBit = 1;
        bitMask1 = 0x80;  
        bitMask2 = 0x3f;
      }
      /*
       * PSC emulation prevention.
       */
      else if ((c & 0xfc) == 0x44)
      {
        foundPrevBit = 1;
        bitMask1 = 0xf8;  
        bitMask2 = 0x03;
      }
      /*
       * EOS emulation prevention.
       */
      else if ((c & 0xfc) == 0x78)
      {
        foundPrevBit = 1;
        bitMask1 = 0xf8;
        bitMask2 = 0x03;
      }

      if (foundPrevBit)
      {
	buf->numEmulPreventBits++;

	/* remove the extra bit */
	c = (c & bitMask1) | ((c & bitMask2) << 1);
	*w++ = 0; p++;
	*w++ = (unsigned int)c >> sh; p++;
	*w++ = (((unsigned int)p[1] << 1) | ((unsigned int)c << 8)) >> sh;
	p++;
	bits++;
	if (--sh == 0)
	{
	  sh = 8;
	  w--;
        }
	i += 2;
	continue;
      }
    }
    *w++ = (((unsigned int)*p << 8) | (unsigned int)p[1]) >> sh;
    p++;
  }

  if (i < buf->size)
  {
    /* process last byte */
    
    *w = *p << (8 - sh);
  }

  buf->bitsLeft -= bits;
  buf->size = (buf->bitsLeft + 7) >> 3;
}
#endif


#ifdef BIB_VDEBUF
static u_int32 bibShowBitsSlave(int numberOfBits, bibBuffer_t *buffer,
   int *numberOfBitsGot, int16 *errorCode)
{
   u_int32 retValue;

   retValue = bibFetchBits(numberOfBits, buffer, 1, 1, numberOfBitsGot, 
      errorCode);

   if (*errorCode == ERR_BIB_NOT_ENOUGH_DATA) {
      int bitCount = *numberOfBitsGot;
      u_int32 retStore;

      *errorCode = 0;

      bibNewBuffer(buffer, 0, errorCode);
      if (*errorCode) {
         return retValue;
      }

      retStore = (retValue <<= (numberOfBits - *numberOfBitsGot));

      retValue = bibShowBitsSlave(numberOfBits - bitCount, buffer, numberOfBitsGot, 
         errorCode);

      retValue |= retStore;

      *numberOfBitsGot += bitCount;
   }
   return retValue;
}

static void bibNewBuffer(bibBuffer_t *buffer, int fFlush, int16 *errorCode)
{
   bibListItem_t *newListItem, *currListItem;
   int
      sncCode,
      numStuffBits;
   int16
      error = 0;
   void *newBuf;

   /* If a newer buffer exists */
   if (dlstNextExists(&buffer->bufList)) {
      /* Fetch the next newer buffer */
      if (dlstNext(&buffer->bufList, &currListItem) < 0) {
         *errorCode = ERR_BIB_BUFLIST;
         return;
      }

      buffer->numBitsBeforeCurrListItem += buffer->size * 8;

      /* Update the current buffer data according to the fetched buffer */
      #ifdef BIB_H223_BUF
      {
         h223Buffer_t *h223Buf = (h223Buffer_t *) currListItem->buf;
         buffer->baseAddr = h223Buf->dataptr;
         buffer->size = h223Buf->datalen;
         buffer->getIndex = 0;
         buffer->bitIndex = 7;
         buffer->bitsLeft = (u_int32) h223Buf->datalen << 3;
      }
      #endif

   }

   /* Else (no newer buffer exists) */
   else {
      int fEmptyBuffer = 0;

      do {
         /* Get buffer from the buffer handler */
         newBuf = vdeBufGetDataBuffer();
         if (!newBuf) {
            *errorCode = ERR_BIB_NOT_ENOUGH_DATA;
            return;
         }

         #ifdef BIB_H223_BUF
         {
            h223Buffer_t *h223Buf = (h223Buffer_t *) newBuf;

            /* If the buffer contains no data, get the next one */
            if (h223Buf->datalen == 0) {
               if (vdeBufReturnDataBuffer(h223Buf->retId, h223Buf) < 0) {
                  *errorCode = ERR_BIB_BUFLIST;
                  return;
               }
               fEmptyBuffer = 1;
            }
            else
               fEmptyBuffer = 0;
         }
         #endif
      } while (fEmptyBuffer);

      buffer->numBitsBeforeCurrListItem += buffer->size * 8;

      /* Update the current buffer data according to the new buffer */
      #ifdef BIB_H223_BUF
      {
         h223Buffer_t *h223Buf = (h223Buffer_t *) newBuf;

         buffer->baseAddr = h223Buf->dataptr;
         buffer->size = h223Buf->datalen;
         buffer->getIndex = 0;
         buffer->bitIndex = 7;
         buffer->bitsLeft = (u_int32) h223Buf->datalen << 3;
      }
      #endif

      /* If the buffer is large enough to contain a synchronization code 
         (22 bits) and if the buffer size exceeds the minimum rewindable
         size (32 bits, for bibShowBits) */
      if (buffer->size >= 4) {

         /* Get the synchronization code from the beginning of the buffer */
         sncCode = sncCheckSync(buffer, &numStuffBits, &error);
         if (error) {
            *errorCode = ERR_BIB_BUFLIST;
            return;
         }

         /* If the code is PSC or EOS */
         if (sncCode == SNC_PSC || sncCode == SNC_EOS) {
            /* Remove the previous buffers
               (in bibShowBits, remove the previous buffers excluding 
               the second latest */
            if (dlstHead(&buffer->bufList, &currListItem) < 0) {
               *errorCode = ERR_BIB_BUFLIST;
               return;
            }

            while ((fFlush && currListItem) || 
               (!fFlush && dlstNextExists(&buffer->bufList))) {
               if (dlstRemove(&buffer->bufList, &currListItem) < 0) {
                  *errorCode = ERR_BIB_BUFLIST;
                  return;
               }

               if (currListItem) {
                  #ifdef BIB_H223_BUF
                     buffer->numBitsBeforeCurrListItem -= 
                        ((h223Buffer_t *) currListItem->buf)->datalen * 8;
                     if (vdeBufReturnDataBuffer(
                        ((h223Buffer_t *) currListItem->buf)->retId, 
                        currListItem->buf) < 0) {
                        *errorCode = ERR_BIB_BUFLIST;
                        return;
                     }
                  #endif

                  bibDealloc(currListItem);
               }
            }
         }
      }

      /* Add the new buffer to the buffer list */
      newListItem = bibMalloc(sizeof(bibListItem_t));
      if (newListItem == NULL) {
         *errorCode = ERR_BIB_BUFLIST;
         return;
      }

      newListItem->buf = newBuf;

      if (dlstTail(&buffer->bufList, &currListItem) < 0) {
         *errorCode = ERR_BIB_BUFLIST;
         return;
      }

      if (dlstAddAfterCurr(&buffer->bufList, newListItem) < 0) {
         *errorCode = ERR_BIB_BUFLIST;
         return;
      }
   }
}
#endif
