/*
 * fil.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for deblocking filter.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "nrctypes.h"
#include "nccglob.h"
#include "mainFlags.h"
#include "deblockFilt.h"

/*
 * Global variables
 */

u_char annexTlookup[32] =
  { 0,  1,  2,  3,  4,  5,  6,  6,  7,  8,  9,  9, 10, 10, 11, 11,
   12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15};

/*
 * Module scope variables
 */

u_char
  filDiffActTable[32][9],
  filOverallActTable[32][256];

/* LogTable32 is 65536 * Log(i) for i = 0 to 31 */
int32 LogTable[32] = { 0, 45426, 71998, 90852, 105476, 117424, 
							  127527, 136278, 143997, 150902, 157148, 
							   162850, 168096, 172953, 177474, 181704, 
							   185677, 189423, 192966, 196328, 199525,
                               202574, 205487, 208276, 210952, 213522, 
							   215995, 218379, 220679, 222900, 225049 };

/*
 * Define local function prototypes
 */

#define filMax(X,Y) ((X) > (Y) ? (X) : (Y))

#define filDiff(X,Y) ((X) > (Y) ? (X) - (Y) : (Y) - (X))

#define ChooseSmall(X,Y) ((X) <= (Y) ? (X) : (Y))

static int32 filFindFinalActFunc(u_char *image, int32 location, int32 dir,
                                 int32 max, int32 qp);

#define filFindFinalAct( dest, image, location, step, max, qp )             \
  {                                                                         \
    int32 i;                                                                \
                                                                            \
    for(i = 1; i < max; i++)                                              \
      if(filDiff((int32)image[location], (int32)image[location + i * step]) \
         > filDiffActTable[qp][i])                                          \
        break;                                                              \
                                                                            \
    dest = i;                                                               \
  }

#define filFindOverallAct(image, location, step, qp )                       \
  filOverallActTable[qp][filDiff((int32)image[location],                    \
                         (int32)image[location + step])]

/*
 * Function implementations
 */

/*
 *
 * filLoopFilter:
 *
 * Parameters:
 *      orig      I: Original frame
 *      picSizeX  I: Luminance frame width
 *      picSizeY  I: Luminance frame height
 *      qp        I: Quantization parameter
 *      result    O: Result of filtering
 *
 * Function:
 *      Calls filCompFilt() function, which performs loop filtering for
 *      luminance and chrominance frames.
 *
 * Returns:
 *      Nothing
 */


void filLoopFilter(struct frame *orig, int32 picSizeX, int32 picSizeY,
                   int32 qp, struct frame *result, int useLoopFilter)
{
  int32
    qpchr;

  qpchr = annexTlookup[qp];
  
  filCompFilt(orig->yptr, picSizeX, picSizeY, qp, result->yptr, 1, 0, NULL, NULL, useLoopFilter); /* enco_flag has no importance here */
  filCompFilt(orig->uptr, picSizeX / 2, picSizeY / 2, qpchr, result->uptr, 1, 0, NULL, NULL, useLoopFilter);
  filCompFilt(orig->vptr, picSizeX / 2, picSizeY / 2, qpchr, result->vptr, 1, 0, NULL, NULL, useLoopFilter);

}


/*
 *
 * filCompFilt:
 *
 * Parameters:
 *      orig      I: Original frame
 *      picSizeX  I: Luminance frame width
 *      picSizeY  I: Luminance frame height
 *      qp        I: Quantization parameter
 *      result    O: Result of filtering
 *
 * Function:
 *      Performs loop filtering for frame using average filter with varying 
 *      filter support.
 *
 * Returns:
 *      Nothing
 */

#define stepX 8
#define stepY 8
#define mbSizeX 16
#define mbSizeY 16
#define mbShift 4
#define mbHalf 128
#define wb  16 
#define wb2 32
#define wb3 48
#define wb4 64
#define wb5 80

/*
 * Round fixed-point number x with y fraction bits to nearest integer.
 */
#define toUchar(x,y) (u_char)( ((int32)(x) + (1<<((y)-1))) >> (y) )
/*
 * Decrease precision of fixed-point number x by y bits and round to nearest
 * value.
 */
#define decPrec(x,y) ( ((x) + (1<<((y)-1))) >> (y) )


void filCompFilt(u_char *orig, int32 picSizeX, int32 picSizeY,
                 int32 qp, u_char *result, int intra_flag, int enco_flag,
                 object *object_list1, objectBuffer_t *object_list2,
                 int useLoopFilter)
{
  int32
    i1, j1, i2, j2,               /* Loop counters */
    offSetX, offSetY,
    rpb, lpb, rpf, lpf,           /* Location counters for pixels at block */
                                  /* boundaries */
    rightActivity, leftActivity,  /* Local activities for both sides of the 
                                     boundary */
    rightTrunc, leftTrunc,        /* Maximum search supports according to 
                                     region types */
    overallActivity,              /* Maximum search support according to 
                                     overall activity */
    tmpInt, tmpInt2,              /* Auxiliary variables */
    wf, wf2, wf3, wf4, wf5,       /* Frame wraping values for y direction */
    *aop, corner,
    aux[256],                     /* Auxiliary frame buffer */
    aux2[16];

  u_char  
    *cp1, *cp2;
    
  int 
    regmodes[5], pecmodes[5];

  static int32
    inv3 = 65536/3,
    inv5 = 65536/5,
    inv7 = 65536/7,
    inv9 = 65536/9,
    inv25 = 65536/25;


  wf = picSizeX;
  wf2 = 2 * picSizeX;
  wf3 = 3 * picSizeX;
  wf4 = 4 * picSizeX;
  wf5 = 5 * picSizeX;
    
  /* 
   * Scan the original image macroblock by macroblock. First filter the
   * vertical 8x8 block boundaries and then the horizontal 8x8 block 
   * boundaries.
   */
  for (j1 = 0; j1 < picSizeY - stepY; j1 += stepY)
  {
    for (i1 = 0; i1 < picSizeX - stepX; i1 += stepX)
    {
      /*
       * Copy one macroblock of the original image to 16x16 auxiliary buffer.
       */
      offSetX = (i1 > 0 ? stepX : 0);
      offSetY = (j1 > 0 ? stepY : 0);
      
      corner = j1 * picSizeX + i1;  /* top-left corner coordinate of the macroblock */
      cp1 = (offSetX || offSetY ? result : orig) + corner; 
	  /* if not on any frame border, cp1 points to filtered data else original data */
      cp2 = (offSetY ? result : orig) + corner;
	  /* if not top border of frame, cp2 points to filtered data else original data */

	  /* Get the region modes and the prediction error coding modes of the	
	     8x8 blocks in the macroblock region being filtered 						*/
	  if (!intra_flag)
      {
		  if (enco_flag) /* If filtering an encoded frame */
            fil_form_region_and_pec_info(i1, j1, mbSizeX, picSizeX, picSizeY, object_list1, NULL,
                                     intra_flag, enco_flag, regmodes, pecmodes);
          else           /* if filtering a decoded frame */
            fil_form_region_and_pec_info(i1, j1, mbSizeX, picSizeX, picSizeY, NULL, object_list2,
                                     intra_flag, enco_flag, regmodes, pecmodes);
	  } 
	  /* The following 4 for loops copies the necessary pixels quadrant (8x8) by quadrant */
	  /* to the auxiliary buffer aux[] */
	  for (j2 = 0; j2 < stepY; j2++, cp1 += picSizeX, cp2 += picSizeX)
      { 
        for (i2 = 0; i2 < stepX; i2 += 2)
        {
          aux[(j2 << mbShift) + i2] = (int32)cp1[i2] << 16;   /* form the left half macroblock by taking pixel */
          aux[(j2 << mbShift) + i2 + 1] = (int32)cp1[i2 + 1] << 16; /* values from where cp1 and */
          aux[(j2 << mbShift) + stepX + i2] = (int32)cp2[stepX + i2] << 16;             /* cp2 points */
          aux[(j2 << mbShift) + stepX + i2 + 1] = (int32)cp2[stepX + i2 + 1] << 16;
        }/* aux[] holds the macroblock to be filtered */
      }

      cp1 = (offSetX ? result : orig) + corner + stepY * picSizeX;
      cp2 = orig + corner + stepY * picSizeX;       
      for (j2 = stepY; j2 < mbSizeX; j2++, cp1 += picSizeX, cp2 += picSizeX)
      { 
        for (i2 = 0; i2 < stepX; i2 += 2)
        {
          aux[(j2 << mbShift) + i2] = (int32)cp1[i2] << 16;  /* form the right half macroblock by taking pixel */
          aux[(j2 << mbShift) + i2 + 1] = (int32)cp1[i2 + 1] << 16; /* values from where cp1 and */
          aux[(j2 << mbShift) + stepX + i2] = (int32)cp2[stepX + i2] << 16; /* cp2 points */
          aux[(j2 << mbShift) + stepX + i2 + 1] = (int32)cp2[stepX + i2 + 1] << 16;
        }
      }
      /* HORIZONTAL FILTERIG FIRST!!! */
      /*
       * Calculate pixel locations in the auxiliary 16x16 buffer, pixel with 
       * higher x - coordinate is referred as right pixel.
       */
      rpb = stepX + (offSetY << mbShift); /* points to the first pixel at the right of the boundary */
      lpb = rpb - 1;                                    /* points to the first pixel at the left of boundary */

      /*
       * Scan the macroblock row by row.
       */
      cp1 = (offSetX ? result : orig) + corner + offSetY * picSizeX;
      cp2 = orig + corner + offSetY * picSizeX;

      for(j2 = offSetY; j2 < mbSizeY; j2++, rpb += mbSizeX, lpb += mbSizeX)
      {
        /*
         * Copy the original 8-bit pixel values of the row
         * to 32-bit integer buffer.
         */
		/* The following for loop copies the necessary 16 pixel wide row into auxiliary */
		/* bufer aux2[]. cp1 and cp2 points to necessary pixel locations in origianl or filtered  */
		/* frames */
        for(i2 = 0; i2 < stepX; i2 += 2)
        {
          aux2[i2]     = cp1[i2];
          aux2[i2 + 1] = cp1[i2 + 1];
          aux2[stepX + i2]     = cp2[stepX + i2];
          aux2[stepX + i2 + 1] = cp2[stepX + i2 + 1];
        }
       
        cp1 += picSizeX;
        cp2 += picSizeX;

        aop = aux2 + stepX; /* points to the middle pixel of the auxiliary buffer aux2[] */
          
        /*
         * Get search support truncations for both sides of the boundary
         * according to region types.
         */
        rightTrunc = 6;
        leftTrunc = 6;
          
        /*
         * Update search support truncations for both sides
         * of the boundary according to pixel value difference at the
         * block boundary
         */
        overallActivity = filFindOverallAct(aop, 0, -1, qp);
        rightTrunc = overallActivity;
        leftTrunc = overallActivity;
        
        /* If intra flag is not set, make truncations on the 
		     * left and right activity parameters according to region mode
         * and pecmode data.
         */
        if(useLoopFilter){
          if (!intra_flag)
          { /* Make the further truncation on the overall activity parameter stored in 
		           rightTrunc and leftTrunc */
            fil_overall_activity_truncator(1, j2, regmodes, pecmodes, &rightTrunc, &leftTrunc);
          }
        }

        /*
         * Find final activities for both sides of the boundary
         * according to pixel value differences within blocks
         */
        filFindFinalAct(rightActivity, aop, 0, 1, rightTrunc, qp);
        filFindFinalAct(leftActivity, aop, -1, -1, leftTrunc, qp);
          
        /*
         * Check if filtering is needed
         */
        if(rightActivity > 1 && leftActivity > 1){ 
          /* if there is activity at both sides of the boundary */
          /* The following Table shows the rightActivity vs filter window lengths for particular 
		       pixels at the right of the boundary. aop[0] corresponds to the first pixel at the
			   right side of the boundary. R1 = aop[0]. 

										*		Filter Window Size for Pixels 	   *
              
			  rightActivity        |         R1            |        R2      |        R3       
                       1                         x                       x                   x
                       2                         1                       x                   x  
					   3                         1                       1*                 x
					   4                         2                       2                   x
					   5                         2                       2                   2**
					   6                       3 or 2 ***           3                   3   
 

              x        : no filtering done
			  *        : When updating R2 the updated value of R1 is used.  
			  **      : When updating R3 the updated values of R2 and R1 are used.
              ***    : The support window for R1 is 3 if leftAvtivity > 2, otherwise it is 2.            
		  
		  NOTE : The above table can also be used for the window size of the pixels 
		              on the left of the boundary!!!. 
					  Replace the folowing : 
					   rightActivity <--> leftActivity, 
					   R1,2,3 <--> L1,2,3 				   
					  
					The above table is also valid for the vertical filtering!!!!
				    Right <--> BELOW the boundary  ( pixels with higher y coordinate) 
					Left  <--> ABOVE the boundary   ( pixels with lowe y coordinate)
			*/
           /*
           * Filter right side of the boundary with proper filter
		   * The filter is averaging filter of window size selected according to the table above.
		   * The filtering is done using Integer Arithmetic only!!!!!
           */
          if(rightActivity > 5) 
          { /* rightActivity = 6 */
            if(leftActivity > 2) /* window size for R1 = 3 */
            { /* aop[0] corresponds first pixel on the right side of the boundary */
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2] + aop[3];
			  /* put the filtered values to the auxiliary buffer aux[]*/
              aux[rpb] = (aop[-3] + tmpInt) * inv7;  /* 1st pixel on the right of  boundary */
              aux[rpb + 1] = (tmpInt + aop[4]) * inv7; /* 2nd  pixel on the right*/
              aux[rpb + 2] = aux[rpb + 1] + (aop[5] - aop[-2]) * inv7; /* 3rd  pixel on the right*/
            }
            else /* window size for R1 = 2 */
            {
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
              aux[rpb] = tmpInt * inv5;

              aux[rpb + 1] = (tmpInt + aop[3] + aop[4]) * inv7;
              
              aux[rpb + 2] = aux[rpb + 1] + (aop[5] - aop[-2]) * inv7;
            }
          }
          else if(rightActivity == 5)
          {
            tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
            aux[rpb] = tmpInt * inv5;
            tmpInt2 = tmpInt + aop[3] - aop[-2];
            aux[rpb + 1] = tmpInt2 * inv5;
            aux[rpb + 2] = (tmpInt + tmpInt2) * inv25 +
			               (aop[2] + aop[3] + aop[4]) * inv5;
          }
          else if(rightActivity > 3) 
          { /* rightActivity = 4 */
            tmpInt = aop[-1] + aop[0] + aop[1] + aop[2];
            aux[rpb] = (aop[-2] + tmpInt) * inv5;
            aux[rpb + 1] = (tmpInt + aop[3]) * inv5;
          }
          else if(rightActivity == 3)
          {
            tmpInt = aop[-1] + aop[0] + aop[1];
            aux[rpb] = tmpInt * inv3;
            aux[rpb + 1] = tmpInt * inv9 + (aop[1] + aop[2]) * inv3;
          }
          else
          { /* rightActivity = 2 */
            aux[rpb] = (aop[-1] + aop[0] + aop[1]) * inv3;
          }
          /*
           * Filter left side of the boundary with proper filter
           */

          aop--; 
		  /* Now aop[0] points to the first pixel on the left of the boundary!!! */

          if(leftActivity > 5)
          {/* leftActivity =6*/
            if(rightActivity > 2)
            {/* window size for L1 = 3 */
              tmpInt = aop[-3] + aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
              aux[lpb] = (tmpInt + aop[3]) * inv7;
              aux[lpb - 1] = (aop[-4] + tmpInt) * inv7;
              aux[lpb - 2] = aux[lpb - 1] + (aop[-5] - aop[2]) * inv7;
            }
            else
            {/* window size for L1 = 2 */
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
              aux[lpb] = tmpInt * inv5;
              aux[lpb - 1] = (tmpInt + aop[-3] + aop[-4]) * inv7;
              aux[lpb - 2] = aux[lpb - 1] + (aop[-5] - aop[2]) * inv7;
            }
          }
          else if (leftActivity == 5)
          {
            tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
            aux[lpb] = tmpInt * inv5;
            tmpInt2 = tmpInt + aop[-3] - aop[2];
            aux[lpb - 1] = tmpInt2 * inv5;
            aux[lpb - 2] = (tmpInt  + tmpInt2) * inv25 +
                           (aop[-4] + aop[-3] + aop[-2]) * inv5;
          }
          else if(leftActivity > 3)
          {/* leftActivity =4*/
            tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1];
            aux[lpb] = (tmpInt + aop[2]) * inv5; 
            aux[lpb - 1] = (aop[-3] + tmpInt) * inv5;
          }
          else if(leftActivity==3)
          {
            tmpInt = aop[-1] + aop[0] + aop[1];
            aux[lpb] = tmpInt * inv3;
            aux[lpb - 1] = tmpInt * inv9 + (aop[-2] + aop[-1]) * inv3;
          }
          else
          {/* leftActivity =2*/
            aux[lpb] = (aop[-1] + aop[0] + aop[1]) * inv3;
          }
        }
      }

      /*
       * Copy the macroblock after horizontal filtering to the original
       * image.
	   * Vertical filtering uses the output of the horizontal filter for
	   * pixels on the right of the boundary. 
       */
      cp1 = result + corner + offSetY * picSizeX;

      for (j2 = offSetY; j2 < mbSizeY; j2++, cp1 += picSizeX)
      { 
        for (i2 = 0; i2 < mbSizeX; i2 += 4)
        {
           cp1[i2]     = toUchar(aux[(j2 << mbShift) + i2], 16);
           cp1[i2 + 1] = toUchar(aux[(j2 << mbShift) + i2 + 1], 16);
           cp1[i2 + 2] = toUchar(aux[(j2 << mbShift) + i2 + 2], 16);
           cp1[i2 + 3] = toUchar(aux[(j2 << mbShift) + i2 + 3], 16);
        }
      }
      /* VERTICAL FILTERING STARTS */
	  /* It is the same as the previously done horizontal filtering.
	      Now rpb points to the first pixel BELOW the boundary and
		  lpb points to the first pixel ABOVE the boundary. */
      /*
       * Calculate pixel locations in the 16x16 auxiliary buffer and in the 
       * original frame, pixel with higher y - coordinate is referred as 
       * right pixel.
       */
      rpb = mbHalf + offSetX;
      lpb = rpb - mbSizeX;
      rpf = corner + stepY * picSizeX + offSetX;
      lpf = rpf - picSizeX;
           
      /*
       * Go through vertical block boundaries
       */
      for (i2 = offSetX; i2 < mbSizeX; i2++, rpb++, lpb++, rpf++, lpf++)
      {   
        /*
         * Get search support truncations for both sides of the boundary
         * according to region types
         */
        rightTrunc = 6;
        leftTrunc = 6;

        /*
         * Update search support truncations for both sides
         * of the boundary according to pixel value difference at the
         * block boundary
         */
        overallActivity = filFindOverallAct(orig, rpf, -picSizeX, qp);
        rightTrunc = overallActivity;
        leftTrunc = overallActivity;
		    /* If intra flag is not set, make truncations on the 
		     * left and right activity parameters according to region
         * mode and pecmode data.
         */
        if(useLoopFilter){
          if (!intra_flag)
          {
            fil_overall_activity_truncator(0, i2, regmodes, pecmodes, &rightTrunc, &leftTrunc);
          }
        }

        /*
         * Find activities for both sides of the boundary
         * according to pixel value differences within blocks
         */
        filFindFinalAct(rightActivity, orig, rpf, picSizeX, 
                        rightTrunc, qp);
        filFindFinalAct(leftActivity, orig, lpf, -picSizeX, 
                        leftTrunc, qp);
      
        /*
         * Check if filtering is needed
         */
        if(rightActivity > 1 && leftActivity > 1)
        {
          /*
           * Filter lower side of the boundary with proper filter
		   * The filter is averaging filter of window size selected according to the table above.
		   * The filtering is done using Integer Arithmetic only!!!!!
           */
          if(rightActivity > 5)
          {
            if(leftActivity > 2)
            {
              tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] +
                       aux[rpb + wb] + aux[rpb + wb2] + aux[rpb + wb3];
              result[rpf] = toUchar(decPrec(aux[rpb - wb3] + tmpInt, 10) *
			                                inv7, 22);

              tmpInt += aux[rpb + wb4];
              result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

              tmpInt += (aux[rpb + wb5] - aux[rpb - wb2]);
              result[rpf + wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
            }
            else
            {
              tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] + 
                       aux[rpb + wb] + aux[rpb + wb2];
              result[rpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

              tmpInt += (aux[rpb + wb3] + aux[rpb + wb4]);
              result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

              tmpInt += (aux[rpb + wb5] - aux[rpb - wb2]);
              result[rpf + wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
            }
          }
          else if(rightActivity == 5)
          {
            tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] + 
                     aux[rpb + wb] + aux[rpb + wb2];
            result[rpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

            tmpInt += (aux[rpb + wb3] - aux[rpb - wb2]);
            result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);
            
            result[rpf + wf2] = toUchar(((int32)result[rpf] +
                                        result[rpf + wf] + result[rpf + wf2] +
                                        result[rpf + wf3] + 
                                        result[rpf + wf4]) * inv5, 16);
          }
          else if(rightActivity > 3)
          { 
            tmpInt = aux[rpb - wb] + aux[rpb] +
                     aux[rpb + wb] + aux[rpb + wb2];
            result[rpf] = toUchar(decPrec(aux[rpb - wb2] + tmpInt, 10) *
			                      inv5, 22);

            result[rpf + wf] = toUchar(decPrec(tmpInt + aux[rpb + wb3], 10) *
			                           inv5, 22);
          }
          else if(rightActivity == 3)
          {
            result[rpf] = toUchar(decPrec(aux[rpb - wb] + aux[rpb] + 
                                          aux[rpb + wb], 10) * inv3, 22);

            result[rpf + wf] = toUchar(((int32)result[rpf] + result[rpf + wf] +
                                        result[rpf + wf2]) * inv3, 16);
          }
          else
          {  
            result[rpf] = toUchar(decPrec(aux[rpb - wb] + aux[rpb] +
                                          aux[rpb + wb], 10) * inv3, 22);
          }
          /*
           * Filter upper side of the boundary with proper filter
		   * The filter is averaging filter of window size selected according to the table above.
		   * The filtering is done using Integer Arithmetic only!!!!!
           */
          if(leftActivity > 5) 
          {
            if(rightActivity > 2)
            {
              tmpInt = aux[lpb - wb3] + aux[lpb - wb2] + aux[lpb - wb] +
                       aux[lpb] + aux[lpb + wb] + aux[lpb + wb2];
              result[lpf] = toUchar(decPrec(tmpInt + aux[lpb + wb3], 10) *
			                        inv7, 22);

              tmpInt += aux[lpb - wb4];
              result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

              tmpInt += (aux[lpb - wb5] - aux[lpb + wb2]);
              result[lpf - wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
            }
            else
            {
              tmpInt = aux[lpb - wb2] + aux[lpb - wb] + aux[lpb] +
                       aux[lpb + wb] + aux[lpb + wb2];
              result[lpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

              tmpInt += (aux[lpb - wb4] + aux[lpb - wb3]);
              result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

              tmpInt += (aux[lpb - wb5] - aux[lpb + wb2]);
              result[lpf - wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
            }
          }
          else if(leftActivity == 5)
          {
            tmpInt = aux[lpb - wb2] + aux[lpb - wb] + aux[lpb] + 
                     aux[lpb + wb] + aux[lpb + wb2];
            result[lpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

            tmpInt += (aux[lpb - wb3] - aux[lpb + wb2]);
            result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

            result[lpf - wf2] = toUchar(((int32)result[lpf - wf4] +
                                         result[lpf - wf3] +
										 result[lpf - wf2] +
                                         result[lpf - wf] + 
                                         result[lpf]) * inv5, 16);
          }
          else if(leftActivity > 3)
          {
            tmpInt = aux[lpb - wb2] + aux[lpb - wb] +
                     aux[lpb] + aux[lpb + wb];
            result[lpf] = toUchar(decPrec(tmpInt + aux[lpb + wb2], 10) *
			                      inv5, 22);

            tmpInt += aux[lpb - wb3];
            result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);
          }
          else if(leftActivity == 3)
          {
            result[lpf] = toUchar(decPrec(aux[lpb - wb] + aux[lpb] + 
                                          aux[lpb + wb], 10) * inv3, 22);

            result[lpf - wf] = toUchar(((int32)result[lpf - wf2] +
			                            result[lpf - wf] + 
                                        result[lpf]) * inv3, 16);
          }
          else
          {
            result[lpf] = toUchar(decPrec(aux[lpb - wb] + aux[lpb] +
                                          aux[lpb + wb], 10) * inv3, 22);
          }
        }
      }
    }
  }
}

/*
 *
 * filCompFiltIntra:
 *
 * Parameters:
 *      orig      I: Original frame
 *      picSizeX  I: Luminance frame width
 *      picSizeY  I: Luminance frame height
 *      qp        I: Quantization parameter
 *      upLeftX,
 *      upLeftY   I: Corner coordinates of the 8x8 block to be filtered
 *      result    O: Result of filtering
 *
 * Function:
 *      Performs loop filtering for one 8x8 block inside the frame
 *      using average filter with varying filter support.
 *
 * Returns:
 *      Nothing
 */

void filCompFiltIntra(u_char *orig, int32 picSizeX, int32 picSizeY,
                      int32 qp, u_char *result,
                      int32 upLeftX, int32 upLeftY)
{
  int32
    filterAtLeft, filterAtUp,
    i1, j1, i2, j2,               /* Loop counters */
    rpb, lpb, rpf, lpf,           /* Location counters for pixels at block */
                                  /* boundaries */
    rightActivity, leftActivity,  /* Local activities for both sides of the 
                                     boundary */
    rightTrunc, leftTrunc,        /* Maximum search supports according to 
                                     region types */
    overallActivity,              /* Maximum search support according to 
                                     overall activity */
    tmpInt, tmpInt2,              /* Auxiliary variables */
    wf, wf2, wf3, wf4, wf5,       /* Frame wraping values for y direction */
    *aop, corner,
    aux[256],                     /* Auxiliary frame buffer */
    aux2[16];

  u_char  
    *cp1, *cp2, orig_aux[256];
    
  static int32
    inv3 = 65536/3,
    inv5 = 65536/5,
    inv7 = 65536/7,
    inv9 = 65536/9,
    inv25 = 65536/25;


  wf = picSizeX;
  wf2 = 2 * picSizeX;
  wf3 = 3 * picSizeX;
  wf4 = 4 * picSizeX;
  wf5 = 5 * picSizeX;
    

  filterAtLeft = (upLeftX == 0) ? 0 : 1;
  filterAtUp   = (upLeftY == 0) ? 0 : 1;

  if (filterAtLeft == 0)
  {
    for (j2 = upLeftY; j2 < upLeftY + stepY; j2++)
      for (i2 = upLeftX; i2 < upLeftX + stepX; i2++)
        result[j2 * picSizeX + i2] = orig[j2 * picSizeX + i2];

    if (filterAtUp == 0)
      return;
  }

  {
    /* 
     * First filter the
     * vertical 8x8 block boundaries and then the horizontal 8x8 block 
     * boundaries.
     */

      j1 = upLeftY - stepY;
      i1 = upLeftX - stepX;

      /*
       * Copy one macroblock of the original image to 16x16 auxiliary buffer.
       */
      
      corner = j1 * picSizeX + i1;

      if (filterAtUp)
      {
        cp2 = result + corner;
        cp1 = orig   + corner;

        for (j2 = 0; j2 < stepY; j2++, cp2 += picSizeX, cp1 += picSizeX)
        { 
          for (i2 = 0; i2 < stepX; i2 += 2)
          {
            aux[(j2 << mbShift) + stepX + i2] = (int32)cp2[stepX + i2] << 16;
            aux[(j2 << mbShift) + stepX + i2 + 1] = (int32)cp2[stepX + i2 + 1] << 16;
            orig_aux[(j2 << mbShift) + stepX + i2] = cp1[stepX + i2];
            orig_aux[(j2 << mbShift) + stepX + i2 + 1] = cp1[stepX + i2 + 1];
            orig_aux[((j2 + stepY) << mbShift) + stepX + i2] =
              cp1[stepY * picSizeX + stepX + i2];
            orig_aux[((j2 + stepY) << mbShift) + stepX + i2 + 1] =
              cp1[stepY * picSizeX + stepX + i2 + 1];
          }
        }
      }

      cp1 = result + corner + stepY * picSizeX;
      cp2 = orig + corner + stepY * picSizeX;       
      for (j2 = stepY; j2 < mbSizeX; j2++, cp1 += picSizeX, cp2 += picSizeX)
      {
        for (i2 = 0; i2 < stepX; i2 += 2)
        {
          if (filterAtLeft)
          {
            aux[(j2 << mbShift) + i2] = (int32)cp1[i2] << 16;
            aux[(j2 << mbShift) + i2 + 1] = (int32)cp1[i2 + 1] << 16;
          }
          aux[(j2 << mbShift) + stepX + i2] = (int32)cp2[stepX + i2] << 16;
          aux[(j2 << mbShift) + stepX + i2 + 1] = (int32)cp2[stepX + i2 + 1] << 16;
        }
      }

      if (filterAtLeft)
      {
        /*
         * Calculate pixel locations in the auxiliary 16x16 buffer, pixel with 
         * higher x - coordinate is referred as right pixel.
         */
        rpb = stepX + (stepY << mbShift);
        lpb = rpb - 1;

        /*
         * Scan the macroblock row by row.
         */
        cp1 = result + corner + stepY * picSizeX;
        cp2 = orig + corner + stepY * picSizeX;

        for(j2 = stepY; j2 < mbSizeY; j2++, rpb += mbSizeX, lpb += mbSizeX)
        {
          /*
           * Copy the original 8-bit pixel values of the row
           * to 32-bit integer buffer.
           */
          for(i2 = 0; i2 < stepX; i2 += 2)
          {
            aux2[i2]     = cp1[i2];
            aux2[i2 + 1] = cp1[i2 + 1];
            aux2[stepX + i2]     = cp2[stepX + i2];
            aux2[stepX + i2 + 1] = cp2[stepX + i2 + 1];
          }
       
          cp1 += picSizeX;
          cp2 += picSizeX;

          aop = aux2 + stepX;
          
          /*
           * Get search support truncations for both sides of the boundary
           * according to region types.
           */
          rightTrunc = 6;
          leftTrunc = 6;
          
          /*
           * Update search support truncations for both sides
           * of the boundary according to pixel value difference at the
           * block boundary
           */
          overallActivity = filFindOverallAct(aop, 0, -1, qp);
          rightTrunc = overallActivity;
          leftTrunc = overallActivity;
                /* if intra flag is not set, make truncations on the 
                   left and right activity parameters according to region mode and pecmode data. */
          /*
           * Find activities for both sides of the boundary
           * according to pixel value differences within blocks
         */
          filFindFinalAct(rightActivity, aop, 0, 1, rightTrunc, qp);
          filFindFinalAct(leftActivity, aop, -1, -1, leftTrunc, qp);
          
          /*
           * Check if filtering is needed
           */
          if(rightActivity > 1 && leftActivity > 1){
          
            /*
             * Filter right side of the boundary with proper filter
             */
            if(rightActivity > 5)
            {
              if(leftActivity > 2)
              {
                tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2] + aop[3];
                aux[rpb] = (aop[-3] + tmpInt) * inv7;
                aux[rpb + 1] = (tmpInt + aop[4]) * inv7;
                aux[rpb + 2] = aux[rpb + 1] + (aop[5] - aop[-2]) * inv7;
              }
              else
              {
                tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
                aux[rpb] = tmpInt * inv5;

                aux[rpb + 1] = (tmpInt + aop[3] + aop[4]) * inv7;
              
                aux[rpb + 2] = aux[rpb + 1] + (aop[5] - aop[-2]) * inv7;
              }
            }
            else if(rightActivity == 5)
            {
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
              aux[rpb] = tmpInt * inv5;
              tmpInt2 = tmpInt + aop[3] - aop[-2];
              aux[rpb + 1] = tmpInt2 * inv5;
              aux[rpb + 2] = (tmpInt + tmpInt2) * inv25 +
                                       (aop[2] + aop[3] + aop[4]) * inv5;
            }
            else if(rightActivity > 3)
            {
              tmpInt = aop[-1] + aop[0] + aop[1] + aop[2];
              aux[rpb] = (aop[-2] + tmpInt) * inv5;
              aux[rpb + 1] = (tmpInt + aop[3]) * inv5;
            }
            else if(rightActivity == 3)
            {
              tmpInt = aop[-1] + aop[0] + aop[1];
              aux[rpb] = tmpInt * inv3;
              aux[rpb + 1] = tmpInt * inv9 + (aop[1] + aop[2]) * inv3;
            }
            else
            {
              aux[rpb] = (aop[-1] + aop[0] + aop[1]) * inv3;
            }
            /*
             * Filter left side of the boundary with proper filter
             */

            aop--;

            if(leftActivity > 5)
            {
              if(rightActivity > 2)
              {
                tmpInt = aop[-3] + aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
                aux[lpb] = (tmpInt + aop[3]) * inv7;
                aux[lpb - 1] = (aop[-4] + tmpInt) * inv7;
                aux[lpb - 2] = aux[lpb - 1] + (aop[-5] - aop[2]) * inv7;
              }
              else
              {
                tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
                aux[lpb] = tmpInt * inv5;
                aux[lpb - 1] = (tmpInt + aop[-3] + aop[-4]) * inv7;
                aux[lpb - 2] = aux[lpb - 1] + (aop[-5] - aop[2]) * inv7;
              }
            }
            else if (leftActivity == 5)
            {
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1] + aop[2];
              aux[lpb] = tmpInt * inv5;
              tmpInt2 = tmpInt + aop[-3] - aop[2];
              aux[lpb - 1] = tmpInt2 * inv5;
              aux[lpb - 2] = (tmpInt  + tmpInt2) * inv25 +
                             (aop[-4] + aop[-3] + aop[-2]) * inv5;
            }
            else if(leftActivity > 3)
            {
              tmpInt = aop[-2] + aop[-1] + aop[0] + aop[1];
              aux[lpb] = (tmpInt + aop[2]) * inv5; 
              aux[lpb - 1] = (aop[-3] + tmpInt) * inv5;
            }
            else if(leftActivity==3)
            {
              tmpInt = aop[-1] + aop[0] + aop[1];
              aux[lpb] = tmpInt * inv3;
              aux[lpb - 1] = tmpInt * inv9 + (aop[-2] + aop[-1]) * inv3;
            }
            else
            {
              aux[lpb] = (aop[-1] + aop[0] + aop[1]) * inv3;
            }
          }
        }

        /*
         * Copy the macroblock after horizontal filtering to the original
         * image.
         */
        cp1 = result + corner + stepY * picSizeX;

        for (j2 = stepY; j2 < mbSizeY; j2++, cp1 += picSizeX)
        { 
          for (i2 = 0; i2 < mbSizeX; i2 += 4)
          {
             cp1[i2]     = toUchar(aux[(j2 << mbShift) + i2], 16);
             cp1[i2 + 1] = toUchar(aux[(j2 << mbShift) + i2 + 1], 16);
             cp1[i2 + 2] = toUchar(aux[(j2 << mbShift) + i2 + 2], 16);
             cp1[i2 + 3] = toUchar(aux[(j2 << mbShift) + i2 + 3], 16);
          }
        }
      }

      if (filterAtUp)
      {
        /*
         * Calculate pixel locations in the 16x16 auxiliary buffer and in the 
         * original frame, pixel with higher y - coordinate is referred as 
         * right pixel.
         */
        rpb = mbHalf + stepX;
        lpb = rpb - mbSizeX;
        rpf = corner + stepY * picSizeX + stepX;
        lpf = rpf - picSizeX;
           
        /*
         * Go through vertical block boundaries
         */
        for (i2 = stepX; i2 < mbSizeX; i2++, rpb++, lpb++, rpf++, lpf++)
        {   
          /*
           * Get search support truncations for both sides of the boundary
           * according to region types
           */
          rightTrunc = 6;
          leftTrunc = 6;

          /*
           * Update search support truncations for both sides
           * of the boundary according to pixel value difference at the
           * block boundary
           */
          overallActivity = filFindOverallAct(orig_aux, rpb, -mbSizeX, qp);
          rightTrunc = overallActivity;
          leftTrunc = overallActivity;
                /* if intra flag is not set, make truncations on the 
                   left and right activity parameters according to region mode and pecmode data. */
          /*
           * Find activities for both sides of the boundary
           * according to pixel value differences within blocks
           */
          filFindFinalAct(rightActivity, orig_aux, rpb, mbSizeX, 
                          rightTrunc, qp);
          filFindFinalAct(leftActivity, orig_aux, lpb, -mbSizeX, 
                          leftTrunc, qp);
      
          /*
           * Check if filtering is needed
           */
          if(rightActivity > 1 && leftActivity > 1)
          {
            /*
             * Filter right side of the boundary with proper filter
             */
            if(rightActivity > 5)
            {
              if(leftActivity > 2)
              {
                tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] +
                         aux[rpb + wb] + aux[rpb + wb2] + aux[rpb + wb3];
                result[rpf] = toUchar(decPrec(aux[rpb - wb3] + tmpInt, 10) *
                                                        inv7, 22);

                tmpInt += aux[rpb + wb4];
                result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

                tmpInt += (aux[rpb + wb5] - aux[rpb - wb2]);
                result[rpf + wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
              }
              else
              {
                tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] + 
                         aux[rpb + wb] + aux[rpb + wb2];
                result[rpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

                tmpInt += (aux[rpb + wb3] + aux[rpb + wb4]);
                result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

                tmpInt += (aux[rpb + wb5] - aux[rpb - wb2]);
                result[rpf + wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
              }
            }
            else if(rightActivity == 5)
            {
              tmpInt = aux[rpb - wb2] + aux[rpb - wb] + aux[rpb] + 
                       aux[rpb + wb] + aux[rpb + wb2];
              result[rpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

              tmpInt += (aux[rpb + wb3] - aux[rpb - wb2]);
              result[rpf + wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);
            
              result[rpf + wf2] = toUchar(((int32)result[rpf] +
                                          result[rpf + wf] + result[rpf + wf2] +
                                          result[rpf + wf3] + 
                                          result[rpf + wf4]) * inv5, 16);
            }
            else if(rightActivity > 3)
            { 
              tmpInt = aux[rpb - wb] + aux[rpb] +
                       aux[rpb + wb] + aux[rpb + wb2];
              result[rpf] = toUchar(decPrec(aux[rpb - wb2] + tmpInt, 10) *
                                              inv5, 22);

              result[rpf + wf] = toUchar(decPrec(tmpInt + aux[rpb + wb3], 10) *
                                                   inv5, 22);
            }
            else if(rightActivity == 3)
            {
              result[rpf] = toUchar(decPrec(aux[rpb - wb] + aux[rpb] + 
                                            aux[rpb + wb], 10) * inv3, 22);

              result[rpf + wf] = toUchar(((int32)result[rpf] + result[rpf + wf] +
                                          result[rpf + wf2]) * inv3, 16);
            }
            else
            {  
              result[rpf] = toUchar(decPrec(aux[rpb - wb] + aux[rpb] +
                                          aux[rpb + wb], 10) * inv3, 22);
            }
            /*
             * Filter left side of the boundary with proper filter
             */
            if(leftActivity > 5) 
            {
              if(rightActivity > 2)
              {
                tmpInt = aux[lpb - wb3] + aux[lpb - wb2] + aux[lpb - wb] +
                         aux[lpb] + aux[lpb + wb] + aux[lpb + wb2];
                result[lpf] = toUchar(decPrec(tmpInt + aux[lpb + wb3], 10) *
                                                inv7, 22);

                tmpInt += aux[lpb - wb4];
                result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

                tmpInt += (aux[lpb - wb5] - aux[lpb + wb2]);
                result[lpf - wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
              }
              else
              {
                tmpInt = aux[lpb - wb2] + aux[lpb - wb] + aux[lpb] +
                         aux[lpb + wb] + aux[lpb + wb2];
                result[lpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

                tmpInt += (aux[lpb - wb4] + aux[lpb - wb3]);
                result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv7, 22);

                tmpInt += (aux[lpb - wb5] - aux[lpb + wb2]);
                result[lpf - wf2] = toUchar(decPrec(tmpInt, 10) * inv7, 22);
              }
            }
            else if(leftActivity == 5)
            {
              tmpInt = aux[lpb - wb2] + aux[lpb - wb] + aux[lpb] + 
                       aux[lpb + wb] + aux[lpb + wb2];
              result[lpf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

              tmpInt += (aux[lpb - wb3] - aux[lpb + wb2]);
              result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);

              result[lpf - wf2] = toUchar(((int32)result[lpf - wf4] +
                                         result[lpf - wf3] +
                                         result[lpf - wf2] +
                                         result[lpf - wf] + 
                                         result[lpf]) * inv5, 16);
            }
            else if(leftActivity > 3)
            {
              tmpInt = aux[lpb - wb2] + aux[lpb - wb] +
                       aux[lpb] + aux[lpb + wb];
              result[lpf] = toUchar(decPrec(tmpInt + aux[lpb + wb2], 10) *
                                              inv5, 22);

              tmpInt += aux[lpb - wb3];
              result[lpf - wf] = toUchar(decPrec(tmpInt, 10) * inv5, 22);
            }
            else if(leftActivity == 3)
            {
              result[lpf] = toUchar(decPrec(aux[lpb - wb] + aux[lpb] + 
                                            aux[lpb + wb], 10) * inv3, 22);

              result[lpf - wf] = toUchar(((int32)result[lpf - wf2] +
                                                 result[lpf - wf] + 
                                        result[lpf]) * inv3, 16);
            }
            else
            {
              result[lpf] = toUchar(decPrec(aux[lpb - wb] + aux[lpb] +
                                            aux[lpb + wb], 10) * inv3, 22);
            }
          }
        }
      }
  }
}


#undef stepX
#undef stepY



/*
 *
 * filFindOverallAct:
 *
 * Parameters:
 *      image     I: Luminance frame
 *      location  I: Pixel location
 *      step      I: Step to adjacent pixel
 *      qp        I: Quantization parameter
 *
 * Function:
 *      Finds search support corresponding to overall activity, defined by
 *      pixel value difference at block boundary and quantization parameter.
 *
 * Returns:
 *      Search support according to overall activity
 */
#ifdef filFindOverallActFunc
static int32 filFindOverallAct(u_char *image, int32 location, int32 step, 
                               int32 qp)
{
  int32
    pelVal1, pelVal2, /* Pixel values */
    diff;             /* Pixel value difference */

  pelVal1 = (int32)image[location];
  pelVal2 = (int32)image[location + step];
  diff = filDiff(pelVal1,pelVal2);

  return(filOverallActTable[qp][diff]);
}
#endif


/*
 *
 * filFindFinalAct:
 *
 * Parameters:
 *      image     I: Luminance frame
 *      location  I: Pixel location
 *      step      I: Step to adjacent pixel
 *      max       I: Maximum allowed support size
 *      qp        I: Quantization parameter
 *
 * Function:
 *      Finds local activity, defined by pixel value difference within block 
 *      and quantization parameter value.
 *
 * Returns:
 *      Local activity
 */

static int32 filFindFinalActFunc(u_char *image, int32 location, int32 step, 
                                 int32 max, int32 qp)
{
  int32
    i,                /* Loop counter */
    pelVal1, pelVal2, /* Pixel values */
    diff;             /* Pixel value difference */

  pelVal1 = (int32)image[location];

  for(i = 1; i < max; i++){
    pelVal2 = (int32)image[location + i * step];
    diff = filDiff(pelVal1,pelVal2);
    if(diff > filDiffActTable[qp][i])
      return(i);
  }
  return(i);
}

/*
 *
 * filInitDiffActTable:
 *
 * Parameters:
 *      Nothing
 *
 * Function:
 *      Initializes table for local activity parameter, defined by pixel 
 *      value difference within block and quantization parameter value.
 *
 * Returns:
 *      Nothing
 */

void filInitDiffActTable()
{
  int32
    i, j, lim32;

  for(i = 1; i < 32; i++){
	lim32 =  LogTable[i-1] << 2; /*4.0 * log(i) */
    for(j = 1; j < 9; j++)
	  filDiffActTable[i][j] = (u_char)((lim32 / j) >> 16); 
  }
}


/*
 *
 * filInitOverallActTable:
 *
 * Parameters:
 *      Nothing
 *
 * Function:
 *      Initializes table for maximum search support corresponding to
 *      overall activity, defined by pixel value difference at the block
 *      boundary and quantization parameter value.
 *
 * Returns:
 *      Nothing
 */

void filInitOverallActTable(void)
{
  int32
    i, j, lim32;

  for(i = 1; i < 32; i++){
	lim32 = i * LogTable[i-1]; /* i * log(i)*/
    for(j = 0; j < 256; j++)
    {
      if((4 * j * 65536) < lim32) /*(j < 0.25 * lim)*/
        filOverallActTable[i][j] = 6;
      else if((5 * j * 65536) < (2 * lim32))   /*(j < 0.4 * lim)*/
        filOverallActTable[i][j] = 5;
      else if((50 * j * 65536) < (33 * lim32))  /*(j < 0.66 * lim)*/
        filOverallActTable[i][j] = 4;
      else if((j * 65536) < lim32) /*(j < lim)*/
        filOverallActTable[i][j] = 3;
      else if(( 2 * j * 65536) < (3 * lim32)) /*(j < 1.5 * lim)*/
        filOverallActTable[i][j] = 2;
      else if((j * 65536) < (2 * lim32))  /*(j < 2.0 * lim)*/
        filOverallActTable[i][j] = 1;
      else
        filOverallActTable[i][j] = 0;
	}
  }
}


/*
 *
 * fil_overall_activity_truncator()
 *
 * Parameters:
 *      index       : the line being processed 
 *      regmodes[4] : contanes the regionmodes of the neighboring blocks
 *      pecmodes[4] : containes the pecmodes of the neighboring blocks
 *      rightTrunc : overall activity truncation value for the rightside of the block boundary
 *      leftTrunc  : overall activity truncation value for the leftside of the block boundary
 *      direction_flag : indicates the direction of filtering. 1 : horizontal 0 : vertical filtering
 *
 * Function:
 *      Truncates the overall activity parameter according to the regionMode and pecMode info
 *       of the neighboring quadrants. The truncation is done to the previous truncation 
 *		values which were already in rightTrunc and leftTrunc
 *.
 *      CAUTION     : macroblock size is assumed to be 16!!!
 *
 *
 * Returns:
 *      rightTrunc : overall activity truncation value for the rightside of the block boundary
 *      leftTrunc  : overall activity truncation value for the leftside of the block boundary
 */

void fil_overall_activity_truncator(int direction_flag, int index, int *regmodes, int *pecmodes, int *rightTrunc, int *leftTrunc)

{
  int mbsize = 16, r_regmode, l_regmode, r_pecmode, l_pecmode;

  if (direction_flag == 1)  
  {                            /* if it is Horizontal Filtering */ 
    if (index < (mbsize >> 1))  
    {                          /* compare regmodes[1] and [2] and pecmodes[1] and [2] */
        l_regmode = regmodes[1];
        r_regmode = regmodes[2];         
        l_pecmode = pecmodes[1];
        r_pecmode = pecmodes[2];         
    }
    else
    {                          /* compare regmodes[3] and [4] and pecmodes[3] and [4] */ 
        l_regmode = regmodes[3];
        r_regmode = regmodes[4];            
        l_pecmode = pecmodes[3];
        r_pecmode = pecmodes[4];         
    }
  }
  else                          /* if it is Vertical Filtering */ 
  {
    if (index < (mbsize >> 1))  
    {                          /* compare regmodes[1] and [3] and pecmodes[1] and [3] */
        l_regmode = regmodes[1];
        r_regmode = regmodes[3];         
        l_pecmode = pecmodes[1];
        r_pecmode = pecmodes[3];         
    }
    else
    {                          /* compare regmodes[2] and [4] and pecmodes[2] and [4] */ 
        l_regmode = regmodes[2];
        r_regmode = regmodes[4];            
        l_pecmode = pecmodes[2];
        r_pecmode = pecmodes[4];         
    }
  }
  if (l_regmode == REGMODE_INTRA)  
  {  /* if the two neighboring quadrants are both INTRA */
    if (r_regmode == REGMODE_INTRA)
    {
      rightTrunc[0] = leftTrunc[0]; /* No truncation for INTRA regions!! */
    }
    else
    if ((r_regmode == REGMODE_INTER_COPY)&&(r_pecmode == PECMODE_SKIPPED))
    { /* if the neighboring quadrants are INTRA and COPY with no PEC coding */
      rightTrunc[0] = leftTrunc[0] = ChooseSmall(rightTrunc[0],2);   /* truncate both to at most 2 pixels*/
    }     
    else
    if (((r_regmode == REGMODE_INTER_COPY)||(r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_CODED))
    {  /* if the neighboring quadrants are INTRA and COPY */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],4);      /* truncate right border to at most 4 pixels */
      leftTrunc[0] =  ChooseSmall(leftTrunc[0],6);        /* truncate left border to at most 6 pixels */
    }
    else
    if (((r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_SKIPPED))
    {  /* if the neighboring quadrants are INTRA and PEC not coded */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],2);      /* truncate right border to at most 2 pixels */
      leftTrunc[0] =  ChooseSmall(leftTrunc[0],6);        /* truncate left bprder to at most 6 pixels */
    }
  }       
  else  
  if ((l_regmode == REGMODE_INTER_COPY)&&(l_pecmode == PECMODE_SKIPPED))
  {
    if (r_regmode == REGMODE_INTRA)
    {  /* if the neighboring quadrants are COPY and INTRA */
      rightTrunc[0] = leftTrunc[0] = ChooseSmall(leftTrunc[0],2); /* truncate both to at most 2 pixels */
    }
    else
    if ((r_regmode == REGMODE_INTER_COPY)&&(r_pecmode == PECMODE_SKIPPED))
    { /*  if the neighboring quadrants are COPY and COPY */
       rightTrunc[0] = leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate both to at most 2 pixels */
    }
    else
    if (((r_regmode == REGMODE_INTER_COPY)||(r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_CODED))
    {  /* if the neighboring quadrants are COPY and PEC coded */
      leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate left to at most 2 pixels */
	  rightTrunc[0] = ChooseSmall(rightTrunc[0],4); /* truncate right to at most 4 pixels */
	}
    else
    if (((r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_SKIPPED))
    { /* if the neighboring quadrants are COPY and PEC not coded */
      leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate left border to at most 2 pixels */
	  rightTrunc[0] = ChooseSmall(rightTrunc[0],2); /* truncate right to at most 2 pixels */
	}
  }
  else
  if (((l_regmode == REGMODE_INTER_COPY)||(l_regmode == REGMODE_INTER_NOPRED)||(l_regmode == REGMODE_INTER_PRED)|| 
    (l_regmode == REGMODE_INTER_PRED_DELTA))&&(l_pecmode == PECMODE_CODED))
  { 
    if (r_regmode == REGMODE_INTRA)
    { /* if the neighboring quadrantsa are PEC Coded and INTRA */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],6);    /* truncate right to at most 6 pixels */
      leftTrunc[0] = ChooseSmall(leftTrunc[0],4);      /* truncate left to at most 4 pixels */
    }
    else
    if ((r_regmode == REGMODE_INTER_COPY)&&(r_pecmode == PECMODE_SKIPPED))
    {  /* if the neighboring quadrantsa are PEC Coded and COPY */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate right to at most 2 pixels*/
	  leftTrunc[0] = ChooseSmall(leftTrunc[0],4);	   /* truncate left to at most 4 pixels */ 
	}
    else
    if (((r_regmode == REGMODE_INTER_COPY)||(r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_CODED))
    {  /* if the neighboring quadrantsa are PEC Coded and PEC Coded */ 
      rightTrunc[0] = leftTrunc[0] = ChooseSmall(rightTrunc[0],4);  /* truncate both to at most 4 pixels*/
    }
    else
    if (((r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_SKIPPED))
    {  /* if the neighboring quadrantsa are PEC Coded and PEC not Coded */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate right to at most 2 pixels*/
	  leftTrunc[0] = ChooseSmall(rightTrunc[0],4);   /* truncate left to at most 4 pixels */
    }   
  }
  else
  if (((r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
      (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_SKIPPED))
  {
    if (r_regmode == REGMODE_INTRA)
    { /* if the neighboring quadrants are PEC Not Coded and INTRA */
      rightTrunc[0] = ChooseSmall(rightTrunc[0],6);    /* truncate right to at most 6 pixels*/
      leftTrunc[0] = ChooseSmall(leftTrunc[0],2);       /* truncate left to at most 2 pixels */
    }
    else
    if ((r_regmode == REGMODE_INTER_COPY)&&(r_pecmode == PECMODE_SKIPPED))
    { /* if the neighboring quadrants are PEC Not Coded and COPY */
      rightTrunc[0] = leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate both to at most 2 pixels */
	}
    else
    if (((r_regmode == REGMODE_INTER_COPY)||(r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_CODED))
    { /* if the neighboring quadrants are PEC Not Coded and PEC Coded */
      leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate left to at most 2 pixels*/
	  rightTrunc[0] = ChooseSmall(rightTrunc[0],4); /* truncate right to at most 4 pixels */
    }
    else
    if (((r_regmode == REGMODE_INTER_NOPRED)||(r_regmode == REGMODE_INTER_PRED)|| 
        (r_regmode == REGMODE_INTER_PRED_DELTA))&&(r_pecmode == PECMODE_SKIPPED))
    { /* if the neighboring quadrants are PEC Not Coded and PEC Not Coded */
      rightTrunc[0] = leftTrunc[0] = ChooseSmall(rightTrunc[0],2);  /* truncate both to at most 2 pixels */
    }   
  }
}

/*
 *
 * fil_form_region_and_pec_info()
 *
 * Parameters:
 *      i1          : x coordinate parameter for the current macroblock (corressponds to the left upper corner coordinate)
 *      j1          : y coordinate parameter for the current macroblock (corressponds to the left upper corner coordinate)
 *      mbsize      : macroblock size parameter 
 *      picsizex    : picture size x
 *      picsizey    : picture size y
 *      object_list1 : macroblock information list  of encoder. It is assigned to NULL in decoder filtering.
 *		object_list2 : macroblock information list  of decoder It is assigned to NULL in encoder filtering.
 *      intra_flag  : intra frame indication (if 1 => intra frame ( means no usage of object_list info))
 *      enco_flag   : encoder indication flag ( if 1 => encoder running , use object_list1 else object_list2)
 *      regmodes[4] : contains the regionmodes of the neighboring blocks
 *      pecmodes[4] : contains the pecmodes of the neighboring blocks
 * 
 *     NOTE : 
 *		  Splitted macroblocks are indexed as follows:
 *
 *      -------
 * 		 |  1 | 2 |
 *		 |-------|
 *		 |  3 | 4 |
 *      --------
 *  
 *
 *
 * Function:
 *      Creates the region type and prediction error coding type information of the quadrants 
 *		(8x8 ) inside the macroblock sized (16x16)  region being filtered.
 *      CAUTION : The fixed point operations ( like bit shifts ) assume that  Macroblock size is 16!!!
 *      
 *
 * Returns:
 *      regmodes[4] : contains the region mode information of the neighboring blocks
 *
 *      pecmodes[4] : contains the prediction error coding mode information of the neighboring blocks
 */


void fil_form_region_and_pec_info(int i1, int j1, int mbsize, int picsizex, int picsizey, 
											  object *object_list1, objectBuffer_t *object_list2, 
											  int intra_flag, int enco_flag, int *regmodes, int *pecmodes)
{
    int temp1, temp2,smode;
    int mbnumber1, mbnumber2, mbnumber3, mbnumber4, counter;

  if (!intra_flag)
  {               /* FILTERING INTER TYPE FRAME */
    temp1 = j1;
    temp2 = i1;
    while ( temp1 >= mbsize ) 
       temp1 = temp1 - mbsize;   /* temp1 = temp1 mod mbsize */
    while ( temp2 >= mbsize ) 
       temp2 = temp2 - mbsize;  /* temp2 = temp2 mod mbsize */

    if ((temp1 == 0) && (temp2 == 0))  
      smode = 0;   /* filtered 16x16 region's corner coordinates correspond 
							legal macroblock grid */
    else 
    if ((temp1 == 0) && (temp2 == (mbsize >> 1)))
      smode = 1;   /* filtered 16x16 region is a combination of 2 macroblocks, 
						    16x8 half comes from left macroblock, and 16x8 half from 
							right of the  coordinates (i1+8,j1)*/
    else 
    if ((temp1 == (mbsize >> 1)) && (temp2 == 0))
      smode = 2;   /* filtered 16x16 region is a combination of 2 macroblocks, 
							8x16 half comes from the upper macroblock,
							and 8x16 half from the lower of the coordinates (i1,j1+8) */
    else 
    if ((temp1 == (mbsize >> 1)) && (temp2 == (mbsize >> 1)))
      smode = 3;   /* filtered region is a combination of 4 macroblocks , 
							each quadrant (8x8 block) belongs to a different macroblock 
							where they all intersect at coordinate (i1+8, j1+8)*/

    if (enco_flag)
    {               /* FILTER IS USED IN ENCODER */
      /* The following line assumes that macroblock size is 16!!!         */
	  /* mbnumber1 gives the index number of the macroblock whose  */
	  /* upper left corner coordinates are (i1,j1) in object_list1          */
	    mbnumber1 = ((picsizex >> 4) * (j1 >> 4) + (i1 >> 4)) * 5; 
        switch(smode)
        {
          case 0 :	/* filtered 16x16 region's corner coordinates correspond legal macroblock grid */
          {
            if (object_list1[mbnumber1].regionMode != REGMODE_SPLITTED) /* if the macroblock is not a splitted macroblock*/
            { /* all the region modes and PEC modes are the same for all quadrants 
				   in the 16x16 region selected for filtering */
               regmodes[1] = regmodes[2] = regmodes[3] = regmodes[4] = object_list1[mbnumber1].regionMode;
               pecmodes[1] = pecmodes[2] = pecmodes[3] = pecmodes[4] = object_list1[mbnumber1].pecMode;        
            }
            else /* if the macroblock is a splitted mmacroblock */
            for (counter = 1; counter < 5 ; counter ++) 
            { /* The region modes and PEC modes of each quadrant is different and have different indexes  */
			  /*  like mbnumber1 + 1, 2, 3, 4 in the object_list1*/  
              regmodes[counter] = object_list1[mbnumber1+counter].regionMode;
              pecmodes[counter] = object_list1[mbnumber1+counter].pecMode;
            }
			break;
		  }  

          case 1 : /* filtered 16x16 region is a combination of 2 macroblocks, */		  
          {	/*16x8 half comes from left macroblock, and 16x8 half from right of the  coordinates (i1+8,j1) */
            mbnumber2 =  mbnumber1 + 5; 
			/* mbnumber1 corresponds to the macroblock index of the left macroblock and mbnumber2 corresponds 
				to the right macroblock */
            if (object_list1[mbnumber1].regionMode != REGMODE_SPLITTED) /* if the left macroblock is not splitted */
            { /* all the region modes and PEC modes of the quadrants of the left half are the same */
              regmodes[1] = regmodes[3] = object_list1[mbnumber1].regionMode;
              pecmodes[1] = pecmodes[3] = object_list1[mbnumber1].pecMode; 
            }
            else /* if the left macroblock is splitted, */
            {	  /* the left half of the filtered region contains the quadrants 2 and 4 of this macroblock*/
              regmodes[1] = object_list1[mbnumber1 + 2].regionMode;
              pecmodes[1] = object_list1[mbnumber1 + 2].pecMode;
              regmodes[3] = object_list1[mbnumber1 + 4].regionMode;
              pecmodes[3] = object_list1[mbnumber1 + 4].pecMode;
            }
            if (object_list1[mbnumber2].regionMode != REGMODE_SPLITTED) /* if the right macroblock is not splitted */
            { /* all the region modes and PEC modes of the quadrants of the right half are the same */
              regmodes[2] = regmodes[4] = object_list1[mbnumber2].regionMode;
              pecmodes[2] = pecmodes[4] = object_list1[mbnumber2].pecMode; 
            }
            else /* if the right macroblock is splitted, */
            {     /* the right half of the filtered region contains the quadrants 1 and 3 of this macroblock*/
              regmodes[2] = object_list1[mbnumber2 + 1].regionMode;
              pecmodes[2] = object_list1[mbnumber2 + 1].pecMode;
              regmodes[4] = object_list1[mbnumber2 + 3].regionMode;
              pecmodes[4] = object_list1[mbnumber2 + 3].pecMode;
            }
            break;
		  }

          case 2 : /* filtered 16x16 region is a combination of 2 macroblocks, */
			{ 		 /*8x16 half comes from the upper macroblock, and 8x16 half 
						 from the lower of the coordinates (i1,j1+8) */   
            mbnumber2 =  mbnumber1 + (picsizex >> 4) * 5;
			/* mbnumber1 corresponds to the macroblock index of the upper 
				macroblock and mbnumber2 corresponds to the lower macroblock */
            if (object_list1[mbnumber1].regionMode != REGMODE_SPLITTED) /* if the upper macroblock is not splitted */
            {  /* all the region modes and PEC modes of the upper 2 quadrants are the same */
              regmodes[1] = regmodes[2] = object_list1[mbnumber1].regionMode;
              pecmodes[1] = pecmodes[2] = object_list1[mbnumber1].pecMode; 
            }
            else /* if the upper macroblock is splitted */
            {     /* the upper half of the filtered region contains the quadrants 3 and 4 of this macroblock */
              regmodes[1] = object_list1[mbnumber1 + 3].regionMode;
              pecmodes[1] = object_list1[mbnumber1 + 3].pecMode;
              regmodes[2] = object_list1[mbnumber1 + 4].regionMode;
              pecmodes[2] = object_list1[mbnumber1 + 4].pecMode;
            }
            if (object_list1[mbnumber2].regionMode != REGMODE_SPLITTED) /* if the lower macroblock is not splitted */
            {  /* all the region modes and PEC modes of the lower 2 quadrants are the same */
              regmodes[3] = regmodes[4] = object_list1[mbnumber2].regionMode;
              pecmodes[3] = pecmodes[4] = object_list1[mbnumber2].pecMode; 
            }
            else /* if the lower macroblock is splitted */
            {    /* the lower half of the filtered region contains the quadrants 1 and 2 of this macroblock */
              regmodes[3] = object_list1[mbnumber2 + 1].regionMode;
              pecmodes[3] = object_list1[mbnumber2 + 1].pecMode;
              regmodes[4] = object_list1[mbnumber2 + 2].regionMode;
              pecmodes[4] = object_list1[mbnumber2 + 2].pecMode;
            }
			break;
          }
          
          case 3 :  /* filtered region is a combination of 4 macroblocks , */
          {	/*	each quadrant (8x8 block) belongs to a different macroblock 
				  where they all intersect at coordinate (i1+8, j1+8)*/
            mbnumber2 = mbnumber1 + 5;  
			/* mbnumber2 is the index of the macroblock which is right-adjacent 
				to macroblock with index mbnumber1 */
            mbnumber3 = mbnumber1 + (picsizex >> 4)*5; 
			/* mbnumber3 is the index of the macroblock which is bottom-adjacent 
				to macroblok with index mbnumber1 */
            mbnumber4 = mbnumber3 + 5;
			/* mbnumber4 is the index of the macroblock which is right-adjacebt to 
				macroblock with index mbnumber3 */
            if (object_list1[mbnumber1].regionMode != REGMODE_SPLITTED) /* if macroblock number mbnumber1 is not splitted */
            { /* the quadrant number 1 of the filtered 16x16 region gets the region mode 
				   and PEC mode from macroblock number mbnumber1 */
              regmodes[1] = object_list1[mbnumber1].regionMode;
              pecmodes[1] = object_list1[mbnumber1].pecMode;
            }
            else /* if macroblock number mbnumber1 is splitted */
            { /* the quadrant number 1 of the filtered 16x16 region gets the region mode 
				   and PEC mode of macroblock number mbnumber1's 4th quadrant */
              regmodes[1] = object_list1[mbnumber1+4].regionMode;
              pecmodes[1] = object_list1[mbnumber1+4].pecMode;
            }
            if (object_list1[mbnumber2].regionMode != REGMODE_SPLITTED) /* if macroblock number mbnumber2 is not splitted */
            {/* the quadrant number 2 of the filtered 16x16 region gets the region mode 
			      and PEC mode from macroblock number mbnumber2 */
              regmodes[2] = object_list1[mbnumber2].regionMode;
              pecmodes[2] = object_list1[mbnumber2].pecMode;
            }
            else /* if macroblock number mbnumber2 is splitted */
            {  /* the quadrant number 2 of the filtered 16x16 region gets the region mode 
			        and PEC mode of macroblock number mbnumber2's 3th quadrant */
              regmodes[2] = object_list1[mbnumber2+3].regionMode;
              pecmodes[2] = object_list1[mbnumber2+3].pecMode;
            }
            if (object_list1[mbnumber3].regionMode != REGMODE_SPLITTED) /* if macroblock number mbnumber3 is not splitted */
            { /* the quadrant number 3 of the filtered 16x16 region gets the region mode 
			       and PEC mode from macroblock number mbnumber3 */
              regmodes[3] = object_list1[mbnumber3].regionMode;
              pecmodes[3] = object_list1[mbnumber3].pecMode;
            }
            else /* if macroblock number mbnumber3 is splitted */
            { /* the quadrant number 3 of the filtered 16x16 region gets the region mode 
			       and PEC mode of macroblock number mbnumber3's 2nd quadrant */
              regmodes[3] = object_list1[mbnumber3+2].regionMode;
              pecmodes[3] = object_list1[mbnumber3+2].pecMode;
            }
            if (object_list1[mbnumber4].regionMode != REGMODE_SPLITTED)  /* if macroblock number mbnumber4 is not splitted */
            { /* the quadrant number 4 of the filtered 16x16 region gets the region mode 
			      and PEC mode from macroblock number mbnumber4 */
              regmodes[4] = object_list1[mbnumber4].regionMode;
              pecmodes[4] = object_list1[mbnumber4].pecMode;
            }
            else /* if macroblock number mbnumber4 is splitted */
            { /* the quadrant number 4 of the filtered 16x16 region gets the region mode 
			      and PEC mode of macroblock number mbnumber4's 1st quadrant */
              regmodes[4] = object_list1[mbnumber4+1].regionMode;
              pecmodes[4] = object_list1[mbnumber4+1].pecMode;
            }
            break;
		  }
        }
    }
    else  /* FILTER USED IN DECODER  */
    {
      /* The following line assumes that macroblock size is 16!!!         */
	  /* mbnumber1 gives the index number of the macroblock whose  */
	  /* upper left corner coordinates are (i1,j1) in object_list2          */
       mbnumber1 = ((picsizex >> 4) * (j1 /*temp1 */>> 4) + (i1 >> 4)) * 4 /*5*/;
       switch(smode)
       {
         case 0 : /* filtered 16x16 region's corner coordinates correspond legal macroblock grid */
         {
           if (object_list2[mbnumber1].regionMode < REGMODE_SPLITTED) /* if the macroblock is not a splitted macroblock*/
           { /* all the region modes and PEC modes are the same for all quadrants in the 16x16 
				  region selected for filtering */
              regmodes[1] = regmodes[2] = regmodes[3] = regmodes[4] = object_list2[mbnumber1].regionMode;
              pecmodes[1] = pecmodes[2] = pecmodes[3] = pecmodes[4] = object_list2[mbnumber1].pecMode;        
           }
           else  /* if the macroblock is a splitted mmacroblock */
           for (counter = 1; counter < 5 ; counter ++)
           {  /* The region modes and PEC modes of each quadrant is different and have different indexes  */
			  /*  The first quadrant index corresponds to mbnumber1 in the object_list2*/  
             regmodes[counter] = (object_list2[mbnumber1+counter-1].regionMode) - REGMODE_INTER_NOPRED_8x8;
             pecmodes[counter] = object_list2[mbnumber1+counter-1].pecMode;
           }
		   break;
         }  
         case 1 : /* filtered 16x16 region is a combination of 2 macroblocks, */		
         {	/*16x8 half comes from left macroblock, and 16x8 half from right of the  coordinates (i1+8,j1) */
           mbnumber2 =  mbnumber1 + 4; 
		   /* mbnumber1 corresponds to the macroblock index of the left macroblock 
		       and mbnumber2 corresponds to the right macroblock */
           if (object_list2[mbnumber1].regionMode < REGMODE_SPLITTED)  /* if the left macroblock is not splitted */
           {  /* all the region modes and PEC modes of the quadrants of the left half are the same */
             regmodes[1] = regmodes[3] = object_list2[mbnumber1].regionMode;
             pecmodes[1] = pecmodes[3] = object_list2[mbnumber1].pecMode; 
           }
           else  /* if the left macroblock is splitted, */
           {     /* the left half of the filtered region contains the quadrants 2 and 4 of this macroblock*/
             regmodes[1] = object_list2[mbnumber1 + 2-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[1] = object_list2[mbnumber1 + 2-1].pecMode;
             regmodes[3] = object_list2[mbnumber1 + 4-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[3] = object_list2[mbnumber1 + 4-1].pecMode;
           }
           if (object_list2[mbnumber2].regionMode < REGMODE_SPLITTED)  /* if the right macroblock is not splitted */
           {   /* all the region modes and PEC modes of the quadrants of the right half are the same */
             regmodes[2] = regmodes[4] = object_list2[mbnumber2].regionMode;
             pecmodes[2] = pecmodes[4] = object_list2[mbnumber2].pecMode; 
           }
           else   /* if the right macroblock is splitted, */
           {      /* the right half of the filtered region contains the quadrants 1 and 3 of this macroblock*/
             regmodes[2] = object_list2[mbnumber2 + 1-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[2] = object_list2[mbnumber2 + 1-1].pecMode;
             regmodes[4] = object_list2[mbnumber2 + 3-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[4] = object_list2[mbnumber2 + 3-1].pecMode;
           }
		   break;
         }

         case 2 :   /* filtered 16x16 region is a combination of 2 macroblocks, */
         {	/* 8x16 half comes from the upper macroblock, and 8x16 half from the 
			    lower of the coordinates (i1,j1+8) */   
           mbnumber2 =  mbnumber1 + (picsizex >> 4) * 4 /*5*/;
		   /* mbnumber1 corresponds to the macroblock index of the upper macroblock and 
		       mbnumber2 corresponds to the lower macroblock */
           if (object_list2[mbnumber1].regionMode < REGMODE_SPLITTED)   /* if the upper macroblock is not splitted */
           {     /* all the region modes and PEC modes of the upper 2 quadrants are the same */
             regmodes[1] = regmodes[2] = object_list2[mbnumber1].regionMode;
             pecmodes[1] = pecmodes[2] = object_list2[mbnumber1].pecMode; 
           }
           else  /* if the upper macroblock is splitted */
           {      /* the upper half of the filtered region contains the quadrants 3 and 4 of this macroblock */
             regmodes[1] = object_list2[mbnumber1 + 3-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[1] = object_list2[mbnumber1 + 3-1].pecMode;
             regmodes[2] = object_list2[mbnumber1 + 4-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[2] = object_list2[mbnumber1 + 4-1].pecMode;
           }
           if (object_list2[mbnumber2].regionMode < REGMODE_SPLITTED)   /* if the lower macroblock is not splitted */
           {  /* all the region modes and PEC modes of the lower 2 quadrants are the same */
             regmodes[3] = regmodes[4] = object_list2[mbnumber2].regionMode;
             pecmodes[3] = pecmodes[4] = object_list2[mbnumber2].pecMode; 
           }
           else  /* if the lower macroblock is splitted */
           {     /* the lower half of the filtered region contains the quadrants 1 and 2 of this macroblock */
             regmodes[3] = object_list2[mbnumber2 + 1-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[3] = object_list2[mbnumber2 + 1-1].pecMode;
             regmodes[4] = object_list2[mbnumber2 + 2-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[4] = object_list2[mbnumber2 + 2-1].pecMode;
           }
		   break;
         }
         
         case 3 : /* filtered region is a combination of 4 macroblocks , */
         {	/*	each quadrant (8x8 block) belongs to a different macroblock 
				  where they all intersect at coordinate (i1+8, j1+8)*/
           mbnumber2 = mbnumber1 + 4;
		   /* mbnumber2 is the index of the macroblock which is right-adjacent to 
			   macroblock with index mbnumber1 */
           mbnumber3 = mbnumber1 + (picsizex >> 4)*4;
		   /* mbnumber3 is the index of the macroblock which is bottom-adjacent to 
		       macroblok with index mbnumber1 */
           mbnumber4 = mbnumber3 + 4;
		   /* mbnumber4 is the index of the macroblock which is right-adjacebt to 
		       macroblock with index mbnumber3 */
           if (object_list2[mbnumber1].regionMode < REGMODE_SPLITTED)  /* if macroblock number mbnumber1 is not splitted */
           {  /* the quadrant number 1 of the filtered 16x16 region gets the region mode 
			       and PEC mode from macroblock number mbnumber1 */
             regmodes[1] = object_list2[mbnumber1].regionMode;
             pecmodes[1] = object_list2[mbnumber1].pecMode;
           }
           else  /* if macroblock number mbnumber1 is splitted */
           {  /* the quadrant number 1 of the filtered 16x16 region gets the region mode 
			       and PEC mode of macroblock number mbnumber1's 4th quadrant */
             regmodes[1] = object_list2[mbnumber1+4-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[1] = object_list2[mbnumber1+4-1].pecMode;
           }
           if (object_list2[mbnumber2].regionMode < REGMODE_SPLITTED)  /* if macroblock number mbnumber2 is not splitted */
           {  /* the quadrant number 2 of the filtered 16x16 region gets the region mode 
			       and PEC mode from macroblock number mbnumber2 */
             regmodes[2] = object_list2[mbnumber2].regionMode;
             pecmodes[2] = object_list2[mbnumber2].pecMode;
           }
           else  /* if macroblock number mbnumber2 is splitted */
           {  /* the quadrant number 2 of the filtered 16x16 region gets the region mode 
			       and PEC mode of macroblock number mbnumber2's 3th quadrant */
             regmodes[2] = object_list2[mbnumber2+3-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[2] = object_list2[mbnumber2+3-1].pecMode;
           }
           if (object_list2[mbnumber3].regionMode < REGMODE_SPLITTED)  /* if macroblock number mbnumber3 is not splitted */
           {  /* the quadrant number 3 of the filtered 16x16 region gets the region mode 
			       and PEC mode from macroblock number mbnumber3 */
             regmodes[3] = object_list2[mbnumber3].regionMode;
             pecmodes[3] = object_list2[mbnumber3].pecMode;
           }
           else  /* if macroblock number mbnumber3 is splitted */
           {  /* the quadrant number 3 of the filtered 16x16 region gets the region mode 
			      and PEC mode of macroblock number mbnumber3's 2nd quadrant */
             regmodes[3] = object_list2[mbnumber3+2-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[3] = object_list2[mbnumber3+2-1].pecMode;
           }
           if (object_list2[mbnumber4].regionMode < REGMODE_SPLITTED)  /* if macroblock number mbnumber4 is not splitted */
           {  /* the quadrant number 4 of the filtered 16x16 region gets the region mode 
			      and PEC mode from macroblock number mbnumber4 */
             regmodes[4] = object_list2[mbnumber4].regionMode;
             pecmodes[4] = object_list2[mbnumber4].pecMode;
           }
           else  /* if macroblock number mbnumber4 is splitted */
           {    /* the quadrant number 4 of the filtered 16x16 region gets the region mode 
			         and PEC mode of macroblock number mbnumber4's 1st quadrant */
             regmodes[4] = object_list2[mbnumber4+1-1].regionMode - REGMODE_INTER_NOPRED_8x8;
             pecmodes[4] = object_list2[mbnumber4+1-1].pecMode;
           }
           break;
		 }
       }
    }
  }
  else  /* FILTERING INTRA FRAME */
  {   /* If the frame being filtered is an INTRA frame, regmode[] ad pecmode[] has no usage. */
	   /*The following lines are put for a possible usage in future*/
    regmodes[1] = regmodes[2] = regmodes[3] = regmodes[4] = REGMODE_INTRA;
    pecmodes[1] = pecmodes[2] = pecmodes[3] = pecmodes[4] = PECMODE_SKIPPED;
  }
  
}


