/*
 * encDct.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for prediction error coding.
 *
 */


#include <limits.h>
#include <math.h>

#include "mainFlags.h"
#include "toolDct.h"
#include "encDct.h"
#include "decDemux.h"
#include "encMux.h"
#include "PredErrorCoder.h"
#include "PredErrorDecoder.h"
#include "encMCC.h"

/* Coded pattern overhead arrays below contain one bit less
 * for each symbol than the original arrays. This is due to
 * the fact that region mode defines already 'macroblock' CNC
 */

int
  allqp_inter64_huff_length[64] = {0, 9, 9, 9, 3, 9, 9, 9, 
                                   3, 9, 9, 9, 5, 9, 9, 9,
                                   3, 9, 9, 9, 5, 9, 9, 9,
                                   5, 9, 9, 9, 5, 9, 9, 9,
                                   3, 9, 9, 9, 5, 9, 9, 9,
                                   5, 9, 9, 9, 5, 9, 9, 9,
                                   5, 9, 9, 9, 5, 9, 9, 9,
                                   5, 9, 9, 9, 4, 9, 9, 9},

  allqp_inter8_huff_length[8]   = {0,  5,  4,  5,  1,  3,  3,  3},

  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
  /* Note: Zero CNC is not removed from these ones !!!! */
  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
  allqp_intra64_huff_length[64] = {4, 10, 10, 10,  6, 10, 10, 10,
                                   6, 10, 10, 10,  5, 10, 10, 10,
                                   6, 10, 10, 10,  6, 10, 10, 10,
                                  10, 10, 10, 10,  5, 10, 10, 10,
                                   6, 10, 10, 10,  6, 10, 10, 10,
                                   6, 10, 10, 10,  5, 10, 10, 10,
                                   6, 10, 10, 10,  5, 10, 10, 10,
                                   5, 10, 10, 10,  2,  4,  5,  2},
  allqp_intra8_huff_length[8]   = {3,  7,  7,  6,  1,  5,  4,  2};

/*
 * Define extern tables.
 */
extern int 
  *region_intra_dc,
  triplet_counter,
  triplet_counter_Y;

extern triplet 
  *region_triplets;

extern u_char 
  zig8x8[64],
  zig4x4[16];

extern int
  tableInter[65],
  tableIntra[65];

extern u_char 
  annexTlookup[32];

static void pecStoreDctInfo(int *dctCoeffs, int siz, int mode, int skipDc);

static void pecCopyBlkToImage(int *block, Byte *image, int x, int y,
                              int picSizeX, int sizeBlk);

static int Quant_DC4x4(int dc, int *bits, int *skipDc);

static int Quant_DC8x8(int dc, int *bits, int *skipDc);

/*
 * This flag is used for signalling whether the 
 * bits, error and CNC data calculated during 
 * 16x16 block INTRA can be used for 8x8 subblocks.
 *
 */

static int lagrangianFlag;

void setLagrangianFlag(int flag)
{
  lagrangianFlag=flag;
}

int getLagrangianFlag(void)
{
	return lagrangianFlag;
}


/*
 * This flag is used for signalling whether the 
 * bits, error and CNC data calculated during 
 * 16x16 block INTRA can be used for 8x8 subblocks.
 *
 */

static int lagrangianCopyRegionFlag;

void setLagrangianCopyRegionFlag(int flag)
{
  lagrangianCopyRegionFlag=flag;
}

int getLagrangianCopyRegionFlag(void)
{
	return lagrangianCopyRegionFlag;
}




int calcDctBits(int *source, int cols, int rows, u_char *zig, int qp, 
                int chroma, int mode, int start);


void encForwardDct(int *diffBlk, int *predBlk, int *origBlk, int *recoBlk,
                   int *sqrErr, int *bitsForCoef, int *quantDctCoeffs,
                   int qp, int sizedx, int sizedy, u_char *zigzagtable,
                   int chroma, int mode, int start, int doReco)
{
  double
    diffBlkFloat[64];
  int
    i, sizeOfBlk, 
    recoBlkTmp[64], errorBlk[64],
    VoidPTR[64];
   

  sizeOfBlk = sizedx*sizedy;


  varSizeDctAndQuant(diffBlk,quantDctCoeffs,sizeOfBlk,qp,chroma,mode,start,
                     sizedx,sizedy,diffBlkFloat,VoidPTR);

  /* Calculate the amount of DCT bits */

  *bitsForCoef = calcDctBits(quantDctCoeffs, sizedx, sizedy,
                             zigzagtable, qp, chroma, mode, start);
 
  if(*bitsForCoef > 0) {

	  if(doReco)
	  {
      /* 
       * Inverse Quantize the coefficients and perform 
       * squared error calculation.
       */
      decDct(quantDctCoeffs,recoBlkTmp,qp,sizedx,sizedy,chroma,mode);
      mceMakeDiff(origBlk, predBlk, recoBlkTmp, errorBlk, sizeOfBlk);
      *sqrErr = calcBlockEnergy(errorBlk, sizeOfBlk);
	  }
	  else 
	  {
      /* 
       * Calculate error in DCT domain.
       */
		  newInvQuantCoefint(quantDctCoeffs, recoBlkTmp, sizedx*sizedy, qp, chroma, mode);
      for(i=0;i<sizeOfBlk;i++) 
        errorBlk[i]=VoidPTR[i]-recoBlkTmp[i];
		  *sqrErr = calcBlockEnergy(errorBlk, sizeOfBlk);
	  }

    if(recoBlk != NULL) {
      for(i=0; i<sizeOfBlk; i++) {
        recoBlk[i] = origBlk[i] - errorBlk[i];
      }
    }
  }
  else {
    if(recoBlk != NULL) {
      for(i=0; i<sizeOfBlk; i++) {
        recoBlk[i] = predBlk[i];
      }
    }
  }
}

/***************************************************************************
 * Function    : SendDctRegion()
 *
 * Purpose     : Codes prediction error and dumps the region layer bitstream
 *               for Inter mode region. DCT used for prediction error coding.
 *
 * Arguments   : See the function call
 *
 * Returns     : Lagrangian cost for the region to be coded/analysed
 *
 * Description : Tests DCT coding mode vs. no prediction error coding mode
 *               if codePredErr flag is set and dumps the bits for chosen
 *               mode. Otherwise just dumps the mode and motion bits to the
 *               bitstream.
 *               Function can be used also to calculate Lagrangian cost for
 *               a region if send flag is set to 0.
 *
 *         NOTE: Lagrangian is returned only if codePredErr flag is set
 *         NOTE: region_triplets has to be allocated before calling this
 *               function
 *         NOTE: Uses B-Frame functions to carry out coding
 *         NOTE: Bits for chroma Dct are not exact
 *               
 *                 Copyright (c) 1997 Nokia Research Center
 ***************************************************************************/

long SendDctRegion(struct frame *currFr, /* In:  Current (original) frame  */
                   struct frame *predFr, /* In:  Prediction frame          */
                   int blockInd,         /* In:  Index of the pred block   */
                   struct frame *recoFr, /* Out: Reconstruction frame      */
                   object *currRegion,   /* In:  Ptr to the region to code */
				           object *object_list,
                   int picSizeX,         /* In:  Width of the image        */
                   int picSizeY,         /* In:  Height of the image       */
                   int regionSize,       /* In:  Region size (16 or 8)     */
                   int regionMode,       /* In:  Region mode               */
				           int regionIndex,
                   int predPossible,     /* In:  Is motion pred possible   */
                   int predDirsAvail,    /* In:  Available pred directions */
                   int predDir,          /* In:  Chosen pred direction     */
                   int predSelNeeded,    /* In:  Pred selection needed     */
                   int predSel,          /* In:  Chosen pred selection     */
                   int *motCoefs,        /* In:  (Delta) Motion coeffs     */
                   u_char *pecMode,      /* Out: Pred error coding mode    */
                   int *bitsMode,        /* Out: Bits for mode             */
                   int *bitsMotion,      /* Out: Bits for motion vect field*/
                   int *bitsOvr,         /* Out: Bits for DCT CNC patterns */
                   int *bitsDct,         /* Out: Bits for luma DCT         */
                   int *bitsDctC,        /* Out: Bits for chrom DCT        */
                   int qp,               /* In:  Quant. param. for luma    */
                   int aQP,              /* In:  QP for affine parameters  */
                   float lambda_Y,       /* In:  Lambda param. for luma    */
                   int codePredErr,      /* In:  Test DCT for this region  */
                   int buildReco,        /* In:  Build reconstruction      */
                   int send)             /* In:  Dump output to bitstream  */
{
  Byte
    *currY = currFr->yptr,
    *currU = currFr->uptr,
    *currV = currFr->vptr,
    *predY = predFr->yptr,
    *predU = predFr->uptr,
    *predV = predFr->vptr,
    *recoY,
    *recoU,
    *recoV;

  long
    sqrErrY      = 0,
    sqrErrC      = 0,
    sqrErrNoDctY = 0,
    sqrErrNoDctC = 0,
    lagrDct,
    lagrNoDct,
    finalLagrangian,
    energy;

  static int first_invo=1;
  static vlc_item Intra8[8];
  static vlc_item Intra64[19];
  extern vlc_item  vlcEncCncInter64[16],
						vlcEncCncInter8[7];
  int
    thresholdY,
    threshold8x8C,
    threshold4x4C, 
    lambda = (int)lambda_Y,
    numCodedBlocks,
    bitsModeDct,
    bitsModeNoDct,
    bits, err,
    ddump,
    x, y, xPred, yPred,
    blkIndex,
    qpC, lambdaC,
    sxmin, sxmax, symax, symin, /* boudning points of the segment */
    picSizeXC, picSizeYC,       /* chroma frame size */
    chu_CNC, chv_CNC,           /* chroma coded/not_coded */
    lum_CNC[4],                 /* luminance codec/not_coded */
    k, l,
    dctCoeffs[64],
    org[64],
    prd[64],
    dif[64],
    rec[64],
    blockCoord, predCoord, frameCoord,
    *inter64_huff_length = allqp_inter64_huff_length,
    *inter8_huff_length  = allqp_inter8_huff_length;

  double
    chromaWeight;

 

  *bitsMode       = 0;
  *bitsMotion     = 0;
  *bitsOvr        = 0;
  *bitsDct        = 0;
  *bitsDctC       = 0;

  sxmin      = currRegion->xmin;
  sxmax      = currRegion->xmax;
  symax      = currRegion->ymax;
  symin      = currRegion->ymin;
  picSizeXC  = picSizeX/2;
  picSizeYC  = picSizeY/2;

  qpC = annexTlookup[qp];
  lambdaC = (int) (0.85 * (double)(qpC * qpC));
  if (lambdaC < 1)
    lambdaC = 1;
  chromaWeight = lambda / lambdaC;  

  if(buildReco){
    recoY = recoFr->yptr;
    recoU = recoFr->uptr;
    recoV = recoFr->vptr;
  }

  if(codePredErr){

    /* Calculate lagrangian thresholds. DCT coding is tested
     * for blocks having larger not coded Lagrangian (energy)
     * than these thresholds.
     *
     * Thresholds are calculated as follows:
     *
     * threshold = quantizationErrorEstimate + lambda*minBitsForVLC
     *
     * As quantizationErrorEstimate LMSE estimate for quantization
     * squared error is utilized. This is a bit pessimistic but a
     * good trade-of between complexity and performance.
     *
     */

    thresholdY    = (64*qp*qp)/3   + lambda*3;
    threshold8x8C = (64*qpC*qpC)/3 + lambdaC*3;
    threshold4x4C = (16*qpC*qpC)/3 + lambdaC*4;

    if(send)
      triplet_counter = 0;

    /*
     * luminance coding
     */

    blkIndex = 0;

    for (y = symin, yPred = (blockInd & 2) << 2; y < symax; 
         y += 8, yPred += 8) {
      for (x = sxmin, xPred = (blockInd & 1) << 3; x < sxmax; 
           x += 8, xPred += 8, blkIndex++) {

        for (k = 0; k < 8; k++)
        {
          blockCoord = k << 3;
          predCoord = ((yPred + k) << 4) + xPred;
          frameCoord = (y + k) * picSizeX + x; 

          for (l = 0; l < 8; l++, blockCoord++, predCoord++, frameCoord++)
          {
            dif[blockCoord] = ((org[blockCoord] = currY[frameCoord]) -
                               (prd[blockCoord] = predY[predCoord]));
          }
        }
        energy        = calcBlockEnergy(dif, 64);
        sqrErrNoDctY += energy;


		  /*
		   * If lagrangian was already calculated during 16x16 INTER COPY,
		   * get the bits,error and CNC data from the 16x16 object struct.
		   *
		   */
		  if((regionSize==8) && (send==0) &&(buildReco==0) && (getLagrangianCopyRegionFlag()) 
			                 && (regionMode==REGMODE_INTER_COPY)) 
		 {
			*bitsDct += object_list[(regionIndex/5)*5].bitsCopy[(regionIndex%5)-1];
			sqrErrY += object_list[(regionIndex/5)*5].errCopy[(regionIndex%5)-1];
			lum_CNC[blkIndex] = object_list[(regionIndex/5)*5].cncCopy[(regionIndex%5)-1];

				
		  }
		  else
		  {

			/* Try coding only if energy is larger than a threshold */
			if(energy >= thresholdY){

			  encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
							qp, 8, 8, zig8x8, LUMA, INTER_BLOCK, 0,buildReco);

			  /* Decide is this block going to be coded or not */
			  if(bits && ((err + lambda*bits) < energy)){
				*bitsDct         += bits;
				sqrErrY          += err;
				lum_CNC[blkIndex] = 1;

				/* Store DCT coefficients if block is worth coding */
				if(send)
				  pecStoreDctInfo(dctCoeffs, 64, INTER_BLOCK, 0);
				if(buildReco)
				  pecCopyBlkToImage(rec, recoY, x, y, picSizeX, 8);

				
				/* Save lagrangian info for reuse with 8x8 blocks */
				if((regionSize==16) && (send==0) &&(buildReco==0) && (regionMode==REGMODE_INTER_COPY) )
				{
					currRegion->bitsCopy[blkIndex]=bits;
					currRegion->errCopy[blkIndex]=err;
					currRegion->cncCopy[blkIndex]=lum_CNC[blkIndex];			
				}

			  }
			  else{
				sqrErrY          += energy;
				lum_CNC[blkIndex] = 0;

				if(buildReco)
				  pecCopyBlkToImage(prd, recoY, x, y, picSizeX, 8);

				/* Save lagrangian info for reuse with 8x8 blocks */
				if((regionSize==16) && (send==0) &&(buildReco==0)  && (regionMode==REGMODE_INTER_COPY))
				{
					currRegion->bitsCopy[blkIndex]=0;
					currRegion->errCopy[blkIndex]=energy;
					currRegion->cncCopy[blkIndex]=lum_CNC[blkIndex];			
				}
			  }
			} 
			else{
			  sqrErrY          += energy;
			  lum_CNC[blkIndex] = 0;

			  if(buildReco)
				pecCopyBlkToImage(prd, recoY, x, y, picSizeX, 8);

			  /* Save lagrangian info for reuse with 8x8 blocks */
				if((regionSize==16) && (send==0) &&(buildReco==0) && (regionMode==REGMODE_INTER_COPY) )
				{
					currRegion->bitsCopy[blkIndex]=0;
					currRegion->errCopy[blkIndex]=energy;
					currRegion->cncCopy[blkIndex]=lum_CNC[blkIndex];			
				}
			}
		  }
	  }
    }


			

    if(send)
      triplet_counter_Y = triplet_counter;

    /*
     * Chrominance coding
     */

    x = sxmin >> 1, xPred = (blockInd & 1) << 2;
    y = symin >> 1, yPred = (blockInd & 2) << 1;

    /* 8x8 chroma coding */
    if(regionSize == 16){

      /* U coding */
      for (k = 0; k < 8; k++)
      {
        blockCoord = k << 3;
        frameCoord = (y + k) * picSizeXC + x; 

        for (l = 0; l < 8; l++, blockCoord++, frameCoord++)
        {
          dif[blockCoord] = ((org[blockCoord] = currU[frameCoord]) -
                             (prd[blockCoord] = predU[blockCoord]));
        }
      }
      energy        = calcBlockEnergy(dif, 64);
      sqrErrNoDctC += energy;

      if(energy >= threshold8x8C) {

        encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                      qpC, 8, 8, zig8x8, CHROMA, INTER_BLOCK, 0,buildReco);

        /* Decide is this block going to be coded or not */
        if(bits && ((err + lambdaC*bits) < energy)){
          *bitsDctC += bits;
          sqrErrC   += err;
          chu_CNC    = 1;

          /* Store DCT coefficients if block is worth coding */
          if(send)
            pecStoreDctInfo(dctCoeffs, 64, INTER_BLOCK, 0);
          if(buildReco)
            pecCopyBlkToImage(rec, recoU, x, y, picSizeXC, 8);
        }
        else{
          sqrErrC += energy;
          chu_CNC  = 0;

          if(buildReco)
            pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 8);
        }
      } 
      else{
        sqrErrC += energy;
        chu_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 8);
      }

      /* V coding */
      for (k = 0; k < 8; k++)
      {
        blockCoord = k << 3;
        frameCoord = (y + k) * picSizeXC + x; 

        for (l = 0; l < 8; l++, blockCoord++, frameCoord++)
        {
          dif[blockCoord] = ((org[blockCoord] = currV[frameCoord]) -
                             (prd[blockCoord] = predV[blockCoord]));
        }
      }
      energy        = calcBlockEnergy(dif, 64);
      sqrErrNoDctC += energy;

      if(energy >= threshold8x8C) {

        encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                      qpC, 8, 8, zig8x8, CHROMA, INTER_BLOCK, 0,buildReco);

        /* Decide is this block going to be coded or not */
        if(bits && ((err + lambdaC*bits) < energy)){
          *bitsDctC += bits;
          sqrErrC   += err;
          chv_CNC    = 1;

          /* Store DCT coefficients if block is worth coding */

          if(send)
            pecStoreDctInfo(dctCoeffs, 64, INTER_BLOCK, 0);
          if(buildReco)
            pecCopyBlkToImage(rec, recoV, x, y, picSizeXC, 8);
        }
        else{
          sqrErrC += energy;
          chv_CNC  = 0;

          if(buildReco)
            pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 8);
        }
      } 
      else{
        sqrErrC += energy;
        chv_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 8);
      }
    }

    /* 4x4 chroma coding */
    else{

      /* U coding */
      for (k = 0; k < 4; k++)
      {
        blockCoord = k << 2;
        predCoord = ((yPred + k) << 3) + xPred;
        frameCoord = (y + k) * picSizeXC + x; 

        for (l = 0; l < 4; l++, blockCoord++, predCoord++, frameCoord++)
        {
          dif[blockCoord] = ((org[blockCoord] = currU[frameCoord]) -
                             (prd[blockCoord] = predU[predCoord]));
        }
      }
      energy        = calcBlockEnergy(dif, 16);
      sqrErrNoDctC += energy;

      if(energy >= threshold4x4C) {

        encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                      qpC, 4, 4, zig4x4, CHROMA, INTER_BLOCK, 0,buildReco);

        /* Decide is this block going to be coded or not */
        if(bits && ((err + lambdaC*bits) < energy)){
          *bitsDctC += bits;
          sqrErrC   += err;
          chu_CNC    = 1;

          /* Store DCT coefficients if block is worth coding */
          if(send)
            pecStoreDctInfo(dctCoeffs, 16, INTER_BLOCK, 0);
          if(buildReco)
            pecCopyBlkToImage(rec, recoU, x, y, picSizeXC, 4);
        }
        else{
          sqrErrC += energy;
          chu_CNC  = 0;

          if(buildReco)
            pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 4);
        }
      } 
      else{
        sqrErrC += energy;
        chu_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 4);
      }

      /* V coding */
      for (k = 0; k < 4; k++)
      {
        blockCoord = k << 2;
        predCoord = ((yPred + k) << 3) + xPred;
        frameCoord = (y + k) * picSizeXC + x; 

        for (l = 0; l < 4; l++, blockCoord++, predCoord++, frameCoord++)
        {
          dif[blockCoord] = ((org[blockCoord] = currV[frameCoord]) -
                             (prd[blockCoord] = predV[predCoord]));
        }
      }
      energy        = calcBlockEnergy(dif, 16);
      sqrErrNoDctC += energy;

      if(energy >= threshold4x4C) {

        encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                      qpC, 4, 4, zig4x4, CHROMA, INTER_BLOCK, 0,buildReco);

        /* Decide is this block going to be coded or not */
        if(bits && ((err + lambdaC*bits) < energy)){
          *bitsDctC += bits;
          sqrErrC   += err;
          chv_CNC    = 1;

          /* Store DCT coefficients if block is worth coding */
          if(send)
            pecStoreDctInfo(dctCoeffs, 16, INTER_BLOCK, 0);
          if(buildReco)
            pecCopyBlkToImage(rec, recoV, x, y, picSizeXC, 4);
        }
        else{
          sqrErrC += energy;
          chv_CNC  = 0;

          if(buildReco)
            pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 4);
        }
      } 
      else{
        sqrErrC += energy;
        chv_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 4);
      }
    }
    
    /*
     * Overhead coding for DCT Yuv
     */

    numCodedBlocks = 0;

    if(regionSize == 16) {
      ddump = (lum_CNC[0] << 5) + (lum_CNC[1] << 4) +
              (lum_CNC[2] << 3) + (lum_CNC[3] << 2) +
              (chu_CNC << 1) + chv_CNC;
      if (ddump)
        numCodedBlocks++;
    }
    else {
      ddump = (lum_CNC[0] << 2) + (chu_CNC << 1) + chv_CNC;
      if (ddump)
        numCodedBlocks++;
    }

    if(numCodedBlocks > 0) {
	  if(regionSize == 16){
      *bitsOvr = vlcEncCncInter64[tableInter[ddump]].len;
	  if (tableInter[ddump] ==15)
		*bitsOvr += 6;
      }
	  else{
		*bitsOvr = vlcEncCncInter8[ddump-1].len; 
      }
    }
    else {
      *bitsOvr = 0;
    }

    /*
     * Calculate Lagrangian for this region if coded with or without DCT
     */

    /* Bits for motion */
    if(predPossible &&
       (regionMode == REGMODE_INTER_PRED ||
        regionMode == REGMODE_INTER_PRED_DELTA))
      *bitsMotion = SendMotionPredInfo(predDirsAvail, predDir,
                                       predSelNeeded, predSel, 0);
    else
      *bitsMotion = 0;

    if (regionMode == REGMODE_INTER_NOPRED)
      *bitsMotion += dumpAffineMotionParams(motCoefs, aQP, 0);
    else if (regionMode == REGMODE_INTER_PRED_DELTA)
      *bitsMotion += dumpAffineMotionParams(motCoefs, aQP, 0);

    if (numCodedBlocks > 0){

      /* Lagrangian with DCT */
      bitsModeDct = SendRegionMode(regionMode, PECMODE_CODED, predPossible,
                                   regionSize, 0);

      lagrDct     = (long)((sqrErrY + chromaWeight * sqrErrC) +
                           lambda*(bitsModeDct + *bitsMotion +
                                   *bitsOvr + *bitsDct + *bitsDctC));

      /* Lagrangian without DCT */
      bitsModeNoDct = SendRegionMode(regionMode, PECMODE_SKIPPED, predPossible,
                                     regionSize, 0);

      lagrNoDct     = (long)((sqrErrNoDctY + chromaWeight * sqrErrNoDctC) +
                             lambda*(bitsModeNoDct + *bitsMotion));

      /* Make the decision */
      if(lagrDct < lagrNoDct){
        *pecMode        = PECMODE_CODED;
        *bitsMode       = bitsModeDct;
        finalLagrangian = lagrDct;
      }
      else{
        *pecMode        = PECMODE_SKIPPED;
        *bitsMode       = bitsModeNoDct;
        *bitsOvr        = 0;
        *bitsDct        = 0;
        *bitsDctC       = 0;
        finalLagrangian = lagrNoDct;
      }
    }
    else{

      /* Nothing waz coded */
      *pecMode        = PECMODE_SKIPPED;

      *bitsMode = SendRegionMode(regionMode, PECMODE_SKIPPED,
                                 predPossible, regionSize, 0);

      finalLagrangian = (long)((sqrErrNoDctY + chromaWeight * sqrErrNoDctC) +
                               lambda*(*bitsMode + *bitsMotion));
      *bitsOvr        = 0;
    }
  } /* End of prediction error coding */

  else{

    /*
     * Else we are just going to dump bitstream for a region
     * with no prediction error coding (no need for lagrangian
     * calculation)
     */

    *pecMode        = PECMODE_SKIPPED;
    finalLagrangian = 0;
  }

  /*
   * Build reconstruction if requested and not done yet
   */

  if(buildReco){

    /* At this point Reconstruction is built with DCT coded
     * prediction error. Thus if DCT was not tested (codePredErr
     * was not set) or DCT coding was discarded due to bad R-D
     * performance (*modeDct == 0 and numCodedBlocks > 0) it has
     * to be built here.
     */

    if(!codePredErr ||
       (codePredErr && *pecMode == PECMODE_SKIPPED && numCodedBlocks > 0))
      pedRecoCopyRegionYUV(predFr, recoFr, currRegion, picSizeX, blockInd);

  }

  /*
   * Dump bitstream if send flag is set
   */

  if(send){

    int ovrovr = *bitsOvr;

    /* Send mode */
    *bitsMode = SendRegionMode(regionMode, *pecMode, predPossible, regionSize,
                               send);

    /* Send prediction information for motion */
    if(predPossible &&
       (regionMode == REGMODE_INTER_PRED ||
        regionMode == REGMODE_INTER_PRED_DELTA))
      *bitsMotion = SendMotionPredInfo(predDirsAvail, predDir,
                                       predSelNeeded, predSel, send);
    else
      *bitsMotion = 0;

    /* Send motion information */
    if (regionMode == REGMODE_INTER_NOPRED)
      *bitsMotion += dumpAffineMotionParams(motCoefs, aQP, send);
    else if (regionMode == REGMODE_INTER_PRED_DELTA)
      *bitsMotion += dumpAffineMotionParams(motCoefs, aQP, send);

    /* Send DCT info (if DCT utilized) */
    if(*pecMode == PECMODE_CODED){

      /* Send Coded/Not Coded information */
      *bitsOvr = SendCncInfoDct(lum_CNC, chu_CNC, chv_CNC, regionMode,
                                regionSize);

      /* Send triplet (Run-Length-Last) information for luminace */
      *bitsDct = SendYInfoDct(lum_CNC, regionMode, regionSize);

      /* Send triplet (Run-Length-Last) information for chrominance */
      *bitsDctC = SendCInfoDct(chu_CNC, chv_CNC, regionMode, regionSize);
    }

    if(*bitsOvr != ovrovr){
      fprintf(stdout, "  Ovr mismatch (inter%d): %d -> %d\n",
              regionSize, ovrovr, *bitsOvr);
      exit(0);
    }

  }

  return finalLagrangian;

}
long 
SendIntraDctRegion(struct frame *currFr, /* In:  Current (original) frame  */
                   struct frame *recoFr, /* Out: Reconstruction frame      */
                   object *currRegion,   /* In:  Ptr to the region to code */
				           object *object_list,
				           int regionIndex,
                   int picSizeX,         /* In:  Width of the image        */
                   int picSizeY,         /* In:  Height of the image       */
                   int regionSize,       /* In:  Region size (16 or 8)     */
                   int predPossible,     /* In:  Is motion pred possible   */
                   u_char *pecMode,      /* Out: Pred error coding mode    */
                   int *bitsMode,        /* Out: Bits for mode             */
                   int *bitsOvr,         /* Out: Bits for DCT CNC patterns */
                   int *bitsDct,         /* Out: Bits for luma DCT         */
                   int *bitsDctC,        /* Out: Bits for chrom DCT        */
                   int qp,               /* In:  Quant. param. for luma    */
                   float lambda_Y,       /* In:  Lambda param. for luma    */
                   int buildReco,        /* In:  Build reconstruction      */
                   int send)             /* In:  Dump output to bitstream  */
{
  Byte
    *currY = currFr->yptr,
    *currU = currFr->uptr,
    *currV = currFr->vptr,
    *recoY,
    *recoU,
    *recoV;

  long
    sqrErrY      = 0,
    sqrErrC      = 0,
    finalLagrangian,
    energy;

  int
    thresholdY,
    threshold8x8C,
    threshold4x4C,
    lambda = (int)lambda_Y,
    numCodedBlocks,
    bits, bitsDc, err,
    ddump,
    x, y, k, l,
    blkIndex,
    qpC, lambdaC,
    blkMode = INTER_BLOCK,
    sxmin, sxmax, symax, symin, /* boudning points of the segment */
    picSizeXC, picSizeYC,       /* chroma frame size */
    chu_CNC, chv_CNC,           /* chroma coded/not_coded */
    lum_CNC[4],                 /* luminance codec/not_coded */
    dctCoeffs[64],
    org[64],
    prd[64],
    dif[64],
    rec[64],
    *intra64_huff_length = allqp_intra64_huff_length,
    *intra8_huff_length  = allqp_intra8_huff_length,
    dc, predDCvalue, distance, start,
    predDcY[4] = {-1, -1, -1, -1},
    predDcU    = 128,
    predDcV    = 128;

  double
    chromaWeight;
  static int counter=0;

  *bitsMode       = 0;
  *bitsOvr        = 0;
  *bitsDct        = 0;
  *bitsDctC       = 0;

  sxmin      = currRegion->xmin;
  sxmax      = currRegion->xmax;
  symax      = currRegion->ymax;
  symin      = currRegion->ymin;
  picSizeXC  = picSizeX/2;
  picSizeYC  = picSizeY/2;

  qpC     = annexTlookup[qp];
  lambdaC = (int)(0.85 * qpC * qpC);

  if (lambdaC < 1)
    lambdaC = 1;
  chromaWeight = lambda / lambdaC;  

  if(buildReco){
    recoY = recoFr->yptr;
    recoU = recoFr->uptr;
    recoV = recoFr->vptr;
  }

  /* Calculate lagrangian thresholds. DCT coding is tested
   * for blocks having larger not coded Lagrangian (energy)
   * than these thresholds.
   */

  thresholdY    = (64*qp*qp)/3   + lambda*2;
  threshold8x8C = (64*qpC*qpC)/3 + lambdaC*2;
  threshold4x4C = (16*qpC*qpC)/3 + lambdaC*2;

  if(send){
    triplet_counter  = 0;
  }

  /*
   * luminance coding
   */

  blkIndex = 0;

  /*
   * Check if the 8x8 lagrangian was already calculated
   * If so, just get it from the object list
   *
   */
  for (y = symin; y < symax; y += 8) {
    for (x = sxmin; x < sxmax; x += 8, blkIndex++) {

      if(blkIndex == 0)
        predDCvalue = 128;
      else if(blkIndex == 1 || blkIndex == 2)
        predDCvalue = predDcY[0];
      else
        predDCvalue = predictDCValue(predDcY[2], predDcY[1], predDcY[0],
                                     LUMA);

      dc = 0;
      for(k = 0; k < 8; k++)
        for(l = 0; l < 8; l+=2) {
          dc += org[k*8+l] = (int)(currY[(y + k)*picSizeX + x + l]);
          dc += org[k*8+l+1] = (int)(currY[(y + k)*picSizeX + x + l+1]);
        }

      dc = Quant_DC8x8(dc, &bitsDc, &start);

      predDcY[blkIndex] = dc;

      if(blkIndex != 0) {
        distance = dc - predDCvalue;
        bitsDc   = bitsForDCDist(distance, LUMA);
      }

      *bitsDct += bitsDc;

      if(send){
        if(blkIndex == 0)
          region_intra_dc[blkIndex] = dc + 100;
        else if(abs(distance) < 64)
          region_intra_dc[blkIndex] = distance;
        else
          region_intra_dc[blkIndex] = dc + 400;
      }

      for(k = 0; k < 64; k+=2){
        prd[k] = dc;
        dif[k] = org[k] - prd[k];
        prd[k+1] = dc;
        dif[k+1] = org[k+1] - prd[k+1];
      }

	  /*
	   * If lagrangian was already calculated during 16x16 INTRA,
	   * get the bits,error and CNC data from the 16x16 object struct.
	   *
	   */
	  if((regionSize==8) && (send==0) &&(buildReco==0) && (getLagrangianFlag())) 
	  {
		*bitsDct += object_list[(regionIndex/5)*5].bitsIntra[(regionIndex%5)-1];
		sqrErrY += object_list[(regionIndex/5)*5].errIntra[(regionIndex%5)-1];
		lum_CNC[blkIndex] = object_list[(regionIndex/5)*5].cncIntra[(regionIndex%5)-1];
		
		
	  }
	  else
	  {
		
		  energy = calcBlockEnergy(dif, 64);

		  /* Try coding only if energy is larger than a threshold */
		  if(energy >= thresholdY){

			encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
						  qp, 8, 8, zig8x8, LUMA, blkMode, start,buildReco);

			/* Decide is this block going to be coded or not */
			if(bits && ((err + lambda*bits) < energy)){
			  *bitsDct         += bits;
			  sqrErrY          += err;
			  lum_CNC[blkIndex] = 1;

			  /* Store DCT coefficients if block is worth coding */
			  if(send)
				pecStoreDctInfo(dctCoeffs, 64, blkMode, start);
			  if(buildReco)
				pecCopyBlkToImage(rec, recoY, x, y, picSizeX, 8);

				/* Save partial lagrangians if we're doing a 16x16 block*/
		  
			  if((regionSize==16) && (send==0) &&(buildReco==0) )
			  {
				currRegion->bitsIntra[blkIndex]=bits;
				currRegion->errIntra[blkIndex]=err;
				currRegion->cncIntra[blkIndex]=lum_CNC[blkIndex];			
			  }

			}
			else{
			  sqrErrY          += energy;
			  lum_CNC[blkIndex] = 0;

			  if(buildReco)
				pecCopyBlkToImage(prd, recoY, x, y, picSizeX, 8);

				/* Save partial lagrangians if we're doing a 16x16 block*/
		  
			  if((regionSize==16) && (send==0) &&(buildReco==0) )
			  {
				currRegion->errIntra[blkIndex]=energy;
				currRegion->bitsIntra[blkIndex]=0;
				currRegion->cncIntra[blkIndex]=lum_CNC[blkIndex];
			  }
		
			}
		  } 
		  else{
			sqrErrY          += energy;
			lum_CNC[blkIndex] = 0;

			if(buildReco)
			  pecCopyBlkToImage(prd, recoY, x, y, picSizeX, 8);

			/* Save partial lagrangians if we're doing a 16x16 block*/
		  
			  if((regionSize==16) && (send==0) &&(buildReco==0) )
			  {
				currRegion->errIntra[blkIndex]=energy;
				currRegion->bitsIntra[blkIndex]=0;
				currRegion->cncIntra[blkIndex]=lum_CNC[blkIndex];
			  }
		  }
		

	  }
		  
	}
  }

  if(send)
    triplet_counter_Y = triplet_counter;

  /*
   * Chrominance coding
   */

  x = sxmin/2;
  y = symin/2;

  /* 8x8 chroma coding */
  if(regionSize == 16){

    /* U coding */
    dc = 0;
    for(k = 0; k < 8; k++)
      for(l = 0; l < 8; l+=2) {
        dc += org[k*8+l] = (int)(currU[(y + k)*picSizeXC + x + l]);
        dc += org[k*8+l+1] = (int)(currU[(y + k)*picSizeXC + x + l+1]);
      }

    dc         = Quant_DC8x8(dc, &bitsDc, &start);
    distance   = dc - predDcU;
    bitsDc     = bitsForDCDist(distance, CHROMA);
    *bitsDctC += bitsDc;

    if(send){
      if(abs(distance) < 64)
        region_intra_dc[4] = distance;
      else
        region_intra_dc[4] = dc + 400;
    }

    for(k = 0; k < 64; k+=2){
      prd[k] = dc;
      dif[k] = org[k] - prd[k];
      prd[k+1] = dc;
      dif[k+1] = org[k] - prd[k+1];
    }
              
    energy = calcBlockEnergy(dif, 64);

    if(energy >= threshold8x8C) {

      encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                    qpC, 8, 8, zig8x8, CHROMA, blkMode, start,buildReco);

      /* Decide is this block going to be coded or not */
      if(bits && ((err + lambdaC*bits) < energy)){
        *bitsDctC += bits;
        sqrErrC   += err;
        chu_CNC    = 1;

        /* Store DCT coefficients if block is worth coding */
        if(send)
          pecStoreDctInfo(dctCoeffs, 64, blkMode, start);
        if(buildReco)
          pecCopyBlkToImage(rec, recoU, x, y, picSizeXC, 8);
      }
      else{
        sqrErrC += energy;
        chu_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 8);
      }
    } 
    else{
      sqrErrC += energy;
      chu_CNC  = 0;

      if(buildReco)
        pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 8);
    }

    /* V coding */
    dc = 0;
    for(k = 0; k < 8; k++)
      for(l = 0; l < 8; l+=2) {
        dc += org[k*8+l] = (int)(currV[(y + k)*picSizeXC + x + l]);
        dc += org[k*8+l+1] = (int)(currV[(y + k)*picSizeXC + x + l+1]);
      }

    dc         = Quant_DC8x8(dc, &bitsDc, &start);
    distance   = dc - predDcV;
    bitsDc     = bitsForDCDist(distance, CHROMA);
    *bitsDctC += bitsDc;

    if(send){
      if(abs(distance) < 64)
        region_intra_dc[5] = distance;
      else
        region_intra_dc[5] = dc + 400;
    }

    for(k = 0; k < 64; k+=2){
      prd[k] = dc;
      dif[k] = org[k] - prd[k];
      prd[k+1] = dc;
      dif[k+1] = org[k+1] - prd[k+1];
    }

    energy = calcBlockEnergy(dif, 64);

    if(energy >= threshold8x8C) {

      encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                    qpC, 8, 8, zig8x8, CHROMA, blkMode, start,buildReco);

      /* Decide is this block going to be coded or not */
      if(bits && ((err + lambdaC*bits) < energy)){
        *bitsDctC += bits;
        sqrErrC   += err;
        chv_CNC    = 1;

        /* Store DCT coefficients if block is worth coding */

        if(send)
          pecStoreDctInfo(dctCoeffs, 64, blkMode, start);
        if(buildReco)
          pecCopyBlkToImage(rec, recoV, x, y, picSizeXC, 8);
      }
      else{
        sqrErrC += energy;
        chv_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 8);
      }
    } 
    else{
      sqrErrC += energy;
      chv_CNC  = 0;

      if(buildReco)
        pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 8);
    }
  }

  /* 4x4 chroma coding */
  else{

    /* U coding */
    dc = 0;
    for(k = 0; k < 4; k++) {
      dc += org[k*4]   = (int)(currU[(y + k)*picSizeXC + x]);
      dc += org[k*4+1] = (int)(currU[(y + k)*picSizeXC + x + 1]);
      dc += org[k*4+2] = (int)(currU[(y + k)*picSizeXC + x + 2]);
      dc += org[k*4+3] = (int)(currU[(y + k)*picSizeXC + x + 3]);
    }

    dc         = Quant_DC4x4(dc, &bitsDc, &start);
    distance   = dc - predDcU;
    bitsDc     = bitsForDCDist(distance, CHROMA);
    *bitsDctC += bitsDc;

    if(send){
      if(abs(distance) < 64)
        region_intra_dc[4] = distance;
      else
        region_intra_dc[4] = dc + 400;
    }

    for(k = 0; k < 16; k+=2){
      prd[k] = dc;
      dif[k] = org[k] - prd[k];
      prd[k+1] = dc;
      dif[k+1] = org[k+1] - prd[k+1];
    }

    energy = calcBlockEnergy(dif, 16);

    if(energy >= threshold4x4C) {

      encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                    qpC, 4, 4, zig4x4, CHROMA, blkMode, start,buildReco);

      /* Decide is this block going to be coded or not */
      if(bits && ((err + lambdaC*bits) < energy)){
        *bitsDctC += bits;
        sqrErrC   += err;
        chu_CNC    = 1;

        /* Store DCT coefficients if block is worth coding */
        if(send)
          pecStoreDctInfo(dctCoeffs, 16, blkMode, start);
        if(buildReco)
          pecCopyBlkToImage(rec, recoU, x, y, picSizeXC, 4);
      }
      else{
        sqrErrC += energy;
        chu_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 4);
      }
    } 
    else{
      sqrErrC += energy;
      chu_CNC  = 0;

      if(buildReco)
        pecCopyBlkToImage(prd, recoU, x, y, picSizeXC, 4);
    }

    /* V coding */
    dc = 0;
    for(k = 0; k < 4; k++) {
      dc += org[k*4] = (int)(currV[(y + k)*picSizeXC + x]);
      dc += org[k*4+1] = (int)(currV[(y + k)*picSizeXC + x + 1]);
      dc += org[k*4+2] = (int)(currV[(y + k)*picSizeXC + x + 2]);
      dc += org[k*4+3] = (int)(currV[(y + k)*picSizeXC + x + 3]);
    }

    dc         = Quant_DC4x4(dc, &bitsDc, &start);
    distance   = dc - predDcV;
    bitsDc     = bitsForDCDist(distance, CHROMA);
    *bitsDctC += bitsDc;

    if(send){
      if(abs(distance) < 64)
        region_intra_dc[5] = distance;
      else
        region_intra_dc[5] = dc + 400;
    }

    for(k = 0; k < 16; k+=2){
      prd[k] = dc;
      dif[k] = org[k] - prd[k];
      prd[k+1] = dc;
      dif[k+1] = org[k+1] - prd[k+1];
    }

    energy = calcBlockEnergy(dif, 16);

    if(energy >= threshold4x4C) {

      encForwardDct(dif, prd, org, rec, &err, &bits, dctCoeffs,
                    qpC, 4, 4, zig4x4, CHROMA, blkMode, start,buildReco);

      /* Decide is this block going to be coded or not */
      if(bits && ((err + lambdaC*bits) < energy)){
        *bitsDctC += bits;
        sqrErrC   += err;
        chv_CNC    = 1;

        /* Store DCT coefficients if block is worth coding */
        if(send)
          pecStoreDctInfo(dctCoeffs, 16, blkMode, start);
        if(buildReco)
          pecCopyBlkToImage(rec, recoV, x, y, picSizeXC, 4);
      }
      else{
        sqrErrC += energy;
        chv_CNC  = 0;

        if(buildReco)
          pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 4);
      }
    } 
    else{
      sqrErrC += energy;
      chv_CNC  = 0;

      if(buildReco)
        pecCopyBlkToImage(prd, recoV, x, y, picSizeXC, 4);
    }
  }
    
  /*
   * Overhead coding for DCT Yuv
   */

  numCodedBlocks = 0;

  if(regionSize == 16){
    ddump = 32*lum_CNC[0] + 16*lum_CNC[1] +
             8*lum_CNC[2] +  4*lum_CNC[3] +
             2*chu_CNC    +    chv_CNC;
    if (ddump)
      numCodedBlocks++;
    *bitsOvr = intra64_huff_length[ddump];
  }
  else{
    ddump = 4*lum_CNC[0] + 2*chu_CNC + chv_CNC;
    if (ddump)
      numCodedBlocks++;
    *bitsOvr = intra8_huff_length[ddump];
  }

  /*
   * Calculate Lagrangian for this region if coded with or without DCT (i.e
   * with DC prediction only)
   */

  if (numCodedBlocks > 0){

    /* Lagrangian with DCT */
    *pecMode        = PECMODE_CODED;

    *bitsMode = SendRegionMode(REGMODE_INTRA, PECMODE_CODED,
                               predPossible, regionSize, 0);

    finalLagrangian = (long)((sqrErrY + chromaWeight * sqrErrC) +
                             lambda*(*bitsMode + *bitsOvr + *bitsDct + *bitsDctC));

  }
  else{

    /* Nothing waz coded */
    *pecMode        = PECMODE_SKIPPED;
    *bitsOvr        = 0;

    *bitsMode = SendRegionMode(REGMODE_INTRA, PECMODE_SKIPPED,
                               predPossible, regionSize, 0);

    finalLagrangian = (long)((sqrErrY + chromaWeight * sqrErrC) +
                              lambda*(*bitsMode + *bitsDct + *bitsDctC));
  }

  /*
   * Dump bitstream if send flag is set
   */

  if(send){

    /* Send mode */
    *bitsMode = SendRegionMode(REGMODE_INTRA, *pecMode, predPossible,
                               regionSize, send);

    /* Send DCT info (At least DC is allways sent) */
    if(*pecMode == PECMODE_CODED){
      /* Send Coded/Not Coded information */
      *bitsOvr = SendCncInfoDct(lum_CNC, chu_CNC, chv_CNC, REGMODE_INTRA,
                                regionSize);
    }
    else
      *bitsOvr = 0;
 
    /* Send DC + triplet (Run-Length-Last) information for luminance */
    *bitsDct = SendYInfoDct(lum_CNC, REGMODE_INTRA, regionSize);

    /* Send DC + triplet (Run-Length-Last) information for chrominance */
    *bitsDctC = SendCInfoDct(chu_CNC, chv_CNC, REGMODE_INTRA, regionSize);
  }

  return finalLagrangian;

}

static void pecStoreDctInfo(int *dctCoeffs, int siz, int mode, int skipDc)
{
  int 
    codedCoeffs, coe, run, las, lev, dctCoeffsZigZag[64];

  u_char
    *zig;

  if(siz == 64)
    zig = zig8x8;
  else
    zig = zig4x4;

  for(coe = 0;coe < siz;coe++)
    dctCoeffsZigZag[zig[coe]] = dctCoeffs[coe];

  coe = skipDc;

  codedCoeffs = 0;
  while(coe < siz)
    if(dctCoeffsZigZag[coe++] != 0) codedCoeffs++;
  
  coe = skipDc;
  run = 0;
  while(codedCoeffs > 0) {

    while(dctCoeffsZigZag[coe] == 0) { coe++; run++; }

    if(codedCoeffs > 1)
      las = 0;
    else
      las = 1;

    lev = dctCoeffsZigZag[coe];

    region_triplets[triplet_counter].last  = las;
    region_triplets[triplet_counter].run   = run;
    region_triplets[triplet_counter].level = lev;

    triplet_counter++;
    coe++;
    run = 0;
    codedCoeffs--;
  }
}

static void pecCopyBlkToImage(int *block, Byte *image, int x, int y, int picSizeX,
                              int sizeBlk)
{
  int
    i, j,
    indBlk = 0;
  int32
    indImg,
    indStart = y * picSizeX + x;

  for( i = 0; i < sizeBlk;i++) {
    indImg = indStart;
    indStart += picSizeX;
    for (j = 0; j < sizeBlk; j++)
      image[indImg++] = block[indBlk++];
  }
}

#define DC_BITS             8
#define DC_LEVELS           (1 << DC_BITS)

static int Quant_DC8x8(int dc, int *bits, int *skipDc)
{

  double
    v = (double) dc / 16320.0;

  int
    index = (int) floor(v * (DC_LEVELS-1) + 0.5);

  *bits = DC_BITS;
  *skipDc = (DC_BITS >= 8) ? 1 : 0;

  return (int) floor(0.5 + (0.5 + index) * (255.0 / DC_LEVELS));
}

#define DC_BITS4x4          8
#define DC_LEVELS4x4        (1<<DC_BITS4x4)

static int Quant_DC4x4(int dc, int *bits, int *skipDc)
{

  double
    v = (double) dc / 4080.0;

  int
    index = (int) floor(v * (DC_LEVELS4x4-1) + 0.5);

  *bits = DC_BITS4x4;
  *skipDc = (DC_BITS4x4 >= 8) ? 1 : 0;

  return (int) floor(0.5 + (0.5 + index) * (255.0 / DC_LEVELS4x4));
}
