/*
 * encMCC.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Functions for prediction error coding.
 *
 */


#include "mainFlags.h"
#include "encMCC.h"
#include "decDemux.h"
#include "encMux.h"
#include "toolMCC.h"
#include "encMCC.h"
#include <limits.h>
#include <math.h>

extern void varSizeDctFixedp(int *source, int *destination, int cols, int rows );

extern void varSizeInvDctFixedp(int *source, int *destination, int cols, int rows );

/*
 * Define extern tables.
 */

extern u_char
  zig4x8[SIZE_8x4],
  zig4x4[SIZE_4x4],
  zig4x4Hori[SIZE_4x4],
  zig4x4Vert[SIZE_4x4],
  zig4x4Intra[SIZE_4x4],
  zig4x4IntraHori[SIZE_4x4],
  zig4x4IntraVert[SIZE_4x4],
  zig8x8[SIZE_8x8],
  zig8x8Hori[SIZE_8x8],
  zig8x8Vert[SIZE_8x8],
  zig8x8Intra[SIZE_8x8],
  zig8x8IntraHori[SIZE_8x8],
  zig8x8IntraVert[SIZE_8x8];

extern u_char
  mirrorVert4x4[SIZE_4x4], 
  mirrorDiag4x4[SIZE_4x4],
  rotateLeft4x4[SIZE_4x4], 
  rotateRight4x4[SIZE_4x4];

extern u_char
  mirrorVert8x8[SIZE_8x8],
  mirrorDiag8x8[SIZE_8x8],
  rotateLeft8x8[SIZE_8x8], 
  rotateRight8x8[SIZE_8x8];

extern u_char
  kltLookUpTable[10];

#define MCE_BIG_LAGRANGIAN                   1048576

#define MCE_CRITICAL_BITS_4x4CBS             0
#define MCE_CRITICAL_BITS_8x8DCT             2
#define MCE_CRITICAL_BITS_4x8DCT             2
#define MCE_CRITICAL_BITS_4x4DCT             2
#define MCE_CRITICAL_BITS_8x8KLT             2
#define MCE_CRITICAL_BITS_4x4KLT             2
#define MCE_CRITICAL_BITS_4x4RDC             1

#define MCE_MIN_IMPROVEMENT_4x4                    50
/*
 * Define local function prototypes.
 */

static int mceFindAbsMaxCoefIndex(int *dctCoefs, int sizeOfBlk,
                                  int *sign, int qp);

static int mceFindFirstIndex(u_char *zigZagTable, int size);

static void mceEnc8x4Dct(int *diffBlk, int *predBlk, int *origBlk, int *seAfter,
                         int *bitsForCoef, int *quantDctCoeffs, int method,
                         int qp, int sizedx, int sizedy, u_char *zigZagTable,
                         int chroma, int mode, int complexityMode);

static int mceEncExtrapol(int *predBlk, int *origBlk, int method);

static void mceEnc8x8Dct(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantDctCoeffs,
                         int method, int qp, int chroma, int mode, int ncEnergy,
                         int *memFlag, int complexityMode);

static void mceEnc8x8Klt(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantDctCoeffs,
                         int method, int qp, int chroma, int mode, int ncEnergy, int complexityMode);

static void mceEnc4x4Dct(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantDctCoeffs,
                         int dctMet, int qp, int chroma, int mode, int blkSize,
                         int *memFlag, int complexityMode);

static void mceEnc4x4Klt(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantKltCoeffs,
                         int kltNum, int qp, int chroma, int mode,
                         int ncEnergy, int complexityMode);

int calcDctBits(int *source, int cols, int rows, u_char *zig, int qp, 
                int chroma, int mode, int start);

/*
 * mceMakeDiff:
 *
 * Parameters:
 *      origBlk         I: Original block pointer
 *      predBlk         I: Prediction block pointer
 *      recoBlk         I: Reconstruction block pointer
 *      diffBlk         O: Difference block pointer
 *      blkSize         I: Size of source block
 *
 * Function:
 *      Produces difference block between original block and real
 *      reconstructed block (clipped prediction + reconstruction).
 *
 * Returns:
 *      Nothing.
 */

void mceMakeDiff(int *origBlk, int *predBlk, int *recoBlk, int *diffBlk,
                 int blkSize)
{
  int 
    i,                /* Loop increment counter                       */
    outValue;         /* Real reconstructed value                     */

  for (i = 0; i < blkSize; i++) {
    outValue = predBlk[i] + recoBlk[i];

    /*
     * Clipping to byte level.
     */
    if (outValue > 255)
      outValue = 255;

    else if (outValue < 0)
      outValue = 0;

    diffBlk[i] = origBlk[i] - outValue;
  }
}
/*
 * mceFindAbsMaxCoefIndex:
 *
 * Parameters:
 *      dctCoefs        I: DCT coefficients
 *      sizeOfBlk       I: Block size
 *      sign            O: Coefficient sign
 *      qp              I: QP
 *
 * Function:
 *      Finds the DCT coefficient that has the largest absolute value.
 *      Function is used in coefficient refinement.
 *
 * Returns:
 *      If the largest absolute coefficient value is larger than QP
 *      (i.e. level can be rounded to -1 or +1), returns the index of
 *      the coefficient, otherwise returns -1.
 */

static int mceFindAbsMaxCoefIndex(int *dctCoefs, int sizeOfBlk,
                                  int *sign, int qp)
{
  int
    i, index = 0,
    max, tmp;

  max = abs(dctCoefs[0]);

  if (qp > max) {
    *sign = dctCoefs[0] < 0 ? -1 : 1;
    max = qp;
    index = -1;
  }

  for (i = 1; i < sizeOfBlk; i++) {
    tmp = abs(dctCoefs[i]);
    if (tmp > max) {
      max = tmp;
      index = i;
    }
  }

  if (index != -1)
    *sign = dctCoefs[index] < 0 ? -1 : 1;
 
  return index;
}


/*
 * mceFindFirstIndex:
 *
 * Parameters:
 *      zigZagTable     I: DCT scanning order
 *      size            I: Block size
 *
 * Function:
 *      Finds the first coefficient in scanning order.
 *
 * Returns:
 *      Coefficient index.
 */

static int mceFindFirstIndex(u_char *zigZagTable, int size)
{
  int
    i = 0, 
    index = -1;

  while (index == -1 && i < size) {
    if (zigZagTable[i] == 0)
      index = i;

    i++;
  }

  return index;
}


/*
 * mceEnc8x4Dct:
 *
 * Parameters:
 *      diffBlk         I: Prediction error block
 *      predBlk         I: Prediction block
 *      origBlk         I: Original block
 *      seAfter         O: SE after DCT
 *      bitsForCoef     O: Coding bits for coefficients
 *      quantDctCoeffs  O: Quantized DCT coefficients
 *      method          I: DCT method
 *      qp              I: Quantization parameter
 *      sizeX           I: Block width
 *      sizeY           I: Block height
 *      zigZagTable     I: DCT scanning table 
 *      chroma          I: Chrominance flag
 *                            0 = luminance
 *                            1 = chrominance
 *      mode            I: Mode flag
 *                            0 = INTER
 *                            1 = INTRA
 *
 * Function:
 *      Calculates 4x4 DCT using refinement step if quantized coefficients
 *      are initially all zero. Usage of state flag allows calculation for 
 *      different coefficient scannings without recalculation of DCT.
 *
 * Returns:
 *      Nothing.
 */

static void mceEnc8x4Dct(int *diffBlk, int *predBlk, int *origBlk, int *seAfter,
                         int *bitsForCoef, int *quantDctCoeffs, int method,
                         int qp, int sizeX, int sizeY, u_char *zigZagTable,
                         int chroma, int mode, int complexityMode)
{
  int 
    nonZeros,             /* Amount of non-zero coefficients */
    sign,                 /* Sign of refined coefficient */
    index,                /* Index of refined coefficient */
    sizeOfBlk,            /* Block size */
    dctCoef[SIZE_8x8],    /* Quantized coefficients */
    invCoef[SIZE_8x8],    /* Inverse quantized coefficients */
    recoBlkTmp[SIZE_8x8], /* Temporary reconstruction block */
    errorBlk[SIZE_8x8];   /* Coding error block */
  
  sizeOfBlk = sizeX * sizeY;
  
  /*
   * Calculate DCT and quantize coefficients.
   */
  varSizeDctFixedp(diffBlk, dctCoef, sizeX, sizeY);
  if (mode != INTER_BLOCK && complexityMode == COMPLEXITY_MODE_ANCHOR)
    nonZeros = RD_Optimized_Quantize_Coef(dctCoef, quantDctCoeffs, sizeOfBlk, qp, chroma, 
                                        mode, DCT_TRANSFORM, zigZagTable, sizeX, sizeY, 0);  
  else
    nonZeros = MCCQuantCoef(dctCoef, quantDctCoeffs, sizeOfBlk, qp,
                         chroma, mode, DCT_TRANSFORM);
  /*
   * If all coefficients are quantized to zero,
   * do refinement.
   */
  if (nonZeros == 0) {

    /*
     * Find largest coefficient value.
     */
    index = mceFindAbsMaxCoefIndex(dctCoef, sizeOfBlk, &sign, qp);

    if (index < 0) {

      /*
       * If largest coefficient value was smaller than QP 
       * find the first coefficient's sign.
       */
      index = mceFindFirstIndex(zigZagTable, sizeOfBlk);
      sign = dctCoef[index] < 0 ? -1 : 1;
    }

    /*
     * Set refined quantized coefficient to -1 or +1.
     */
    quantDctCoeffs[index] = sign;
    nonZeros = 1;
  }

  /*
   * Calculate the amount of DCT bits.
   */
  bitsForCoef[method] = calcDctBits(quantDctCoeffs, sizeX, sizeY,
                                    zigZagTable, qp, chroma, mode, 0);

  /* 
   * Inverse Quantize the coefficients.
   */
  MCCInvQuantCoef(quantDctCoeffs, invCoef, sizeOfBlk, qp, chroma, mode);

  /*
   * Perform actual IDCT and clip the reconstruction
   * values before calculating coding error.
   */
  varSizeInvDctFixedp(invCoef, recoBlkTmp, sizeX, sizeY);
  mceMakeDiff(origBlk, predBlk, recoBlkTmp, errorBlk, sizeOfBlk);
  seAfter[method] = calcBlockEnergy(errorBlk, sizeOfBlk);
}


/*
 * mceEncExtrapol:
 *
 * Parameters:
 *      predBlk         I: Prediction block
 *      origBlk         I: Original block
 *      method          I: Extrapolation method
 *
 * Function:
 *      Calculates extrapolation for one 8x4 half.
 *
 * Returns:
 *      SE for extrapolation
 */

static int mceEncExtrapol(int *predBlk, int *origBlk, int method)
{
  int
    i,                      /* Loop counter */
    seAfter,                /* SE for extrapolation */
    origBlkTmp[SIZE_8x4],   /* Buffer for original pixels */
    errorBlk[SIZE_8x4],     /* Buffer for coding error pixels */
    extValue,               /* Pixel value for extrapolation */
    extrBlk[9],             /* Buffer for pixels used for extrapolation */
    extPels;                /* Number of pixels used for extrapolation */

  /*
   * Get the pixels for extrapolation for upper half
   * of 8x8 block.
   */
  switch (method) {

    case METH_EXTRAPOL_12:
      getIntBlock(origBlk, origBlkTmp, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_1, START_ROW_8x4_1, ROWS_8x8);
      extPels = getExtSou8x8(predBlk, extrBlk, HALF_1);
      break;

    /*
     * Get the pixels for extrapolation for left half
     * of 8x8 block.
     */
    case METH_EXTRAPOL_14:
      getIntBlock(origBlk, origBlkTmp, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
      extPels = getExtSou8x8(predBlk, extrBlk, HALF_3);
      break;

    /*
     * Get the pixels for extrapolation for right half
     * of 8x8 block.
     */
    case METH_EXTRAPOL_28:
      getIntBlock(origBlk, origBlkTmp, COLS_4x8, ROWS_4x8,
                  START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
      extPels = getExtSou8x8(predBlk, extrBlk, HALF_4);
      break;

    /*
     * Get the pixels for extrapolation for lower half
     * of 8x8 block.
     */
    default:
      getIntBlock(origBlk, origBlkTmp, COLS_8x4, ROWS_8x4,
                  START_COL_8x4_2, START_ROW_8x4_2, ROWS_8x8);
      extPels = getExtSou8x8(predBlk, extrBlk, HALF_2);
      break;
  }

  /*
   * Calculate the median of pixels.
   */
  extValue = median(extrBlk, extPels);

  /*
   * Create the coding error block.
   */
  for (i = 0; i < SIZE_8x4; i++)
    errorBlk[i] = origBlkTmp[i] - extValue;

  /*
   * Calculate the SE
   */
  seAfter = calcBlockEnergy(errorBlk, SIZE_8x4);

  return seAfter;
}


/*
 * mceEnc8x8Dct:
 *
 * Parameters:
 *      diffBlkInput    I: Prediction error block
 *      predBlkInput    I: Prediction block
 *      origBlkInput    I: Original block
 *      seAfter         O: SE after DCT
 *      bitsForCoef     O: Coding bits for coefficients
 *      quantDctCoeffs  O: Quantized DCT coefficients
 *      method          I: DCT method
 *      qp              I: Quantization parameter
 *      chroma          I: Chrominance flag
 *                            0 = luminance
 *                            1 = chrominance
 *      mode            I: Mode flag
 *                            0 = INTER
 *                            1 = INTRA
 *      ncEnergy        I: SE for not coded prediction error
 *      memFlag         I: State flag
 *                            0 = First call for this block
 *                            2 = Later call for this block, non-zero
 *                                coefficients do not exist, no need
 *                                for actual transform calculation
 *                            1 = Later call for this block, non-zero
 *                                coefficients do not exist
 *
 * Function:
 *      Calculates 8x8 DCT using refinement step if quantized coefficients are
 *      initially all zero. Usage of state flag allows calculation for 
 *      different coefficient scannings without recalculation of DCT.
 *
 * Returns:
 *      Nothing.
 */

static void mceEnc8x8Dct(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantDctCoeffs,
                         int method, int qp, int chroma, int mode, int ncEnergy,
                         int *memFlag, int complexityMode)
{
  static int 
    memDct[SIZE_8x8],     /* Memory for DCT coefficients */
    memCoe[SIZE_8x8],     /* Memory for quantized coefficients */
    memSe = 0;            /* Memory for SE after DCT */

  int
    value,                /* Refined coefficient before quantization */
    invValue,             /* Refined coefficient after inverse quantization */
    diffBlk[SIZE_8x8],    /* Result of IDCT */
    dctCoeffs[SIZE_8x8],  /* Quantized coefficients */
    dctInvs[SIZE_8x8],    /* Inverse quantized coefficients */
    errorBlk[SIZE_8x8],   /* Coding error block */
    i,                    /* Loop counter */
    nonZeros,             /* Amount of non-zero coefficients */
    sign,                 /* Sign of refined coefficient */
    index;                /* Index of refined coefficient */

  u_char
    *zigZagTable;

  /*
   * Check if this is the first call for the block.
   */
  if (*memFlag == 0) {

    /*
     * Set state flag for later calls, state 2 assumes non-zero
     * coefficients.
     */
    *memFlag = 2;
    
    /*
     * Calculate DCT and quantize coefficients and store them
     * in memory for later calls.
     */
/* NEW ADDITION 14.4.2000 */
      switch (method) {

    case METH_8x8_DCT:
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8 : zig8x8Intra;
      break;

    case METH_DCT_8x8_N_0_8:
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8Hori : zig8x8IntraHori;
      break;

    default: 
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8Vert : zig8x8IntraVert;
      break;
  }

    varSizeDctFixedp(diffBlkInput, memDct, COLS_8x8, ROWS_8x8);

    /*
     * For intra 8x8 DCT blocks the DC has been sent as DCD
     * together with intra prediction method and Coded/Not-coded bit
     * therefore the DC resulting from MCC is deliberately skipped.
     * This is also accounted form in the later zig8x8Intra,
     * zig8x8IntraHori, zig8x8IntraVert tables.
     */
    if (mode != INTER_BLOCK)
      memDct[0] = 0;
    
    if (mode != INTER_BLOCK && complexityMode == COMPLEXITY_MODE_ANCHOR)
      nonZeros = RD_Optimized_Quantize_Coef(memDct, memCoe, SIZE_8x8, qp, chroma, mode, 
                            DCT_TRANSFORM, zigZagTable, 8, 8, 0);
    else
      nonZeros = MCCQuantCoef(memDct, memCoe, SIZE_8x8, qp, chroma, mode,
                            DCT_TRANSFORM);


    /*
     * If all coefficients are quantized to zero,
     * do refinement.
     */
    if (nonZeros == 0) {

      /*
       * Find largest coefficient value.
       */
      index = mceFindAbsMaxCoefIndex(memDct, SIZE_8x8, &sign, qp);

      if (index < 0) {

        /*
         * If largest coefficient value was smaller than QP refinement 
         * is done for the first coefficient in scanning order, 
         * set state to 1.
         */
        *memFlag = 1;
      }
      else {

        /*
         * Set refined quantized coefficient to -1 or +1.
         */
        memCoe[index] = sign;
        nonZeros = 1;
      }
    }

    /*
     * Perform inverse quantization if non-zero coefficients
     * exist.
     */
    if (nonZeros > 0) {
      MCCInvQuantCoef(memCoe, dctInvs, SIZE_8x8, qp, chroma, mode);

      /*
       * Perform actual IDCT and clip the reconstruction
       * values before calculating coding error.
       */
      varSizeInvDctFixedp(dctInvs, diffBlk, COLS_8x8, ROWS_8x8);
      mceMakeDiff(origBlkInput, predBlkInput, diffBlk, errorBlk, SIZE_8x8);
      memSe = calcBlockEnergy(errorBlk, SIZE_8x8);
    }
    else

      /*
       * All coefficients are zero, coding error is set to 
       * original prediction error.
       */
      memSe = ncEnergy;
  }

  /*
   * Select correct scanning order for bit calculation.
   */
  switch (method) {

    case METH_8x8_DCT:
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8 : zig8x8Intra;
      break;

    case METH_DCT_8x8_N_0_8:
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8Hori : zig8x8IntraHori;
      break;

    default: 
      zigZagTable = (mode == INTER_BLOCK) ? zig8x8Vert : zig8x8IntraVert;
      break;
  }

  /*
   * State is 1, refinement is done for DC coefficient.
   */
  if (*memFlag == 1) {
    for (i = 0; i < SIZE_8x8; i++)
      dctCoeffs[i] = 0;

    /*
     * Find the first coefficient and set the quantized 
     * coefficient value to -1 or +1.
     */
    index = mceFindFirstIndex(zigZagTable, SIZE_8x8);
    value = memDct[index];
    dctCoeffs[index] = value < 0 ? -1 : 1;

    /*
     * Perform inverse quantization.
     */
    MCCInvQuantCoef(dctCoeffs, dctInvs, SIZE_8x8, qp, chroma, mode);
    invValue = dctInvs[index];

    /*
     * Perform actual IDCT.
     */
    varSizeInvDctFixedp(dctInvs, diffBlk, COLS_8x8, ROWS_8x8);
  }
  else if (*memFlag == 2)

    /*
     * State is 2, use coefficients in memory.
     */
    memcpy(dctCoeffs, memCoe, sizeof(int) * SIZE_8x8);

  /*
   * Scan coefficients.
   */
  for (i = 0; i < SIZE_8x8; i++)
    quantDctCoeffs[zigZagTable[i]] = dctCoeffs[i];

  /* 
   * Calculate the amount of DCT bits.
   */
  bitsForCoef[method] = calcDctBits(dctCoeffs, COLS_8x8, ROWS_8x8, zigZagTable,
                                    qp, chroma, mode, 0);

  /* 
   * If state is 2 use coding error in memory.
   */
  if (*memFlag == 2)
    seAfter[method] = memSe;

  else {

    /*
     * State is 1, coding error must be evaluated.
     * Clip the reconstruction values and calculate coding error.
     */
    mceMakeDiff(origBlkInput, predBlkInput, diffBlk, errorBlk, SIZE_8x8);
    seAfter[method] = calcBlockEnergy(errorBlk, SIZE_8x8);
  }
}


/*
 * calcCoef:
 *
 * Parameters:
 *      dctcoef                 I: Dct coefficient pointer
 *      size                    I: Amount of dct coefficients
 *
 * Function:
 *      Calculates amount of nonzero coefficients.
 *
 * Returns:
 *      coeffs                  Amount of nonzero coefficients.
 */

static int calcCoef(int *dctcoef, int size)
{
  int i,                        /* Loop increment counter               */
      coeffs = 0;               /* Amount of nonzero coefficient        */

  for(i=0; i<size; i++) {
    if(dctcoef[i] != 0) coeffs++;
  }

  return coeffs;
}


/*
 * calcDctBits:
 *
 * Parameters:
 *      source          I: Source coefficient block pointer
 *      cols            I: Columns in coefficient block
 *      rows            I: Rows in coefficient block
 *      zig             I: Zigzag scanning order tables pointer
 *      qp              I: Quantizer parameter value 
 *      chroma          I: Mode of block (chroma/luma) 
 *      mode            I: Mode of block (intra/inter) 
 *      start           I: First coded coefficient number 
 *
 * Function:
 *      Calculates the amount of bits used in given coefficient
 *      block.
 *
 * Returns:
 *      bits            Amount of bits used.
 */

int calcDctBits(int *source, int cols, int rows, u_char *zig, int qp, 
                int chroma, int mode, int start)
{
	int 
    i,			                /* Loop increment counter */
		bits = 0,               /* Amount of bits */
		size,                   /* Size of coefficient block */
		last,                   /* Last coefficient index in scanning */
		second_last,            /* Second-last coefficient index in scanning*/
		run = 0,                /* Amount of zeros before coefficient		*/
		tableNo,                /* Number of (lrl) table */
		runBits,                /* Amount of runbits if escaped */
    levelBits,              /* Amount of levelbits if escaped */
	  escapeInd;

	vlc_item 
    *vlcTable;
	
  /*
   * Find the beginning of the inverse zigzag  scan table.
   */
	size = cols * rows;
	zig += size;

	/* 
   * Find the index (in zigzag scan order) of the last non-zero element.
   */
	last = size - 1;
	while ((last >= start) && (source[zig[last]] == 0))
		last--;


	/* 
   * Compute number of bits used to code coefficients.
   */
  if (last >= start) {    /* There is at least one coeff to code. */

		/* 
     * Translate the block parameters into index
		 * of triplet VLC table and it also return the amount of bits
		 * by which RUN and LEVEL should be encoded with.
     */
		vdxGetDctTable(cols, rows, qp, chroma, mode,
						       &tableNo, &runBits, &levelBits);
		vlcTable = &((*vectorOfvlcTable[tableNo])[0]);
		escapeInd = (*vectorOftriplet2IndexCubes[tableNo])[0][0][0];
		
		/* 
     * Find index (in zigzag scan order) of second-last non-zero element.
     */
		second_last = last-1;
		while ((second_last >= start) && (source[zig[second_last]] == 0))
			second_last--;

    if (second_last >= start) { /* There is at least two coeff to code. */

      /* 
       * Go through all non-zero coefficients in the buffer from start.
       */
			i	=	start; 
		
			/* 
       * Repeat for all coefficients up to second-last. 
       */
			while (i < second_last) {

				/* If non-zero coefficient add the variable length code bits*/
				/* otherwise add run of zeroes before coefficient.			*/
				if (source[zig[i]] != 0) {
					bits += SendDctOrKlt_EventXxX_FAST(0, run, source[zig[i]], 
														                 tableNo, runBits,
														                 levelBits,escapeInd,
														                 vlcTable);
					run = 0;
				}
				else 
					run++;

				i++;
			} 

			/* 
       * Add variable length code bits for second-last non-zero coeff.
       */
			bits += SendDctOrKlt_EventXxX_FAST(0, run, source[zig[i]], 
												                 tableNo, runBits,
												                 levelBits,escapeInd,
												                 vlcTable);

			/* 
       * Compute run of 0 between second-last and last non-zero coeff.
       */
			run = last - second_last - 1;

			/* 
       * Add variable length code bits for last non-zero coeff.
       */
			bits += SendDctOrKlt_EventXxX_FAST(	1, run, source[zig[last]], 
												tableNo, runBits,
												levelBits,escapeInd,
												vlcTable);
		}
    else {		/* There is only one coeff to code (the last one)	*/

      /*
       * Compute run of zeroes between start and last coeff.
       */
			run = last - start;

			/*
       * Add variable length code bits for last non-zero coeff.
       */
			bits += SendDctOrKlt_EventXxX_FAST(1, run, source[zig[last]], tableNo,
                                         runBits, levelBits, escapeInd, 
                                         vlcTable);
		}
	}
	
	return bits;
}


/*
 * calcKltBits:
 *
 * Parameters:
 *      coeffsVect      I: Coefficient value pointer
 *      cols            I: Columns in coefficient block
 *      rows            I: Rows in coefficient block
 *      qp              I: Quantizer parameter value 
 *      chroma          I: Mode of block (chroma/luma) 
 *      mode            I: Mode of block (intra/inter) 
 *
 * Function:
 *      Calculates the amount of bits used in given coefficient
 *      block.
 *
 * Returns:
 *      bits            Amount of bits used.
 */

int calcKltBits(int *coeffsVect, int cols, int rows, int qp, 
                       int chroma, int mode, int kltNum)
{
  int 
    i,                        /* Loop increment counter */
    bits = 0,                 /* Amount of bits */
    size,                     /* Size of coefficient block */
    last,                     /* Last coefficient flag */
    run = 0,                  /* Amount of zeros before coefficient */
    level,                    /* Absolut level of coefficient */
    coeffs,                   /* Amount of nonzero coefficients */
    tableNo,                  /* Number of TCOEFF table */
    runBits,                  /* Amount of runbits if escaped */
    levelBits,                /* Amount of levelbits if escaped */
	  escapeInd;                /* Index of escape VLC */

	vlc_item 
    *vlcTable;                /* Pointer to TCOEFF table */

  /*
   * Find the amount of coefficients.
   */
  size = cols * rows;
  coeffs = calcCoef(coeffsVect, size);

  /*
   * If nonzero coefficients found then calculate amount of bits,
   * otherwise return zero.
   */
  if (coeffs > 0) {
    vdxGetKltTable(cols, rows, qp, chroma, mode, kltNum,
                   &tableNo, &runBits, &levelBits);
		vlcTable = &((*vectorOfvlcTable[tableNo])[0]);
		escapeInd = (*vectorOftriplet2IndexCubes[tableNo])[0][0][0];

    /*
     * Go through all coefficients in the buffer.
     */
    for (i = 0; i < size; i++) {

      /*
       * For each coefficient add the variable length codes bits,
       * otherwise add run of zeros before coefficient.
       */
      if (coeffsVect[i] != 0) {
        last = coeffs == 1 ? 1 : 0;
        level = abs(coeffsVect[i]);
				bits += SendDctOrKlt_EventXxX_FAST(last, run, level,tableNo, runBits,
														               levelBits, escapeInd, vlcTable);

        if (last) 
          break;
        coeffs--;
        run = 0;
      }
      else 
        run++;
    }
  }

  return bits;
}


/*
 * mceEnc8x8Klt:
 *
 * Parameters:
 *      diffBlkInput    I: Prediction error block
 *      predBlkInput    I: Prediction block
 *      origBlkInput    I: Original block
 *      seAfter         O: SE after KLT
 *      bitsForCoef     O: Coding bits for coefficients
 *      quantKltCoeffs  O: Quantized KLT coefficients
 *      method          I: KLT method
 *      qp              I: Quantization parameter
 *      chroma          I: Chrominance flag
 *                            0 = luminance
 *                            1 = chrominance
 *      mode            I: Mode flag
 *                            0 = INTER
 *                            1 = INTRA
 *      ncEnergy        I: SE for not coded prediction error
 *
 * Function:
 *      Calculates 8x8 KLT using refinement step if quantized coefficients are
 *      initially all zero.
 *
 * Returns:
 *      Nothing.
 */

static void mceEnc8x8Klt(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantKltCoeffs,
                         int method, int qp, int chroma, int mode, int ncEnergy, int complexityMode)
{
  int
    i, tmp,                            /* Loop counters */
    kltNum, kltMet,                    /* Auxiliary variables */
    nonZeros,                          /* Amount of non-zero coefficients */
    sign,                              /* Sign of refined coefficient */
    index,                             /* Index of refined coefficient */
    kltCoef[SIZE_8x8] = {0},           /* KLT coefficients */
    invCoef[SIZE_8x8],                 /* Inverse quantized coefficients */
    recoBlkTmp[SIZE_8x8],              /* Result of IKLT */
    errorBlk[SIZE_8x8],                /* Coding error block */
    rotateFlag,                        /* Flag for block manipulation */
    diffBlkMirror[SIZE_8x8], *diffBlk, /* Rotated prediction error block */
    predBlkMirror[SIZE_8x8], *predBlk, /* Rotated prediction block */
    origBlkMirror[SIZE_8x8], *origBlk; /* Rotated original block */
 
  /*
   * Check the KLT method and mirror/rotate blocks
   * for KLT calculation.
   */
  switch (method) {

    case METH_KLT_8x8_N_1_8:
      rotateFlag = UNIT_OPER;
      kltMet = METH_KLT_8x8_N_1_8;
      break;
 
    case METH_KLT_8x8_N_2_8:
      rotateFlag = UNIT_OPER;
      kltMet = METH_KLT_8x8_N_2_8;
      break;

    case METH_KLT_8x8_N_3_8:
      rotateFlag = MIRROR_DIAG;
      kltMet = METH_KLT_8x8_N_1_8;
      break;

    case METH_KLT_8x8_N_5_8:
      rotateFlag = ROTATE_LEFT;
      kltMet = METH_KLT_8x8_N_1_8;
      break;
 
    case METH_KLT_8x8_N_6_8:
      rotateFlag = MIRROR_VERT;
      kltMet = METH_KLT_8x8_N_2_8;
      break;
 
    default:
      rotateFlag = MIRROR_VERT;
      kltMet = METH_KLT_8x8_N_1_8;
      break;
  }

  switch (rotateFlag) {

    case UNIT_OPER:
      diffBlk = diffBlkInput;
      predBlk = predBlkInput;
      origBlk = origBlkInput;
      break;

    case ROTATE_LEFT:
      for (i = 0; i < SIZE_8x8; i++) {
        tmp = rotateLeft8x8[i];
        diffBlkMirror[i] = diffBlkInput[tmp];
        predBlkMirror[i] = predBlkInput[tmp];
        origBlkMirror[i] = origBlkInput[tmp];
      }
      diffBlk = diffBlkMirror;
      predBlk = predBlkMirror;
      origBlk = origBlkMirror;
      break;

    case MIRROR_VERT:
      for (i = 0; i < SIZE_8x8; i++) {
        tmp = mirrorVert8x8[i];
        diffBlkMirror[i] = diffBlkInput[tmp];
        predBlkMirror[i] = predBlkInput[tmp];
        origBlkMirror[i] = origBlkInput[tmp];
      }
      diffBlk = diffBlkMirror;
      predBlk = predBlkMirror;
      origBlk = origBlkMirror;
      break;

    default:
      for (i = 0; i < SIZE_8x8; i++) {
        tmp = mirrorDiag8x8[i];
        diffBlkMirror[i] = diffBlkInput[tmp];
        predBlkMirror[i] = predBlkInput[tmp];
        origBlkMirror[i] = origBlkInput[tmp];
      }
      diffBlk = diffBlkMirror;
      predBlk = predBlkMirror;
      origBlk = origBlkMirror;
      break;
  }

  kltNum = kltLookUpTable[kltMet - METH_KLT_8x8_N_0_8]; 

  /*
   * Calculate KLT and quantize coefficients.
   */
  KltTransform8x8(diffBlk, kltCoef, kltNum, mode);

  if (mode != INTER_BLOCK && complexityMode == COMPLEXITY_MODE_ANCHOR)
    nonZeros = RD_Optimized_Quantize_Coef(kltCoef, quantKltCoeffs, SIZE_8x8, qp, chroma, 
                                        mode, KLT_TRANSFORM, NULL, 8, 8,kltNum);    
  else
    nonZeros = MCCQuantCoef(kltCoef, quantKltCoeffs, SIZE_8x8, qp,
                          chroma, mode, KLT_TRANSFORM);


  /*
   * If all coefficients are quantized to zero,
   * do refinement.
   */
  if(nonZeros == 0) {

    /*
     * Find largest coefficient value.
     */
    index = mceFindAbsMaxCoefIndex(kltCoef, SIZE_8x8, &sign, qp);

    if(index < 0) {

      /*
       * If largest coefficient value was smaller than QP refinement 
       * is done for the first coefficient in scanning order.
       */
      index = 0;
      sign = quantKltCoeffs[index] < 0 ? -1 : 1;
    }
    quantKltCoeffs[index] = sign;
    nonZeros = 1;
  }

  /*
   * Perform inverse quantization if non-zero coefficients
   * exist.
   */
  if(nonZeros > 0) {

    MCCInvQuantCoef(quantKltCoeffs, invCoef, SIZE_8x8, qp, chroma, mode);

    /* 
     * Calculate the amount of KLT bits.
     */
    bitsForCoef[method] = calcKltBits(quantKltCoeffs, COLS_8x8, ROWS_8x8,
                                      qp, chroma, mode, kltNum);

    /*
     * Perform actual IKLT and clip the reconstruction
     * values before calculating coding error.
     */
    KltInverseTransform8x8(invCoef, recoBlkTmp, kltNum, mode);   
    mceMakeDiff(origBlk, predBlk, recoBlkTmp, errorBlk, SIZE_8x8);
    seAfter[method] = calcBlockEnergy(errorBlk, SIZE_8x8);
  }
  else {
    bitsForCoef[method] = 0;
    seAfter[method] = ncEnergy;
  }
}


/*
 * methSelectOver4x4New:
 *
 * Parameters:
 *      predno          I: Prediction number from 4x4 block
 *      method          I: Method number which overhead is asked
 *      indexIn         I: Index of table (if known, else < 0)
 *
 * Function:
 *      Finds from table the amount of overhead bits given method and
 *      prediction number. If overhead is 1000 this method is not used
 *      in 4x4 method list for that prediction number. 
 *
 * Returns:
 *      overhead        Amount of bits used for 4x4 method selection.
 */

static int methSelectOver4x4New(int predno, int method, int indexIn)
{
  int
    index = 0,          /* Index of method on table                     */
    gen_bitstr_flag = 0,/* Don't send bits (calculate)                  */
    ovhNo,              /* Number of huffman                            */
    overhead;           /* Amount of bits used for method selection     */

  if(indexIn < 0) {
    /*
     * Scan method list one by one and check if the method in the
     * list is same as asked method.
     */
    index = methSelectNum4x4New(predno, method);
  }
  else {
    index = indexIn;
  }

  /*
   * If method is in the method selection list the amount
   * of overhead is defined, otherwise return 1000.
   */
  if(index > 0) {
    ovhNo    = overheadNumbers4x4[predno];
    overhead = SendMethSel_Event(ovhNo, index, gen_bitstr_flag);
  }
  else{   /* Method is not used */
    overhead = -1;
  }

  return overhead;
}


/*
 * methSelectOver8x8New:
 *
 * Parameters:
 *      class           I: Prediction class from 8x8 block
 *      method          I: Method number which overhead is asked
 *      indexIn         I: Index of table (if known, else < 0)
 *
 * Function:
 *      Finds from table the amount of overhead bits given method and class.
 *      If overhead is 1000 this method is not used in 8x8 method list for
 *      that class.
 *
 * Returns:
 *      overhead        Amount of bits used for 8x8 method selection.
 */

static int methSelectOver8x8New(int class, int method, int indexIn)
{
  int
    index = 0,          /* Index of method on table                     */
    gen_bitstr_flag = 0,/* Don't send bits (calculate)                  */
    ovhNo,              /* Number of huffman                            */
    overhead;           /* Amount of bits used for method selection     */

  if (method == NOT_CODED) {
    overhead = 0;
  }
  else {

    if(indexIn < 0) {
      /*
       * Scan method list one by one and check if the method in the
       * list is same as asked method.
       */
      index = methSelectNum8x8New(class, method);
    }
    else {
      index = indexIn;
    }

    /*
     * If method is in the method selection list the amount
     * of overhead is defined, otherwise return 1000.
     */
    if(index > 0) {
      ovhNo    = overheadNumbers8x8[class];
      overhead = SendMethSel_Event(ovhNo, index, gen_bitstr_flag);
    }
    else{   /* Method is not used       */
      overhead = -1;
    }

  }

  return overhead;
}



/*
 * enc8x8Block:
 *
 * Parameters:
 *      methInfo        O: 8x8 coefficients or 4x4 method info
 *      bitsForSel4x4   O: Amount of 4x4 methods selection bits
 *      bitsForMethod   O: Amount of method coding bits (coeffs)
 *      bitsForSel8x8   O: Amount of 8x8 method selection bits
 *      seAfter         O: Square error after oding
 *      dctCoeffs       O: 4x4 methods coefficients or rdc value
 *      diffBlkInput    I: Difference block
 *      qp              I: Quantization parameter value
 *      lambda          I: Lambda value
 *      methodInput     I: Information about used method(s)
 *      classInput      I: Information about blocks class
 *      predBlkInput    I: Prediction block
 *      chroma          I: Block type (luma/chroma)
 *      mode            I: Mode of block (intra/inter/first)
 *      predDirIn       I: Prediction direction (value/number)
 *      fastBest        I: Optimization or final search
 *      prevLagr        I: Previously used best lagrangian
 *      cncBits         I: Difference bits between coded/nc if known
 *
 * Function:
 *      Search best 8x8 method for given 8x8 block with given parameters.
 *
 * Returns:
 *      bestMethod      Number of best 8x8 method.
 */

int enc8x8Block(int *diffBlkInput, int *methInfo, int qp, int lambda,
                int methodInput, int classInput, int *bitsForSel4x4,
                int *bitsForMethod, int *bitsForSel8x8, int *seAfter,
                int *predBlkInput, int *dctCoeffs, int chroma, int mode,
                u_char predDirIn, int fastBest, int prevLagr, int cncBits, 
                int interFlag, int complexityMode)
{
  static int
    tmpCoeffs[AMOUNT_OF_8x8_METH * SIZE_8x8],  /* Coefficient buffer */
    info4x4[SIZE_8x8];                         /* 4x4 information buffer */

  int
    i, j,                           /* Increment loop counters */
    tmp,                            /* Temporal value */
    predNumber,                     /* Prediction class for 8x8 block */
    rotation,                       /* Rotation number of block */
    methInd,                        /* Temporal selection number index */
    methNum,                        /* Current codingmethod number */
    methLagr,                       /* Current lagrangian value */
    squareError,                    /* Current square error value */
    codingBits,                     /* Current method coding bits */
    selectionBits8x8,               /* Current 8x8 method selection bits */
    selectionBits4x4,               /* Current 4x4 methods selection bits */
    selBitsQuad,                    /* Quadrant selection bits */
    seAfterTmp[AMOUNT_OF_8x8_METH], /* Temporal error energy buffer */
    coeffBits[AMOUNT_OF_8x8_METH],  /* Temporal bits buffer */
    methDone[4] = {0},              /* Has method done earlier */
    origBlkInput[SIZE_8x8],         /* Original block */
    diffBlk[SIZE_8x8],              /* Difference block after classification */
    predBlk[SIZE_8x8],              /* Prediction block after classification */
    origBlk[SIZE_8x8],              /* Original block after classification */
    *diffBlkPtr,                    /* Difference block pointer */
    *predBlkPtr,                    /* Prediction block pointer */
    *origBlkPtr,                    /* Original block pointer */
    diffBlkRot[SIZE_8x8],           /* Left rotated diffBlk */
    predBlkRot[SIZE_8x8],           /* Left rotated predBlk */
    origBlkRot[SIZE_8x8],           /* Left rotated origBlk */
    diffBlkTmp[SIZE_8x8],           /* Temporal diff subblock */ 
    predBlkTmp[SIZE_8x8],           /* Temporal pred subblock */
    origBlkTmp[SIZE_8x8],           /* Temporal orig subblock */
    extrBlk[9],                     /* Pixels for extrapolation */
    extPels,                        /* Amount of surrounded pels */
    quadEne[5],                     /* Energy of quadrants */
    totalNumOfMethods,              /* Total number of methods to be search */
    methLimit = 3,                  /* Maximum amount of methods for limited 
                                       search */
    bestMethod,                     /* Best method number */
    bestMethodLagr,                 /* Best method lagrangian */
    prevMin,                        /* Previous minimum lagrangian */
    minLagrLimit,                   /* Lagrangian limit which method have 
                                       to win */
    ncEnergy,                       /* Not coded energy (square error) */
    predMeths,                      /* Amount of methods in prediction list */
    memFlag = 0,                    /* Are DCT coeffs in memory */
    compLimit;
	
  u_char
    methPtr[AMOUNT_OF_8x8_METH],    /* Search method buffer */
    selNPtr[AMOUNT_OF_8x8_METH] =   /* Selection number buffer */
      {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};

  /*
   * Classify the block according to coding mode (INTER,INTRA), QP 
   * and component (Y,u/v). INTRA blocks are also classified according
   * to directionality of the prediction block using gradient based
   * classification. 
   */
  if (predDirIn == UCHAR_MAX && mode != INTER_BLOCK)
    blockEdge(predBlkInput, COLS_8x8, ROWS_8x8, &predDirIn);

  predNumber = changeClassNew(classInput, predDirIn, qp,
                              chroma, mode, &rotation);

  /*
   * Rotate the blocks if needed. Rotation is used to fit the error block
   * to one of the predefined coding method subsets.
   */
  if (rotation == UNIT_OPER) {
    predBlkPtr = predBlkInput;
    diffBlkPtr = diffBlkInput;
    for (i = 0; i < SIZE_8x8; i++)
      origBlkInput[i] = predBlkInput[i] + diffBlkInput[i];
 
    origBlkPtr = origBlkInput;
  }
  else {
    copyBlock8x8(predBlkInput, predBlk, rotation);
    copyBlock8x8(diffBlkInput, diffBlk, rotation);
    predBlkPtr = predBlk;
    diffBlkPtr = diffBlk;
    for (i = 0; i < SIZE_8x8; i++)
      origBlk[i] = predBlk[i] + diffBlk[i];

    origBlkPtr = origBlk;
  }
  for (i = 0; i < SIZE_8x8; i++) {
    tmp = rotateLeft8x8[i];
    predBlkRot[i] = predBlkPtr[tmp];
    diffBlkRot[i] = diffBlkPtr[tmp];
    origBlkRot[i] = origBlkPtr[tmp];
  }

  /* 
   * Calculate error energies for each 4x4 quadrant of the
   * 8x8 block and total error energy for NOT_CODED 8x8 block.
   */
  getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
              START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);
  quadEne[0] = calcBlockEnergy(diffBlkTmp, SIZE_4x4);
  getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
              START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);
  quadEne[1] = calcBlockEnergy(diffBlkTmp, SIZE_4x4);
  getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
              START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);
  quadEne[2] = calcBlockEnergy(diffBlkTmp, SIZE_4x4);
  getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
              START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
  quadEne[3] = calcBlockEnergy(diffBlkTmp, SIZE_4x4);
  ncEnergy = quadEne[0] + quadEne[1] + quadEne[2] + quadEne[3];

  /*
   * Initialize variables for NOT_CODED method, which is used as
   * original reference.
   */
  bestMethod = NOT_CODED;
  *bitsForSel8x8 = *bitsForSel4x4 = *bitsForMethod = 0;
  *seAfter = ncEnergy;
  bestMethodLagr = *seAfter - lambda * cncBits;

  /* 
   * If possible previous method is better than NOT_CODED,
   * use that as original reference.
   */
  if (prevLagr < bestMethodLagr) {
    bestMethod = -1;
    *seAfter = bestMethodLagr = prevLagr;
  }
  /*
   * Check the amount of coding methods in the subset
   * and set the counter to first method.
   */
  tmp = predNumber * (MAX_OF_8x8_METH_OVH + 1);
  predMeths = overheadOrder8x8[tmp];
  tmp++;

  /*
   * Several methods in the method list are tried. Search
   * the methods from the method list.
   */
  if (methodInput > AMOUNT_OF_8x8_METH) {

    /*
     * Check if limited search of coding methods is needed. Limited
     * search is used when the best INTRA prediction method is 
     * searched for.
     */
    if (!fastBest || predMeths < methLimit) 
      switch (complexityMode) {

        case COMPLEXITY_MODE_ANCHOR:
          compLimit = predMeths;
          break;

        default:
          compLimit = predMeths >> 1;
          break;
      }

    else
      switch (complexityMode) {

        case COMPLEXITY_MODE_ANCHOR:
          compLimit = methLimit;
          break;

        default:
          compLimit = methLimit >> 1;
          break;
      }

    /*
     * Copy the selected coding methods to search buffer.
     */
    memcpy(methPtr, overheadOrder8x8 + tmp, compLimit);
    totalNumOfMethods = compLimit;
  }
  else {

    /*
     * Only one method is tried out. Find the index of the method
     * in the method selection list.
     */
    methNum = methodInput;
    methInd = methSelectNum8x8New(predNumber, methNum);
    if (methInd > 0) {
      methPtr[0] = methodInput;
      selNPtr[0] = methInd;
      totalNumOfMethods = 1;
    }
  }

  /* 
   * Scan through all the methods in the search buffer.
   */
  for (j = 0; j < totalNumOfMethods; j++) {

    /*
     * Pick up the jth method from the buffer and find 
     * out the 8x8 method selection bits.
     */
    methNum = methPtr[j];
    methInd = selNPtr[j];
    selectionBits8x8 = methSelectOver8x8New(predNumber, methNum, methInd);

    /*  
     * Set the minimum Lagrangian cost so far for
     * cost margin evaluation. 
     */
    prevMin = bestMethodLagr;

    /* 
     * Initialize values for the jth coding method.
     */
    squareError = codingBits = selectionBits4x4 = 0;
    methLagr = lambda * selectionBits8x8;
    minLagrLimit = bestMethodLagr - methLagr;

    /*
     * Perform coding.
     */
    switch(methNum) {

      /*
       * 8x8 DCT.
       */
      case METH_DCT_8x8_N_0_8: case METH_DCT_8x8_N_4_8: 
      case METH_8x8_DCT:

        /*
         * Update the cost margin by the minimum number of bits
         * needed for DCT coefficients and check if some cost
         * margin still exists.
         */
        minLagrLimit -= lambda * MCE_CRITICAL_BITS_8x8DCT;
        if (minLagrLimit > 0) {

          /*
           * Try to code the block using 8x8 DCT.
           */
          mceEnc8x8Dct(diffBlkPtr, predBlkPtr, origBlkPtr, seAfterTmp,
                       coeffBits, tmpCoeffs + methNum * SIZE_8x8, methNum,
                       qp, chroma, mode, ncEnergy, &memFlag, complexityMode);

          /*
           * Evaluate SE, coding bits and Lagrangian cost if 8x8 DCT coded
           * something.
           */
          if (coeffBits[methNum] > 0) {
            squareError += seAfterTmp[methNum];
            codingBits += coeffBits[methNum];
            methLagr += seAfterTmp[methNum] + lambda * coeffBits[methNum];
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        break;

      /* 
       * 8x8 KLT.
       */
      case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_2_8:
      case METH_KLT_8x8_N_3_8: case METH_KLT_8x8_N_5_8:
      case METH_KLT_8x8_N_6_8: case METH_KLT_8x8_N_7_8:

        /*
         * Update the cost margin by the minimum number of bits
         * needed for KLT coefficients and check if some cost
         * margin still exists.
         */
        minLagrLimit -= lambda * MCE_CRITICAL_BITS_8x8KLT;
        if (minLagrLimit > 0) {

          /*
           * Try to code the block with 8x8 KLT.
           */
          mceEnc8x8Klt(diffBlkPtr, predBlkPtr, origBlkPtr, seAfterTmp,
                       coeffBits, tmpCoeffs + methNum * SIZE_8x8, methNum,
                       qp, chroma, mode, ncEnergy, complexityMode);

          /*
           * Evaluate SE, coding bits and Lagrangian cost if 8x8 KLT coded
           * something.
           */
          if (coeffBits[methNum] > 0) {
            squareError += seAfterTmp[methNum];
            codingBits += coeffBits[methNum];
            methLagr += seAfterTmp[methNum] + lambda * coeffBits[methNum];
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        break;

      /* 
       * 8x4 DCT applied for left or/and right half of the block.
       */
      case METH_4x8_DCT_1: case METH_4x8_DCT_2: case METH_4x8_DCT_3:

        /*
         * Evaluate SE for the half that is not coded.
         */
        if (methNum == METH_4x8_DCT_2) {
          squareError += tmp = quadEne[0] + quadEne[2];
          methLagr += tmp;
        }
        else if (methNum == METH_4x8_DCT_1) {
          squareError += tmp = quadEne[1] + quadEne[3];
          methLagr += tmp;
        }

        /*
         * Calculate 8x4 DCT for the left half.
         */
        if (methNum != METH_4x8_DCT_2) {

          /*
           * Update the cost margin by the minimum number of bits
           * needed for DCT coefficients and check if some cost
           * margin still exists.
           */
          if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {

            /*
             * Check.
             */
            if(prevMin > methDone[0]) {

              /*
               * Try to code with 8x4 DCT.
               */
              getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              getIntBlock(predBlkPtr, predBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              getIntBlock(origBlkPtr, origBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp, seAfterTmp,
                           coeffBits, tmpCoeffs + METH_4x8_DCT_1 * SIZE_8x8,
                           METH_4x8_DCT_1, qp, COLS_4x8, ROWS_4x8, zig4x8,
                           chroma, mode, complexityMode);
              methDone[0] = prevMin;
            }

            /*
             * Evaluate SE, coding bits and Lagrangian cost if 8x4 DCT coded
             * something.
             */
            if (coeffBits[METH_4x8_DCT_1] > 0) {
              squareError += seAfterTmp[METH_4x8_DCT_1];
              codingBits += coeffBits[METH_4x8_DCT_1];
              methLagr += seAfterTmp[METH_4x8_DCT_1] +
                          lambda * coeffBits[METH_4x8_DCT_1];
            }
            else
              methLagr = MCE_BIG_LAGRANGIAN;
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }

        /*
         * Calculate 8x4 DCT for the right half.
         */
        if (methNum != METH_4x8_DCT_1) {
          if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {
            if (prevMin > methDone[1]) {

              /*
               * Try to code with 8x4 DCT.
               */
              getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              getIntBlock(predBlkPtr, predBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              getIntBlock(origBlkPtr, origBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp, seAfterTmp,
                           coeffBits, tmpCoeffs + METH_4x8_DCT_2 * SIZE_8x8,
                           METH_4x8_DCT_2, qp, COLS_4x8, ROWS_4x8, zig4x8,
                           chroma, mode, complexityMode);
              methDone[1] = prevMin;
            }

            /*
             * Evaluate coding bits and Lagrangian cost if 8x4 DCT coded
             * something.
             */
            if (coeffBits[METH_4x8_DCT_2] > 0) {
              squareError += seAfterTmp[METH_4x8_DCT_2];
              codingBits += coeffBits[METH_4x8_DCT_2];
              methLagr += seAfterTmp[METH_4x8_DCT_2] +
                          lambda * coeffBits[METH_4x8_DCT_2];
            }
            else
              methLagr = MCE_BIG_LAGRANGIAN;
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        break;

      /* 
       * 8x4 DCT applied for vertical middle part of the block.
       */
      case METH_4x8_DCT_4:

        /*
         * Evaluate SE for the part that is not coded.
         */
        getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x8, ROWS_4x8,
                          2, 0, ROWS_8x8);
        quadEne[4] = calcBlockEnergy(diffBlkTmp, SIZE_4x8);
        squareError += tmp = ncEnergy - quadEne[4];
        methLagr += tmp;

        if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {

          /*
           * Try to code with 8x4 DCT.
           */
          getIntBlock(predBlkPtr, predBlkTmp, COLS_4x8, ROWS_4x8,
                      2, 0, ROWS_8x8);
          getIntBlock(origBlkPtr, origBlkTmp, COLS_4x8, ROWS_4x8,
                      2, 0, ROWS_8x8);
          mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp,
                       seAfterTmp, coeffBits, tmpCoeffs + methNum * SIZE_8x8,
                       methNum, qp, COLS_4x8, ROWS_4x8, zig4x8, chroma, mode, complexityMode);

          /*
           * Evaluate SE, coding bits and Lagrangian cost if 8x4 DCT coded
           * something.
           */
          if (coeffBits[methNum] > 0) {
            squareError += seAfterTmp[methNum];
            codingBits += coeffBits[methNum];
            methLagr += seAfterTmp[methNum] + lambda * coeffBits[methNum];
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        break;

      /* 
       * 8x4 DCT applied for upper or/and lower half of the block.
       */
      case METH_8x4_DCT_1: case METH_8x4_DCT_2: case METH_8x4_DCT_3: 

        /*
         * Evaluate SE for the half that is not coded.
         */
        if (methNum == METH_8x4_DCT_2) {
          squareError += tmp = quadEne[0] + quadEne[1];
          methLagr += tmp;
        }
        else if (methNum == METH_8x4_DCT_1) {
          squareError += tmp = quadEne[2] + quadEne[3];
          methLagr += tmp;
        }

        /*
         * Calculate 8x4 DCT for the upper half.
         */
        if (methNum != METH_8x4_DCT_2) {
          if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {
            if (prevMin > methDone[2]) {

              /*
               * Try to code with 8x4 DCT.
               */
              getIntBlock(diffBlkRot, diffBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              getIntBlock(predBlkRot, predBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              getIntBlock(origBlkRot, origBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_1, START_ROW_4x8_1, ROWS_8x8);
              mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp, seAfterTmp,
                           coeffBits, tmpCoeffs + METH_8x4_DCT_1 * SIZE_8x8,
                           METH_8x4_DCT_1, qp, COLS_4x8, ROWS_4x8, zig4x8, 
                           chroma, mode, complexityMode);
              methDone[2] = prevMin;
            }

            /*
             * Evaluate SE, coding bits and Lagrangian cost if 8x4 DCT coded
             * something.
             */
            if (coeffBits[METH_8x4_DCT_1] > 0) {
              squareError += seAfterTmp[METH_8x4_DCT_1];
              codingBits += coeffBits[METH_8x4_DCT_1];
              methLagr += seAfterTmp[METH_8x4_DCT_1] +
                          lambda * coeffBits[METH_8x4_DCT_1];
            }
            else
              methLagr = MCE_BIG_LAGRANGIAN;
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }

        /*
         * Calculate 8x4 DCT for the lower half.
         */
        if (methNum != METH_8x4_DCT_1) {
          if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {
            if (prevMin > methDone[3]) {

              /*
               * Try to code with 8x4 DCT.
               */
              getIntBlock(diffBlkRot, diffBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              getIntBlock(predBlkRot, predBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              getIntBlock(origBlkRot, origBlkTmp, COLS_4x8, ROWS_4x8,
                          START_COL_4x8_2, START_ROW_4x8_2, ROWS_8x8);
              mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp, seAfterTmp,
                           coeffBits, tmpCoeffs + METH_8x4_DCT_2 * SIZE_8x8,
                           METH_8x4_DCT_2, qp, COLS_4x8, ROWS_4x8, zig4x8, 
                           chroma, mode, complexityMode);
              
              methDone[3] = prevMin;
            }

            /*
             * Evaluate SE, coding bits and Lagrangian cost if 8x4 DCT coded
             * something.
             */
            if (coeffBits[METH_8x4_DCT_2] > 0) {
              squareError += seAfterTmp[METH_8x4_DCT_2];
              codingBits += coeffBits[METH_8x4_DCT_2];
              methLagr += seAfterTmp[METH_8x4_DCT_2] +
                          lambda * coeffBits[METH_8x4_DCT_2];
            }
            else
              methLagr = MCE_BIG_LAGRANGIAN;
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        break;

      /* 
       * 8x4 DCT applied for horizontal middle part of the block.
       */
      case METH_8x4_DCT_4:

        /*
         * Evaluate SE for the part that is not coded.
         */
        getIntBlock(diffBlkRot, diffBlkTmp, COLS_4x8, ROWS_4x8,
                    2, 0, ROWS_8x8);
        quadEne[4] = calcBlockEnergy(diffBlkTmp, SIZE_4x8);
        squareError += tmp = ncEnergy - quadEne[4];
        methLagr += tmp;

        if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x8DCT) {

          /*
           * Try to code with 8x4 DCT.
           */
          getIntBlock(predBlkRot, predBlkTmp, COLS_4x8, ROWS_4x8,
                      2, 0, ROWS_8x8);
          getIntBlock(origBlkRot, origBlkTmp, COLS_4x8, ROWS_4x8,
                      2, 0, ROWS_8x8);
          mceEnc8x4Dct(diffBlkTmp, predBlkTmp, origBlkTmp, seAfterTmp,
                       coeffBits, tmpCoeffs + methNum * SIZE_8x8,
                       methNum, qp, COLS_4x8, ROWS_4x8, zig4x8, chroma, mode, complexityMode);

          /*
           * Evaluate SE, coding bits and Lagrangian cost if 8x4 DCT coded
           * something.
           */
          if (coeffBits[methNum] > 0) {
            squareError += seAfterTmp[methNum];
            codingBits += coeffBits[methNum];
            methLagr += seAfterTmp[methNum] + lambda * coeffBits[methNum];
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        break;

      /*
       * 8x4 extrapolation applied for upper half of the block.
       */
      case METH_EXTRAPOL_12: 

        /*
         * Evaluate SE and Lagrangian cost for the part that is not 
         * extrapolated.
         */
        squareError += tmp = quadEne[2] + quadEne[3];
        methLagr += tmp;

        /*
         * Check if cost margin exists and extrapolate.
         */
        if (methLagr < prevMin) {
          seAfterTmp[methNum] = mceEncExtrapol(predBlkPtr, origBlkPtr,
                                               methNum);
          squareError += seAfterTmp[methNum];
          methLagr += seAfterTmp[methNum];
        }
        break;
      
      /*
       * 8x4 extrapolation applied for left half of the block.
       */
      case METH_EXTRAPOL_14:

        /*
         * Evaluate SE and Lagrangian cost for the part that is not 
         * extrapolated.
         */
        squareError += tmp = quadEne[1] + quadEne[3];
        methLagr += tmp;

        /*
         * Check if cost margin exists and extrapolate.
         */
        if (methLagr < prevMin) {
          seAfterTmp[methNum] = mceEncExtrapol(predBlkPtr, origBlkPtr,
                                               methNum);
          squareError += seAfterTmp[methNum];
          methLagr += seAfterTmp[methNum];
        }
        break;

      /*
       * 8x4 extrapolation applied for right half of the block.
       */
      case METH_EXTRAPOL_28: 
    
        /*
         * Evaluate SE and Lagrangian cost for the part that is not
         * extrapolated.
         */
        squareError += tmp = quadEne[0] + quadEne[2];
        methLagr += tmp;

        /*
         * Check if cost margin exists and extrapolate.
         */
        if (methLagr < prevMin) {
          seAfterTmp[methNum] = mceEncExtrapol(predBlkPtr, origBlkPtr,
                                               methNum);
          squareError += seAfterTmp[methNum];
          methLagr += seAfterTmp[methNum];
        }
        break;

      /*
       * 8x4 extrapolation applied for lower half of the block.
       */
      case METH_EXTRAPOL_48:

        /*
         * Evaluate SE and Lagrangian cost for the part that is not 
         * extrapolated.
         */
        squareError += tmp = quadEne[0] + quadEne[1];
        methLagr += tmp;

        /*
         * Check if cost margin exists and extrapolate.
         */
        if (methLagr < prevMin) {
          seAfterTmp[methNum] = mceEncExtrapol(predBlkPtr, origBlkPtr,
                                               methNum);
          squareError += seAfterTmp[methNum];
          methLagr += seAfterTmp[methNum];
        }
        break;

      /*
       * 4x4 quadrants of the 8x8 block coded separately.
       */
      case METH_4x4_15:

        /*
         * Upper left quadrant.
         */
        if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x4CBS - 
            MCE_MIN_IMPROVEMENT_4x4) {

          /*
           * Try to code the error using 4x4 method.
           */
          getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);
          getIntBlock(predBlkPtr, predBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);
          getIntBlock(origBlkPtr, origBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);
          extPels = getExtSou8x8(predBlkPtr, extrBlk, QUAD_1);
          selBitsQuad = enc4x4Block(diffBlkTmp, predBlkTmp, origBlkTmp,
                                    seAfterTmp, coeffBits,info4x4,
                                    METH_4x4_1, lambda, qp,
                                    tmpCoeffs + METH_4x4_1 * SIZE_8x8, 
                                    extrBlk,
                                    extPels, chroma, mode, prevMin,
                                    SIZE_8x8, fastBest, 
                                    0, 0, quadEne[0], interFlag, 
                                    complexityMode);

          /*
           * Evaluate SE, coding bits and Lagranfian cost if
           * 4x4 coding was able to code  something.
           */
          if (selBitsQuad != -1) {
            squareError += seAfterTmp[METH_4x4_1];
            codingBits += coeffBits[METH_4x4_1];
            selectionBits4x4 += selBitsQuad;
            methLagr += seAfterTmp[METH_4x4_1] + lambda *
                        (coeffBits[METH_4x4_1] + selBitsQuad);
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        /*
         * Upper right quadrant.
         */
        if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x4CBS - 
            MCE_MIN_IMPROVEMENT_4x4) {

          /*
           * Try to code the error using 4x4 method.
           */
          getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);
          getIntBlock(predBlkPtr, predBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);
          getIntBlock(origBlkPtr, origBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);
          extPels = getExtSou8x8(predBlkPtr, extrBlk, QUAD_2);
          selBitsQuad = enc4x4Block(diffBlkTmp, predBlkTmp, origBlkTmp,
                                    seAfterTmp, coeffBits, info4x4 + 16,
                                    METH_4x4_2, lambda, qp,
                                    tmpCoeffs + METH_4x4_2 * SIZE_8x8, 
                                    extrBlk,
                                    extPels, chroma, mode, prevMin,
                                    SIZE_8x8, fastBest, 
                                    0, 0, quadEne[1], interFlag,
                                    complexityMode);

          /*
           * Evaluate SE, coding bits and Lagranfian cost if
           * 4x4 coding was able to code  something.
           */
          if (selBitsQuad != -1) {
            squareError += seAfterTmp[METH_4x4_2];
            codingBits += coeffBits[METH_4x4_2];
            selectionBits4x4 += selBitsQuad;
            methLagr += seAfterTmp[METH_4x4_2] + lambda *
                        (coeffBits[METH_4x4_2] + selBitsQuad);
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        /*
         * Bottom left quadrant.
         */
        if (methLagr < prevMin - lambda*MCE_CRITICAL_BITS_4x4CBS - 
            MCE_MIN_IMPROVEMENT_4x4) {

          /*
           * Try to code the error using 4x4 method.
           */
          getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);
          getIntBlock(predBlkPtr, predBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);
          getIntBlock(origBlkPtr, origBlkTmp, COLS_4x4, ROWS_4x4,
                      START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);
          extPels = getExtSou8x8(predBlkPtr, extrBlk, QUAD_3);
          selBitsQuad = enc4x4Block(diffBlkTmp, predBlkTmp, origBlkTmp,
                                    seAfterTmp, coeffBits, info4x4 + 32,
                                    METH_4x4_4, lambda, qp,
                                    tmpCoeffs + METH_4x4_4 * SIZE_8x8, 
                                    extrBlk,
                                    extPels, chroma, mode, prevMin,
                                    SIZE_8x8, fastBest, 
                                    0, 0, quadEne[2], interFlag, 
                                    complexityMode);

          /*
           * Evaluate SE, coding bits and Lagranfian cost if
           * 4x4 coding was able to code  something.
           */
          if (selBitsQuad != -1) {
            squareError += seAfterTmp[METH_4x4_4];
            codingBits += coeffBits[METH_4x4_4];
            selectionBits4x4 += selBitsQuad;
            methLagr += seAfterTmp[METH_4x4_4] + lambda *
                        (coeffBits[METH_4x4_4] + selBitsQuad);
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        /* 
         * Bottom right quadrant. 
         */
        if (methLagr < prevMin - lambda * MCE_CRITICAL_BITS_4x4CBS - 
            MCE_MIN_IMPROVEMENT_4x4) {

          /*
           * Try to code the error using 4x4 method.
           */
          getIntBlock(diffBlkPtr, diffBlkTmp, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
          getIntBlock(predBlkPtr, predBlkTmp, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
          getIntBlock(origBlkPtr, origBlkTmp, COLS_4x4, ROWS_4x4,
                  START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
          extPels = getExtSou8x8(predBlkPtr, extrBlk, QUAD_4);
          selBitsQuad = enc4x4Block(diffBlkTmp, predBlkTmp, origBlkTmp,
                                    seAfterTmp, coeffBits, info4x4 + 48,
                                    METH_4x4_8, lambda, qp,
                                    tmpCoeffs + METH_4x4_8 * SIZE_8x8, 
                                    extrBlk,
                                    extPels, chroma, mode, prevMin,
                                    SIZE_8x8, fastBest, 
                                    0, 0, quadEne[3], interFlag,
                                    complexityMode);

          /*
           * Evaluate SE, coding bits and Lagranfian cost if
           * 4x4 coding was able to code  something.
           */
          if (selBitsQuad != -1) {
            squareError += seAfterTmp[METH_4x4_8];
            codingBits += coeffBits[METH_4x4_8];
            selectionBits4x4 += selBitsQuad;
            methLagr += seAfterTmp[METH_4x4_8] + lambda *
                        (coeffBits[METH_4x4_8] + selBitsQuad);
          }
          else
            methLagr = MCE_BIG_LAGRANGIAN;
        }
        else
          methLagr = MCE_BIG_LAGRANGIAN;

        break;
    }

    /* 
     * If the method is better than the previous one,
     * update method information.
     */
    if (methLagr < bestMethodLagr) {
      bestMethod = methNum;
      *bitsForSel8x8 = selectionBits8x8;
      *bitsForSel4x4 = selectionBits4x4;
      *bitsForMethod = codingBits;
      *seAfter = squareError;
      bestMethodLagr = methLagr;
    }
  }

  /*
   * Store the information about the best method.
   * For 8x8 and 8x4 tranform coding methods, coefficients
   * are stored in methInfo.  For 4x4 coding, suplementary informaton 
   * (e.g. class, coding method is stored in methInfo and possible
   * 4x4 transform coefficients are stored in dctCoeffs. 
   */
  for (i = 0; i < SIZE_8x8; i++) 
    methInfo[i] = dctCoeffs[i] = 0; 

  switch(bestMethod) {

    case -1: case 0:
    case METH_EXTRAPOL_12: case METH_EXTRAPOL_14:
    case METH_EXTRAPOL_28: case METH_EXTRAPOL_48:
      break;

    case METH_DCT_8x8_N_0_8 : case METH_DCT_8x8_N_4_8 : 
    case METH_8x8_DCT :
      memcpy(methInfo, tmpCoeffs + bestMethod * SIZE_8x8, 
             sizeof(int) * SIZE_8x8);
      break;

    case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_2_8:
    case METH_KLT_8x8_N_3_8: case METH_KLT_8x8_N_5_8:
    case METH_KLT_8x8_N_6_8: case METH_KLT_8x8_N_7_8:
      dctCoeffs[1] = tmpCoeffs[bestMethod * SIZE_8x8 + 63];
      memcpy(methInfo, tmpCoeffs + bestMethod * SIZE_8x8, 
             sizeof(int) * 63);
      methInfo[63] = 0;
      break;

    case METH_4x8_DCT_1: case METH_8x4_DCT_1:
      for (i = 0; i < SIZE_8x4; i++) 
        methInfo[zig4x8[i]] = tmpCoeffs[bestMethod * SIZE_8x8 + i];
      break;

    case METH_4x8_DCT_2: case METH_8x4_DCT_2:
      for (i = 0; i < SIZE_8x4; i++) 
        methInfo[SIZE_8x4 + zig4x8[i]] = tmpCoeffs[bestMethod * SIZE_8x8 + i];
      break;

    case METH_4x8_DCT_4: case METH_8x4_DCT_4:
      for (i = 0; i < SIZE_8x4; i++) 
        methInfo[zig4x8[i]] = tmpCoeffs[bestMethod * SIZE_8x8 + i];
      break;

    case METH_4x8_DCT_3:
      for (i = 0; i < SIZE_8x4; i++) {
        methInfo[zig4x8[i]] = tmpCoeffs[METH_4x8_DCT_1 * SIZE_8x8 + i];
        methInfo[SIZE_8x4 + zig4x8[i]] = tmpCoeffs[METH_4x8_DCT_2 * SIZE_8x8 + i];
      }
      break;

    case METH_8x4_DCT_3:
      for (i = 0; i < SIZE_8x4; i++) {
        methInfo[zig4x8[i]] = tmpCoeffs[METH_8x4_DCT_1 * SIZE_8x8 + i];
        methInfo[SIZE_8x4 + zig4x8[i]] = tmpCoeffs[METH_8x4_DCT_2 * SIZE_8x8 + i];
      }
      break;

    case METH_4x4_15:
      memcpy(dctCoeffs, tmpCoeffs + METH_4x4_1 * SIZE_8x8, 
             sizeof(int) * SIZE_4x4);
      memcpy(dctCoeffs + 16, tmpCoeffs + METH_4x4_2 * SIZE_8x8, 
             sizeof(int) * SIZE_4x4);
      memcpy(dctCoeffs + 32, tmpCoeffs + METH_4x4_4 * SIZE_8x8, 
             sizeof(int) * SIZE_4x4);
      memcpy(dctCoeffs + 48, tmpCoeffs + METH_4x4_8 * SIZE_8x8, 
             sizeof(int) * SIZE_4x4);
      memcpy(methInfo, info4x4, sizeof(int) * SIZE_8x8);
      break;
  }

  return bestMethod;
}


/*
 * enc4x4Block:
 *
 * Parameters:
 *      diffBlk         I: Difference block
 *      predBlk         I: Prediction block
 *      origBlk         I: Original block
 *      seAfter         O: Square error after oding
 *      codingBits      O: Amount of method coding bits (coeffs)
 *      info4x4         O: 4x4 method information
 *      methodInput     I: Number of output buffers
 *      lambda          I: Lambda value
 *      qp              I: Quantization parameter value
 *      dctCoeffs       O: 4x4 methods coefficients
 *      extrBlk         I: Surrounded pels for extrapolation buffer
 *      extPels         I: Amount of surrounded pels
 *      chroma          I: Block type (luma/chroma)
 *      mode            I: Mode of block (intra/inter/first)
 *      prevLagr        I: Previously used best lagrangian
 *      blkSize         I: Size of coded block (8x8/4x4)
 *      fastBest        I: Optimization or final search
 *      predDir         I: Intra 4x4 prediction direction number
 *      cncBits         I: Difference bits between coded/nc if known
 *      ncEnergy        I: Energy of difference block if known
 *
 * Function:
 *      Search best 4x4 method for given 4x4 block with given parameters.
 *
 * Returns:
 *      Amount 4x4 method selection bits or -1 if 4x4 coding does
 *      not provide better Lagrangian cost than previous method.
 */

int enc4x4Block(int *diffBlk, int *predBlk, int *origBlk, int *seAfter,
                int *codingBits, int *info4x4, int methodInput, int lambda,
                int qp, int *coeffs, int *extrBlk, int extPels,
                int chroma, int mode, int prevLagr, int blkSize, int fastBest,
                int predDir, int cncBits, int ncEnergy, int interFlag, 
                int complexityMode)
{
  int
    i, j,                         /* Loop counters */
    tmp,                          /* Temporal value */
    predNumber,                   /* Prediction class for 4x4 block */
    rotation,                     /* Rotation number of block */
    methInd,                      /* Temporal selection number index */
    methNum,                      /* Current codingmethod number */
    methLagr,                     /* Current lagrangian value */
    squareError,                  /* Current square error value */
    coeffBits,                    /* Temporal method coding bits */
    bestSelBits,                  /* Current 4x4 methods selection bits */
    selBits,                      /* Temporal selectionBits buffer */
    diffBlkRot[SIZE_4x4],         /* Difference block after classification */
    predBlkRot[SIZE_4x4],         /* Prediction block after classification */
    origBlkRot[SIZE_4x4],         /* Original block after classification */
    *diffBlkPtr,                  /* Difference block pointer */
    *predBlkPtr,                  /* Prediction block pointer */
    *origBlkPtr,                  /* Original block pointer */
    totalNumOfMethods,            /* Total number of methods to be search */
    methLimit = 3,                /* Maximum amount of methods for limited 
                                     search */
    bestMethod,                   /* Best method number */
    bestMethodLagr,               /* Best method lagrangian */
    bestDcError,                  /* Index of best method dc-difference */
    minLagrLimit,                 /* Lagrangian cost margin */ 
    errorBlk[SIZE_4x4],           /* Error Block after reco and clip */
    tmpCoeffs[SIZE_4x4],          /* Temporal 4x4 methods coefficients */
    dcBits,                       /* Bits used for dc removing */
    kltNum,                       /* Number of used KLT */
    *diffBlockPtr,                /* Pointer of used difference block */
    *predBlockPtr,                /* Pointer of used prediction block */
    *origBlockPtr,                /* Pointer of used original block */
    mirrored = 0,                 /* Is diff block done */
    predBlkRotateLeft[SIZE_4x4],  /* Prediction block buffer */
    predBlkMirrorDiag[SIZE_4x4],  /* Prediction block buffer */
    predBlkMirrorVert[SIZE_4x4],  /* Prediction block buffer */
    origBlkRotateLeft[SIZE_4x4],  /* Original block buffer */
    origBlkMirrorDiag[SIZE_4x4],  /* Original block buffer */
    origBlkMirrorVert[SIZE_4x4],  /* Original block buffer */
    diffBlkRotateLeft[SIZE_4x4],  /* Difference block buffer */
    diffBlkMirrorDiag[SIZE_4x4],  /* Difference block buffer */
    diffBlkMirrorVert[SIZE_4x4],  /* Difference block buffer */
    predMeths,                    /* Amount of methods in prediction list */
    memFlag = 0,                  /* Are DCT coeffs in memory */
    compLimit;

  u_char
    methPtr[AMOUNT_OF_4x4_METH],  /* Search method buffer */
    selNPtr[AMOUNT_OF_4x4_METH] = /* Selection number buffer */
      {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12};

  /* 
   * Initialize 4x4 method information and calculate
   * NOT_CODED energy if needed.
   */
  for (i = 0; i < SIZE_4x4; i++)
    info4x4[i] = 0;

  bestMethod = AMOUNT_OF_4x4_METH;
  bestDcError = bestSelBits = codingBits[methodInput] = 0;
  bestMethodLagr = seAfter[methodInput] = prevLagr + 1;
  if (ncEnergy < 0)
    ncEnergy = calcBlockEnergy(diffBlk, SIZE_4x4);

  /*
   * If the 4x4 block is chrominance block corresponding to
   * single 8x8 luminance block, previous coding method does
   * not exist. In this case, CNC information can be sent
   * separately and the reference Lagrangian cost for
   * coding is set to NOT_CODED energy with CNC bits subtracted.
   */
  if (blkSize == SIZE_4x4) {
    tmp = ncEnergy - lambda * cncBits;
    if (bestMethodLagr > tmp)
      bestMethodLagr = tmp;
  }
  /*
   * Classify the block according to coding mode (INTER,INTRA), QP 
   * and component (Y,u/v). INTRA blocks are also classified according
   * to directionality of the prediction block using gradient based
   * classification. 
   */
  predNumber = findPredNum4x4(predBlk, qp, chroma, mode, blkSize,
                              predDir, &rotation);

  /*
   * Rotate the blocks if needed. Rotation is used to fit the error block
   * to one of the predefined coding method subsets.
   */
  if (rotation == UNIT_OPER) {
    predBlkPtr = predBlk;
    diffBlkPtr = diffBlk;
    origBlkPtr = origBlk;
  }
  else {
    copyBlock4x4(predBlk, predBlkRot, rotation);
    copyBlock4x4(diffBlk, diffBlkRot, rotation);
    predBlkPtr = predBlkRot;
    diffBlkPtr = diffBlkRot;
    for (i = 0; i < SIZE_4x4; i++)
      origBlkRot[i] = predBlkPtr[i] + diffBlkPtr[i];

    origBlkPtr = origBlkRot;
  }

  /*
   * Find the correct subset of coding methods based on the 
   * classification result.
   */
  tmp = predNumber * (MAX_OF_4x4_METH_OVH + 1);
  predMeths = overheadOrder4x4[tmp];
  tmp++;

  /*
   * Check if limited search of coding methods is needed. Limited
   * search is used when the best INTRA prediction method is 
   * searched for.
   */
  if (!fastBest || predMeths < methLimit)
    switch (complexityMode) {

      case COMPLEXITY_MODE_ANCHOR:
        compLimit = predMeths;
        break;

      default:
        compLimit = predMeths >> 1;
        break;
    }

  else
    switch (complexityMode) {

      case COMPLEXITY_MODE_ANCHOR:
        compLimit = methLimit;
        break;

      default:
        compLimit = methLimit >> 1;
        break;
    }

  /*
   * Copy the selected coding methods to search buffer.
   */
	memcpy(methPtr, overheadOrder4x4 + tmp, compLimit);
  totalNumOfMethods = compLimit;

  /* 
   * Scan through all the methods in the search buffer.
   */
  for (j = 0; j < totalNumOfMethods; j++) {

    /*
     * Pick up the jth method from the buffer and find out 
     * the 4x4 method selection bits.
     */
    methNum = methPtr[j];
    methInd = selNPtr[j];
    selBits = methSelectOver4x4New(predNumber, methNum, methInd);

    /* 
     * Initialize values for the jth coding method.
     */
    dcBits = coeffBits = 0;
    methLagr = MCE_BIG_LAGRANGIAN;

     /* 
      * Calculate the margin for the Lagrangian cost of the coding method.  
      */
     minLagrLimit = bestMethodLagr - lambda * selBits;

    /*
     * Perform coding.
     */
    switch (methNum) {

      /* 
       * NOT_CODED method.
       */
      case NOTCODED_METH:

        /*
         * Check if there is a cost margin.
         */
        if (minLagrLimit > 0) {

          /* 
           * Evaluate SE and Lagrangian cost of the method.
           */
          squareError = ncEnergy;
          methLagr = lambda * selBits + squareError;
        }
        break;

      /* 
       * 4x4 DCT.
       */
      case METH_DCT_4x4: case METH_DCT_4x4_N_0_8: case METH_DCT_4x4_N_4_8:

        /*
         * Update the cost margin by the minimum number of bits
         * needed for DCT coefficients and check if some cost
         * margin still exists.
         */
        minLagrLimit -= lambda * MCE_CRITICAL_BITS_4x4DCT;
        if (minLagrLimit > 0) {

          /*
           * Try to code the block using 4x4 DCT.
           */
          mceEnc4x4Dct(diffBlkPtr, predBlkPtr, origBlkPtr, &squareError,
                       &coeffBits, tmpCoeffs, methNum, qp, chroma, mode,
                       blkSize, &memFlag, complexityMode);

          /*
           * Evaluate Lagrangian cost if 4x4 DCT coded
           * something.
           */
          if (coeffBits > 0)
            methLagr = lambda * (selBits + coeffBits) + squareError;
        }
        break;

      /* 
       * 4x4 extrapolation.
       */
      case EXTRAPOL_METH:

        /*
         * Check if there is a cost margin.
         */
        if (minLagrLimit > 0) {

          /*
           * Calculate the median of the 12 boundary pixels
           * and fill the block with the median value.
           */
          tmp = median(extrBlk, extPels);
          for (i = 0; i < SIZE_4x4; i++)
            errorBlk[i] = origBlkPtr[i] - tmp;

          /* 
           * Evaluate SE and Lagrangian cost of the method.
           */
          squareError = calcBlockEnergy(errorBlk, SIZE_4x4);
          methLagr = lambda * selBits + squareError;
        }
        break;

      /* 
       * 4x4 KLT.
       */
      case KLT_4x4_N_1_8: case KLT_4x4_N_2_8: case KLT_4x4_N_3_8:
      case KLT_4x4_N_5_8: case KLT_4x4_N_6_8: case KLT_4x4_N_7_8:

        /*
         * Update the cost margin by the minimum number of bits
         * needed for KLT coefficients and check if some cost
         * margin still exists.
         */
        minLagrLimit -= lambda * MCE_CRITICAL_BITS_4x4KLT;
        if(minLagrLimit > 0) {

          /*
           * Check the KLT method number and rotate/mirror
           * the blocks if needed. This is required, since only
           * -22,5 deg and -45 deg basis functions are implemented.
           * Block manipulations are equivalent to rotation of basis
           * functions.
           */
          switch (methNum) {

            case KLT_4x4_N_1_8: case KLT_4x4_N_2_8:
              diffBlockPtr = diffBlkPtr;
              predBlockPtr = predBlkPtr;
              origBlockPtr = origBlkPtr;
              break;

            case KLT_4x4_N_3_8:

              /*
               * Mapping of -22,5 deg to -67,5 deg.
               */
              for (i = 0; i < SIZE_4x4; i++) {
                tmp = mirrorDiag4x4[i];
                diffBlkMirrorDiag[i] = diffBlkPtr[tmp];
                predBlkMirrorDiag[i] = predBlkPtr[tmp];
                origBlkMirrorDiag[i] = origBlkPtr[tmp];
              }
              diffBlockPtr = diffBlkMirrorDiag;
              predBlockPtr = predBlkMirrorDiag;
              origBlockPtr = origBlkMirrorDiag;
              break;

            case KLT_4x4_N_5_8:

              /*
               * Mapping of -112,5 deg to -22,5 deg.
               */
              for (i = 0; i < SIZE_4x4; i++) {
                tmp = rotateLeft4x4[i];
                diffBlkRotateLeft[i] = diffBlkPtr[tmp];
                predBlkRotateLeft[i] = predBlkPtr[tmp];
                origBlkRotateLeft[i] = origBlkPtr[tmp];
              }
              diffBlockPtr = diffBlkRotateLeft;
              predBlockPtr = predBlkRotateLeft;
              origBlockPtr = origBlkRotateLeft;
              break;

            default:

              /*
               * Mapping of -135 deg to -45 deg and
               * -157,5 deg to -22,5 deg.
               */
              if (!mirrored) {
                for (i = 0; i < SIZE_4x4; i++) {
                  tmp = mirrorVert4x4[i];
                  diffBlkMirrorVert[i] = diffBlkPtr[tmp];
                  predBlkMirrorVert[i] = predBlkPtr[tmp];
                  origBlkMirrorVert[i] = origBlkPtr[tmp];
                }
                mirrored = 1;
              }
              diffBlockPtr = diffBlkMirrorVert;
              predBlockPtr = predBlkMirrorVert;
              origBlockPtr = origBlkMirrorVert;
              break;
          }

          /*
           * Find out the used KLT basis functions and
           * try to code the block with 4x4 KLT.
           */
          kltNum = kltLookUpTable[methNum - KLT_4x4_N_0_8];
          mceEnc4x4Klt(diffBlockPtr, predBlockPtr, origBlockPtr,
                       &squareError, &coeffBits, tmpCoeffs,
                       kltNum, qp, chroma, mode, ncEnergy, complexityMode);

          /*
           * Evaluate Lagrangian cost if 4x4 KLT coded
           * something.
           */
          if (coeffBits > 0)
            methLagr = lambda * (selBits + coeffBits) + squareError;
        }
        break;
    }

    /* 
     * If the method is better than the previous one,
     * update method information.
     */
    if (methLagr < bestMethodLagr) {
      bestMethodLagr = methLagr;
      bestMethod = methNum;
      bestSelBits = selBits;
      codingBits[methodInput] = coeffBits + dcBits;
      seAfter[methodInput] = squareError;

      if (methNum == METH_DCT_4x4 ||
          methNum == METH_DCT_4x4_N_0_8 || methNum == METH_DCT_4x4_N_4_8 ||
          (methNum >= KLT_4x4_N_1_8 && methNum <= KLT_4x4_N_7_8))
        memcpy(coeffs, tmpCoeffs, sizeof(int) * SIZE_4x4);
    }
  }

  /* 
   * Save the 4x4 information of the best method.
   */
  info4x4[10] = bestMethod;
  info4x4[13] = predNumber;
  info4x4[14] = rotation;

  if (bestMethod == AMOUNT_OF_4x4_METH)
    return -1;

  else if (bestMethod == NOTCODED_METH ||
           bestMethod == EXTRAPOL_METH)
    for (i = 0; i < SIZE_4x4; i++)
      coeffs[i] = 0;

  return bestSelBits;
}


/*
 * mceEnc4x4Klt:
 *
 * Parameters:
 *      diffBlkInput    I: Prediction error block
 *      predBlkInput    I: Prediction block
 *      origBlkInput    I: Original block
 *      seAfter         O: SE after KLT
 *      bitsForCoef     O: Coding bits for coefficients
 *      quantKltCoeffs  O: Quantized KLT coefficients
 *      kltNum          I: KLT number
 *      qp              I: Quantization parameter
 *      chroma          I: Chrominance flag
 *                            0 = luminance
 *                            1 = chrominance
 *      mode            I: Mode flag
 *                            0 = INTER
 *                            1 = INTRA
 *      ncEnergy        I: SE for not coded prediction error
 *
 * Function:
 *      Calculates 4x4 KLT using refinement step if quantized coefficients are
 *      initially all zero.
 *
 * Returns:
 *      Nothing.
 */

static void mceEnc4x4Klt(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantKltCoeffs,
                         int kltNum, int qp, int chroma, int mode,
                         int ncEnergy, int complexityMode)
{
  int
    nonZeros,                /* Amount of non-zero coefficients */
    sign,                    /* Sign of refined coefficient */
    index,                   /* Index of refined coefficient */
    kltCoef[SIZE_4x4] = {0}, /* KLT coefficients */
    invCoef[SIZE_4x4],       /* Inverse quantized coefficients */
    recoBlkKlt[SIZE_4x4],    /* Result of IKLT */
    errorBlk[SIZE_4x4];      /* Coding error block */

  /*
   * Calculate KLT and quantize coefficients.
   */
  KltTransform4x4(diffBlkInput, kltCoef, kltNum, mode);
  
  if (mode != INTER_BLOCK && complexityMode == COMPLEXITY_MODE_ANCHOR)
    nonZeros = RD_Optimized_Quantize_Coef(kltCoef, quantKltCoeffs, SIZE_4x4, qp, chroma, 
                                        mode, KLT_TRANSFORM, NULL, 4, 4,kltNum);  
  else
    nonZeros = MCCQuantCoef(kltCoef, quantKltCoeffs, SIZE_4x4, qp,
                          chroma, mode, KLT_TRANSFORM);

  /*
   * If all coefficients are quantized to zero,
   * do refinement.
   */
  if (nonZeros == 0) {

    /*
     * Find largest coefficient value.
     */
    index = mceFindAbsMaxCoefIndex(kltCoef, SIZE_4x4, &sign, qp);

    /*
     * If largest coefficient value was smaller than QP refinement 
     * is done for the first coefficient in scanning order.
     */
    if (index < 0) {
      index = 0;
      sign = kltCoef[index] < 0 ? -1 : 1;
    }
    quantKltCoeffs[index] = sign;
    nonZeros = 1;
  }

  /*
   * Perform inverse quantization if non-zero coefficients
   * exist.
   */
  if (nonZeros > 0) {

    MCCInvQuantCoef(quantKltCoeffs, invCoef, SIZE_4x4, qp, chroma, mode);

    /* 
     * Calculate the amount of KLT bits.
     */
    *bitsForCoef = calcKltBits(quantKltCoeffs, COLS_4x4, ROWS_4x4,
                               qp, chroma, mode, kltNum);

    /*
     * Perform actual IKLT and clip the reconstruction
     * values before calculating coding error.
     */
    KltInverseTransform4x4(invCoef, recoBlkKlt, kltNum, mode);
    mceMakeDiff(origBlkInput, predBlkInput, recoBlkKlt, errorBlk, SIZE_4x4);
    *seAfter = calcBlockEnergy(errorBlk, SIZE_4x4);
  }
  else {
    *bitsForCoef = 0;
    *seAfter = ncEnergy;
  }
}


/*
 * mceEnc4x4Dct:
 *
 * Parameters:
 *      diffBlkInput    I: Prediction error block
 *      predBlkInput    I: Prediction block
 *      origBlkInput    I: Original block
 *      seAfter         O: SE after DCT
 *      bitsForCoef     O: Coding bits for coefficients
 *      quantDctCoeffs  O: Quantized DCT coefficients
 *      dctMet          I: DCT method
 *      qp              I: Quantization parameter
 *      chroma          I: Chrominance flag
 *                            0 = luminance
 *                            1 = chrominance
 *      mode            I: Mode flag
 *                            0 = INTER
 *                            1 = INTRA
 *      blkSize         I: Size of source block
 *      memFlag         I: State flag
 *                            0 = First call for this block
 *                            2 = Later call for this block, non-zero
 *                                coefficients do not exist, no need
 *                                for actual transform calculation
 *                            1 = Later call for this block, non-zero
 *                                coefficients do not exist
 *
 * Function:
 *      Calculates 4x4 DCT using refinement step if quantized coefficients are
 *      initially all zero. Usage of state flag allows calculation for 
 *      different coefficient scannings without recalculation of DCT.
 *
 * Returns:
 *      Nothing.
 */

static void mceEnc4x4Dct(int *diffBlkInput, int *predBlkInput, int *origBlkInput,
                         int *seAfter, int *bitsForCoef, int *quantDctCoeffs,
                         int dctMet, int qp, int chroma, int mode, int blkSize,
                         int *memFlag, int complexityMode)
{

  static int 
    memDct[SIZE_4x4],     /* Memory for DCT coefficients */
    memCoe[SIZE_4x4],     /* Memory for quantized coefficients */
    memSe = 0;            /* Memory for SE after DCT */

  int
    value,                /* Refined coefficient before quantization */
    invValue,             /* Refined coefficient after inverse quantization */
    diffBlk[SIZE_4x4],    /* Result of IDCT */
    dctCoeffs[SIZE_4x4],  /* Quantized coefficients */
    dctInvs[SIZE_4x4],    /* Inverse quantized coefficients */
    errorBlk[SIZE_4x4],   /* Coding error block */
    i,                    /* Loop counter */
    nonZeros,             /* Amount of non-zero coefficients */
    sign,                 /* Sign of refined coefficient */
    index;                /* Index of refined coefficient */

  u_char
    *zigZagTable;         /* Scanning table */

  /*
   * Check if this is the first call for the block.
   */
  if (*memFlag == 0) {

    /*
     * Set state flag for later calls, state 2 assumes non-zero
     * coefficients.
     */
    *memFlag = 2;
    
    /*
     * Calculate DCT and quantize coefficients and store them
     * in memory for later calls.
     */
    switch (dctMet) {

      case METH_DCT_4x4:
        zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                      zig4x4Intra : zig4x4;
        break;

      case METH_DCT_4x4_N_0_8:
        zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                      zig4x4IntraHori : zig4x4Hori;
        break;

      default:
        zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                      zig4x4IntraVert : zig4x4Vert;
        break;
    }

    varSizeDctFixedp(diffBlkInput, memDct, COLS_4x4, ROWS_4x4);

    /*
     * For intra single 4x4 DCT chroma blocks the DC has been sent as DCD
     * together with intra prediction method and Coded/Not-coded bit
     * therefore the DC resulted from MCC is deliberately skipped.
     * This is also accounted form in the later zig4x4Intra,
     * zig4x4IntraHori, zig4x4IntraVert tables.
     */
    if (mode != INTER_BLOCK && blkSize == SIZE_4x4)
      memDct[0] = 0;
    
    if (mode != INTER_BLOCK && complexityMode == COMPLEXITY_MODE_ANCHOR)
      nonZeros = RD_Optimized_Quantize_Coef(memDct, memCoe, SIZE_4x4, qp, chroma, 
                                        mode, DCT_TRANSFORM, zigZagTable, 4, 4, 0);  
    else
      nonZeros = MCCQuantCoef(memDct, memCoe, SIZE_4x4, qp, chroma, mode,
                            DCT_TRANSFORM);

    /*
     * If all coefficients are quantized to zero,
     * do refinement.
     */
    if (nonZeros == 0) {

      /*
       * Find largest coefficient value.
       */
      index = mceFindAbsMaxCoefIndex(memDct, SIZE_4x4, &sign, qp);

      if (index < 0)

        /*
         * If largest was smaller than QP refinement is done for
         * the first coefficient in scanning order, set state to 1.
         */
        *memFlag = 1;

      else {

        /*
         * Set set refined quantized coefficient to -1 or +1.
         */
        memCoe[index] = sign;
        nonZeros = 1;
      }
    }

    /*
     * Perform inverse quantization if non-zero coefficients
     * exist.
     */
    if(nonZeros > 0) {
      MCCInvQuantCoef(memCoe, dctInvs, SIZE_4x4, qp, chroma, mode);

      /*
       * Perform actual IDCT and clip the reconstruction
       * values before calculating coding error.
       */
      varSizeInvDctFixedp(dctInvs, diffBlk, COLS_4x4, ROWS_4x4);
      mceMakeDiff(origBlkInput, predBlkInput, diffBlk, errorBlk, SIZE_4x4);
      memSe = calcBlockEnergy(errorBlk, SIZE_4x4);
    }
    else {

      /*
       * All coefficients are zero, coding error is set to 
       * original prediction error.
       */
      memSe = calcBlockEnergy(diffBlkInput, SIZE_4x4);
    }
  }

  /*
   * Select correct scanning order for bit calculation.
   */
  switch (dctMet) {

    case METH_DCT_4x4:
      zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                    zig4x4Intra : zig4x4;
      break;

    case METH_DCT_4x4_N_0_8:
      zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                    zig4x4IntraHori : zig4x4Hori;
      break;

    default:
      zigZagTable = (mode != INTER_BLOCK && blkSize == SIZE_4x4) ?
                    zig4x4IntraVert : zig4x4Vert;
      break;
  }

  /*
   * State is 1, refinement is done for the first coefficient
   * in the scanning order.
   */
  if (*memFlag == 1) {
    for (i = 0; i < SIZE_4x4; i++)
      dctCoeffs[i] = 0;

    /*
     * Find the first coefficient and set the quantized 
     * coefficient value to -1 or +1.
     */
    index = mceFindFirstIndex(zigZagTable, SIZE_4x4);
    value = memDct[index];
    dctCoeffs[index] = value < 0 ? -1 : 1;

    /*
     * Perform inverse quantization.
     */
    MCCInvQuantCoef(dctCoeffs, dctInvs, SIZE_4x4, qp, chroma, mode);
    invValue = dctInvs[index];

    /*
     * Perform actual IDCT.
     */
    varSizeInvDctFixedp(dctInvs, diffBlk, COLS_4x4, ROWS_4x4);    
  }
  else

    /*
     * State is 2, use coefficients in memory.
     */
    memcpy(dctCoeffs, memCoe, sizeof(int) * SIZE_4x4);

  /*
   * Scan coefficients.
   */
  for (i = 0; i < SIZE_4x4; i++) {
    quantDctCoeffs[zigZagTable[i]] = dctCoeffs[i];
  }

  /* 
   * Calculate the amount of DCT bits. 
   */
  *bitsForCoef = calcDctBits(dctCoeffs, COLS_4x4, ROWS_4x4, zigZagTable,
                             qp, chroma, mode, 0);

  /* 
   * If state is 2 use coding error in memory.
   */
  if (*memFlag == 2)
    *seAfter = memSe;

  else {

    /*
     * State is 1, coding error must be evaluated.
     * Clip the reconstruction values and calculate coding error.
     */
    mceMakeDiff(origBlkInput, predBlkInput, diffBlk, errorBlk, SIZE_4x4);
    *seAfter = calcBlockEnergy(errorBlk, SIZE_4x4);
  }
}



int RD_Optimized_Quantize_Coef(int *dctCoef, int *quantDctCoef, int size, int qp, int chroma, 
                                      int mode, int transform, u_char *zigZagTable, int sizeX, int sizeY, int kltNum)
{
  int 
    Level0[64], Level1[64], tmpCoef[64], diflevel_index[64],
    quantDctCoef_Level1[64], coefarray_all[8][64], quantcoefarray_all[8][64],
    i, counter1, counter2, maxLevel,
    Double_ctr = 0, last_coef_index = size -1,
    coef_index, nonZeros = 0,
    numofpossibleCoefs =0,
    best_choice_index = 99, Bits, position = 0, all_Positive = 1;
  double 
    lambda, RD_Gain[8], RD_Minimum;
  int32 
    Squared_error = 0;

  /* initializations */
  for (i = 0; i < size;i++)
  {
     diflevel_index[i] = 0;
     RD_Gain[i % 8] = 0;
     quantDctCoef_Level1[i] = 0;
  }
 
  /* Put the dct coefs in tmpCoef in proper scanning order */
  if (transform != KLT_TRANSFORM)
   for (i = 0; i < size;i++)
     tmpCoef[zigZagTable[i]] = dctCoef[i]; 
  else
   for (i = 0; i < size;i++)
     tmpCoef[i] = dctCoef[i]; 

  /* Quantize the dct coefficients */
  MCCQuantCoef(tmpCoef, quantDctCoef,size,qp,chroma,mode,transform);
  /* inverse quantize the coefficients to obtain Level0[] array */
  MCCInvQuantCoef(quantDctCoef, Level0, size, qp, chroma, mode);

  if(chroma == CHROMA) {
    if(qp ==  0)      maxLevel = LEVEL_09;
    else if(qp <=  1) maxLevel = LEVEL_08;
    else if(qp <=  2) maxLevel = LEVEL_07;
    else if(qp <=  5) maxLevel = LEVEL_06;
    else if(qp <= 11) maxLevel = LEVEL_05;
    else if(qp <= 23) maxLevel = LEVEL_04;
    else              maxLevel = LEVEL_03;
  }
  else {        /* LUMA */
    if(qp ==  0)      maxLevel = LEVEL_10;
    else if(qp <=  1) maxLevel = LEVEL_09;
    else if(qp <=  2) maxLevel = LEVEL_08;
    else if(qp <=  5) maxLevel = LEVEL_07;
    else if(qp <= 11) maxLevel = LEVEL_06;
    else if(qp <= 23) maxLevel = LEVEL_05;
    else              maxLevel = LEVEL_04;
  }

  /* add one to the quantized coefficients for Level1[] calculation. Don't touch the zero coeffs */
  for (i = 0;i < size; i++)
  {
    if (tmpCoef[i] > 0 ){
      if(quantDctCoef[i] + 1 >= maxLevel)
        quantDctCoef_Level1[i] = quantDctCoef[i];
      else
        quantDctCoef_Level1[i] = quantDctCoef[i] + 1;
    }
    else if (tmpCoef[i] < 0 ){
      if(quantDctCoef[i] - 1 <= -maxLevel)
        quantDctCoef_Level1[i] = quantDctCoef[i];
      else
        quantDctCoef_Level1[i] = quantDctCoef[i] - 1;
    }
  }

  /* inverse quantize the one added coefficients to obtain Level1[] array */
  MCCInvQuantCoef(quantDctCoef_Level1, Level1, size, qp, chroma, mode);
 
  /* Now we have Level0[], Level1[] and quantDctCoef[] in proper scanning order*/
 
  for (coef_index = 0; coef_index <= last_coef_index; coef_index++)
  {
    switch(Double_ctr)
    {
      case 0: case 1: case 2:
      break;
      
      default:
         Level1[coef_index] = Level0[coef_index];
      break;
    }
    if (Double_ctr <3)
      if (fabs((float)tmpCoef[coef_index])> (fabs(0.55 * (float)Level1[coef_index]) + fabs(0.45 * (float)Level0[coef_index])))
        diflevel_index[Double_ctr++] = coef_index;
  }
  /* End of Loop1 */ 

  switch (Double_ctr)
  {
     case 0: /* number of different levelled coeffs is 0 */
       numofpossibleCoefs = 0;
     break;
     case 1: /* number of different levelled coeffs is 1 */
       numofpossibleCoefs = 1;
     break;
     case 2: /* number of different levelled coeffs is 2 */
       numofpossibleCoefs = 3;
     break;
     case 3: /* number of different levelled coeffs is 3 */
       numofpossibleCoefs = 7;
     break;
  }

  for (counter1 = 0; counter1 <= numofpossibleCoefs; counter1++)
    for (counter2 = 0; counter2 <= last_coef_index; counter2++)
    {
       coefarray_all[counter1][counter2] = Level0[counter2];     
       quantcoefarray_all[counter1][counter2] = quantDctCoef[counter2];
    }

  if (numofpossibleCoefs == 0)
    best_choice_index = 0;
  else
  if (numofpossibleCoefs == 1)
  {
    coefarray_all[1][diflevel_index[0]] = Level1[diflevel_index[0]];
    quantcoefarray_all[1][diflevel_index[0]] = quantDctCoef_Level1[diflevel_index[0]];
  }
  else
  if (numofpossibleCoefs == 3)
  {
    coefarray_all[2][diflevel_index[0]] = 
    coefarray_all[3][diflevel_index[0]] =
    Level1[diflevel_index[0]];
    quantcoefarray_all[2][diflevel_index[0]] = 
    quantcoefarray_all[3][diflevel_index[0]] =
    quantDctCoef_Level1[diflevel_index[0]] ;

    coefarray_all[1][diflevel_index[1]] = 
    coefarray_all[3][diflevel_index[1]] =
    Level1[diflevel_index[1]];
    quantcoefarray_all[1][diflevel_index[1]] = 
    quantcoefarray_all[3][diflevel_index[1]] =
    quantDctCoef_Level1[diflevel_index[1]] ;  
  }
  else
  if (numofpossibleCoefs == 7)
  {
    coefarray_all[4][diflevel_index[0]] = 
    coefarray_all[5][diflevel_index[0]] =
    coefarray_all[6][diflevel_index[0]] = 
    coefarray_all[7][diflevel_index[0]] =
    Level1[diflevel_index[0]];
    quantcoefarray_all[4][diflevel_index[0]] = 
    quantcoefarray_all[5][diflevel_index[0]] =
    quantcoefarray_all[6][diflevel_index[0]] = 
    quantcoefarray_all[7][diflevel_index[0]] =
    quantDctCoef_Level1[diflevel_index[0]]; 

    coefarray_all[2][diflevel_index[1]] = 
    coefarray_all[3][diflevel_index[1]] =
    coefarray_all[6][diflevel_index[1]] = 
    coefarray_all[7][diflevel_index[1]] =
    Level1[diflevel_index[1]];
    quantcoefarray_all[2][diflevel_index[1]] = 
    quantcoefarray_all[3][diflevel_index[1]] =
    quantcoefarray_all[6][diflevel_index[1]] = 
    quantcoefarray_all[7][diflevel_index[1]] =
    quantDctCoef_Level1[diflevel_index[1]];  

    coefarray_all[1][diflevel_index[2]] = 
    coefarray_all[3][diflevel_index[2]] =
    coefarray_all[5][diflevel_index[2]] = 
    coefarray_all[7][diflevel_index[2]] =
    Level1[diflevel_index[2]];
    quantcoefarray_all[1][diflevel_index[2]] = 
    quantcoefarray_all[3][diflevel_index[2]] =
    quantcoefarray_all[5][diflevel_index[2]] = 
    quantcoefarray_all[7][diflevel_index[2]] =
    quantDctCoef_Level1[diflevel_index[2]];  
  }
  /* Now, all the possible coefficient combinations are ready!! */
  /* Time to Accumulate The RD_GAIN */
  if (best_choice_index !=0)
  {
    if (mode == INTER_BLOCK)
      lambda = 0.85 * qp * qp;
    else
      lambda = 0.85 * qp * qp;
    for (counter1 = 0; counter1 <= numofpossibleCoefs; counter1++)
    {
       Squared_error = 0;
       if (transform != KLT_TRANSFORM)
         Bits = calcDctBits(quantcoefarray_all[counter1], sizeX, sizeY, zigZagTable, qp, chroma, mode, 0);
       else
         Bits = calcKltBits(quantcoefarray_all[counter1], sizeX, sizeY, qp, chroma, mode, kltNum);

       for (counter2 = 0; counter2 <= last_coef_index; counter2++)
       {
          Squared_error += (tmpCoef[counter2] - coefarray_all[counter1][counter2]) *
                                (tmpCoef[counter2] - coefarray_all[counter1][counter2]);
       }
       RD_Gain[counter1] = (float)Bits * lambda + (float)Squared_error;
    }
    /* Now the RD Gains of different coefficient combinations are ready in RD_Gain[] */


    all_Positive = 0; 
    if (!all_Positive)
    {  
       RD_Minimum = RD_Gain[0];
       for (i = 1; i <= numofpossibleCoefs; i++)
         if (RD_Gain[i] < RD_Minimum) 
         {
           RD_Minimum = RD_Gain[i];
           position = i;
         }
      for (i = 0; i<=  last_coef_index; i++)
      {
         if (quantcoefarray_all[position][i] != 0)
           nonZeros++;
         if (transform != KLT_TRANSFORM)
            quantDctCoef[zigZagTable[i+size]] = quantcoefarray_all[position][i];
         else
            quantDctCoef[i] = quantcoefarray_all[position][i];
      }
    }
    else
      for (i = 0; i<=  last_coef_index; i++)
         if (transform != KLT_TRANSFORM)
            quantDctCoef[zigZagTable[i+size]] =  0;
         else
            quantDctCoef[i] = 0;
  }
  else
  {
     for (i = 0; i<=  last_coef_index; i++)
      {
         if (quantcoefarray_all[0][i] != 0)
           nonZeros++;
         if (transform != KLT_TRANSFORM)
            quantDctCoef[zigZagTable[i+size]] = quantcoefarray_all[0][i];
         else
            quantDctCoef[i] = quantcoefarray_all[0][i];
      }
  } 
  return nonZeros;
}
