/*
 * imo.c
 *
 *    Copyright (c) 2000 Nokia Research Center, Tampere, FINLAND
 *
 * Project:
 *    Nokia Very Low Bitrate Codec
 *
 * Contains:
 *    High level functions for affine motion estimation
 *
 * Author:
 *    Nokia Research Center
 */

#include <stdio.h>

#include "tools.h"
#include "mainFlags.h"
#include "mcp.h"
#include "encMux.h"

double mseRegionLum(double A[], object *region, Byte *curr, Byte *prev);

void findPredRegion(object *currRegion, object *objectList, int regionIndex,
                    struct frame *currfr, struct frame *prevfr,
                    struct frame *predfr, double lambda,
                    int picSizeX, int picSizeY, int bitsOverheadUp,
                    int bitsOverheadLeft, int complexityMode, int refFrame);

int intcompare ( const void *x, const void *y )
{
  int *a, *b;
  
  a = (int *) x;
  b = (int *) y;
  return *a - *b;
}

/*
 * CalculateVectorMedian:
 *
 * Function:
 *     Calculates median of n translational motion vectors
 *
 * Returns:
 *     Nothing
 */

void CalculateVectorMedian(int *vectors, int n, int *x,int *y)
{
  int i;
  int tx[16];
  int ty[16];

  for(i=0;i<n;i++)
  {
    tx[i]=vectors[2*i];
    ty[i]=vectors[2*i+1];
  }

  qsort ( tx, n, sizeof(int), intcompare );
  qsort ( ty, n, sizeof(int), intcompare );
  *x=(tx[n/2]+tx[(n/2)+1])/2;
  *y=(ty[n/2]+ty[(n/2)+1])/2;
}

/*
 * transLoc2OrthCoeffs:
 *
 * Function:
 *     Transforms motion parameters in local affine system to
 *     orthonormalized system.
 *
 * Returns:
 *     Nothing
 */

void transLoc2OrthCoeffs(double *Loc, double *Orth, int regionSize)
{
  if(regionSize == 16){
    Orth[0] = 16.0*Loc[0];
    Orth[1] = -9.83418075444575*Loc[1];
    Orth[2] = -9.83418075444575*Loc[2];
    Orth[3] = 16.0*Loc[3];
    Orth[4] = -9.83418075444575*Loc[4];
    Orth[5] = -9.83418075444575*Loc[5];
  }
  else{
    Orth[0] =  8.0*Loc[0];
    Orth[1] = -5.23722936566383*Loc[1];
    Orth[2] = -5.23722936566383*Loc[2];
    Orth[3] =  8.0*Loc[3];
    Orth[4] = -5.23722936566383*Loc[4];
    Orth[5] = -5.23722936566383*Loc[5];
  }
}

/*
 * reoptimiseMotGauss:
 *
 * Function:
 *     Motion estimation starting from the input motion coefficients.
 *     This function uses Gauss-Newton steps to minimize prediction
 *     MSE and replaces input motion coefficients by the refined ones.
 *
 * Returns:
 *     Prediction MSE of the region
 */

double reoptimiseMotGauss(Byte *prev, Byte *curr, object *current_region,
                          int picSizeX, int picSizeY, double motCoef[6],
                          int complexityMode)
{

  int 
    numIter = 0, stopCondMaxNumIters;

  double 
    mse, mseTmp, motCoefTmp[6], stopCondMinMseDiff, diff;

  /* Conditions to stop Gauss-Newton iterations */
  if(complexityMode == COMPLEXITY_MODE_ANCHOR){
    stopCondMinMseDiff  = 0.1;
    stopCondMaxNumIters = 5;
  }
  else if(complexityMode == COMPLEXITY_MODE_PRACTICAL){
    stopCondMinMseDiff  = 10.0;
    stopCondMaxNumIters = 3;
  }
  else{
    stopCondMinMseDiff  = 10.0;
    stopCondMaxNumIters = 1;
  }

  memcpy(motCoefTmp, motCoef, 6 * sizeof(double));

  mse  = mseRegionLum(motCoef, current_region, curr, prev);

  diff = stopCondMinMseDiff + 100.0;

  /* Gauss-Newton iterations */
  while(diff > stopCondMinMseDiff && numIter < stopCondMaxNumIters){

    /* Take a step */
    gnsGaussNewtonStep(current_region, picSizeX, picSizeY, 
                       curr, prev, motCoef, motCoefTmp);

    /* Evaluate prediction error */
    mseTmp = mseRegionLum(motCoef, current_region, curr, prev);

    diff   = mse - mseTmp;

    /* Accept or reject result of this step */
    if(mseTmp < mse){   
      memcpy(motCoefTmp, motCoef, 6 * sizeof(double));
      mse = mseTmp;
    } 
    else
      memcpy(motCoef, motCoefTmp, 6 * sizeof(double));
    
    numIter++;
  }

  return(mse);
}

/*
 *
 * InitialMotionRefinement:
 *
 * Parameters:
 *       currRegion       B: List for segment attributes.
 *       objectList       B: List of all the segments
 *       currFr           I: Current image
 *       prevFr           I: Previous image
 *       picSizeX         I: Horizontal image size
 *       picSizeY         I: Vertical image size
 *       aQp              I: Quantization param. for motion params.
 *       lambda           I: rate-distortion trade-off parameter
 *       bitsOverheadUp   I: Prediction ovr bits (up direction)
 *       bitsOverheadLeft I: Prediction ovr bits (left direction)
 *       complexityMode   I: Complexity mode
 *
 * Function:
 *       Finds a prediction candidate for motion coefficient
 *       prediction and runs a motion estimation starting from
 *       the predicted parameters. Stores the motion information
 *       to the currRegion's motion coefficient fields.
 *
 * Returns:
 *       Nothing
 *
 */

void InitialMotionRefinement(object *currRegion,
                             object *objectList,
                             int regionIndex,
                             struct frame *currFr,
                             struct frame *prevFr,
                             int picSizeX, int picSizeY, 
                             int aQp, double lambda,
                             int bitsOverheadUp,
                             int bitsOverheadLeft,
                             int complexityMode,
                             int refFrame)
{

  int i, motCoefOrthQuant[6], motionBits, regionSize;

  double
    mse, lagrPred, lagrNoPred,
    motCoefOrth[6],
    motCoefPredLoc[6],
    motCoefLoc[6];

  static Byte
    predY[256], predU[64], predV[64];

  static struct frame
    predBuffer;

  predBuffer.yptr = predY;
  predBuffer.uptr = predU;
  predBuffer.vptr = predV;

  currRegion->zeroRefinement = 1;
  regionSize                 = currRegion->no_of_pels == 256 ? 16 : 8;

  /* Find prediction motion parameters (store to
   * currRegion->motCoefPredGlob)
   */
  findPredRegion(currRegion, objectList, regionIndex, currFr, prevFr,
                 &predBuffer, lambda, picSizeX, picSizeY, bitsOverheadUp,
                 bitsOverheadLeft, complexityMode, refFrame);

  if(currRegion->zero_motion){
    memcpy(currRegion->motCoefPredRefinedGlob, currRegion->motCoefPredGlob,
           6*sizeof(double));
  }
  else{

    /* Refinement search around the predictor */
    mcpTransformGlob2LocCoeffs(currRegion->motCoefPredGlob, currRegion,
                               picSizeX, picSizeY, motCoefLoc);
    memcpy(motCoefPredLoc, motCoefLoc, 6 * sizeof(double));
    mse = reoptimiseMotGauss(prevFr->yptr, currFr->yptr, currRegion,
                             picSizeX, picSizeY, motCoefLoc, complexityMode);
    mcpTransformLoc2GlobCoeffs(motCoefLoc, currRegion, picSizeX, picSizeY, 
                               currRegion->motCoefPredRefinedGlob);

    /* In low complexity modes approximated lagrangian costs
     * are evaluated and only one coefficient removal will be
     * performed - either with the motion parameters obtained
     * by the full motion search or the ones obtained by the
     * refinement search above (the one that looks more
     * promising at this point in lagrangian sense)
     */
    if(complexityMode == COMPLEXITY_MODE_ANCHOR){
      /* Test refinement */
      currRegion->zeroRefinement = 0;
    }
    else{
 
      /* Lagrangian for no prediction case */
      transLoc2OrthCoeffs(currRegion->motCoefEstimLoc[refFrame], motCoefOrth, regionSize);
      
      orthQuantParams(motCoefOrth, motCoefOrthQuant, aQp);

      if(motCoefOrthQuant[0] != 0 ||
         motCoefOrthQuant[1] != 0 ||
         motCoefOrthQuant[2] != 0 ||
         motCoefOrthQuant[3] != 0 ||
         motCoefOrthQuant[4] != 0 ||
         motCoefOrthQuant[5] != 0)
        motionBits = dumpAffineMotionParams(motCoefOrthQuant, aQp, 0);
      else
        motionBits = 0;
      
      lagrNoPred = lambda*motionBits + currRegion->mseMin[refFrame]*currRegion->no_of_pels;
      
      /* Lagrangian for prediction case */
      for(i = 0;i < 6;i++)
        motCoefLoc[i] -= motCoefPredLoc[i];
      
      transLoc2OrthCoeffs(motCoefLoc, motCoefOrth, regionSize);
      
      orthQuantParams(motCoefOrth, motCoefOrthQuant, aQp);
      
      if(motCoefOrthQuant[0] != 0 ||
         motCoefOrthQuant[1] != 0 ||
         motCoefOrthQuant[2] != 0 ||
         motCoefOrthQuant[3] != 0 ||
         motCoefOrthQuant[4] != 0 ||
         motCoefOrthQuant[5] != 0)
        motionBits = dumpAffineMotionParams(motCoefOrthQuant, aQp, 0);
      else
        motionBits = 0;

      lagrPred   = lambda*motionBits + mse*currRegion->no_of_pels;
      
      if(lagrPred < lagrNoPred){
        currRegion->zeroRefinement = 0;
        currRegion->zero_motion    = 1;
      }
      else{
        currRegion->zeroRefinement = 1;
      }
    }
  }
}


/*
 *
 * Initial_Motion_Estimation:
 *
 * Parameters:
 *       currRegion       B: List for segment attributes.
 *       currY/U/V        I: Current image
 *       prevY/U/V        I: Previous image
 *       subsCurrY        I: Subsampled current image
 *       subsPrevY        I: Subsampled previous image
 *       picSizeX         I: Horizontal image size
 *       picSizeY         I: Vertical image size
 *       stage            I: quadtree depth. can take the values
 *                                    1 -  32x32 + residues
 *                                    2 -  16x16
 *                                    3 -  8x8
 *       lambda           I: rate-distortion trade-off parameter used
 *                           for 'zero motion' decisions 
 *       zeroMotionFlag   O: Indicator for 'zero motion'
 *       motCoefLoc       I: Inherited motion parameters (if any)
 *                        O: Output motion parameters
 *       mseOut           O: Output prediction MSE
 *       complexityMode   I: Complexity mode
 *       BMvectors           A pointer to saved vectors from BM
 *
 * Function:
 *       Estimates the affine motion parameters associated with the input
 *       region. Records the associated prediction mse, and sets zeroMotionFlag 
 *       to 1 if the estimated motion coefficients are zero.
 * 
 *       Uses block matching + Gauss-Newton minimizations on current and
 *       previous images on subsampled and regular images
 *
 * Returns:
 *       Nothing
 *
 */

void Initial_Motion_Estimation(object *currRegion,
                               Byte *currY, Byte *prevY,
                               Byte *currU, Byte *prevU,
                               Byte *currV, Byte *prevV,
                               Byte *subsCurrY, Byte *subsPrevY,
                               int picSizeX, int picSizeY, 
                               int stage, int qp, double lambda,
                               Byte *zeroMotionFlag, double *motCoefLoc,
                               double *mseOut, int complexityMode, int *BMvectors)
{

  int 
    i, j, m, n, sx, sy, ex, ey, vecX, vecY,
    stageNumber, picIndex, picSizeXChr = picSizeX/2,
    stopCondMaxNumIters,blkSizeX,blkSizeY,ci,cj;

  double 
    A[6], A_temp[6], A_init[6], A_in[6],
    mse, mse_temp, mse_init, mseNoMotionChroma, mse_blk_match,
    est_red, difference, zeroMotionThr, zeroMotionThrChroma,
    stopCondMinMseDiff;

  /* Thresholds to stop Gauss-Newton iterations */
  if(complexityMode == COMPLEXITY_MODE_ANCHOR){
    stopCondMinMseDiff  = 0.1;
    stopCondMaxNumIters = 5;
  }
  else if(complexityMode == COMPLEXITY_MODE_PRACTICAL){
    stopCondMinMseDiff  = 10.0;
    stopCondMaxNumIters = 3;
  }
  else{
    stopCondMinMseDiff  = 10.0;
    stopCondMaxNumIters = 1;
  }

  /* Thresholds for declaring region a zero motion one */
  zeroMotionThr       = (qp*qp/3)/2;
  zeroMotionThrChroma = zeroMotionThr/2;

  if(zeroMotionThr < 5.0)       zeroMotionThr       = 5.0;
  if(zeroMotionThrChroma < 5.0) zeroMotionThrChroma = 5.0;

  for(j = 0; j<6; j++)
    A[j] = A_temp[j] = 0.0;
      
  sx = currRegion->xmin;
  sy = currRegion->ymin;
  ex = currRegion->xmax;
  ey = currRegion->ymax;

  /* Set zero_motion flag to FALSE by default */
  *zeroMotionFlag = 0;

  /* Find the MSE of the region with zero motion coefficients */
  for(mse=0.0, i=sy; i<=ey; i++)
    for(j=sx; j<=ex; j++){
      picIndex   = i*picSizeX+j;
      difference = (double)currY[picIndex] - (double)prevY[picIndex];
      mse       += difference * difference;
    }

  mse    /= currRegion->no_of_pels;
  *mseOut = mse;

  /* Include also chrominance component to the zero motion decision */
  mse_temp = 0.0;
  for(i=sy/2; i<=ey/2; i++)
    for(j=sx/2; j<=ex/2; j++){
      picIndex   = i*picSizeXChr+j;
      difference = (double)currU[picIndex] - (double)prevU[picIndex];
      mse_temp  += difference * difference;
      difference = (double)currV[picIndex] - (double)prevV[picIndex];
      mse_temp  += difference * difference;
    }

  mseNoMotionChroma = mse_temp / (currRegion->no_of_pels/2);

  /* Check if zero motion parameters can be assigned for the region */  
  if (mseNoMotionChroma < zeroMotionThrChroma &&
      *mseOut           < zeroMotionThr)
  {
    memcpy(motCoefLoc, A, 6 * sizeof(double));
    
    /* Set zero_motion flag to TRUE */
    *zeroMotionFlag = 1;  
    
    /* Set BM vectors to 0 */
    if(stage==1)
      for(i=currRegion->xmin,ci=0;i<currRegion->xmax;i+=8,ci++)
      {
        for(j=currRegion->ymin,cj=0;j<currRegion->ymax;j+=8,cj++)
        {
          BMvectors[2*(4*cj+ci)]   = 0;
          BMvectors[2*(4*cj+ci)+1] = 0;
        }
      }
  }
  else
  {
    /* Find nonzero motion coefficients of the region */
    
    /* Find the MSE of the region with inherited motion parameters
     * (Done for 8x8 and 16x16 regions only)
     */
    if (stage > 1){
      memcpy(A_init, motCoefLoc, 6 * sizeof(double));
      mse_init = mseRegionLum(A_init, currRegion, currY, prevY);
    }

    /* For the Anchor complexity mode all the blocks are first estimated
     * by block matching
     */
    if(complexityMode == COMPLEXITY_MODE_ANCHOR)
    {
      /* Find initial motion vector by block matching */
      subsBlkMatching(currRegion, subsPrevY, subsCurrY,
                      picSizeX>>1, picSizeY>>1, &vecX, &vecY);
      
      A_temp[0] = vecX;
      A_temp[1] = vecY;
    }
    else
    {
      /* For other complexity modes block matching is carried out
       * for 8x8 blocks. Translational motion parameters for the
       * larger blocks are obtained by calculating the median of
       * the child vectors.
       */
      blkSizeX=(currRegion->xmax-currRegion->xmin+1);
      blkSizeY=(currRegion->ymax-currRegion->ymin+1);
      
      /* If current block is 32x32, 32x16, 16x32 or the NASTY corner
       * 16x16 block, save the BM vectors for 8x8 subblocks. Initial
       * motion for current block is the median of the child blocks
       */
      if(stage == 1){

        /* Do block matching for all the 8x8 blocks inside this block */
        for(i=currRegion->xmin,ci=0;i<currRegion->xmax;i+=8,ci++)
        {
          for(j=currRegion->ymin,cj=0;j<currRegion->ymax;j+=8,cj++)
          {
            /* Block matching is done for here with subsamled images.
             * Thus, vectors are multiplied by 2 before storing
             */
            MatchBlock(subsPrevY, subsCurrY, (i>>1), (j>>1), (int)4,
                       picSizeX>>1, picSizeY>>1, &BMvectors[2*(4*cj+ci)],
                       &BMvectors[2*(4*cj+ci)+1]);
            BMvectors[2*(4*cj+ci)]   *= 2;
            BMvectors[2*(4*cj+ci)+1] *= 2;
          }
        }

        /* Calculate median of the child motion vectors */
        CalculateVectorMedian(BMvectors,(blkSizeX*blkSizeY)/64,&vecX,&vecY);
      }

      /* For 16x16 macroblocks translational motion is obtained by
       * calculating the median of the child vectors
       */
      else if(stage == 2)
        CalculateVectorMedian(currRegion->BMvectors,(blkSizeX*blkSizeY)/64,&vecX,&vecY);
      
      /* For 8x8 blocks the motion vectors calculated in stage 1
       * are retrieved
       */
      else
      {
        vecX = currRegion->BMvectors[0];
        vecY = currRegion->BMvectors[1];
      }

      A_temp[0]=vecX;
      A_temp[1]=vecY;
    }

    /* Find the MSE of the region with motion parameters obtained
     * by block matching
     */
    if(sx+ vecX >= 0 && ex+ vecX < picSizeX){
      if(sy+ vecY >= 0 && ey+ vecY < picSizeY)
        /* If there is no boundary violation */
        for(mse_temp=0.0, i=sy; i<=ey; i++){
          m = i + vecY;
          for(j=sx; j<=ex; j++){
            n = j + vecX;
            difference = (double)currY[i*picSizeX+j] - 
              (double)prevY[m*picSizeX+n];
            mse_temp += difference * difference;
          }
        }
        else
          /* Vertical boundary violation */
          for(mse_temp=0.0, i=sy; i<=ey; i++){
            m = i + vecY;
            m = ( m < 0 ? 0 : ( m >= picSizeY ? picSizeY - 1 : m)) ;
            for(j=sx; j<=ex; j++){
              n = j + vecX;
              difference = (double)currY[i*picSizeX+j] - 
                (double)prevY[m*picSizeX+n];
              mse_temp += difference * difference;
            }
          }
    }
    else{ /* Horizontal boundary violation */
      if(sy+ vecY >= 0 && ey+ vecY < picSizeY)
        /* Horizontal boundary violation only */
        for(mse_temp=0.0,j=sx; j<=ex; j++){
          n = j + vecX;
          n = ( n < 0 ? 0 : ( n >= picSizeX ? picSizeX - 1 : n)) ;
          for(i=sy; i<=ey; i++){
            m = i + vecY;
            difference = (double)currY[i*picSizeX+j] - 
              (double)prevY[m*picSizeX+n];
            mse_temp += difference * difference;
          }
        }
        else
          /* Horizontal and vertical boundary violation */
          for(mse_temp=0.0, i=sy; i<=ey; i++){
            m = i + vecY;
            m = ( m < 0 ? 0 : ( m >= picSizeY ? picSizeY - 1 : m)) ;
            for(j=sx; j<=ex; j++){
              n = j + vecX;
              n = ( n < 0 ? 0 : ( n >= picSizeX ? picSizeX - 1 : n)) ;
              difference = (double)currY[i*picSizeX+j] - 
                (double)prevY[m*picSizeX+n];
              mse_temp += difference * difference;
            }
          }
    }

    mse_temp     /= currRegion->no_of_pels;
    mse_blk_match = mse_temp;
    
    /* Choose either zero motion parameters or the ones from
     * block matching
     */
    if(mse_temp < mse)
    {
      /* Accept block matching */
      memcpy(A, A_temp, 6 * sizeof(double));
      mse = mse_temp;
    }   
    else
    {
      /* Reject block matching */
      memcpy(A_temp, A, 6 * sizeof(double));
    }
    
    A[3] = A[1];
    A[1] = 0.0;
    
    memcpy(A_temp, A, 6 * sizeof(double));

    /* For 16x16 and 8x8 blocks compare the MSE to the one with
     * inherited motion parameters. Choose the better one
     */
    if (stage > 1  && mse_init < mse)
    {
      mse = mse_init;
      memcpy(A_temp, A_init, 6 * sizeof(double));
      memcpy(A,      A_init, 6 * sizeof(double));
    }

    /*
     * MINIMIZATION WITH GAUSS-NEWTON STEPS
     */
    stageNumber = 0;
    
    while(stageNumber < stopCondMaxNumIters){
      
      memcpy(A_in, A_temp, 6 * sizeof(double));

      /* Take a Gauss-Newton step from A_in to A_temp */
      est_red  = gnsGaussNewtonStep(currRegion, picSizeX, picSizeY, 
                                    currY, prevY, A_temp, A_in);
      
      mse_temp = mseRegionLum(A_temp, currRegion, currY, prevY);
      
      /* If the MSE dropped less than stopCondMinMseDiff,
       * quit minimizations
       */
      if(mse - mse_temp < stopCondMinMseDiff)
        stageNumber = stopCondMaxNumIters;
      
      stageNumber++;
      
      /* Compare the MSE of the region to the best MSE so far and
       * store the MSE and the motion parameters if the MSE is
       * lowered. If the MSE did not drop restore the best
       * parameters so far
       */
      if(mse_temp < mse)
      {
        memcpy(A, A_temp, 6 * sizeof(double));
        mse = mse_temp;
      }
      else
        memcpy(A_temp, A, 6 * sizeof(double));
    }
    
    *mseOut = mse;
    memcpy(motCoefLoc, A, 6 * sizeof(double));   
    
    /* Set zero motion flag if zero motion parameters yielded the
     * minimum mse
     */
    if(motCoefLoc[0] == 0.0 &&
       motCoefLoc[1] == 0.0 &&
       motCoefLoc[2] == 0.0 &&
       motCoefLoc[3] == 0.0 &&
       motCoefLoc[4] == 0.0 &&
       motCoefLoc[5] == 0.0)
      *zeroMotionFlag = 1;
  }

}

