#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "mainFlags.h"
#include "encMux.h"
#include "deblockFilt.h"
#include "mcp.h"
#include "IOimg.h"
#include "PredErrorCoder.h"
#include "motionOpt.h"

/**********************************************************************
  *
  * Code_P_Frame() performs the complete prediction & reconstruction
  *                of one frame (recoFr) given one or more previous
  *                refenrence frames (longTermMem). If deblocking
  *                filter is active then filteredfr holds the filtered
  *                reconstruction otherwise its content is the same
  *                as recoFr.
  *                All pictures are structures having Y, u and v
  *                components..
  * 
  *                It returns the total number of bits needed.
  *                Is also returns detailed information about the
  *                predicted/coded and filtered frame such as:
  *
  *                   - final object list of regions
  *                   - all the available bit counters
  *
  *********************************************************************/

int Code_P_Frame(struct frame **longTermMem, struct frame *currFr,
                 struct frame *prev_orig, struct frame *predFr,
                 struct frame *recoFr, struct frame *filteredfr,
                 int picSizeX, int picSizeY, int frNo,
                 object *objectList, int qp, double lmb,
                 Bits *bits, int numRefFrames,
                 int16 *slicePredMap, int sliceMode, int sliceBits,
                 int complexityMode, int useLoopFilter)
{
  int
    picSize = picSizeX * picSizeY,
    sliceXmin, sliceYmin,
    sliceYmax, sliceXmax,
    currSliceID, frameBitsAtStartSlice,
    mbPerLine = picSizeX/16,
    mbPosX, mbPosY, mbIndex,
    i, aQP,
    refFrame,
    numActiveRefFrames,
    orderedRefFrames[LONG_TERM_MEM_MAX_SIZE],
    mbPerFrame = picSizeX*picSizeY/(16*16),
    BMvectorstable[2*16],
    *BMvectors = BMvectorstable,
    frameLagr, frameLagrNoLTM, lagrNoLTM;

  static Byte
    predBufferY[256],
    predBufferU[64],
    predBufferV[64];

  static struct
    frame predBuffer =
      {&predBufferY[0], &predBufferU[0], &predBufferV[0]};

  struct frame *prevFr;

  static Byte
    *prevSubsampled[LONG_TERM_MEM_MAX_SIZE],
    *currSubsampled;

  frameLagr      = 0;
  frameLagrNoLTM = 0;

  fprintf(stderr, "Coding P-Frame %d\n", frNo);

  /*
   * slicePredMap[] holds the labels of each
   * 8x8 luma block according to how the slices
   * are constructed during the encoding.
   * If sliceMode is 0 then the whole frame
   * is treated as a big slice.
   */
  for (i = (picSize >> 6) - 1; i >= 0; i--)
    slicePredMap[i] = SHRT_MAX;

  /* Zeroing the prediction and reconstruction frames */
  for(i = 0; i < picSize; i++) {
    predFr->yptr[i] = (Byte) 0;
    recoFr->yptr[i] = (Byte) 0;
  }

  for(i = 0; i < picSize / 4; i++) {
    predFr->uptr[i] = (Byte) 0;
    predFr->vptr[i] = (Byte) 0;
    recoFr->uptr[i] = (Byte) 0;
    recoFr->vptr[i] = (Byte) 0;
  }

  numActiveRefFrames = numRefFrames;

  /* This is dummy information. Currently it is assumed that all the
   * reference frames are available.
   */

  orderedRefFrames[0] = 0;
  prevFr              = longTermMem[0];

  /*
   * Reset the internals of bitstream functions
   * and set all subfileds in "bits" to zero.
   */
  reset_all_bits();
  ResetBitCount(bits);

  /* The affine QP parameter is a function of the frame QP */
  aQP = affineQPs[qp];

  /* Send P picture header (delayed in long-term memory case) */
  if(numActiveRefFrames == 1)
    bits->header = SendPicHead(PCT_INTER, frNo, sliceMode,
                               useLoopFilter, qp,
                               picSizeX, picSizeY,
                               numActiveRefFrames,
                               orderedRefFrames);

  /*
   * Initialize parameters for the current slice.
   * These will be updated as the coding goes
   * Macroblock by Macroblock.
   */
  currSliceID           = 0;
  frameBitsAtStartSlice = 0,
  sliceXmin             = picSizeX;
  sliceYmin             = picSizeY;
  sliceYmax             = 0;
  sliceXmax             = 0;

  /* 
   * Send slice header (if slice mode is enabled)
   *                   (delayed in LTM case) 
   */

  if(numActiveRefFrames == 1)
    if (sliceMode)
    {
      /*
	     * Dump the header of very first Slice without
	     * any SSC since PSC has just been sent.
	     */
      bits->header += SendSliceHead(currSliceID, 0, qp,
                                    frNo, picSizeX, picSizeY,
                                    0, 0, 0);

      fprintf(stderr, "First slice start at P frame %3d, id %2d\n",
		          frNo, currSliceID);
    }

  /*
   * Initialize the subsampled images
   * which are used for block matching
   */

  motPrepareSubsampledImages(prevSubsampled, &currSubsampled,
                             longTermMem, currFr, picSizeX, picSizeY);

  /*
   * Main Macroblock by Macroblock coding loop.
   * mbIndex runs from 0, 1, to (Last MB number in frame).
   * (mbPosX, mbPosY) are the coordinates of currently
   * coded Macroblock.
   */

  for(mbIndex = 0, mbPosY = 0; mbPosY < picSizeY; mbPosY += 16){
    for(mbPosX = 0; mbPosX < picSizeX; mbPosX += 16, mbIndex++)
	  {

      /*
       * Dump slice header if slice mode is enabled and
       * a new slice is decided. (delayed in LTM case)
       */
      if(numActiveRefFrames == 1){
        if (sliceMode && sliceBits)
          SendSliceHeaderIfNeeded(&currSliceID, &frameBitsAtStartSlice,
                                  &sliceXmin, &sliceYmin,
                                  &sliceYmax, &sliceXmax,
                                  picSizeX, picSizeY, mbPerLine, mbIndex,
                                  qp, frNo, sliceBits, bits);
      }

      /*
       * Initial 32x32 motion estimation and
       * initialization of macroblock data.
       */
      if((mbPosX % 32 == 0) && (mbPosY % 32 == 0))
        motInitialEstimation32x32(objectList, currFr, longTermMem,
                                  currSubsampled, prevSubsampled, mbPosX, mbPosY,
                                  picSizeX, picSizeY, numActiveRefFrames, qp, lmb,
                                  complexityMode, BMvectors);

      /*
       * Motion estimation for the current macroblock
       * and its four subblocks.
       */
      for(refFrame = 0;refFrame < numActiveRefFrames;refFrame++)
        motEstimateMacroblock(objectList, currFr, longTermMem[refFrame],
                              currSubsampled, prevSubsampled[refFrame],
                              mbPosX, mbPosY, picSizeX, picSizeY, qp, lmb,
                              numActiveRefFrames, complexityMode, BMvectors,
                              refFrame);

      /*
       * Final motion estimation (evaluating prediction possibilities),
       * choosing of the coding mode and building the MC prediction
       * in predFr.
       */
      motFindFinalMotionAndMode(objectList, currFr, longTermMem, predFr,
                                &predBuffer, mbPosX, mbPosY,
                                picSizeX, picSizeY,
                                qp, aQP, lmb, &lagrNoLTM, currSliceID,
                                complexityMode, numActiveRefFrames);

      frameLagr      += (int)objectList[5*mbIndex].lagr;
      frameLagrNoLTM += lagrNoLTM;

      /*
       * Build prediction and code prediction error (delayed in LTM case)
       */
      if(numActiveRefFrames == 1){

        motBuildMBPrediction(&objectList[5*mbIndex],
                             longTermMem[objectList[5*mbIndex].refFrameNumber],
                             predFr, &predBuffer, picSizeX, picSizeY);

        /*
         * Encode prediction error and dump the data to bitstream
         * (including the motion related information) for the current MB.
         */
        SendMacroblockAccumBits(currFr, &predBuffer, recoFr,
                                qp, aQP, objectList,
                                5*mbIndex, picSizeX, picSizeY,
                                (float)lmb, slicePredMap,  currSliceID,
                                &sliceXmin, &sliceXmax,
                                &sliceYmin, &sliceYmax,
                                numActiveRefFrames,
                                bits, complexityMode);
      }

    } /* Macroblock by Macroblock loop */
  }
  
  /*
   * Delayed prediction error coding and bitsream generation for
   * long-term memory case
   */
  if(numActiveRefFrames > 1){

    if(frameLagrNoLTM < frameLagr){
      numActiveRefFrames = 1;
    }

    bits->header = SendPicHead(PCT_INTER, frNo, sliceMode,
                               useLoopFilter, qp,
                               picSizeX, picSizeY,
                               numActiveRefFrames,
                               orderedRefFrames);

    if (sliceMode)
    {
      /*
	     * Dump the header of very first Slice without
	     * any SSC since PSC has just been sent.
	     */
      bits->header += SendSliceHead(currSliceID, 0, qp,
                                    frNo, picSizeX, picSizeY,
                                    0, 0, 0);

      fprintf(stderr, "First slice start at P frame %3d, id %2d\n",
		          frNo, currSliceID);
    }

    for(mbIndex = 0, mbPosY = 0; mbPosY < picSizeY; mbPosY += 16){
      for(mbPosX = 0; mbPosX < picSizeX; mbPosX += 16, mbIndex++){

        /*
         * Dump slice header if slice mode is enabled
         * and a new slice is decided.
         */
        if (sliceMode && sliceBits)
          SendSliceHeaderIfNeeded(&currSliceID, &frameBitsAtStartSlice,
                                  &sliceXmin, &sliceYmin,
                                  &sliceYmax, &sliceXmax,
                                  picSizeX, picSizeY, mbPerLine, mbIndex,
                                  qp, frNo, sliceBits, bits);

        /*
         * If long-term memory did not help reducing total lagrangian
         * cost, final mode and motion are searched again (now with
         * one reference frame only)
         */
        if(frameLagrNoLTM < frameLagr)
          motFindFinalMotionAndMode(objectList, currFr, longTermMem, predFr,
                                    &predBuffer, mbPosX, mbPosY,
                                    picSizeX, picSizeY,
                                    qp, aQP, lmb, &lagrNoLTM, currSliceID,
                                    complexityMode, numActiveRefFrames);

        motBuildMBPrediction(&objectList[5*mbIndex],
                             longTermMem[objectList[5*mbIndex].refFrameNumber],
                             predFr, &predBuffer, picSizeX, picSizeY);

        /*
         * Encode prediction error and dump the data to bitstream
         * (including the motion related information) for the current MB.
         */
        SendMacroblockAccumBits(currFr, &predBuffer, recoFr,
                                qp, aQP, objectList,
                                5*mbIndex, picSizeX, picSizeY,
                                (float)lmb, slicePredMap,  currSliceID,
                                &sliceXmin, &sliceXmax,
                                &sliceYmin, &sliceYmax,
                                numActiveRefFrames,
                                bits, complexityMode);

      } /* Macroblock by Macroblock loop */
    }
  }

  /*
   * Apply deblocking filter.
   */

  filCompFilt(recoFr->yptr, picSizeX, picSizeY, qp, filteredfr->yptr,
              0, 1, objectList, NULL, useLoopFilter);

  if(useLoopFilter){
    for (i = 0; i < picSize; i++)
      (recoFr->yptr)[i] = (filteredfr->yptr)[i];
  }

  /*
   * Chroma is never filtered for P frames.
   */
  for (i = 0; i < picSize / 4; i++){
    (filteredfr->uptr)[i] = (recoFr->uptr)[i];
    (filteredfr->vptr)[i] = (recoFr->vptr)[i];
  }

  /*
   * The total # of bits for current P frame is computed.
   */
  bits->total = bits->motcoef +
                bits->coefy   + bits->coefu + bits->coefv +
                bits->cncy    + bits->cncuv + bits->mode  +
                bits->header;

  /*
   * padding with zeros for byte alignment
   * and flushing bitstream buffer.
   */
  put_bits(0,0,1,1);
  fflush(bitstream_bit);

  return bits->total;
}
