#ifndef _MCP_H_
#define _MCP_H_

/*
 * mcp.h
 *
 *       Copyright (c) 1997 Nokia Research Center, Tampere, FINLAND
 *
 * Project:
 *       Nokia Very Low Bitrate Codec
 *
 * Contains:
 *       Functions used for motion compensated prediction, which is
 *       based on quadtree segmentation and orthonormal polynomial
 *       motion fields.
 *
 */

#include "nrctypes.h"
#include "decDemux.h"

/*
 * Defines
 */

#define MCP_MAX_LOOKUP  45
#define MCP_TINY        1.0e-20

/*
 * Structs and typedefs
 */

typedef struct mcpNLink_s{
  int32 val;
  struct mcpNLink_s *next;
  struct mcpNLink_s *prev;
} mcpNLink_t;

typedef struct{
  int32 xMin;
  int32 yMin;
  int32 xMax;
  int32 yMax;
  int32 corVal;
  mcpNLink_t nIndList;
} mcpNode_t;

/*
 * Global arrays
 */

extern int affineQPs[32];

/*
 * Function prototypes
 */
void mcpCalcFinalParams(object *currRegion);

void mcpOrthToMvDelta(int *motCoefOrth, int32 *affineData, int regionSize);

/*
 *
 * mcpBuildSegmPred:
 *
 * Parameters:
 *      picSizeX    I: Y frame width as pixels
 *      picSizeY    I: Y frame height as pixels
 *      prevY       I: Previous reconstruction Y frame
 *      prevU       I: Previous reconstruction u frame
 *      prevV       I: Previous reconstruction v frame
 *      currRegion  B: Attribute list for segment
 *      predY       B: Updated prediction for Y frame
 *      predU       B: Updated prediction for u frame
 *      predV       B: Updated prediction for v frame
 *
 * Function:
 *      Builds prediction for Y, u, and v frames for one segment using
 *      segmentation information obtained from merging, and motion mode
 *      and motion coefficients for segment, decoded from the bitstream.
 *      The motion parameters are stored in the attribute list of the
 *      segment.
 *
 * Returns:
 *      Nothing
 *
 */

void mcpBuildSegmPred(int32 picSizeX, int32 picSizeY,
                      u_char *prevY, u_char *prevU, u_char *prevV,
                      object *currRegion, u_char *predY, u_char *predU,
                      u_char *predV);

/**************************************************************************
 *  mcpTransformGlob2LocCoeffs() computes the 6 motion parameters from
 *  global (frame wise) to local (region-wise) coordinate system.
 *
 *  Input args:
 *    A (1-d array buffer holding the 6 motion coefficients)
 *    current_region (pointer to object structure of the current region)
 *    pic_sizex, pic_sizey (sizes of the luminance image buffer)
 *
 *  Output args:
 *    B (1-d array buffer holding the resulting 6 motion coefficients)
 *
 *  Returns:
 *    void
 *************************************************************************/

void mcpTransformGlob2LocCoeffs(double A[6], object *current_region,
                                int pic_sizex, int pic_sizey, double B[6]);


/**************************************************************************
 *  mcpTransformLoc2GlobCoeffs() computes the 6 motion parameters from 
 *  local (region-wise) to global (frame wise) coordinate system.
 *
 *  Input args:
 *    A (1-d array buffer holding the 6 motion coefficients)
 *    current_region (pointer to object structure of the current region)
 *    pic_sizex, pic_sizey (sizes of the luminance image buffer)
 *
 *  Output args:
 *    B (1-d array buffer holding the resulting 6 motion coefficients)
 *
 *  Returns:
 *    void
 *************************************************************************/

void mcpTransformLoc2GlobCoeffs(double A[6], object *current_region,
                                int pic_sizex, int pic_sizey, double B[6]);

void mcpFindNeighs(object *object_list, int blkIndex, int regionSize, int dir,
                   int mbPerLine, int posX, int posY, int *numOfNeighs,
                   int neighIndexes[2], int refFrame, int *motionNeighs);

void mcpFindPredAlternatives(object *object_list, int blkIndex, int regionSize,
                             int posX, int posY, int mbPerLine,
                             u_char *predPossible, u_char *predDirsAvail,
                             u_char *predDir, u_char *predSelNeededUp,
                             u_char *predSelNeededLeft, int refFrame,
                             int *motionNeighs);

void mcpFindNeighsDecoder(objectBuffer_t *object_list, int blkIndex, 
                          int regionSize, 
                          int dir, int mbPerLine, int posX, int posY,
                          int *numOfNeighs, int neighIndexes[2],
                          int refFrame, int *motionNeighs);

void mcpFindPredAlternativesDecoder(objectBuffer_t *object_list, int blkIndex, 
                                    int regionSize, int posX, int posY,
                                    int mbPerLine, u_char *predPossible,
                                    u_char *predDirsAvail, u_char *predDir,
                                    u_char *predSelNeededUp,
                                    u_char *predSelNeededLeft,
                                    int refFrame,
                                    int *motionNeighs);

#endif










