/*
 * mvc.c
 *
 *    Copyright (c) 1999 Nokia Research Center, Tampere, FINLAND
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "nrctypes.h"
#include "tools.h"
#include "mvc.h"

#define SPLINE_SUBDIVISIONS            256
#define DENOMINATOR                    65536

#define Div2(X) (((X) >=0) ? ((X) >> 1) : (-((-(X)) >> 1)))

/*
 * Module scope variables
 */

short int mvcKernelInt[MVC_SIZE_SPLINE2];


/*
 * mvcPred4x4NoBoundCheckUV
 *
 *   Differentially calculates the motion compensated chroma coordinates
 *   associated with each pixel in the 4x4 chroma block, and composes
 *   4x4 chroma prediction block by invoking mvcSplIntRndNoBndChkUV() for
 *   each. Assumes that no boundary checking is required for any of the
 *   motion compensated coordinates.
 *
 * Parameters:
 *
 *   jStepHor      I: Used for differential calculation of motion
 *                    compensated coordinates.
 *   jStepVer      I:          "
 *   iStepHor      I:          "
 *   jStepVer      I:          "
 *   prevU         I: U component of previous reconstructed frame
 *   prevV         I: V component of previous reconstructed frame
 *   predU         O: U component of prediction frame
 *   predV         O: V component of prediction frame
 *   i             I: Vertical coordinate of upper left corner
 *                    of the 4x4 chroma prediction block.
 *   j             I: Horizontal coordinate of upper left corner
 *                    of the 4x4 chroma prediction block.
 *   iPrevUpLeft   I: Motion compensated vertical coordinate of
 *                    of the 4x4 chroma block
 *   jPrevUpLeft   I: Motion compensated horizontal coordinate of
 *                    of the 4x4 chroma block
 *   picSizeX      I: Frame width as pixels
 */

static void mvcPred4x4NoBoundCheckUV(int32 jStepHor32, int32 jStepVer32,
                                     int32 iStepHor32, int32 iStepVer32,
                                     u_char *prevU, u_char *prevV,
                                     u_char *predU, u_char *predV,
                                     int32 i, int32 j, int32 iPrevUpLeft32,
                                     int32 jPrevUpLeft32, int32 picSizeX)
{
  int32
    m, GoBack, n, i32, j32, AuxPtr;

  u_char 
    *predPtrU, *predPtrV, *imptrU, *imptrV;

  int32
    jfr, ifr,                     /* Fractional parts of jPrev and iPrev */
    one_minus_jfr, one_minus_ifr,
    resU, resV, zU, zV;           /* Auxiliary variables. */

  
  GoBack = 4;  
  AuxPtr = (i << 3) + j;  
  predPtrU = predU + AuxPtr;
  predPtrV = predV + AuxPtr;

  for(m = 0; m < 4; m++, predPtrU += GoBack, predPtrV += GoBack){
    i32 = iPrevUpLeft32;
    j32 = jPrevUpLeft32;
    for(n = 0; n < 4; n++){
  
      /*
       * Normal 2-D interpolation using 2x2 pixel area
       */
    
      /*
       * Find the upper left corner of the 2x2 image area
       */
      AuxPtr = (j32 >> 16)  + ((i32 >> 16) * picSizeX);
      imptrU = prevU + AuxPtr;
      imptrV = prevV + AuxPtr;

      /* 
       * Bilinear interpolation starts
       */

      jfr = (j32 & 0xffff)>>2;
      ifr = (i32 & 0xffff)>>2;
      one_minus_jfr = 0x4000 - jfr;
      one_minus_ifr = 0x4000 - ifr;

      resU = (imptrU[0] * one_minus_jfr + imptrU[1] * jfr)>>7;
      resV = (imptrV[0] * one_minus_jfr + imptrV[1] * jfr)>>7;

      imptrU+=picSizeX;
      imptrV+=picSizeX;

      zU = (imptrU[0] * one_minus_jfr + imptrU[1] * jfr)>>7;
      zV = (imptrV[0] * one_minus_jfr + imptrV[1] * jfr)>>7;

      resU = resU * one_minus_ifr + zU * ifr;
      resV = resV * one_minus_ifr + zV * ifr;

      /* 
       * Now, resU and resV hold the interpolated values with 21 fraction bits.
       */
      *predPtrU++ = (u_char)((resU + 0x100000)>>21);
      *predPtrV++ = (u_char)((resV + 0x100000)>>21);

      i32 += iStepHor32;
      j32 += jStepHor32;
    }  
    iPrevUpLeft32 += iStepVer32;
    jPrevUpLeft32 += jStepVer32;
  }
}

/*
 * mvcPred4x4BoundCheckUV
 *
 *   Differentially calculates the motion compensated chroma coordinates
 *   associated with each pixel in the 4x4 chroma block, and composes
 *   4x4 chroma prediction block by invoking mvcSplIntRndBndChkUV() for
 *   each.
 *
 * Parameters:
 *
 *   jStepHor      I: Used for differential calculation of motion
 *                   compensated coordinates.
 *   jStepVer      I:              "
 *   iStepHor      I:              "
 *   jStepVer      I:              "
 *   prevU         I: U component of previous reconstructed frame
 *   prevV         I: V component of previous reconstructed frame
 *   predU         O: U component of prediction frame
 *   predV         O: V component of prediction frame
 *   i             I: Vertical coordinate of upper left corner
 *                    of the 4x4 chroma prediction block.
 *   j             I: Horizontal coordinate of upper left corner
 *                    of the 4x4 chroma prediction block.
 *   iPrevUpLeft   I: Motion compensated vertical coordinate of
 *                    of the 4x4 chroma block
 *   jPrevUpLeft   I: Motion compensated horizontal coordinate of
 *                    of the 4x4 chroma block
 *   picSizeX      I: Frame width as pixels
 *   picSizeY      I: Frame height as pixels
 */

static void mvcPred4x4BoundCheckUV(int32 jStepHor32, int32 jStepVer32,
                                   int32 iStepHor32, int32 iStepVer32,
                                   u_char *prevU, u_char *prevV,
                                   u_char *predU, u_char *predV,
                                   int32 i, int32 j, int32 iPrevUpLeft32,
                                   int32 jPrevUpLeft32, int32 picSizeX,
                                   int32 picSizeY)
{
  int32
    m, GoBack, n, i32, j32, iPrev32, jPrev32, AuxPtr;

  u_char 
    *predPtrU, *predPtrV, *imptrU, *imptrV;

  int32
    jfr, ifr,                     /* Fractional parts of jPrev and iPrev */
    one_minus_jfr, one_minus_ifr,
    resU, resV, zU, zV;           /* Auxiliary variables. */

  
  GoBack = 4;  
  AuxPtr = (i << 3) + j;  
  predPtrU = predU + AuxPtr;
  predPtrV = predV + AuxPtr;

  for(m = 0; m < 4; m++, predPtrU += GoBack, predPtrV += GoBack){
    iPrev32 = iPrevUpLeft32;
    jPrev32 = jPrevUpLeft32;
    for(n = 0; n < 4; n++){

      i32 = iPrev32;
      j32 = jPrev32;

      /* 
       * Boundary Checking 
       */
      if(j32 < 0)
        j32 = 0;
      else if(j32 >= ((picSizeX << 16) - 65536))
        j32 = (picSizeX << 16) - 65536 - 1;

      if(i32 < 0)
        i32 = 0;
      else if(i32 >= (picSizeY << 16) - 65536)
        i32 = (picSizeY << 16) - 65536 - 1;

      /*
       * Normal 2-D interpolation using 2x2 pixel area
       */
    
      /*
       * Find the upper left corner of the 2x2 image area
       */
      AuxPtr = (j32 >> 16)  + ((i32 >> 16) * picSizeX);
      imptrU = prevU + AuxPtr;
      imptrV = prevV + AuxPtr;

      /* 
       * Bilinear interpolation starts
       */

      jfr = (j32 & 0xffff)>>2;
      ifr = (i32 & 0xffff)>>2;
      one_minus_jfr = 0x4000 - jfr;
      one_minus_ifr = 0x4000 - ifr;

      resU = (imptrU[0] * one_minus_jfr + imptrU[1] * jfr)>>7;
      resV = (imptrV[0] * one_minus_jfr + imptrV[1] * jfr)>>7;

      imptrU+=picSizeX;
      imptrV+=picSizeX;

      zU = (imptrU[0] * one_minus_jfr + imptrU[1] * jfr)>>7;
      zV = (imptrV[0] * one_minus_jfr + imptrV[1] * jfr)>>7;

      resU = resU * one_minus_ifr + zU * ifr;
      resV = resV * one_minus_ifr + zV * ifr;

      /* 
       * Now, resU and resV hold the interpolated values with 21 fraction bits.
       */
      *predPtrU++ = (u_char)((resU + 0x100000)>>21);
      *predPtrV++ = (u_char)((resV + 0x100000)>>21);

      iPrev32 += iStepHor32;
      jPrev32 += jStepHor32;
    }  
    iPrevUpLeft32 += iStepVer32;
    jPrevUpLeft32 += jStepVer32;
  }
}


/*
 * mvcPred8x8NoBoundCheckY:
 *
 *   Differentially calculates the motion compensated coordinates
 *   associated with each pixel in the 8x8 luminance block, and composes
 *   8x8 luminanance prediction block Assumes that no boundary checking
 *   is required for any of the motion compensated coordinates.
 *
 * Parameters:
 *
 *   jStepHor      I: Used for differential calculation of motion
 *                    compensated coordinates.
 *   jStepVer      I:              "
 *   iStepHor      I:              "
 *   jStepVer      I:              "
 *   prevY         I: Y component of previous reconstructed frame
 *   predY         O: Y component of prediction frame
 *   i             I: Vertical coordinate of upper left corner
 *                   of the 8x8 luminanance prediction block.
 *   j             I: Horizontal coordinate of upper left corner
 *                    of the  8x8 luminanance prediction block.
 *   iPrevUpLeft   I: Motion compensated vertical coordinate of
 *                    of the  8x8 luminanance block
 *   jPrevUpLeft   I: Motion compensated horizontal coordinate of
 *                    of the  8x8 luminanance block
 *   picSizeX      I: Frame width as pixels
 */

static void mvcPred8x8NoBoundCheckY(int32 jStepHor32, int32 jStepVer32,
                                    int32 iStepHor32, int32 iStepVer32,
                                    u_char *prevY, u_char *predY, int32 i,
                                    int32 j, int32 iPrevUpLeft32,
                                    int32 jPrevUpLeft32, int32 picSizeX)
{
  int32
    m, GoBack, n, i32, j32, sum1, sum, result;

  short int
    *spx,*spy;

  u_char
    *predPtrY, *img;


  GoBack = 8;
  predPtrY = predY + (i << 4) + j;

  for(m = 0; m < 8; m++, predPtrY += GoBack){
    i32 = iPrevUpLeft32;
    j32 = jPrevUpLeft32;
    for(n = 0; n < 8; n++){

      /*
       * Normal 2-D interpolation using 4x4 pixel area
       */

      /*
       * Find the upper left corner of the 4x4 image area
       */
      
      img = prevY + ((j32>>16) - 1) + (((i32>>16) - 1) * picSizeX);

      spx = mvcKernelInt + ((j32 & 0xff00) >> 6);
      spy = mvcKernelInt + ((i32 & 0xff00) >> 6);

      /*
       * Horizontal interpolation for the first line
       */
      sum1 = spx[0] * img[0];
      sum1 += spx[1] * img[1];
      sum1 += spx[2] * img[2];
      sum1 += spx[3] * img[3];
      sum = (sum1>>8) * (*spy); /* sum1 shifted right to fit into 16-bit */

      /*
       * Horizontal interpolation for the second line
       */
      sum1 = spx[0] * *(img += picSizeX);
      sum1 += spx[1] * img[1];
      sum1 += spx[2] * img[2];
      sum1 += spx[3] * img[3];
      sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

      /*
       * Horizontal interpolation for the third line
       */
      sum1 = spx[0] * *(img += picSizeX);
      sum1 += spx[1] * img[1];
      sum1 += spx[2] * img[2];
      sum1 += spx[3] * img[3];
      sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

      /*
       * Horizontal interpolation for fourth line
       */
      sum1 = spx[0] * *(img += picSizeX);
      sum1 += spx[1] * img[1];
      sum1 += spx[2] * img[2];
      sum1 += spx[3] * img[3];
      sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

      /* sum has 20 fraction bits 2^20 = 1048576 */
      result = (sum + 0x80000) >> 20;
      result = (result < 0) ? 0 : ((result > 255) ? 255 : result);
      *predPtrY++ = (u_char)result;

      i32 += iStepHor32;
      j32 += jStepHor32;
    }
    iPrevUpLeft32 += iStepVer32;
    jPrevUpLeft32 += jStepVer32;
  }

}


/*
 * mvcPred8x8BoundCheckY
 *
 *   Differentially calculates the motion compensated coordinates
 *   associated with each pixel in the 8x8 luminance block, and composes
 *   8x8 luminanance prediction block by invoking mvcSplIntRndBndChkY() for
 *   each.
 *
 * Parameters:
 *
 *   jStepHor      I: Used for differential calculation of motion
 *                    compensated coordinates.
 *   jStepVer      I:              "
 *   iStepHor      I:              "
 *   jStepVer      I:              "
 *   prevY         I: Y component of previous reconstructed frame
 *   predY         O: Y component of prediction frame
 *   i             I: Vertical coordinate of upper left corner
 *                    of the 8x8 luminanance prediction block.
 *   j             I: Horizontal coordinate of upper left corner
 *                    of the  8x8 luminanance prediction block.
 *   iPrevUpLeft   I: Motion compensated vertical coordinate of
 *                    of the  8x8 luminanance block
 *   jPrevUpLeft   I: Motion compensated horizontal coordinate of
 *                    of the  8x8 luminanance block
 *   picSizeX      I: Frame width as pixels
 *   picSizeY      I: Frame height as pixels
 */

static void mvcPred8x8BoundCheckY(int32 jStepHor32, int32 jStepVer32,
                                  int32 iStepHor32, int32 iStepVer32,
                                  u_char *prevY, u_char *predY, int32 i,
                                  int32 j, int32 iPrevUpLeft32,
                                  int32 jPrevUpLeft32, int32 picSizeX,
                                  int32 picSizeY)
{
  int32
    m, GoBack, n, i32, j32, offsetX, offsetY, sum1, sum, p, q, result;

  short int
    *spx,*spy;

  u_char
    *predPtrY, *img;

  GoBack = 8;
  predPtrY = predY + (i << 4) + j;

  for(m = 0; m < 8; m++, predPtrY += GoBack){
    i32 = iPrevUpLeft32;
    j32 = jPrevUpLeft32;
    for(n = 0; n < 8; n++){

      /*
       * Check if the 4x4 area for interpolation is not inside
       * inside the frame  
       */
      if((i32 < 65536) || (i32 >= ((picSizeY - 2) << 16)) || 
         (j32 < 65536) || (j32 >= ((picSizeX - 2) << 16))){

        /*
         * If there are pixels outside the frame, interpolation
         * is done using projection
         */
        int32 cy, cx, bx, by;   /* Pixel locations for interpolation */


        bx = (j32 >> 16) + 3;
        by = (i32 >> 16) + 3;

        offsetX = ((j32 & 0xff00) >> 6);
        offsetY = ((i32 & 0xff00) >> 6);

        /*
         * Interpolate pixel value by projecting pixels outside
         * the frame to corresponding boundary pixels
         */
        spx = mvcKernelInt + offsetX;
        spy = mvcKernelInt + offsetY;
        for(sum = 0, p = by - 4; p < by; p++){
          cy = p < 0 ? 0 : (p >= picSizeY ? picSizeY - 1 : p);
          cy *= picSizeX;
          q = bx - 4;

          cx = q < 0 ? 0 : (q >= picSizeX ? picSizeX - 1 : q);
          sum1 = spx[0] * prevY[cy + cx];
          q++;
          cx = q < 0 ? 0 : (q >= picSizeX ? picSizeX - 1 : q);
          sum1 += spx[1] * prevY[cy + cx];
          q++;
          cx = q < 0 ? 0 : (q >= picSizeX ? picSizeX - 1 : q);
          sum1 += spx[2] * prevY[cy + cx];
          q++;
          cx = q < 0 ? 0 : (q >= picSizeX ? picSizeX - 1 : q);
          sum1 += spx[3] * prevY[cy + cx];

          sum += (*spy++) * (sum1>>8);
        }
        result = (sum + 0x80000) >> 20;
        result = (result < 0) ? 0 : ((result > 255) ? 255 : result);
        *predPtrY++ = (u_char)result;

      }
      else{

        /*
         * Normal 2-D interpolation using 4x4 pixel area
         */

        /*
         * Find the upper left corner of the 4x4 image area
         */
      
        img = prevY + ((j32>>16) - 1) + (((i32>>16) - 1) * picSizeX);

        spx = mvcKernelInt + ((j32 & 0xff00) >> 6);
        spy = mvcKernelInt + ((i32 & 0xff00) >> 6);

        /*
         * Horizontal interpolation for the first line
         */
        sum1 = spx[0] * img[0];
        sum1 += spx[1] * img[1];
        sum1 += spx[2] * img[2];
        sum1 += spx[3] * img[3];
        sum = (sum1>>8) * (*spy); /* sum1 shifted right to fit into 16-bit */

        /*
         * Horizontal interpolation for the second line
         */
        sum1 = spx[0] * *(img += picSizeX);
        sum1 += spx[1] * img[1];
        sum1 += spx[2] * img[2];
        sum1 += spx[3] * img[3];
        sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

        /*
         * Horizontal interpolation for the third line
         */
        sum1 = spx[0] * *(img += picSizeX);
        sum1 += spx[1] * img[1];
        sum1 += spx[2] * img[2];
        sum1 += spx[3] * img[3];
        sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

        /*
         * Horizontal interpolation for fourth line
         */
        sum1 = spx[0] * *(img += picSizeX);
        sum1 += spx[1] * img[1];
        sum1 += spx[2] * img[2];
        sum1 += spx[3] * img[3];
        sum += (sum1>>8) * *(++spy); /* sum1 shifted right to fit into 16-bit */

        /* sum has 20 fraction bits */
        result = (sum + 0x80000) >> 20;
        result = (result < 0) ? 0 : ((result > 255) ? 255 : result);
        *predPtrY++ = (u_char)result;
      }

      i32 += iStepHor32;
      j32 += jStepHor32;
    }

    iPrevUpLeft32 += iStepVer32;
    jPrevUpLeft32 += jStepVer32;
  }

}

/*
 * mvcBuildPredRegionYUV
 *
 *   For an INTER-type segment, builds the part of prediction frame
 *   that is within segment boundaries, using motion compensation.
 *   The motion coefficients associated with the segment are assumed
 *   to be for global coordinate system.
 *
 * Parameters:
 *
 *   currRegion  I: Attribute list for segment
 *   A           I: Motion field coefficients
 *   prevY       I: Y component of previous reconstructed frame
 *   predY       O: Y component of prediction frame
 *   prevU       I: U component of previous reconstructed frame
 *   predU       O: U component of prediction frame
 *   prevV       I: V component of previous reconstructed frame
 *   predV       O: V component of prediction frame
 *   picSizeX    I: Frame width as pixels
 *   picSizeY    I: Frame height as pixels
 */

void mvcBuildPredRegionYUV(object *currRegion, int32 affineData[6],
                           u_char *prevY, u_char *predY,
                           u_char *prevU, u_char *predU,
                           u_char *prevV, u_char *predV,
                           int32 picSizeX, int32 picSizeY)
{
  int32
    mvx      = affineData[0],
    mvy      = affineData[3],
    jStepHor = affineData[2]+DENOMINATOR,
    jStepVer = affineData[1],
    iStepHor = affineData[5],
    iStepVer = affineData[4]+DENOMINATOR,

    /*
     * allRegionSafe == 1: no boundary checks needed for the all region
     *          == 0: boundary checks needed sometimes
     */
    allRegionSafe = 1,

    /*
     * Meaningful when allRegionSafe == 0.
     *
     * blockSafeArray[n] == 1: no boundary checks needed for the n'th block
     *                 in the segment.
     *             == 0: boundary checks needed
     *
     */
    blockSafeArray[396],

    xMin, xMax, yMin, yMax,  /* Bounding box of the region */
    iPrev, jPrev,          /* Motion compensated coordinates */
    i, j;              /* Loop counters */

  static int32
    /*
     * Used for detecting whether the picture size has changed since last
     * invocation. In case of change, some extra initializations are needed
     */
    lastPicSizeX = 0,

    /*
     * Right and bottom margins for boundary checking.
     */
    RightBnd, BotBnd;

  /* Modify the following values only if picture size has changed */
  if(picSizeX != lastPicSizeX)
  {
    lastPicSizeX = picSizeX;

    RightBnd = (picSizeX >> 1) - 2;
    BotBnd   = (picSizeY >> 1) - 2;
  }

  /* 
   * Doing color components first. This is due to the fact that if boundary
   * checks are not needed for motion compensated color coordinates, then it
   * is not needed for luminance coordinates as well; but not vice versa! 
   */
  /* 
   * Initializations 
   */
  picSizeX >>= 1;
  picSizeY >>= 1;

  xMin = currRegion->xmin >> 1;
  xMax = currRegion->xmax >> 1;
  yMin = currRegion->ymin >> 1;
  yMax = currRegion->ymax >> 1;

  /* 
   * Boundary checks 
   */

  /* 
   * Upper left corner 
   */

  j = currRegion->xmin;
  i = currRegion->ymin;
  jPrev = (j*DENOMINATOR + mvx) >> 1;
  iPrev = (i*DENOMINATOR + mvy) >> 1;
  if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
     (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
    allRegionSafe = 0;

  /* 
   * Upper right corner 
   */
  j = currRegion->xmax + 1;
  i = currRegion->ymin;
  jPrev = Div2(j*DENOMINATOR + mvx + (j-currRegion->xmin)*(jStepHor-DENOMINATOR));
  iPrev = Div2(i*DENOMINATOR + mvy + (j-currRegion->xmin)*iStepHor);
  if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
     (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
    allRegionSafe = 0;

  /* 
   * Lower right corner 
   */
  j = currRegion->xmax + 1;
  i = currRegion->ymax + 1;
  jPrev = Div2(j*DENOMINATOR + mvx + (j-currRegion->xmin)*(jStepHor-DENOMINATOR) +
           (i-currRegion->ymin)*jStepVer) ; 
  iPrev = Div2(i*DENOMINATOR + mvy + (j-currRegion->xmin)*iStepHor +
           (i-currRegion->ymin)*(iStepVer-DENOMINATOR));
  if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
     (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
    allRegionSafe = 0;

  /*
   * Lower left corner
   */
  j = currRegion->xmin;
  i = currRegion->ymax + 1;
  jPrev = Div2(j*DENOMINATOR + mvx + (i-currRegion->ymin)*jStepVer);
  iPrev = Div2(i*DENOMINATOR + mvy + (i-currRegion->ymin)*(iStepVer-DENOMINATOR));
  if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
     (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
    allRegionSafe = 0;

  /* 
   * Now, allRegionSafe = 1 means no need for boundary checks.
   * allRegionSafe=0 means we need boundary checks.        
   */

  if(allRegionSafe)
  {
    for(i = yMin; i <= yMax; i += 4)
    {
      for(j = xMin; j <= xMax; j += 4)
      {
        jPrev = j*DENOMINATOR + Div2(mvx + (2*j-currRegion->xmin)*(jStepHor-DENOMINATOR) +
                (2*i-currRegion->ymin)*jStepVer);
        iPrev = i*DENOMINATOR + Div2(mvy + (2*j-currRegion->xmin)*iStepHor +
                (2*i-currRegion->ymin)*(iStepVer-DENOMINATOR));

        mvcPred4x4NoBoundCheckUV(jStepHor, jStepVer, iStepHor, iStepVer,
                                 prevU, prevV, predU, predV, i-yMin, j-xMin,
                                 iPrev, jPrev, picSizeX);
      }
    }
  } /* allRegionSafe */
  else
  {
    
    /* 
     * Not safe! boundary check necessity for a block-by-block basis 
     */
    int32
    cnt = 0, blockSafe, ii, jj;

    for(i = yMin; i <= yMax; i += 4){
      for(j = xMin; j <= xMax; j += 4){

        blockSafe = 1;

        /* 
         * Boundary checks
         */

        /* 
         * Upper right corner
         */
        jj = j + 4;
        ii = i;
        jPrev = jj*DENOMINATOR + Div2(mvx + (2*jj-currRegion->xmin)*(jStepHor-DENOMINATOR) +
                (2*ii-currRegion->ymin)*jStepVer);
        iPrev = ii*DENOMINATOR + Div2(mvy + (2*jj-currRegion->xmin)*iStepHor +
                (2*ii-currRegion->ymin)*(iStepVer-DENOMINATOR));

        if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
           (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
          blockSafe = 0;

        /*
         * Lower right corner
         */
        jj = j + 4;
        ii = i + 4;
        jPrev = jj*DENOMINATOR + Div2(mvx + (2*jj-currRegion->xmin)*(jStepHor-DENOMINATOR) +
                (2*ii-currRegion->ymin)*jStepVer);
        iPrev = ii*DENOMINATOR + Div2(mvy + (2*jj-currRegion->xmin)*iStepHor +
                (2*ii-currRegion->ymin)*(iStepVer-DENOMINATOR));

        if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
           (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
          blockSafe = 0;

        /* 
         * Lower left corner
         */
        jj = j;
        ii = i + 4;
        jPrev = jj*DENOMINATOR + Div2(mvx + (2*jj-currRegion->xmin)*(jStepHor-DENOMINATOR) +
                (2*ii-currRegion->ymin)*jStepVer);
        iPrev = ii*DENOMINATOR + Div2(mvy + (2*jj-currRegion->xmin)*iStepHor +
                (2*ii-currRegion->ymin)*(iStepVer-DENOMINATOR));

        if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
           (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
          blockSafe = 0;


        /*
         * Upper left corner
         */
        jj = j;
        ii = i;
        jPrev = jj*DENOMINATOR + Div2(mvx + (2*jj-currRegion->xmin)*(jStepHor-DENOMINATOR) +
                (2*ii-currRegion->ymin)*jStepVer);
        iPrev = ii*DENOMINATOR + Div2(mvy + (2*jj-currRegion->xmin)*iStepHor +
                (2*ii-currRegion->ymin)*(iStepVer-DENOMINATOR));

        if((jPrev <= DENOMINATOR) || (jPrev >= RightBnd * DENOMINATOR) ||
           (iPrev <= DENOMINATOR) || (iPrev >= BotBnd   * DENOMINATOR))
          blockSafe = 0;

        /* 
         * Store the result for use in luminance prediction 
         */
        blockSafeArray[cnt++] = blockSafe;

        if(blockSafe)
          mvcPred4x4NoBoundCheckUV(jStepHor, jStepVer, iStepHor, iStepVer,
                                   prevU, prevV, predU, predV, i-yMin, j-xMin, 
                                   iPrev, jPrev, picSizeX);
        else
          mvcPred4x4BoundCheckUV(jStepHor, jStepVer, iStepHor, iStepVer,
                                 prevU, prevV, predU, predV, i-yMin, j-xMin,
                                 iPrev, jPrev, picSizeX, picSizeY);
      }
    }
  }

  /* 
   * Luminance prediction starts 
   */

  /* 
   * Initializations 
   */
  picSizeX *= 2;
  picSizeY *= 2;

  xMin = currRegion->xmin;
  xMax = currRegion->xmax;
  yMin = currRegion->ymin;
  yMax = currRegion->ymax;

  /* 
   * allRegionSafe value has been computed during chroma prediction 
   */

  if(allRegionSafe){
    for(i = yMin; i <= yMax; i += 8)
    {
      for(j = xMin; j <= xMax; j += 8)
      {
        jPrev = j*DENOMINATOR + mvx +
                (j-xMin)*(jStepHor-DENOMINATOR) +(i-yMin)*jStepVer;
        iPrev = i*DENOMINATOR + mvy +
                (j-xMin)*iStepHor +(i-yMin)*(iStepVer-DENOMINATOR);

        mvcPred8x8NoBoundCheckY(jStepHor, jStepVer, iStepHor, iStepVer,
                                prevY, predY, i-yMin, j-xMin, iPrev, jPrev, picSizeX);
      }
    }
  } /* all region safe */
  else{ 
    
    /* 
     * all region is not safe 
     */
    int32
      cnt = 0;

    for(i = yMin; i <= yMax; i += 8)
    {
      for(j = xMin; j <= xMax; j += 8)
      {
        jPrev = j*DENOMINATOR + mvx +
                (j-xMin)*(jStepHor-DENOMINATOR) +(i-yMin)*jStepVer;
        iPrev = i*DENOMINATOR + mvy +
                (j-xMin)*iStepHor +(i-yMin)*(iStepVer-DENOMINATOR);

        if(blockSafeArray[cnt++])
          mvcPred8x8NoBoundCheckY(jStepHor, jStepVer, iStepHor, iStepVer,
                                  prevY, predY, i-yMin, j-xMin, iPrev, jPrev,
                                  picSizeX);
      	else
	        mvcPred8x8BoundCheckY(jStepHor, jStepVer, iStepHor, iStepVer,
                                prevY, predY, i-yMin, j-xMin, iPrev, jPrev,
                                picSizeX, picSizeY);
      }
    }
  }
}


/*
 * fixedPointSpline:
 *
 *      Calculates the 16384 times the cubic spline interpolation value
 *      corresponding to s/256.0 subpixel coordinate. Returned value is
 *      in the range {-1216 .. 16384}.
 *
 * Parameters:
 *      s           I: An integer between {0 .. 511} corresponding to
 *                     the position s/256.0
 */

static int32 fixedPointSpline(int32 s)
{
  int32
    val;

  /* s must be a positive integer in the range 0 .. 511 */

  if (s < 256)
    val = (int32)16384 - ((s * s * ((int32)1280 - 3 * s) + 1024) >> 11);
  else if (s < 512)
  {
    s -= 256;
    val = -((s * ((int32)65536 + s * s - (int32)512 * s) + 1024) >> 11);
  }
  else
    val = 0;

  return val;
}

/*
 * mvcInitKernelInt:
 *
 *   Calculates the 16384 times the cubic spline interpolation values
 *   for each of the 256 subdivisions of a pixel interval and stores
 *   them in a precomputed array to be used during fast calculations.
 */

void mvcInitKernelInt(void)
{
  int32 i,tot;

  for (i=0; i<SPLINE_SUBDIVISIONS; i++)
  {
    tot = mvcKernelInt[4*i]    = (short int) fixedPointSpline(i+256);
    tot +=(mvcKernelInt[4*i+1] = (short int) fixedPointSpline(i));
    tot +=(mvcKernelInt[4*i+2] = (short int) fixedPointSpline(256-i));
    mvcKernelInt[4*i+3]        = (short int) (16384 - tot);
  }
}

void mvcInitKernel(void)
{
  mvcInitKernelInt();
}
