#ifndef _NCCGLOB_H_
#define _NCCGLOB_H_

/*
 * nccglob.h
 *
 *    Copyright (c) 1997-2000 Nokia Research Center
 *
 * Contains:
 *    Common header file for Nokia codec
 *
 */

#include <stdlib.h> /* for memory allocation functions and exit */
#include "debug.h"

/*
 * Common error indication values
 */

#define NCC_OK 1
#define NCC_ERROR -1


/*
 * Wrappers for memory allocation functions
 *
 * These wrappers are intended to replace standard C functions 
 * malloc, calloc, realloac and free. They differ from standard functions
 * in automatically including the check against NULL pointer. If a NULL
 * pointer is detected, nccExit is called. One may also
 * build some debugging features in these functions.
 *
 * Define NCC_ALLOC_NO_NULL_CHECK if the check against NULL pointer is not 
 * wanted. This define may be used e.g. in a Win32 codec backed up with
 * structured error handling (SEH).
 */

#ifdef NCC_ALLOC_NO_NULL_CHECK
   #define nccMalloc malloc
   #define nccCalloc calloc
   #define nccRealloc realloc
   #define nccFree free

#else
   void *nccMalloc(size_t size);
   void *nccCalloc(size_t num, size_t size);
   void *nccRealloc(void *memblock, size_t size);
   #define nccFree free
#endif


/*
 * Wrapper for exit
 *
 * This wrapper is intended to be used instead of standard C function like
 * exit and abort. In a normal command-line executable nccExit is an alias
 * to exit, but in a Win32 DLL nccExit just disables the usage of the codec
 * instance by causing an exception which is then handled by an upper-level
 * exception handler.
 */

#ifdef WIN32DLL
   void nccExitWin32DLL(int status);

   #define nccExit(status) \
   { \
      deb2f(stderr, "nccExit in file %s, line %d.\n", __FILE__, __LINE__); \
      nccExitWin32DLL(status); \
   }

#else
   #define nccExit(status) \
   { \
      deb2f(stderr, "nccExit in file %s, line %d.\n", __FILE__, __LINE__); \
      exit(status); \
   }
#endif


/*
 * nccAssert
 *
 * An assertion is a macro or function that complains loudly if an assumption
 * is not true. In MVC, there is a macro called nccAssert. The macro
 * is defined to 0 for the release version of the software. For the debug
 * version (_DEBUG defined), the macro causes an exit (using nccExit).
 */

#ifdef _DEBUG
   #define nccAssert(booleanExpr) \
      if (booleanExpr) \
         0; \
      else \
         nccExit(2)

#else
   #define nccAssert(booleanExpr) 0
#endif


/*
 * Settings for the bit buffer module
 */

/* Redirection of memory alloctions */
#define BIB_REDIRECT_ALLOCATION
#define BIB_ALLOC_H "nccglob.h"
#define BIB_MALLOC nccMalloc
#define BIB_CALLOC nccCalloc
#define BIB_REALLOC nccRealloc
#define BIB_DEALLOC nccFree

/* Removal of PSC emulation prevention bits is used */
#define BIB_PSC_INSERT

/* No ERR_BIB_NOT_ENOUGH_DATA when a buffer containing a single frame ends,
   no effect if bitbuf.c is used */
#define BIB_NO_EOB /*ACTIVATES checking to omit reading PSC due to BER*/
#ifdef BIB_NO_EOB
#define NO_BIT_ERROR 0      
#define VDX_SAW_SYNC 1 /*Error signal warning bits read from PSC*/
#define PSC_BITS_TOTAL      22
#endif




/* biblin.c is used instead of bitbuf.c */
#define BIBLIN

#endif
