#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mainFlags.h"
#include "tools.h"



/*
 * givens
 *
 * Performs one Givens rotation
 */

void givens(double g[2], double a, double b)
{
  double t;

  if(b == 0.0){
    g[0] = 1.0;
    g[1] = 0.0;
  }
  else if(fabs(b) > fabs(a)){
    t    = -a/b;
    g[1] = 1/sqrt(1.0+t*t);
    g[0] = g[1]*t;
  }
  else{
    t    = -b/a;
    g[0] = 1/sqrt(1.0+t*t);
    g[1] = g[0]*t;
  }
}


/*
 * updateMatrix
 *
 * Updates matrix R and vector z after removing one column from
 * the matrix.
 */

void updateMatrix(double R[6][5], double z[6], int coeffToRemove,
                  int numCoeffsLeft)
{

  double G[2],t1,t2;
  int i,j;

  for (i = coeffToRemove;i < numCoeffsLeft;i++){

    givens(G, R[i][i], R[i+1][i]);

    for (j = i;j < numCoeffsLeft;j++){
      t1        = R[i][j];
      t2        = R[i+1][j];
      R[i][j]   = G[0]*t1 - G[1]*t2;
      R[i+1][j] = G[1]*t1 + G[0]*t2;
    }

    t1     = z[i];
    t2     = z[i+1];
    z[i]   = G[0]*t1 - G[1]*t2;
    z[i+1] = G[1]*t1 + G[0]*t2;

  }
}

/*
 * evalRemovalCost
 *
 * Calculates approximated increase in prediction error when
 * removing given coefficient (amplitude of the 'left over
 * item' in z vector).
 */

double evalRemovalCost(double R[6][6], double z[6],
                       int coeffToRemove, int numCoeffsLeft)
{

  int
    j, k, column;

  double
    cost, Rn[6][5], zn[6];

  for(j = 0;j < numCoeffsLeft;j++)
    zn[j] = z[j];

  for(j = 0, column = 0;j < numCoeffsLeft;j++)
    if(j != coeffToRemove){

      for(k = 0;k < numCoeffsLeft;k++)
        Rn[k][column] = R[k][j];

      column++;
    }

  updateMatrix(Rn, zn, coeffToRemove, numCoeffsLeft-1);

  cost = zn[numCoeffsLeft-1]*zn[numCoeffsLeft-1];

  return cost;

}

/*
 * clearZeros
 *
 * Removes coefficients which would be zero after coefficient
 * quantization and updates matrix system accordingly.
 */

int clearZeros(double R[MATRIX_SIZE][MATRIX_SIZE], double *z, double *motCoeffs,
               int *activeCoeffs, int *numCoeffsLeft, int aQP)
{
  int 
    i, j, cnt=-1, no_zeros=0, column_no;
  static double 
    R_buf[6][6], z_buf[6], E[6][6], y[6], p[6], x[6];

  double
    chx;

  /* First detect the coefficients which will be zero after quantization */

  for(i=0; i<6; i++){
    z_buf[i] = z[i];
    if(motCoeffs[i] < aQP*0.5 && motCoeffs[i] > aQP*(-0.5)){
      activeCoeffs[i] = 0;
      no_zeros++;
    }
  }

  /* Construct R_buf by removing the columns of R associated with the 
   * coefficients-to-be-removed.
   */

  column_no = 6 - no_zeros;
  if(column_no > 0 && column_no < 6){

    for(j=0; j<6; j++)
      if(activeCoeffs[j])
        for(cnt++,i=0;i<6; i++)
          R_buf[i][cnt]=R[i][j];
        
    /* Calculate E = R_buf'*R_buf and y = R_buf'*z_buf. These are
     * needed for least square solution of overdetermined equation
     * R_buf*x = z_buf
     */

    for(i = 0;i < column_no;i++){

      for(j = 0;j < column_no;j++)
        for(E[i][j] = 0.0, cnt = 0;cnt < 6;cnt++)
          E[i][j] += R_buf[cnt][i]*R_buf[cnt][j];

      for(y[i] = 0.0, cnt = 0;cnt < 6;cnt++)
        y[i] += R_buf[cnt][i]*z_buf[cnt];
    }

    /* Cholesky decomposition of E (E = R'R) */
        
    for(i = 0;i < column_no; i++){
      for(j = i; j < column_no; j++){

        chx = E[i][j];

        if(i > 0)
          for(cnt = i - 1 ;cnt >= 0;cnt--)
            chx -= R[cnt][j] * R[cnt][i];

        if(i == j){
          if(chx < 0.00000001)
            chx = 0.00000001;
          p[i] = 1.0/(R[i][i]=sqrt(chx));
        }
        else{
          R[i][j] = chx*p[i];
          R[j][i] = 0.0;
        }                    
      }
    }
      
    /* Solve motion coefficient vector x */

    gnsTransposeBacksubstitution(R, y, z, column_no);
    gnsBacksubstitution(R, z, column_no, x); 
  }
  
  /* Decrement column & row size */

  *numCoeffsLeft -= no_zeros;
  
  /* Update the motion coefficient vector */

  if(column_no != 6){

    for(cnt = 0, i = 0;i < 6;i++){
      if(activeCoeffs[i] == 0)
        motCoeffs[i] = 0.0;
      else
        motCoeffs[i] = x[cnt++];
    }
  }

  return column_no;
}	

/*
 * updateSolution
 *
 * Removes one column from matrix R and one item from z.
 * Diagonalizes R and solves system for new set of motion
 * coefficients.
 */

void updateSolution(double R[6][6], double z[6], int coeffToRemove,
                    int *numCoeffsLeft, double motCoeffs[6])
{
  double
    Rn[6][5], zn[6];

  int
    i, j,
    column = 0;

  for (i=0; i < *numCoeffsLeft; i++)
    if (i != coeffToRemove){

      for (j=0; j < *numCoeffsLeft; j++)
        Rn[j][column] = R[j][i];

      column++;
    }

  for (i=0; i < *numCoeffsLeft; i++)
    zn[i] = z[i];

  updateMatrix(Rn, zn, coeffToRemove, *numCoeffsLeft-1);

  for (i=0; i < *numCoeffsLeft-1; i++)
    for (j=0; j < *numCoeffsLeft-1; j++)
      R[i][j] = Rn[i][j];
    
  for (i=0; i < *numCoeffsLeft; i++)
    z[i] = zn[i];

  *numCoeffsLeft = *numCoeffsLeft-1;

  gnsBacksubstitution(R, z, *numCoeffsLeft, motCoeffs);

}

/*
 * chooseCoeffToRemove
 *
 * Chooses the coefficient to remove by calculating estimate
 * of increase in prediction error for each of the remaining
 * coefficients.
 */

int chooseCoeffToRemove(double R[MATRIX_SIZE][MATRIX_SIZE], double *z,
                        int numCoeffsLeft)
     
{

  double
    eMin = HUGE,
    cost;

  int
    coeffToRemove = 0,
    coeffIndex;

  for (coeffIndex = 0;coeffIndex < numCoeffsLeft;coeffIndex++){

    cost = evalRemovalCost(R, z, coeffIndex, numCoeffsLeft);

    if (eMin > cost || coeffIndex == 0) {
      eMin          = cost;
      coeffToRemove = coeffIndex;
    }
  }

  return coeffToRemove;
}

/*
 * removeCoeff
 *
 * Chooses the coefficient to remove, removes it from the motion
 * model and updates matrix system accordingly.
 */

void removeCoeff(double R[MATRIX_SIZE][MATRIX_SIZE],double *z,
                 double *motCoeffs, int *activeCoeffs, int *numCoeffsLeft)		     
{
  int i, coeffToRemove, number;
  double
    motCoeffsNew[MATRIX_SIZE];

  if(*numCoeffsLeft == 1){

    *numCoeffsLeft = 0;

    for(i = 0;i < MATRIX_SIZE;i++){
      activeCoeffs[i] = 0;
      motCoeffs[i]    = 0.0;	
    }     
  }
  else{

    coeffToRemove = chooseCoeffToRemove(R, z, *numCoeffsLeft);

    updateSolution(R, z, coeffToRemove, numCoeffsLeft, motCoeffsNew);
  
    for(i = 0,number = -1;i < MATRIX_SIZE;i++){

      if(activeCoeffs[i] == 1)
        number += 1;
      if(number == coeffToRemove){
        activeCoeffs[i] = 0;
        break;
      }
    }
            
    for(i = 0,number = 0;i < MATRIX_SIZE;i++)
      motCoeffs[i] = (activeCoeffs[i] == 1) ? motCoeffsNew[number++] : 0.0;
  }  
}

  
