/*
 * toolGaussNewton.c
 *
 *    Copyright (c) 2000 Nokia Research Center, Tampere, FINLAND
 *
 * Project:
 *    Nokia Very Low Bitrate Codec
 *
 * Contains:
 *    Functions for building and solving least square equations for
 *    minimization of region's prediction mse w.r.t. motion coefficients.
 *
 * Author:
 *    Nokia Research Center
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>

#include "mainFlags.h"
#include "tools.h"

#define REG 0.0001

void splDerivativeXxX(int xx, int xy, int yx, int yy, int *pels,
                      int *drx, int *dry, Byte *yin, int i32, int j32,
                      int picSizeX, int picSizeY, int blkSideLen);

void splDerivativeNoAffineXxX(int32 *pels, int32 *dx, int32 *dy, Byte *yin,
                              int32 i, int32 j, int picSizeX, int picSizeY,
                              int blkSideLen);

/*
 * buildLsEqnXxX:
 *
 * Parameters:
 *     LHS              O: Buffer for the upper triangular part of the Left
 *                         Hand Side Matrix of least squares equation
 *     rhs              O: Right Hand Side vector of least squares equation
 *     xValues          I: xValues
 *     yValues          I: yValues
 *     xstepx           I: Step values for interpolating across the block
 *     xstepy           I: -"-
 *     ystepx           I: -"-
 *     ystepy           I: -"-
 *     i,j              I: Coordinates of the block in current frame
 *     prevY            I: Y component of previous reconstructed frame
 *     currY            I: Y component of current frame
 *     iPrevUpLeftD     I: Coordinates of the block in previous frame
 *     jPrevUpLeftD     I: 
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     blkSideLen       I: Length of the side of the block
 *
 * Function:
 *     This function is a part of buildLsEqn (inner loop of the original
 *     gnsBuildLsEqn[Glob] functions). It is used to build matrix LHS and
 *     vector rhs to satisfy: LHS(x-x0) = rhs, where x is the unknown
 *     motion parameter vector and x0 the initial motion parameter vector.
 *
 * Returns:
 *     Nothing
 *
 */

void buildLsEqnXxX(double LHS[], double rhs[],
                   double xValues[], double yValues[],
                   int xstepx, int xstepy, 
                   int ystepx, int ystepy,
                   Byte currY[], int i, int j, 
                   Byte prevY[], double iPrevUpLeftD, double jPrevUpLeftD,
                   int picSizeX, int picSizeY, int blkSideLen)
{
  Byte 
    *curptr = currY + i*picSizeX + j;

  int
    m, n,
    iPrevUpLeft, jPrevUpLeft,
    difA[64], dX[64], dY[64], *dxp, *dyp, *diffp;

  double
    x, y,
    derX, derY, diff,
    *xPtr,
    contribPel[6];

  /* Location where the upper left corner motion vector points
   * in 16-bit fraction fixed point representation
   */
  iPrevUpLeft = (int)(65536.0*iPrevUpLeftD)+128;
  jPrevUpLeft = (int)(65536.0*jPrevUpLeftD)+128;

  /*
   * Find interpolated pixel values (difA) and both horizontal and 
   * vertical derivatives (dX, dY) of the reference frame. difA will
   * have 20 fraction bits and dX, dY will have 19 fraction bits.
   *
   * Interpolations accept coordinates in 16-bit fraction fixed-point
   * representation
   */
  if (((iPrevUpLeft & 0xff00) | (jPrevUpLeft & 0xff00) |
       (xstepx - (1<<16)) | (ystepy - (1<<16)) | ystepx | xstepy) == 0) {
    splDerivativeNoAffineXxX(difA, dX, dY, prevY,
                             iPrevUpLeft>>16, jPrevUpLeft>>16,
                             picSizeX, picSizeY, blkSideLen);
  }
  else {
    splDerivativeXxX(xstepx, xstepy, ystepx, ystepy, difA, dX, dY, prevY,
                     iPrevUpLeft, jPrevUpLeft, picSizeX, picSizeY,
                     blkSideLen);
  }

  /*
   * Calculate difference block
   */
  diffp = difA;
  for(m=0; m<blkSideLen; m++, curptr += picSizeX){
    for(n=0; n<blkSideLen; n++, diffp++){
      *diffp = ((int)curptr[n] << 20) - (*diffp);
    }
  }

  diffp = difA;
  dxp   = dX;
  dyp   = dY;

  for(m=0; m<blkSideLen; m++)
    for(xPtr = xValues, y=yValues[m], n=0; n<blkSideLen ; n++) {

      /*
       * LHS is formed by multiplying matrix E by its transpose from the
       * left: LHS = E'E. E is a 6*(blkSideLen*blkSideLen) matrix formed as:
       * 
       * [ dxp[0] y[0]*dxp[0] x[0]*dxp[0] dyp[0] y[0]*dyp[0] x[0]*dyp[0] ]
       * [ dxp[1] y[1]*dxp[1] x[1]*dxp[1] dyp[1] y[1]*dyp[1] x[1]*dyp[1] ]
       * [                              ...                              ]
       * [ dxp[p] y[p]*dxp[p] x[p]*dxp[p] dyp[p] y[p]*dyp[p] x[p]*dyp[p] ]
       *
       * LHS is a symmetric matrix and thus only upper diagonal part
       * needs to be calculated here.
       *
       * rhs is calculated as: rhs = E'y where y is a vector having the
       * prediction differences as elements
       *
       * dxp and dyp below are in 19-bit fraction and diffp in 20-bit
       * fraction fixed point representation. That is, LHS[i] needs to
       * be scaled by 1/(2^19*2^19) and rhs[i] by 1/(2^19*2^20) before
       * using. This rescaling is done by higher level functions.
       */

      x    = (*xPtr++);
      derX = (*dxp++);
      derY = (*dyp++);
      diff = (*diffp++);

      contribPel[0] = derX;
      contribPel[1] = derX*y;
      contribPel[2] = derX*x;
      
      contribPel[3] = derY;
      contribPel[4] = derY*y;
      contribPel[5] = derY*x;

      /* The following loop is unrolled below
      
      for(cnt=0, p=0; p<6; p++){
        rhs[p] += contribPel[p] * diff;
        for(q=p; q<6; q++)
          LHS[cnt++] += contribPel[p]*contribPel[q];
      } 

      */

      rhs[0]  += contribPel[0] * diff;      
      rhs[1]  += contribPel[1] * diff;      
      rhs[2]  += contribPel[2] * diff;      
      rhs[3]  += contribPel[3] * diff;      
      rhs[4]  += contribPel[4] * diff;      
      rhs[5]  += contribPel[5] * diff;      

      LHS[0]  += contribPel[0] * contribPel[0];
      LHS[1]  += contribPel[0] * contribPel[1];
      LHS[2]  += contribPel[0] * contribPel[2];
      LHS[3]  += contribPel[0] * contribPel[3];
      LHS[4]  += contribPel[0] * contribPel[4];
      LHS[5]  += contribPel[0] * contribPel[5];

      LHS[6 ] += contribPel[1] * contribPel[1];
      LHS[7 ] += contribPel[1] * contribPel[2];
      LHS[8 ] += contribPel[1] * contribPel[3];
      LHS[9 ] += contribPel[1] * contribPel[4];
      LHS[10] += contribPel[1] * contribPel[5];

      LHS[11] += contribPel[2] * contribPel[2];
      LHS[12] += contribPel[2] * contribPel[3];
      LHS[13] += contribPel[2] * contribPel[4];
      LHS[14] += contribPel[2] * contribPel[5];

      LHS[15] += contribPel[3] * contribPel[3];
      LHS[16] += contribPel[3] * contribPel[4];
      LHS[17] += contribPel[3] * contribPel[5];

      LHS[18] += contribPel[4] * contribPel[4];
      LHS[19] += contribPel[4] * contribPel[5];

      LHS[20] += contribPel[5] * contribPel[5];
    }

}

/*
 * gnsBuildLsEqn:
 *
 * Parameters:
 *     LHS              O: Left Hand Side Matrix of least squares equation
 *     rhs              O: Right Hand Side vector of least squares equation
 *     region           I: Attribute list for segment
 *     prevY            I: Y component of previous reconstructed frame
 *     currY            I: Y component of current frame
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     motCoef[6]       I: Input motion coefficients
 *
 * Function:
 *     Reduces the Gauss-Newton minimization step on the segment prediction 
 *     mse to a linear equation: LHS*x = rhs, where x = newMotCoef - motCoef.
 *
 *     NOTE: Local coordinate system is assumed.
 *
 * Returns:
 *     Nothing
 *
 */

void gnsBuildLsEqn(double LHS[][6], double rhs[], object *region,
                   Byte prevY[], Byte currY[],
                   int picSizeX, int picSizeY, double motCoef[])
{
  int 
    i, j,                    /* Looping variables           */
    xMin, xMax, yMin, yMax,  /* Bounding box of the region  */

    xstepx, xstepy,  ystepx, ystepy,
    blkSideLen = 8;

  double 
    invSizeX, invSizeY,      /* Values used for calculating (x,y)           */
    jPrevUpLeft,             /* Motion compensated horizontal coordinate    */
                             /* for the upper left corner of a square block */
    iPrevUpLeft,             /* Motion compensated vertical coordinate      */
                             /* for the upper left corner of a square block */
    bufLHS[21],              /* Buffer that holds upper tri. part of LHS    */
    *xValues, *yValues;      /* Pointers for x&y values                     */

  /* Scaling factors for LHS matrix and rhs vector (they are generated
   * by fixed point operations in buildLsEqnXxX()
   */
  static double
    scaleRhs = 1.0/(1048576.0*524288.0),
    scaleLhs = 1.0/(524288.0*524288.0),

    /* Affine basis functions for local system (for regions of sizes
     * 4, 8, 16 and 32 pixels)
     */
    xyValues3[] = {
      -3/3.0, -1/3.0, 1/3.0,  3/3.0},
    xyValues7[] = {
      -7/7.0, -5/7.0, -3/7.0, -1/7.0,
       1/7.0,  3/7.0,  5/7.0,  7/7.0},
    xyValues15[] = {
      -15/15.0, -13/15.0, -11/15.0, -9/15.0,
       -7/15.0,  -5/15.0,  -3/15.0, -1/15.0,
        1/15.0,   3/15.0,   5/15.0,  7/15.0,
        9/15.0,  11/15.0,  13/15.0, 15/15.0},
    xyValues31[] = {
      -31/31.0, -29/31.0, -27/31.0, -25/31.0,
      -23/31.0, -21/31.0, -19/31.0, -17/31.0,
      -15/31.0, -13/31.0, -11/31.0,  -9/31.0,
       -7/31.0,  -5/31.0,  -3/31.0,  -1/31.0,
        1/31.0,   3/31.0,   5/31.0,   7/31.0,
        9/31.0,  11/31.0,  13/31.0,  15/31.0,
       17/31.0,  19/31.0,  21/31.0,  23/31.0,
       25/31.0,  27/31.0,  29/31.0,  31/31.0};

  /* Initializations */

  rhs[0] = 0.0;
  rhs[1] = 0.0;
  rhs[2] = 0.0;
  rhs[3] = 0.0;
  rhs[4] = 0.0;
  rhs[5] = 0.0;

  for(j=0; j<21; j+=3) {
    bufLHS[j]   = 0.0;
    bufLHS[j+1] = 0.0;
    bufLHS[j+2] = 0.0;
  }  
  
  xMin = region->xmin; xMax = region->xmax;
  yMin = region->ymin; yMax = region->ymax;

  /* Choose the basis functions to be used */
  switch (xMax-xMin) {
  case 7:
    xValues = xyValues7;
    invSizeX = 1.0/7.0;
    break;
  case 15:
    xValues = xyValues15;
    invSizeX = 1.0/15.0;
    break;
  case 31:
    xValues = xyValues31;
    invSizeX = 1.0/31.0;
    break;
  case 3:
    xValues = xyValues3;
    invSizeX = 1.0/3.0;
    blkSideLen = 4;
    break;
  default:
    fprintf(stderr, "Region x-size not supported in gnsBuildLsEqn\n");
    exit(1);
  }

  switch (yMax-yMin) {
  case 7:
    yValues = xyValues7;
    invSizeY = 1.0/7.0;
    break;
  case 15:
    yValues = xyValues15;
    invSizeY = 1.0/15.0;
    break;
  case 31:
    yValues = xyValues31;
    invSizeY = 1.0/31.0;
    break;
  case 3:
    yValues = xyValues3;
    invSizeY = 1.0/3.0;
    blkSideLen = 4;
    break;
  default:
    fprintf(stderr, "Region y-size not supported in gnsBuildLsEqn\n");
    exit(1);
  }

  /* Differences between neighboring motion vectors when generating
   * the motion vector field (in 16-bit fraction fixed point
   * representation)
   */
  ystepy = (int)(65536.0*(1.0 + (2.0*motCoef[4])*invSizeY));
  ystepx = (int)(65536.0*((2.0*motCoef[5])*invSizeX));
  xstepx = (int)(65536.0*(1.0 + (2.0*motCoef[2])*invSizeX));
  xstepy = (int)(65536.0*((2.0*motCoef[1])*invSizeY));

  /* Building of LHS matrix and rhs vector starts */
  for(i=yMin; i<=yMax; i+=blkSideLen){
    for(j=xMin; j<=xMax; j+=blkSideLen)
    {

      /* Location where the motion vector of the upper left
       * corner of this blkSideLen*blkSideLen block points
       */
      jPrevUpLeft = j + motCoef[0] + motCoef[1]*yValues[i-yMin]
                                   + motCoef[2]*xValues[j-xMin];

      iPrevUpLeft = i + motCoef[3] + motCoef[4]*yValues[i-yMin]
                                   + motCoef[5]*xValues[j-xMin];
       
      /* Add the contribution of this blkSideLen*blkSideLen
       * block to the total LHS and rhs
       */
      buildLsEqnXxX(bufLHS, rhs, xValues+j-xMin, yValues+i-yMin,
                    xstepx, xstepy, ystepx, ystepy,
                    currY, i, j, prevY, iPrevUpLeft, jPrevUpLeft,
                    picSizeX, picSizeY, blkSideLen);
    }
  }

  /* Form LHS by symmetrically copying upper triangular part buffer
   * and rescaling elements
   */

  LHS[0][0] =             bufLHS[0]  * scaleLhs;
  LHS[0][1] = LHS[1][0] = bufLHS[1]  * scaleLhs;
  LHS[0][2] = LHS[2][0] = bufLHS[2]  * scaleLhs;
  LHS[0][3] = LHS[3][0] = bufLHS[3]  * scaleLhs;
  LHS[0][4] = LHS[4][0] = bufLHS[4]  * scaleLhs;
  LHS[0][5] = LHS[5][0] = bufLHS[5]  * scaleLhs;

  LHS[1][1] =             bufLHS[6]  * scaleLhs;
  LHS[1][2] = LHS[2][1] = bufLHS[7]  * scaleLhs;
  LHS[1][3] = LHS[3][1] = bufLHS[8]  * scaleLhs;
  LHS[1][4] = LHS[4][1] = bufLHS[9]  * scaleLhs;
  LHS[1][5] = LHS[5][1] = bufLHS[10] * scaleLhs;

  LHS[2][2] =             bufLHS[11] * scaleLhs;
  LHS[2][3] = LHS[3][2] = bufLHS[12] * scaleLhs;
  LHS[2][4] = LHS[4][2] = bufLHS[13] * scaleLhs;
  LHS[2][5] = LHS[5][2] = bufLHS[14] * scaleLhs;

  LHS[3][3] =             bufLHS[15] * scaleLhs;
  LHS[3][4] = LHS[4][3] = bufLHS[16] * scaleLhs;
  LHS[3][5] = LHS[5][3] = bufLHS[17] * scaleLhs;

  LHS[4][4] =             bufLHS[18] * scaleLhs;
  LHS[4][5] = LHS[5][4] = bufLHS[19] * scaleLhs;

  LHS[5][5] =             bufLHS[20] * scaleLhs;

  rhs[0] *= scaleRhs;
  rhs[1] *= scaleRhs;
  rhs[2] *= scaleRhs;
  rhs[3] *= scaleRhs;
  rhs[4] *= scaleRhs;
  rhs[5] *= scaleRhs;
}


/*
 * gnsCholeskyDec:
 *
 * Parameters:
 *     inpMatr          I: 6x6 symmetric input matrix
 *     outMatr          O: 6x6 upper triangular output matrix
 *
 * Function:
 *     Performs Cholesky decomposition of inpMatr, i.e. calculates
 *     upper triangular outMatr such that outMatr'*outMatr = inpMatr
 * Returns:
 *     1 if inpMatr is positive definite, i.e. factorization has been done
 *     0 if inpMatr is not positive definite, i.e. factorization has failed
 *
 */

int gnsCholeskyDec(double inpMatr[6][6] , double outMatr[6][6] )
{ 
  int 
    i, j, k;     /* Looping Variables */
  double 
    scale;       /* scaling factor for each row */
  double 
    invDiag[6];  /* Vector of the inverse of diagonal entries of outMatr */


  /*
   *  Cholesky decomposition starts
   */

  for(i = 0; i < 6; i++)
    for(j = i; j < 6; j++)
    {
      /* Compute the scaling factor */
      scale=inpMatr[i][j];
      if ( i > 0) for( k = i - 1 ; k >= 0 ; k--)
        scale -= outMatr[k][j] * outMatr[k][i];

      /* Compute i'th row of outMatr */
      if(i==j){
        if(scale <= 0 ){  /* If inpMatr is singular */
          return(0);
        }
        else              /* Normal operation */
          invDiag[i] =  1.0/(outMatr[i][i]=sqrt(scale));
      }
      else{
        outMatr[i][j] = scale*invDiag[i]; /* Upper triangular part          */
        outMatr[j][i] = 0.0;              /* Lower triangular part set to 0 */
      }                    
    }
  
  return(1); /* Signal that Cholesky factorization is successfully performed */
}

/*
 * gnsTransposeBacksubstitution:
 *
 * Parameters:
 *     U          I: a 6x6 upper triangular matrix
 *     rhs        I: a vector of size 6
 *     x          O: vector that satisfies U'*x = rhs
 *
 * Function:
 *     Solves U'*x = rhs for x
 *
 * Returns:
 *     Nothing
 *
 * History
 *     8.4.97         Initial version.
 */

void gnsTransposeBacksubstitution(double U[6][6], double rhs[], 
                                  double x[], int order)
{
  int 
    i,j;              /* Looping variables */
  double 
    sum;              /* Holds backsubstitution from already handled rows */

  /* Backsubstitution starts */
  x[0] = rhs[0]/U[0][0];               /* First row of U'                   */
  for (i = 1; i < order; i++){         /* For the rows 1..order-1           */

    for (j = 0, sum = 0.0; j < i; j++) /* Backsubst already solved unknowns */
      sum += x[j]*U[j][i];

    x[i]=(rhs[i] - sum)/U[i][i];       /* i'th component of solution vect.  */
  }
}


/*
 * gnsTransposeBacksubstitution:
 *
 * Parameters:
 *     R          I: a 6x6 upper triangular matrix
 *     z          I: a vector of size 6
 *     A          O: vector that satisfies R*A = z
 *
 * Function:
 *     Solves by backsubstitution the system given by the matrix
 *     equation: R * A = z, with R being upper-triangular.
 *
 * Returns:
 *     Nothing
 *
 */

void gnsBacksubstitution(double R[6][6], double z[6], int R_size, double A[6])
{
  int
    i, j;

  double
    sum;

  R_size--;

  A[R_size] = z[R_size] / R[R_size][R_size];

  for (i = R_size-1; i >= 0; i--) {

    for (j = i+1, sum = 0.0; j <= R_size; j++)
      sum += R[i][j] * A[j];

    A[i] = (z[i] - sum) / R[i][i];

  }
}

/*
 * gnsSolveByChol:
 *
 * Parameters:
 *     LHS              I: Left Hand Side Matrix(6x6) of least squares eqn.
 *     rhs              I: Right Hand Side vector(6x1) of least squares eqn.
 *     x                O: vector that satisfies LHS*x = rhs
 *
 * Function:
 *     Solves LHS*x = rhs for x. Uses Cholesky decomposition + two
 *     backsubstitutions for the solution. In case LHS is singular, 
 *     regularizes LHS by adding a small positive value to the
 *     diagonals of LHS
 *
 * Returns:
 *     Nothing
 *
 */

void gnsSolveByChol(double LHS[][6], double rhs[], double x[])
{ 
  double
    aux[6],     /* Auxiliary vector */
    U[6][6];    /* Upper triangular Cholesky factor of LHS */
  int
    i;          /* Looping variable */
  
  /* The equation to be solved is LHSx = rhs */
  
  /* Compute upper triangular U such that U'*U = LHS */
  if(gnsCholeskyDec(LHS,U)){ /* If Cholesky decomposition has been successful */

    /* Now, the equation is  U'*U*x = rhs, where U is upper triangular
     * Solve U'*aux = rhs for aux
     */
    gnsTransposeBacksubstitution(U, rhs, aux, 6);         

    /* The equation is now U*x = aux, solve it for x (new motion coefficients) */
    gnsBacksubstitution(U, aux, 6, x);   

  }
  else{ /* LHS was singular */ 

    /* Regularize LHS */
    for(i=0; i<6; i++)
      LHS[i][i] += REG;

    /* Compute upper triangular U such that U'*U = regularized LHS */
    gnsCholeskyDec(LHS, U);

    /* Solve  U'*aux = rhs for aux */  
    gnsTransposeBacksubstitution(U, rhs, aux, 6);   

    /* Solve U*x = aux for x */
    gnsBacksubstitution(U, aux, 6, x);
  }  
}

/*
 * gnsGaussNewtonStep:
 *
 * Parameters:
 *     region           I: Attribute list for segment
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     currY            I: Y component of current frame
 *     prevY            I: Y component of previous reconstructed frame
 *     newMotCoef       O: Motion coefficients resulting from the 
 *                         minimization step
 *     oldMotCoef       I: Starting point motion coefficients for the step
 *
 * Function:
 *     Applies a Gauss-Newton minimization step on the segment prediction mse 
 *     as a function of motion coefficients. The input motion coefficients are
 *     read from the attribute list of the segment. Local coordinate system
 *     is assumed.
 *
 * Returns:
 *     estimated reduction in region prediction mse by this Gauss-Newton step.
 *
 */

double gnsGaussNewtonStep(object *region, int picSizeX, int picSizeY,
                          Byte *currY, Byte *prevY, double newMotCoef[],
                          double oldMotCoef[])  
{ 
  double 
    LHS[6][6],      /* Left Hand Side matrix of least squares equation    */
    rhs[6],         /* Right Hand Side vector of least squares equation   */
    estRed = 0.0;   /* Estimated reduction in mse by this GaussNewtonStep */
  int
    i;

  /* Build the Least Squares equation */
  gnsBuildLsEqn(LHS, rhs, region, prevY, currY, picSizeX, picSizeY,oldMotCoef);

  /* Solve the Least Squares equation, i.e. LHS*x = rhs, for x. Store
   * the solution to newMotCoef                                      
   */
  gnsSolveByChol(LHS, rhs, newMotCoef);

  /* Now newMotCoef contains the correction to input motion coefficient
   * vector. Compute estimated mse reduction & new motion coefficient vector.
   */ 
  for(i=0; i<6; i++){
    estRed        += rhs[i]*newMotCoef[i];
    newMotCoef[i] += oldMotCoef[i];
  }

  return(estRed/region->no_of_pels);
}


/*
 * gnsBuildLsEqnCRemXxX:
 *
 * Parameters:
 *     LHS              O: Buffer for the upper triangular part of the Left
 *                         Hand Side Matrix of least squares equation
 *     rhs              O: Right Hand Side vector of least squares equation
 *     xValues          I: xValues
 *     yValues          I: yValues
 *     xstepx           I: Step values for interpolating across the block
 *     xstepy           I: -"-
 *     ystepx           I: -"-
 *     ystepy           I: -"-
 *     i,j              I: Coordinates of the block in current frame
 *     prevY            I: Y component of previous reconstructed frame
 *     currY            I: Y component of current frame
 *     iPrevUpLeftD     I: Coordinates of the block in previous frame
 *     jPrevUpLeftD     I: 
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     blkSideLen       I: Length of the side of the block
 *
 * Function:
 *     This function is a part of buildLsEqn (inner loop of the original
 *     gnsBuildLsEqn[Glob] functions). It is used to build matrix LHS and
 *     vector rhs to satisfy: LHS(x-x0) = rhs, where x is the unknown
 *     motion parameter vector and x0 the initial motion parameter vector.
 *
 * Returns:
 *     Nothing
 *
 */

void buildLsEqnCRemXxX(double LHS[], double rhs[],
                       double xValues[], double yValues[],
                       double f0,
                       int xstepx, int xstepy, 
                       int ystepx, int ystepy,
                       Byte currY[], int i, int j, Byte prevY[],
                       double iPrevUpLeftD, double jPrevUpLeftD,
                       double xMvDelta, double yMvDelta,
                       double xstepxDelta, double xstepyDelta,
                       double ystepxDelta, double ystepyDelta, 
                       int picSizeX, int picSizeY, int blkSideLen)
{
  Byte 
    *curptr = currY + i*picSizeX + j;

  int
    m, n,
    iPrevUpLeft, jPrevUpLeft,
    difA[64], dX[64], dY[64], *dxp, *dyp, *diffp;

  double
    f1, f2, derX, derY, diff, *xPtr,
    contribPel[6], xMv, yMv;

  /* Location where the upper left corner motion vector points
   * in 16-bit fraction fixed point representation
   */
  iPrevUpLeft = (int)(65536.0*iPrevUpLeftD)+128;
  jPrevUpLeft = (int)(65536.0*jPrevUpLeftD)+128;

  /*
   * Find interpolated pixel values (difA) and both horizontal and 
   * vertical derivatives (dX, dY) of the reference frame. difA will
   * have 20 fraction bits and dX, dY will have 19 fraction bits.
   *
   * Interpolations accept coordinates in 16-bit fraction fixed-point
   * representation
   */
  if (((iPrevUpLeft & 0xff00) | (jPrevUpLeft & 0xff00) |
       (xstepx - (1<<16)) | (ystepy - (1<<16)) | ystepx | xstepy) == 0) {
    splDerivativeNoAffineXxX(difA, dX, dY, prevY,
                             iPrevUpLeft>>16, jPrevUpLeft>>16,
                             picSizeX, picSizeY, blkSideLen);
  }
  else {
    splDerivativeXxX(xstepx, xstepy, ystepx, ystepy, difA, dX, dY, prevY,
                     iPrevUpLeft, jPrevUpLeft, picSizeX, picSizeY,
                     blkSideLen);
  }

  /*
   * Calculate difference block
   */
  diffp = difA;
  for(m=0; m<blkSideLen; m++, curptr += picSizeX){
    for(n=0; n<blkSideLen; n++, diffp++){
      *diffp = ((int)curptr[n] << 20) - (*diffp);
    }
  }

  diffp = difA;
  dxp   = dX;
  dyp   = dY;

  xstepxDelta -= 1.0;
  ystepyDelta -= 1.0;

  for(m=0; m<blkSideLen; m++){

    xMv = xMvDelta+m*xstepyDelta-j;
    yMv = yMvDelta+m*ystepyDelta-i;

    for(xPtr = xValues, f1=yValues[m], n=0; n<blkSideLen ; n++) {

      /*
       * LHS is formed by multiplying matrix E by its transpose from the
       * left: LHS = E'E. E is a 6*(blkSideLen*blkSideLen) matrix formed as:
       * 
       * [ (f0*dx)[0] (f1*dx)[0] (f2*dx)[0] (f0*dy)[0] (f1*dy)[0] (f2*dy)[0] ]
       * [ (f0*dx)[1] (f1*dx)[1] (f2*dx)[1] (f0*dy)[1] (f1*dy)[1] (f2*dy)[1] ]
       * [                                ...                                ]
       * [ (f0*dx)[p] (f1*dx)[p] (f2*dx)[p] (f0*dy)[p] (f1*dy)[p] (f2*dy)[p] ]
       *
       * LHS is a symmetric matrix and thus only upper diagonal part
       * needs to be calculated here.
       *
       * rhs is calculated as: rhs = E'y where y is a vector having the
       * prediction differences as elements
       *
       * dxp and dyp below are in 19-bit fraction and diffp in 20-bit
       * fraction fixed point representation. That is, LHS[i] needs to
       * be scaled by 1/(2^19*2^19) and rhs[i] by 1/(2^19*2^20) before
       * using. This rescaling is done by higher level functions.
       */

      f2    = (*xPtr++);

      derX  = (*dxp++);
      derY  = (*dyp++);

      diff  = (*diffp++);
      diff += 2*(xMv*derX + yMv*derY);

      /* Update motion vectors */
      xMv  += xstepxDelta;
      yMv  += ystepxDelta;

      contribPel[0] = derX*f0;
      contribPel[1] = derX*f1;
      contribPel[2] = derX*f2;
      
      contribPel[3] = derY*f0;
      contribPel[4] = derY*f1;
      contribPel[5] = derY*f2;


      /* The following loop is unrolled below
      
      for(cnt=0, p=0; p<6; p++){
        rhs[p] += contribPel[p] * diff;
        for(q=p; q<6; q++)
          LHS[cnt++] += contribPel[p]*contribPel[q];
      } 

      */

      rhs[0]  += contribPel[0] * diff;      
      rhs[1]  += contribPel[1] * diff;      
      rhs[2]  += contribPel[2] * diff;      
      rhs[3]  += contribPel[3] * diff;      
      rhs[4]  += contribPel[4] * diff;      
      rhs[5]  += contribPel[5] * diff;    

      LHS[0]  += contribPel[0] * contribPel[0];
      LHS[1]  += contribPel[0] * contribPel[1];
      LHS[2]  += contribPel[0] * contribPel[2];
      LHS[3]  += contribPel[0] * contribPel[3];
      LHS[4]  += contribPel[0] * contribPel[4];
      LHS[5]  += contribPel[0] * contribPel[5];

      LHS[6 ] += contribPel[1] * contribPel[1];
      LHS[7 ] += contribPel[1] * contribPel[2];
      LHS[8 ] += contribPel[1] * contribPel[3];
      LHS[9 ] += contribPel[1] * contribPel[4];
      LHS[10] += contribPel[1] * contribPel[5];

      LHS[11] += contribPel[2] * contribPel[2];
      LHS[12] += contribPel[2] * contribPel[3];
      LHS[13] += contribPel[2] * contribPel[4];
      LHS[14] += contribPel[2] * contribPel[5];

      LHS[15] += contribPel[3] * contribPel[3];
      LHS[16] += contribPel[3] * contribPel[4];
      LHS[17] += contribPel[3] * contribPel[5];

      LHS[18] += contribPel[4] * contribPel[4];
      LHS[19] += contribPel[4] * contribPel[5];

      LHS[20] += contribPel[5] * contribPel[5];

    }
  }
}

/*
 * gnsBuildLsEqnCRem:
 *
 * Parameters:
 *     LHS              O: Left Hand Side Matrix of least squares equation
 *     rhs              O: Right Hand Side vector of least squares equation
 *     region           I: Attribute list for segment
 *     prevY            I: Y component of previous reconstructed frame
 *     currY            I: Y component of current frame
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     motCoef[6]       I: Input motion coefficients
 *
 * Function:
 *     Reduces the Gauss-Newton minimization step on the segment prediction 
 *     mse to a linear equation: LHS*motCoef = rhs
 *     Linearization of the reference frame is performed in the vicinity of
 *     motion vector field given by the input motion coefficients.
 *
 *     NOTE: Global coordinate system is assumed.
 *
 * Returns:
 *     Nothing
 *
 */

void gnsBuildLsEqnCRem(double LHS[][6], double rhs[], object *region, 
                       Byte prevY[], Byte currY[], int picSizeX, int picSizeY,
                       double motCoef[6], double motCoefDelta[6])
{
  int 
    i, j,                    /* Looping variables                          */
    xMin, xMax, yMin, yMax,  /* Bounding box of the region                 */
    xstepx, xstepy, ystepx, ystepy; /* Motion vector differences           */
  
  double 
    invSizeX, invSizeY,      /* Values used for calculating (x,y)          */
    centX, centY,            /* coordinates of a pixel                     */

    jPrevUpLeft,             /* Motion compensated horizontal coordinate   */
                             /* for the upper left corner of an 8x8 block  */
    iPrevUpLeft,             /* Motion compensated vertical coordinate     */
                             /* for the upper left corner of an 8x8 block  */
    bufLHS[21],              /* Buffer that holds upper tri. part of LHS   */
    *orthBasisFunctions,     /* 2nd and 3rd basis functions in orth system */
    orth1stBasisFunction,    /* First basis function in orthonorm. model   */
    xMvDelta, yMvDelta,      /* Variables defining the delta MV field      */
    xstepxDelta, xstepyDelta,                             
    ystepxDelta, ystepyDelta;

  /* Scaling factors for LHS matrix and rhs vector (they are generated
   * by fixed point operations in buildLsEqnXxX()
   */
  static double
    scaleRhs = 1.0/(1048576.0*524288.0), /* 1/(2^20*2^19) */
    scaleLhs = 1.0/( 524288.0*524288.0), /* 1/(2^19*2^19) */
    
    xValuesGlob[16],
    yValuesGlob[16],

    /* Orthonormalized basis functions */
    orth1stBasisFunction16 = 0.0625,
    orth1stBasisFunction8  = 0.1250,

    orthBasisFunctions16[16] = {
       0.10168615210250,
       0.08812799848883,
       0.07456984487516,
       0.06101169126150,
       0.04745353764783,
       0.03389538403417,
       0.02033723042050,
       0.00677907680683,
      -0.00677907680683,
      -0.02033723042050,
      -0.03389538403417,
      -0.04745353764783,
      -0.06101169126150,
      -0.07456984487516,
      -0.08812799848883,
      -0.10168615210250
    },
    
    orthBasisFunctions8[8] = {
       0.19094065395649,
       0.13638618139750,
       0.08183170883850,
       0.02727723627950,
      -0.02727723627950,
      -0.08183170883850,
      -0.13638618139750,
      -0.19094065395649
    };

  /* Initializations */

  rhs[0] = 0.0;
  rhs[1] = 0.0;
  rhs[2] = 0.0;
  rhs[3] = 0.0;
  rhs[4] = 0.0;
  rhs[5] = 0.0;

  for(j=0; j<21; j+=3) {
    bufLHS[j]   = 0.0;
    bufLHS[j+1] = 0.0;
    bufLHS[j+2] = 0.0;
  }

  xMin = region->xmin; xMax = region->xmax;
  yMin = region->ymin; yMax = region->ymax;

  if(xMax - xMin == 7){
    orth1stBasisFunction = orth1stBasisFunction8;
    orthBasisFunctions   = orthBasisFunctions8;
  }
  else{
    orth1stBasisFunction = orth1stBasisFunction16;
    orthBasisFunctions   = orthBasisFunctions16;
  }

  invSizeX = 1.0/(double)(picSizeX - 1);
  invSizeY = 1.0/(double)(picSizeY - 1);
  centX = (double) (picSizeX - 1)/2.0;
  centY = (double) (picSizeY - 1)/2.0;

  /* Differences between neighboring motion vectors when generating
   * the motion vector field
   */
  ystepy = (int)(65536.0*(1.0 + (2.0*motCoef[4])*invSizeY));
  ystepx = (int)(65536.0*((2.0*motCoef[5])*invSizeX));
  xstepx = (int)(65536.0*(1.0 + (2.0*motCoef[2])*invSizeX));
  xstepy = (int)(65536.0*((2.0*motCoef[1])*invSizeY));

  /* Same for the delta vectors (in double accuracy) */
  ystepyDelta = 1.0 + 2.0*motCoefDelta[4]*invSizeY;
  ystepxDelta =       2.0*motCoefDelta[5]*invSizeX;
  xstepxDelta = 1.0 + 2.0*motCoefDelta[2]*invSizeX;
  xstepyDelta =       2.0*motCoefDelta[1]*invSizeY;

  for(j=xMin; j<=xMax; j++) 
    xValuesGlob[j-xMin] = (2.0*(j - centX))*invSizeX;

  for(i=yMin; i<=yMax; i++) 
    yValuesGlob[i-yMin] = (2.0*(i - centY))*invSizeY;

  /* Building of LHS matrix and rhs vector starts */
  for(i=yMin; i<=yMax; i+=8){
    for(j=xMin; j<=xMax; j+=8)
    {
      /* Location where the motion vector of the upper left
       * corner of this 8x8 block points
       */
      jPrevUpLeft = j + motCoef[0] + motCoef[1]*yValuesGlob[i-yMin]
                                   + motCoef[2]*xValuesGlob[j-xMin];

      iPrevUpLeft = i + motCoef[3] + motCoef[4]*yValuesGlob[i-yMin]
                                   + motCoef[5]*xValuesGlob[j-xMin];

      /* Same for the delta vectors */
      xMvDelta = j + motCoefDelta[0]
                   + motCoefDelta[1]*yValuesGlob[i-yMin]
                   + motCoefDelta[2]*xValuesGlob[j-xMin];

      yMvDelta = i + motCoefDelta[3]
                   + motCoefDelta[4]*yValuesGlob[i-yMin]
                   + motCoefDelta[5]*xValuesGlob[j-xMin];

      /* Add the contribution of this 8x8 block to the
       * total LHS and rhs
       */

      buildLsEqnCRemXxX(bufLHS, rhs,
                        orthBasisFunctions+j-xMin, orthBasisFunctions+i-yMin,
                        orth1stBasisFunction, xstepx, xstepy, ystepx, ystepy,
                        currY, i, j, prevY, iPrevUpLeft, jPrevUpLeft,
                        xMvDelta, yMvDelta, xstepxDelta, xstepyDelta,
                        ystepxDelta, ystepyDelta, picSizeX, picSizeY, 8);
    }
  }

  /* Form LHS by symmetrically copying upper triangular part buffer
   * and rescaling elements
   */

  LHS[0][0] =             bufLHS[0]  * scaleLhs;
  LHS[0][1] = LHS[1][0] = bufLHS[1]  * scaleLhs;
  LHS[0][2] = LHS[2][0] = bufLHS[2]  * scaleLhs;
  LHS[0][3] = LHS[3][0] = bufLHS[3]  * scaleLhs;
  LHS[0][4] = LHS[4][0] = bufLHS[4]  * scaleLhs;
  LHS[0][5] = LHS[5][0] = bufLHS[5]  * scaleLhs;

  LHS[1][1] =             bufLHS[6]  * scaleLhs;
  LHS[1][2] = LHS[2][1] = bufLHS[7]  * scaleLhs;
  LHS[1][3] = LHS[3][1] = bufLHS[8]  * scaleLhs;
  LHS[1][4] = LHS[4][1] = bufLHS[9]  * scaleLhs;
  LHS[1][5] = LHS[5][1] = bufLHS[10] * scaleLhs;

  LHS[2][2] =             bufLHS[11] * scaleLhs;
  LHS[2][3] = LHS[3][2] = bufLHS[12] * scaleLhs;
  LHS[2][4] = LHS[4][2] = bufLHS[13] * scaleLhs;
  LHS[2][5] = LHS[5][2] = bufLHS[14] * scaleLhs;

  LHS[3][3] =             bufLHS[15] * scaleLhs;
  LHS[3][4] = LHS[4][3] = bufLHS[16] * scaleLhs;
  LHS[3][5] = LHS[5][3] = bufLHS[17] * scaleLhs;

  LHS[4][4] =             bufLHS[18] * scaleLhs;
  LHS[4][5] = LHS[5][4] = bufLHS[19] * scaleLhs;

  LHS[5][5] =             bufLHS[20] * scaleLhs;

  rhs[0] *= scaleRhs;
  rhs[1] *= scaleRhs;
  rhs[2] *= scaleRhs;
  rhs[3] *= scaleRhs;
  rhs[4] *= scaleRhs;
  rhs[5] *= scaleRhs;

}

/*
 * gnsBuildTriangLsEqnCRem:
 *
 * Parameters:
 *     prevY            I: Y component of previous reconstructed frame
 *     currY            I: Y component of current frame
 *     currRegion       I: Attribute list for segment
 *     picSizeX         I: Frame width as pixels
 *     picSizeY         I: Frame height as pixels
 *     U[6][6]          O: Upper triangular characteristic matrix
 *     uRhs[6]          O: Right hand side of the characteristic
 *                         equation Ux = uRhs (where x is a coefficient
 *                         vector)
 *     *partialSum, LHS, rhs : Temporary/dummy variables in this version
 *     motCoef[6]       I: Motion coefficient vector
 *
 * Function:
 *     Builds an upper triangular least squares equation for segment's
 *     prediction mse as a function of a motion coefficient vector.
 *     Also provides additional data (partialSum) that can be used in
 *     the computation of estimated increase when the equation matrices
 *     are perturbed.
 *
 *     NOTE: Global coordinate system is assumed.
 *
 * Returns:
 *     Zero (reserved for singularity. Currently, singularities are
 *     overcome by regularizing the Least Square Equation matrix.)
 *
 */

int gnsBuildTriangLsEqnCRem(Byte *prevY, Byte *currY,
                            object *currRegion,
                            int picSizeX, int picSizeY,
                            double U[6][6], double uRhs[6],
                            double *partialSum,
                            double LHS[6][6], double rhs[6],
                            double motCoef[6],
                            double motCoefDelta[6])
{
  double 
    parSum = 0;

  int
    j;

  /* Build the Least Squares equation */
  gnsBuildLsEqnCRem(LHS, rhs, currRegion, prevY, currY,
                    picSizeX, picSizeY, motCoef, motCoefDelta);

  /* Cholesky decomposition */
  if(gnsCholeskyDec(LHS,U)){

    /* At this point our characteristic equation is U'Ux = rhs.
     * Let's create triagular characteristic equation Ux = uRhs.
     * Now, U'Ux = rhs becomes U'uRhs = rhs and uRhs is solved
     * by backsubstitution
     */
    gnsTransposeBacksubstitution(U, rhs, uRhs, 6);   

    for(j=0; j<6; j++)
      parSum += uRhs[j]*uRhs[j];

    *partialSum = parSum;
  }
  else{

    /* If decomposition fails, add regularization terms to
     * the diagonal and perform new decomposition
     */
    for(j=0; j<6; j++)
      LHS[j][j] += REG;

    gnsCholeskyDec(LHS,U);

    gnsTransposeBacksubstitution(U, rhs, uRhs, 6);   

    for(j=0; j<6; j++)
      parSum += uRhs[j]*uRhs[j];

    *partialSum = parSum;
  }

  return 0;
}
