/*
 * toolMCC.c
 *
 *    Copyright (c) 1995-2000 Nokia Research Center
 *
 * Contains:
 *    Auxiliary functions for prediction error coding and 
 *    reconstructing the prediction error from the coding information.
 *
 */

#include <stdlib.h>
#include "table.dat"
#include "mainFlags.h"
#include "encMux.h"

extern u_char
  overheadOrder8x8[AMOUNT_OF_8x8_OVH * (MAX_OF_8x8_METH_OVH + 1)],
  overheadNumbers8x8[AMOUNT_OF_8x8_OVH];

extern u_char
  overheadOrder4x4[AMOUNT_OF_4x4_OVH * (MAX_OF_4x4_METH_OVH + 1)],
  overheadNumbers4x4[AMOUNT_OF_4x4_OVH];

extern char
  motionTableX[AMOUNT_OF_INTRA_MOTION],
  motionTableY[AMOUNT_OF_INTRA_MOTION];

extern int16
  kltBasisFunc4x4Diag1[8 * SIZE_4x4],
  kltBasisFunc4x4Diag2[8 * SIZE_4x4],
  kltBasisFunc8x8Diag1[16 * SIZE_8x8],
  kltBasisFunc8x8Diag2[16 * SIZE_8x8];

extern u_char
  mirrorVert4x4[16], 
  mirrorHori4x4[16], 
  mirrorDiag4x4[16],
  rotate180_4x4[16], 
  rotateLeft4x4[16], 
  rotateRight4x4[16];

extern u_char
  mirrorVert8x8[64],
  mirrorHori8x8[64],
  mirrorDiag8x8[64],
  rotate180_8x8[64], 
  rotateLeft8x8[64], 
  rotateRight8x8[64];


/*
 * Sizes of basisfunctions are defined.
 *
 * The order of codebooks is
 *      0 = Diagonal 1
 *      1 = Diagonal 2
 */

int kltSize4x4[AMOUNT_OF_KLT_4x4] = 
{
     8,    8
};

int kltSize8x8[AMOUNT_OF_KLT_8x8] = 
{
    16,   16
};


/*
 * Define local function prototypes.
 */
static int32 calcUnnormVar(int *source, int size);

static int32 maxInt32(int32 *value, int32 size);

static int32 minInt32(int32 *value, int32 *minimum, int32 size);

static void sortIntBigToSmall(int *value, int size, int sort);

static void sortInt32SmallToBigIndex(int32 *value, int *index,
                                     int size, int sort);

static int findPredNum(int edge, float variance, int qp, int chroma,
                       int blkSize);

static void block_mid_hori(int *block, int *block_mid,
                           int blocksizex, int blocksizey);

static void block_mid_vert(int *block, int *block_mid,
                           int blocksizex, int blocksizey);

static int32 Gradient_angled(int *block, int *block_mid,
                             int sizex, int sizey, int angle);

static int32 calcDirGradient(int *source, int sizex, int sizey, int dir);

int calcDctBits(int *source, int cols, int rows, u_char *zig, int qp, 
                int chroma, int mode, int start);

int calcKltBits(int *coeffsVect, int cols, int rows, int qp, 
                       int chroma, int mode, int kltNum);
/*
 * maxFloat:
 *
 * Parameters:
 *      value           I: Float value pointer
 *      size            I: Amount of values
 *
 * Function:
 *      Find the biggest value of size amount of values.
 *
 * Returns:
 *      max             Maximum value
 */

static int32 maxInt32(int32 *value, int32 size)
{
  int32 
    i,              /* Loop increment counter */
    max;            /* Maximum value */

  max = value[0];
  for (i = 1; i < size; i++)
    if (value[i] > max)
      max = value[i];

  return max;
}


/*
 * minFloat:
 *
 * Parameters:
 *      value           I: Float value pointer
 *      size            I: Amount of values
 *
 * Function:
 *      Find the smallest value of size amount of values.
 *
 * Returns:
 *      min             Minimum value
 */

static int32 minInt32(int32 *value, int32 *minimum, int32 size)
{
  int32 
    i,              /* Loop increment counter */
    position = 0;   /* Position of the minimum value */

  *minimum = value[0];
  for (i = 1; i < size; i++)
    if (value[i] < *minimum) {
      *minimum = value[i];
      position = i;
    }

  return position;
}




/*
 * sortIntBigToSmall:
 *
 * Parameters:
 *      value         I/O: Integer value pointer
 *      size            I: Amount of values
 *      sort            I: Amount of sorted values
 *
 * Function:
 *      Sort the sort amount the biggest values of size amount
 *      of integer values.
 *
 * Returns:
 *      void
 */

static void sortIntBigToSmall(int *value, int size, int sort)
{
  int 
    i, j,             /* Loop increment counters */
    tmp;              /* Temporal value needed for swap */

  for (i = 0; i < sort; i++)
    for (j = i + 1; j < size; j++)
      if (value[j] > value[i]) {
        tmp = value[i];
        value[i] = value[j];
        value[j] = tmp;
      }
}


static void sortInt32SmallToBigIndex(int32 *value, int *index,
                                     int size, int sort)
{
  int32 
    i, j,  /* Loop increment counters */
    tmp;   /* Temporal value needed for swap */

  for (i = 0; i < sort; i++)
    for (j = i + 1; j < size; j++)
      if (value[j] < value[i]) {
        tmp = value[i];
        value[i] = value[j];
        value[j] = tmp;
        tmp = index[i];
        index[i] = index[j];
        index[j] = tmp; 
      }
}


/*
 * median:
 *
 * Parameters:
 *      value           I: Integer value pointer
 *      size            I: Amount of values
 *
 * Function:
 *      Finds the median value from given value buffer.
 *      If the size of buffer is even then the average of two
 *      median values is calculated
 *
 * Returns:
 *      median          Median value of buffer
 */

int median(int *value, int size)
{
  int median,           /* Median value                                 */
      medianIndex;      /* Index which places to median value           */
 
  medianIndex = size >> 1;
  sortIntBigToSmall(value, size, medianIndex + 1);
  if (size & 1)
    median = value[medianIndex];

  else {
	  if ((value[medianIndex-1] + value[medianIndex] + 1) >=0)
	    median = (value[medianIndex-1] + value[medianIndex] + 1) >> 1;

    else
	    median = -((-value[medianIndex-1] - value[medianIndex] - 1) >> 1);
  }

  return median;
}


/*
 * ipdDCquantAndInvQuant:
 *
 * Parameters:
 *      cumDiff           I: The cummulative sum of intra prediction error
 *      pels              I: Amount of pels on which cumDiff has been taken
 *      maxVal            I: Maximum positive pixel error (i.e. 255)
 *      levels            I: Total amount of quantisation levels
 *      index             O: The resulting quantisation level
 *
 * Function:
 *      Performs quantisation and then inverse quantisation for
 *      the DC difference used after INTRA predicted 8x8 or 4x4
 *      blocks.
 *
 * Returns:
 *      value     The inverse quantisation value as a signed number.
 */
int ipdDCquantAndInvQuant(int cumDiff, int pels, int maxVal, int levels, int *index)
{

  ldiv_t
    levelData;

  long
    lastLevel = levels - 1,
     maxSum   = pels * maxVal;

  int
    value, flag = 0;

  /*
   * Store the sign in variable 'flag'
   * and take |cumDiff| as the input sum
   * which is to be quantised.
   */
  if(cumDiff < 0)
  {
    cumDiff  = -cumDiff;
    flag = 1;
  }
  
  /*                  / cumDiff * (levels - 1) \
   *   level = round (  ______________________  )
   *                  \     pels * maxVal      /
   */
  levelData = ldiv(lastLevel * cumDiff, maxSum);
  if (levelData.rem * 2 >= maxSum)
    levelData.quot += 1;

  /*                         / level * maxVal \
   *  invQuantValue = floor (  ______________  )
   *                         \   levels - 1   /
   */
  value = maxVal * levelData.quot / lastLevel;

  /*
   * Restore the sign of cumDiff and return the
   * level for which |level| is one of 0, 1, 2, ..., (levels-1).
   * Return also the inverse quantization value for which
   * |value| is one of 0, 1, ..., maxVal.
   */
  if(flag == 1)
  {
    *index = -(int)levelData.quot;
    value  = -value;
  }
  else
    *index = (int)levelData.quot;
  
  return value;
}



/*
 * ipdDCinvQuant:
 *
 * Parameters:
 *      maxVal            I: Maximum positive pixel error (i.e. 255)
 *      levels            I: Total amount of quantisation levels
 *      level             I: The index of the given quantisation level
 *                           which is to be de-quantised.
 *
 * Function:
 *      Performs inverse quantisation for the DC difference used
 *      after INTRA predicted 8x8 or 4x4 blocks.
 *
 * Returns:
 *      value     The inverse quantisation value as a signed number.
 */
int ipdDCinvQuant(int maxVal, int levels, int level)
{
  long
    lastLevel = levels - 1;

  int
    value, flag = 0;

  /*
   * Store the sign in variable 'flag'
   * and take |level| as the input
   * to be de-quantised.
   */
  if(level < 0)
  {
    level = -level;
    flag  = 1;
  }

  /*                         / level * maxVal \
   *  invQuantValue = floor (  ______________  )
   *                         \   levels - 1   /
   */
  value = maxVal * level / lastLevel;

  /*
   * Restore the sign and rturn inverse quantization value for which
   * |value| is one of 0, 1, ..., maxVal.
   */
  if(flag == 1)
    value = -value;

  return value;
}


/*
 * predictDCValue:
 *
 * Parameters:
 *      left            I: Left DCvalue
 *      upper           O: Upper DCvalue
 *      upperLeft       I: Upperleft DCvalue
 *      chroma          I: Prediction block type (luma or chroma)
 *
 * Function:
 *      Finds the prediction DC value for given block. It uses
 *      three DC values which have to be bigger or equal as 0
 *      and generates the prediction DC value based of those.
 *      If function return 300 it means that prediction isn't used.
 *
 * Returns:
 *      predValue       Prediction DC value.
 */

int predictDCValue(int left, int upper, int upperLeft, int chroma)
{
  int quess,                    /* Quess value for block        */
      predValue = 128;          /* Prediction DC value          */

  if(left>=0 && upper>=0 && upperLeft>=0) {
    quess = left + upper - upperLeft;
    if(quess > 255)    quess = 255;
    else if(quess < 0) quess = 0;

    if(upperLeft==left && upperLeft==upper) {
      predValue = upperLeft;
    }
    else if(upperLeft == upper) {
      predValue = left;
    }
    else if(upperLeft == left) {
      predValue = upper;
    }
    else if(left == upper) {
      if(chroma == CHROMA) {
        predValue = left;
      }
      else {
        if(abs(upperLeft - left) <= 4) {
          predValue = quess;
        }
        else {
          predValue = left;
        }
      }
    }
    else if(upperLeft<left && left<upper) {
      predValue = upper;
    }
    else if(upperLeft<upper && upper<left) {
      predValue = left;
    }
    else if(upperLeft>left && left>upper) {
      predValue = upper;
    }
    else if(upperLeft>upper && upper>left) {
      predValue = left;
    }
    else if(left<upperLeft && upperLeft<upper) {
      predValue = quess;
    }
    else if(upper<upperLeft && upperLeft<left) {
      predValue = quess;
    }
    else {
      predValue = quess;                /* NOT USED */
    }
  }
  else if(left>=0 && upper>=0) {
    predValue = (left+upper+1) >> 1;
  }
  else if(left>=0 && upperLeft>=0) {
    predValue = left;
  }
  else if(upper>=0 && upperLeft>=0) {
    predValue = upper;
  }
  else if(left >= 0) {
    predValue = left;
  }
  else if(upper >= 0) {
    predValue = upper;
  }
  else if(upperLeft >= 0) {
    predValue = upperLeft;
  }
  else {
    predValue = 128;
  }

  return predValue;
}


/*
 * bitsForDCDist:
 *
 * Parameters:
 *      distance        I: Distance between current and predicted DC value
 *      chroma          I: Prediction block type (luma or chroma)
 *
 * Function:
 *      Returns amount of bits needed to send distance information
 *      between current and predicted DC value.
 *
 * Returns:
 *      bits            Bits needed for sending distance.
 */

int bitsForDCDist(int distance, int chroma)
{
  int bits;             /* Bits needed for sending distance */

  distance = abs(distance);
  if(chroma == CHROMA) {
    if(distance < 1) {
      bits = (3 + 0 + 0);               /* (huffman + choice + sign)    */
    }
    else if(distance < 2) {
      bits = (2 + 0 + 1);
    }
    else if(distance < 4) {
      bits = (2 + 1 + 1);
    }
    else if(distance < 8) {
      bits = (2 + 2 + 1);
    }
    else if(distance < 16) {
      bits = (4 + 3 + 1);
    }
    else if(distance < 32) {
      bits = (5 + 4 + 1);
    }
    else if(distance < 64) {
      bits = (6 + 5 + 1);
    }
    else {
      bits = (6 + 8 + 0);               /* Choice is just DC_value */
    }
  }
  else {
    if(distance < 1) {
      bits = (3 + 0 + 0);               /* (huffman + choice + sign)    */
    }
    else if(distance < 2) {
      bits = (3 + 0 + 1);
    }
    else if(distance < 4) {
      bits = (3 + 1 + 1);
    }
    else if(distance < 8) {
      bits = (3 + 2 + 1);
    }
    else if(distance < 16) {
      bits = (2 + 3 + 1);
    }
    else if(distance < 32) {
      bits = (3 + 4 + 1);
    }
    else if(distance < 64) {
      bits = (4 + 5 + 1);
    }
    else {
      bits = (4 + 8 + 0);               /* Choice is just DC_value */
    }
  }
  return bits;
}


/*
 * getExtSouJunct:
 *
 * Parameters:
 *      source          I: Source (Prediction) block pointer
 *      destination     O: Destination (extrapSource) pointer
 *      cols            I: Columns in source block
 *      rows            I: Rows in source block
 *
 * Function:
 *      Finds the edge pels from given cols*rows source block and
 *      returns them as a destination.
 *
 * Returns:
 *      pels            Amount of pels in destination.
 */

int getExtSouJunct(int *source, int *destination, int cols, int rows)
{
  int i, j,             /* Loop increment counters                      */
      pels = 0;         /* Amount of pels in destination                */

  for(j=0; j<rows; j++) {
    for(i=0; i<cols; i++) {

      /*
       * Check if edge pel.
       */
      if((i%cols) == 0 || ((i+1)%cols) == 0 ||
         (j%rows) == 0 || ((j+1)%rows) == 0) {
        destination[pels++] = source[cols*j + i];
      }
    }
  }

  return pels;
}


/*
 * getExtSou8x8:
 *
 * Parameters:
 *      source          I: Source (Prediction) 8x8 block pointer
 *      destination     O: Destination (extrapSource) pointer
 *      location        I: 4x4 blocks quadrant in 8x8 block or
 *                         4x8 blocks half in 8x8 block
 *
 * Function:
 *      Finds the extrapolation source pels from given 8x8 block and
 *      returns them as a destination.
 *
 * Returns:
 *      pels            Amount of pels in destination.
 */

int getExtSou8x8(int *source, int *destination, int location)
{
  int pels,             /* Amount of pels in destination                */
      cols,             /* Amount of columns                            */
      rows,             /* Amount of rows                               */
      startCol,         /* Starting from column                         */
      startRow;         /* Starting from row                            */

  if(location == QUAD_1) {
    cols = 5; rows = 1; startCol = 0; startRow = 4;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, COLS_8x8);
    cols = 1; rows = 4; startCol = 4; startRow = 0;
    pels+= getIntBlock(source, destination+pels, cols, rows,
                       startCol, startRow, COLS_8x8);
  }
  else if(location == QUAD_2) {
    cols = 5; rows = 1; startCol = 3; startRow = 4;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, COLS_8x8);
    cols = 1; rows = 4; startCol = 3; startRow = 0;
    pels+= getIntBlock(source, destination+pels, cols, rows,
                       startCol, startRow, COLS_8x8);
  }
  else if(location == QUAD_3) {
    cols = 5; rows = 1; startCol = 0; startRow = 3;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, COLS_8x8);
    cols = 1; rows = 4; startCol = 4; startRow = 4;
    pels+= getIntBlock(source, destination+pels, cols, rows,
                       startCol, startRow, COLS_8x8);
  }
  else if(location == QUAD_4) {
    cols = 5; rows = 1; startCol = 3; startRow = 3;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, COLS_8x8);
    cols = 1; rows = 4; startCol = 3; startRow = 4;
    pels+= getIntBlock(source, destination+pels, cols, rows,
                       startCol, startRow, COLS_8x8);
  }
  else if(location == HALF_1) {
    cols = 8; rows = 1; startCol = 0; startRow = 4;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, ROWS_8x8);
  }
  else if(location == HALF_2) {
    cols = 8; rows = 1; startCol = 0; startRow = 3;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, ROWS_8x8);
  }
  else if(location == HALF_3) {
    cols = 1; rows = 8; startCol = 4; startRow = 0;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, ROWS_8x8);
  }
  else if(location == HALF_4) {
    cols = 1; rows = 8; startCol = 3; startRow = 0;
    pels = getIntBlock(source, destination, cols, rows,
                       startCol, startRow, ROWS_8x8);
  }
  else {
    printf("Error. Wrong location %d in getExtSou8x8()\n", location);
    exit(1);
  }

  return pels;
}


/*
 * changeClassNew:
 *
 * Parameters:
 *      classIn         I: Original 8x8 prediction class
 *      predDir         I: Prediction block direction
 *      qp              I: Quantization parameter
 *      chroma          I: Prediction block type (luma or chroma)
 *      mode            I: Prediction block mode (intra or inter or first)
 *      rotateFlag      O: Rotation information
 *
 * Function:
 *      Changes the original prediction class to class which is
 *      needed in 8x8 method selection.
 *
 * Returns:
 *      class8x8        Class needed for 8x8 method selection.
 */

int changeClassNew(int classIn, int predDirIn, int qp,
                   int chroma, int mode, int *rotateFlag)
{
  int 
    class8x8 = 0,       /* Class needed for 8x8 method selection        */
    qpRange, tmpClass;

  /*
   * Find the right part of the coding method table according
   * to component type (chrominance/luminance) and QP range.
   */
  if (chroma) {
    class8x8 += START_OF_8x8_CHROMA_OVH;
    if (qp <= 9)       
      qpRange = 0;
    else
      qpRange = 1;
  }
  else {
    if (qp <= 9)       
      qpRange = 0; 
    else if (qp <= 15) 
      qpRange = 1;
    else if (qp <= 22) 
      qpRange = 2;
    else              
      qpRange = 3; 
  }

  class8x8 += qpRange * AMOUNT_OF_8x8_CLASS_OVH;  

  /*
   * Find the right row according to block type (INTER/INTRA)
   * and prediction class in case of INTRA.
   */
  if (mode == INTER_BLOCK) 
    *rotateFlag = UNIT_OPER;

  else {

    /*
     * Prediction class is infered from the prediction method for 
     * DC prediction and pixel extrapolations.
     */
    tmpClass = (classIn - (NON_CLASSIFIED + 1));
    if (tmpClass == 1 || tmpClass == 0) 
    {
      class8x8 += 1;
      *rotateFlag = UNIT_OPER;
    }
    else if (tmpClass == 2) 
    {
      class8x8 += 2;
      *rotateFlag = MIRROR_HORI;
    }
    else if (tmpClass == 3) 
    {
      class8x8 += 3;
      *rotateFlag = MIRROR_HORI;
    }
    else if (tmpClass == 4) 
    {
      class8x8 += 4;
      *rotateFlag = ROTATE_LEFT;
    }
    else if (tmpClass == 5) 
    {
      class8x8 += 5;
      *rotateFlag = MIRROR_DIAG;
    }
    else if (tmpClass == 6) 
    {
      class8x8 += 6;
      *rotateFlag = MIRROR_DIAG;
    }
    else if (tmpClass == 7) 
    {
      class8x8 += 7;
      *rotateFlag = UNIT_OPER;
    }
    else if (tmpClass == 8) 
    {
      class8x8 += 6;
      *rotateFlag = UNIT_OPER;
    }
    else if (tmpClass == 9) 
    {
      class8x8 += 5;
      *rotateFlag = UNIT_OPER;
    }

    /*
     * Prediction was done using block matching, actual classifier
     * result for the prediction block is used.
     */
    else
      *rotateFlag = UNIT_OPER;
  }

  return class8x8;
}


/*
 * copyBlock8x8:
 *
 * Parameters:
 *      source          I: Source block pointer
 *      destination     O: Destination block pointer
 *      rotateFlag      I: Rotation flag
 *
 * Function:
 *      Rotates the source reconstruction block to the destination.
 *
 * Returns:
 *      void
 */

void copyBlock8x8(int *source, int *destination, int rotateFlag)
{
  int 
    i;

  switch (rotateFlag) {

    case UNIT_OPER:
      memcpy(destination, source, sizeof(int) << 6);
      break;

    case ROTATE_LEFT:
      for (i = 0; i < SIZE_8x8; i++)
        destination[i] = source[rotateLeft8x8[i]];
      break;

    case ROTATE_RIGHT:
      for (i = 0; i < SIZE_8x8; i++)
        destination[i] = source[rotateRight8x8[i]];
      break;

    case MIRROR_VERT:
      for (i = 0; i < SIZE_8x8; i++)
        destination[i] = source[mirrorVert8x8[i]];
      break;

    case MIRROR_HORI:
      for (i = 0; i < SIZE_8x8; i++)
        destination[i] = source[mirrorHori8x8[i]];
      break;

    default:
      for (i = 0; i < SIZE_8x8; i++)
        destination[i] = source[mirrorDiag8x8[i]];
      break;
  }
}


/*
 * copyBlock4x4:
 *
 * Parameters:
 *      source          I: Source block pointer
 *      destination     O: Destination block pointer
 *      rotateFlag      I: Rotation flag
 *
 * Function:
 *      Rotates 4x4 source reconstruction block to the destination.
 *
 * Returns:
 *      void
 */

void copyBlock4x4(int *source, int *destination, int rotateFlag)
{
  int 
    i;

  switch (rotateFlag) {

    case UNIT_OPER:
      memcpy(destination, source, sizeof(int) << 4);
      break;

    case ROTATE_LEFT:
      for (i = 0; i < SIZE_4x4; i++)
        destination[i] = source[rotateLeft4x4[i]];
      break;

    case ROTATE_RIGHT:
      for (i = 0; i < SIZE_4x4; i++)
        destination[i] = source[rotateRight4x4[i]];
      break;

    case MIRROR_VERT:
      for (i = 0; i < SIZE_4x4; i++)
        destination[i] = source[mirrorVert4x4[i]];
      break;

    case MIRROR_HORI:
      for (i = 0; i < SIZE_4x4; i++)
        destination[i] = source[mirrorHori4x4[i]];
      break;

    default:
      for (i = 0; i < SIZE_4x4; i++)
        destination[i] = source[mirrorDiag4x4[i]];
      break;
  }
}


int methSelectNum4x4New(int predno, int method)
{
  int i = 1,            /* Loop increment counter                       */
      tablex,           /* X size of table                              */
      tableMeth,        /* Method number in table                       */
      max,              /* Amount of different methods                  */
      index = 0;        /* Number of method in method overhead table    */

  /*
   * Find the amount of different methods for that predno.
   */

  tablex = MAX_OF_4x4_METH_OVH + 1;
  max = overheadOrder4x4[predno * tablex];
 
  /*
   * Scan method list one by one and check if the method in the
   * list is same as asked method.
   */
  while(index == 0 && i <= max) {
    tableMeth = overheadOrder4x4[predno * tablex + i];
    if(tableMeth == method) {
      index = i;
    }
    i++;
  }

  return index;
}


int methSelectNum8x8New(int class, int method)
{
  int i = 1,            /* Loop increment counter                       */
      tablex,           /* X size of table                              */
      tableMeth,        /* Method number in table                       */
      max,              /* Amount of different methods                  */
      index = 0;        /* Number of method in method overhead table    */

  /*
   * Find the amount of different methods for that class.
   */

  tablex = MAX_OF_8x8_METH_OVH + 1;
  max = overheadOrder8x8[class * tablex];
 
  /*
   * Scan method list one by one and check if the method in the
   * list is same as asked method.
   */
  while (index == 0 && i <= max) {
    tableMeth = overheadOrder8x8[class * tablex + i];
    if(tableMeth == method) {
      index = i;
    }
    i++;
  }

  return index;
}

/*
 * MCCQuantCoef:
 *
 * Parameters:
 *      dctCoef         I: Coefficient value buffer
 *      quantDctCoef    O: Quantized coefficient value buffer
 *      size            I: Size of buffer
 *      qp              I: Quantization parameter
 *      chroma          I: Mode of block (chroma/luma) 
 *      mode            I: Mode of block (inter/intra/first) 
 *      transform       I: Type of transform (dct/klt) 
 *
 * Function:
 *      Quantizes the given coefficient values using given qp value.
 *
 * Returns:
 *      nonZeros        Amount of nonzero coefficients
 */

int MCCQuantCoef(int *dctCoef, int *quantDctCoef, int size, int qp,
                 int chroma, int mode, int transform)
{
  int
    i,                  /* Loop increment counter                       */
    maxLevel,           /* Maximum coefficient level value              */
    tmpCoef,            /* Temporal coefficient value                   */
    quantCoefValue,     /* Quantized coefficient value                  */
    nonZeros = 0;       /* Amount of nonzero coefficients               */

  int
    coefValue,          /* Coefficient value                            */
    tmpValue;           /* Temporal coefficient value                   */

  if(chroma == CHROMA) {
    if(qp ==  0)      maxLevel = LEVEL_09;
    else if(qp <=  1) maxLevel = LEVEL_08;
    else if(qp <=  2) maxLevel = LEVEL_07;
    else if(qp <=  5) maxLevel = LEVEL_06;
    else if(qp <= 11) maxLevel = LEVEL_05;
    else if(qp <= 23) maxLevel = LEVEL_04;
    else              maxLevel = LEVEL_03;
  }
  else {        /* LUMA */
    if(qp ==  0)      maxLevel = LEVEL_10;
    else if(qp <=  1) maxLevel = LEVEL_09;
    else if(qp <=  2) maxLevel = LEVEL_08;
    else if(qp <=  5) maxLevel = LEVEL_07;
    else if(qp <= 11) maxLevel = LEVEL_06;
    else if(qp <= 23) maxLevel = LEVEL_05;
    else              maxLevel = LEVEL_04;
  }

  for(i=0; i<size; i++) {
    coefValue = dctCoef[i];
    if(coefValue < 0) {
      tmpValue = -coefValue;
    }
    else {
      tmpValue =  coefValue;
    }
    tmpCoef = tmpValue;

    if(qp == 0) {
      quantCoefValue = tmpCoef;
    }
    else {
      if(mode == INTER_BLOCK) {
        quantCoefValue = (tmpCoef - qp/2) / (2*qp);
      }
      else if(mode == INTRA_BLOCK || mode == FIRST_BLOCK) {
        quantCoefValue = tmpCoef/(2*qp);
      }
      else {
        printf("Wrong mode %d or chroma %d in newQuantCoef()\n", mode, chroma);
        exit(1);
      }
    }

    /* Clipping of quantCoefValue*/ 

    if(quantCoefValue >= maxLevel) {
      quantCoefValue = maxLevel - 1;
    }
    if(quantCoefValue != 0) {
      nonZeros++;
    }

    /* Correct sign for quantCoefValue */

    if(coefValue < 0) {
      quantCoefValue = -quantCoefValue;
    }

    quantDctCoef[i] = quantCoefValue;
  }

  return nonZeros;
}

/*
 * MCCInvQuantCoef:
 *
 * Parameters:
 *      quantDctCoef    I: Quantized coefficient value buffer
 *      dctCoef         O: Coefficient value buffer
 *      size            I: Size of buffer
 *      qp              I: Quantization parameter
 *      chroma          I: Mode of block (chroma/luma) 
 *      mode            I: Mode of block (inter/intra) 
 *
 * Function:
 *      Makes the inverse quantization for the given coefficient values
 *      using given qp value.
 *
 * Returns:
 *      void
 */

void MCCInvQuantCoef(int *quantDctCoef, int *dctCoef, int size, int qp,
                     int chroma, int mode)
{
  int
    i,                  /* Loop increment counter                       */
    quantCoefValue,     /* Quantized coefficient value                  */
    tmpValue,           /* Temporal quantized coefficient value         */
    coefValue,          /* Coefficient value                            */
    add,
    dist;

  add = qp >> 1;      
  dist = qp << 1;

  for (i = 0; i < size; i++) {

    quantCoefValue = quantDctCoef[i];

    if (quantCoefValue) {

      tmpValue = quantCoefValue < 0 ? -quantCoefValue : quantCoefValue;

      switch (mode) {

        case INTER_BLOCK:
          coefValue = tmpValue * dist + add + (qp - 1);
          break;

        default:
          coefValue = tmpValue * dist + (qp - 1);
          break;
      }
      dctCoef[i] = quantCoefValue < 0 ? -coefValue : coefValue;
    }
    else
      dctCoef[i] = 0;
  }
}

/*
 * Function    : KltTransform8x8()
 * Purpose     : Calculation of forward KLT
 * Arguments   : 
 * Returns     : 
 * Description : Brute force implementation by taking inner product of
 *               coded pattern with each basis function.
 * Updates     : 
 */

void KltTransform8x8(int *diff, int *coeffs, int32 kltNum, int32 mode)
{
  int32
    i, j;           /* Loop counters */

  int16
    *basisFuncPtr;  /* Pointer to basis functions */

  /*
   * Find the right KLT basis function set.
   */
  basisFuncPtr = (kltNum == 1 ? kltBasisFunc8x8Diag1 : kltBasisFunc8x8Diag2);

  /*
   * Calculate the fixed-point KLT coefficients.
   */
  for (j = 0; j < kltSize8x8[kltNum - 1]; j++) 
  {
    coeffs[j] = 0;
    for (i = 0; i < SIZE_8x8; i++) 
      coeffs[j] += diff[i] * basisFuncPtr[SIZE_8x8 * j + i];

    /*
     * Truncate coefficient to integer precision.
     */
    coeffs[j] >>= 15;
  }
}


/*
 * Function    : KltTransform4x4()
 * Purpose     : Calculation of forward KLT
 * Arguments   : 
 * Returns     : 
 * Description : Brute force implementation by taking inner product of
 *               coded pattern with each basis function.
 * Updates     : 
 */

void KltTransform4x4(int *diff, int *coeffs, int32 kltNum, int32 mode)
{
  int32
    i, j;           /* Loop counters */

  int16
    *basisFuncPtr;  /* Pointer to basis functions */

  /*
   * Find the right KLT basis function set.
   */
  basisFuncPtr = (kltNum == 1 ? kltBasisFunc4x4Diag1 : kltBasisFunc4x4Diag2);

  /*
   * Calculate the fixed-point KLT coefficients.
   */
  for (j = 0; j < kltSize4x4[kltNum - 1]; j++) 
  {
    coeffs[j] = 0;
    for (i = 0; i < SIZE_4x4; i++) 
      coeffs[j] += diff[i] * basisFuncPtr[SIZE_4x4 * j + i];

    /*
     * Truncate coefficient to integer precision.
     */
    coeffs[j] >>= 15;
  }
}

void KltInverseTransform8x8(int *coeffs, int *reco, int32 kltNum, int32 mode)
{
  int32
    i, j;           /* Loop counters */

  int16
    *basisFuncPtr;  /* Pointer to basis functions */

  /*
   * Initialize reconstruction block and find the right 
   * KLT basis function set.
   */
  for (i = 0; i < SIZE_8x8; i++)
    reco[i] = 0;
  basisFuncPtr = (kltNum == 1 ? kltBasisFunc8x8Diag1 : kltBasisFunc8x8Diag2);

  /*
   * Calculate the inverse KLT using fixed-point basis functions.
   */
  for (j = 0; j < kltSize8x8[kltNum - 1]; j++)
  {
    if (coeffs[j] != 0)
    {
      for (i = 0; i < SIZE_8x8; i++) 
        reco[i] += coeffs[j] * basisFuncPtr[SIZE_8x8 * j + i];
    }
  }

  /*
   * Truncate reconstrution values to integer precision.
   */
  for (i = 0; i < SIZE_8x8; i++)
    reco[i] >>= 15;
}


void KltInverseTransform4x4(int *coeffs, int *reco, int32 kltNum, int32 mode)
{
  int32
    i, j;           /* Loop counters */

  int16
    *basisFuncPtr;  /* Pointer to basis functions */

  /*
   * Initialize reconstruction block and find the right 
   * KLT basis function set.
   */
  for (i = 0; i < SIZE_4x4; i++)
    reco[i] = 0;
  basisFuncPtr = (kltNum == 1 ? kltBasisFunc4x4Diag1 : kltBasisFunc4x4Diag2);

  /*
   * Calculate the inverse KLT using fixed-point basis functions.
   */
  for (j = 0; j < kltSize4x4[kltNum - 1]; j++)
  {
    if (coeffs[j] != 0)
    {
      for (i = 0; i < SIZE_4x4; i++) 
        reco[i] += coeffs[j] * basisFuncPtr[SIZE_4x4 * j + i];
    }
  }

  /*
   * Truncate reconstrution values to integer precision.
   */
  for (i = 0; i < SIZE_4x4; i++)
    reco[i] >>= 15;
}

/*
 * findDctMeth:
 *
 * Parameters:
 *      methInfo        I: Source buffer which consists 8x8 method information
 *      method8x8       I: Chosen 8x8 method
 *
 * Function:
 *      Finds from 8x8 information if 4x4 DCT or KLT is used and returns
 *      the corresponding 8x8 method for bitstream saving.
 *
 *      Quadrants order is      1 2 
 *                              3 4.
 *
 * Returns:
 *      dctmethod       Dct based method value.
 *                      Base value means that 4x4 DCT is not used.
 */

int findDctMeth(int *methInfo, int method8x8, int *save4)
{
  int method,           /* Real method number which do not consist base  */
      dctmethod,        /* Dct based method value with base              */
      quadNum,          /* Number of quadrant                            */
      meth4x4,          /* Actual 4x4 method for given quadrant          */ 
      i, j=0;           /* Loop increment counter                        */

  method = method8x8 - METH_4x4_1 + 1;
  dctmethod = METH_4x4_DCT_1 - 1;
  quadNum = 1;
  for(i = 0; i < 4; i++) {
    if(method&quadNum) {
      meth4x4 = methInfo[(i << 4) + 10];
      if (meth4x4 == METH_DCT_4x4 ||
          meth4x4 == METH_DCT_4x4_N_0_8 ||
          meth4x4 == METH_DCT_4x4_N_4_8 ||
          (meth4x4 >= KLT_4x4_N_0_8 &&
           meth4x4 <= KLT_4x4_N_7_8)) {
        dctmethod += quadNum;
        if (save4 != NULL)
          save4[j++] = meth4x4;
      }
    }
    quadNum *= 2;
  }

  return dctmethod;
}


/*
 * putIntBlock:
 *
 * Parameters:
 *      block           I: Source integer block pointer
 *      image           O: Destination integer image pointer
 *      cols            I: Columns in block
 *      rows            I: Rows in block
 *      colStart        I: Starting column in image
 *      rowStart        I: Starting row in image
 *      colsIma         I: Columns in image
 *
 * Function:
 *      Copies size of cols*rows block to image
 *      starting from (colStart, rowStart).
 *
 * Returns:
 *      pels            Amount of pels in block
 */

int putIntBlock(int *block, int *image, int cols, int rows,
                int colStart, int rowStart, int colsIma)
{
  int
     i, j, wrap = colsIma - cols;  /* Loop increment counters */

  image += rowStart * colsIma + colStart;

  for(j = 0; j < rows; j++, image += wrap)
     for(i = 0; i < cols; i++)
       *image++ = *block++;

   return rows * cols;
}


/*
 * getIntBlock:
 *
 * Parameters:
 *      image           I: Source integer image pointer
 *      block           O: Destination integer block pointer
 *      cols            I: Columns in block
 *      rows            I: Rows in block
 *      colStart        I: Starting column in image
 *      rowStart        I: Starting row in image
 *      colsIma         I: Columns in image
 *
 * Function:
 *      Takes size of cols*rows block from image
 *      starting from (colStart, rowStart).
 *
 * Returns:
 *      pels            Amount of pels in block
 */

int getIntBlock(int *image, int *block, int cols, int rows,
                int colStart, int rowStart, int colsIma)
{
  int
     i, j;   /* Loop increment counters */

  image += rowStart * colsIma + colStart;
  switch (cols) {
    case 4:
      for(j = 0; j < rows; j++, block += 4, image += colsIma) {
        block[0] = image[0];
        block[1] = image[1];
        block[2] = image[2];
        block[3] = image[3];
      }
      break;
    case 1:
      for(j = 0; j < rows; j++, image += colsIma)
        block[j] = *image;
      break;
    case 5:
      for(j = 0; j < rows; j++, block += 5, image += colsIma) {
        block[0] = image[0];
        block[1] = image[1];
        block[2] = image[2];
        block[3] = image[3];
        block[4] = image[4];
      }
      break;
    case 2:
      for(j = 0; j < rows; j++, block += 2, image += colsIma) {
        block[0] = image[0];
        block[1] = image[1];
      }
      break;
    case 8:
      for(j = 0; j < rows; j++, block += 8, image += colsIma) {
        block[0] = image[0];
        block[1] = image[1];
        block[2] = image[2];
        block[3] = image[3];
        block[4] = image[4];
        block[5] = image[5];
        block[6] = image[6];
        block[7] = image[7];
      }
      break;
    default:
      for(j = 0; j < rows; j++, image += colsIma)
        for(i = 0; i < cols; i++)
          *block++ = image[i];
      break;
  }

  return cols * rows;
}


/*
 * calcBlockEnergy:
 *
 * Parameters:
 *      block           I: Source image or block pointer
 *      size            I: Size of elements
 *
 * Function:
 *      Calculates total energy of source block. 
 *
 * Returns:
 *      energy          Amount of energy in block.
 */

int calcBlockEnergy(int *block, int size)
{
  int i,                /* Loop increment counter                       */
      energy = 0;       /* Amount of energy in block                    */

  for(i=0; i<size; i++) {
    energy += block[i]*block[i];
  }

  return energy;
}

/*
 * findPredNums8x8:
 *
 * Parameters:
 *      pred            I: Prediction block
 *      rotateFlag      I: Rotation flag for 8x8
 *      qp              I: Quantization parameter
 *      dctcoef         I: Coefficient pointer where prediction number and
 *                         rotation information are saved
 *      chroma          I: Type of block (chroma or luma)
 *      method          I: Coding method
 *      mode            I: Mode of block (intra/inter)
 *
 * Function:
 *      Finds 4x4 blocks prediction numbers and prediction flags 
 *      for given 8x8 block.
 *
 * Returns:
 *      void
 */

void findPredNums8x8(int *pred, int rotateFlag, int qp, int *dctcoef,
                     int chroma, int method, int mode)
{
  int
    tmpPred[SIZE_8x8],  /* Temporal prediction block            */
    subPred[SIZE_4x4],  /* 4x4 block from 8x8 prediction block  */
    predno,             /* 4x4 prediction number                */
    rotate,             /* Rotation flag                        */
    subMethod,
    predDir=0;

  copyBlock8x8(pred, tmpPred, rotateFlag);

  subMethod = method - METH_4x4_1 + 1;

  if(subMethod&1){
     getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                 START_COL_4x4_1, START_ROW_4x4_1, COLS_8x8);
     predno = findPredNum4x4(subPred, qp, chroma, mode, SIZE_8x8,
                             predDir, &rotate);
     dctcoef[13] = predno;
     dctcoef[14] = rotate;
  }

  if(subMethod&2){
     getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                 START_COL_4x4_2, START_ROW_4x4_2, COLS_8x8);
     predno = findPredNum4x4(subPred, qp, chroma, mode, SIZE_8x8,
                             predDir, &rotate);
     dctcoef[29] = predno;
     dctcoef[30] = rotate;
  }

  if(subMethod&4){
     getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                 START_COL_4x4_3, START_ROW_4x4_3, COLS_8x8);
     predno = findPredNum4x4(subPred, qp, chroma, mode, SIZE_8x8,
                             predDir, &rotate);
     dctcoef[45] = predno;
     dctcoef[46] = rotate;
  }

  if(subMethod&8){
     getIntBlock(tmpPred, subPred, COLS_4x4, ROWS_4x4,
                 START_COL_4x4_4, START_ROW_4x4_4, COLS_8x8);
     predno = findPredNum4x4(subPred, qp, chroma, mode, SIZE_8x8,
                             predDir, &rotate);
     dctcoef[61] = predno;
     dctcoef[62] = rotate;
  }
}


/*
 * findPredNum (Class/NonClass):
 *
 * Parameters:
 *      pred            I: 4x4 prediction block pointer
 *      qp              I: Quantization parameter
 *      chroma          I: Type of block (chroma or luma)
 *      mode            I: Mode of block (inter or intra or first)
 *      blkSize         I: Size of the coded block
 *      rotate          O: Rotation flag pointer
 *
 * Function:
 *      Finds 4x4 block prediction number and rotation flag.
 *
 * Returns:
 *      predno          Prediction number.
 */

int findPredNum4x4(int *pred, int qp, int chroma, int mode, int blockSize,
                   int predDir, int *rotate)
{
  int
    qpRange,                   /* QP range */
    predClass = 0;             /* Number of the final prediction class */

  /*
   * Find the right part of the coding method table according
   * to component type (chrominance/luminance) and QP range.
   */
  if (chroma) {
    predClass += START_OF_4x4_CHROMA_OVH;
    if (qp <= 9)       
      qpRange = 0;
    else 
      qpRange = 1;
  }
  else {
    if (qp <= 9)       
      qpRange = 0;
    else if (qp <= 15) 
      qpRange = 1;
    else if (qp <= 22) 
      qpRange = 2;
    else              
      qpRange = 3;
  }

  predClass += qpRange * AMOUNT_OF_4x4_CLASS_OVH;
  *rotate = UNIT_OPER;

  if (chroma == CHROMA && blockSize == SIZE_4x4) 
    predClass++;

  return predClass;
}


/*
 * calcUnnormVar:
 *
 * Parameters:
 *      source                  I: Source block pointer
 *      size                    I: Size of source block
 *
 * Function:
 *      Calculates unnormalized variance value.
 *
 * Returns:
 *      variance                Unnormalized variance value.
 */

static int32 calcUnnormVar(int *source, int size)
{
/*
 * Precomputed values for squared bytes
 */
  int 
     i,               /* Loop increment counter                 */
     sum = 0,      /* Sum of source pels                                */
     power = 0; /* Energy of source pels                        */

  int32 
	variance32;	  /* Unnormalized variance value                  */

  for(i=0; i<size; i++) {
    sum   += source[i];
    power += source[i]*source[i];
  }
    
  variance32 = (int32)power - ((int32)(sum * sum)) / size;  /* integer division */

  return variance32;
}


int getIntBlkFromByteImage(u_char *image, int *block, int cols, int rows,
                           int colStart, int rowStart, int colsIma)
{
  int
     i, j, wrap = colsIma - cols;  /* Loop increment counters */

  image += rowStart * colsIma + colStart;
  for(j = 0; j < rows; j++, image += wrap) {
    for(i = 0; i < cols; i++) {
      *block++ = (int)(*image++);
    }
  }

  return cols * rows;
}


int putIntBlktoByteImage(int *block, u_char *image, int cols, int rows,
                         int colStart, int rowStart, int colsIma)
{
  int
     i, j, wrap = colsIma - cols;  /* Loop increment counters */

  image += rowStart * colsIma + colStart;

  for(j = 0; j < rows; j++, image += wrap) {
    for(i = 0; i < cols; i++) {
      *image++ = (u_char)(*block++);
    }
  }

  return rows * cols;
}


/*
 *
 * block_mid_hori :
 *
 * Parameters:
 *      block           I: Block to be processed.
 *      block_mid       O: Mid Block.
 *      blocksizex      I: width of the block.
 *      blocksizey      I: length of the block.
 * Function:
 *      Computes the inter pixel values in 
 *      Horizontal Direction.
 *
 * Returns:
 *      Nothing.
 */

static void block_mid_hori(int *block, int *block_mid,
                           int blocksizex, int blocksizey)
{
  int
    x, y,          /* x- and y-coordinates of pixel       */
    end_x,         /* End of block x                      */
    end_y,         /* End of block y                      */
    *ptrDest,      /* Pointer to destination              */
    *ptrSou1,      /* Pointer to source 1 pixel location  */
    *ptrSou2;      /* Pointer to source 1 pixel location  */


  end_x = blocksizex - 1;
  end_y = blocksizey;
  ptrDest = block_mid;
  ptrSou1 = block;
  ptrSou2 = block + 1;

  for(y=0; y<end_y; y++, ptrSou1++, ptrSou2++) {
    for(x=0; x<end_x; x++, ptrDest++, ptrSou1++, ptrSou2++) {
      if ((*ptrSou1 + *ptrSou2) >=0)
		*ptrDest = (*ptrSou1 + *ptrSou2) >> 1;
	  else
		*ptrDest = -((-(*ptrSou1 + *ptrSou2)) >> 1);
    }
  }
}


/*
 *
 * block_mid_vert :
 *
 * Parameters:
 *      block           I: Block to be processed.
 *      block_mid       O: Mid Block.
 *      blocksizex      I: width of the block.
 *      blocksizey      I: length of the block.
 * Function:
 *      Computes the inter pixel values in 
 *      Vertical Direction.
 *
 * Returns:
 *      Nothing.
 */

static void block_mid_vert(int *block, int *block_mid,
                           int blocksizex, int blocksizey)
{
  int
    x, y,          /* x- and y-coordinates of pixel       */
    end_x,         /* End of block x                      */
    end_y,         /* End of block y                      */
    *ptrDest,      /* Pointer to destination              */
    *ptrSou1,      /* Pointer to source 1 pixel location  */
    *ptrSou2;      /* Pointer to source 1 pixel location  */


  end_x = blocksizex;
  end_y = blocksizey - 1;
  ptrDest = block_mid;
  ptrSou1 = block;
  ptrSou2 = block + blocksizex;

  for(y=0; y<end_y; y++) {
    for(x=0; x<end_x; x++, ptrDest++, ptrSou1++, ptrSou2++) {
	  if ((*ptrSou1 + *ptrSou2) >=0)
		*ptrDest = (*ptrSou1 + *ptrSou2) >> 1;
	  else
		*ptrDest = -((-(*ptrSou1 + *ptrSou2)) >> 1);
    }
  }
}


/*
 *
 *  Gradient_angled:
 *
 * Parameters:
 *      block           I: Block to be processed.
 *      block_mid       I: Averaged middle block to be processed.
 *      sizex           I: Width of the block.
 *      sizey           I: Length of the block.
 *      angle           I: Angle of gradient.
 * Function:
 *      Computes the gradient at the direction
 *      given by angle, the 4 middle angles.
 *      (-3pi/8,-pi/8,pi/8,3pi/8).
 * Returns:
 *      Gradient Value
 */

static int32 Gradient_angled(int *block, int *block_mid,
                             int sizex, int sizey, int angle)
{
  int
    x, y,          /* x- and y-coordinates of pixel      */
    add_x,         /* Other pixels x-difference from pel */
    add_y,         /* Other pixels y-difference from pel */
    row_x,         /* Addition to block after row        */
    mid_x,         /* Addition to block_mid after row    */
    sta_x,         /* Start of block x                   */
    sta_y,         /* Start of block y                   */
    end_x,         /* End of block x                     */
    end_y,         /* End of block y                     */
    sizeMidx,      /* Block_mid sizex                    */
    sizeMidy,      /* Block_mid sizey                    */
    *ptrBlk,       /* Pointer to current pixels location */
    *ptrMid,       /* Pointer to other pixels location   */
    gradsum = 0;   /* Sum of absolute fradient values    */

  int32
    amount;    /* Amount of correct gradient values  */


  /* Angle Dependent Constants */
  if(angle == N_3_8) {
    sizeMidx = sizex - 1;
    sizeMidy = sizey;
    add_x = 0;
    add_y = 1;
    sta_x = 0;
    sta_y = 0;
    end_x = sizex - 1;
    end_y = sizey - 1;
    row_x = 1;  /* sizex    - (end_x - sta_x) */
    mid_x = 0;  /* sizeMidx - (end_x - sta_x) */
  }
  else if(angle == N_5_8) {
    sizeMidx = sizex - 1;
    sizeMidy = sizey;
    add_x = -1;
    add_y = 1;
    sta_x = 1;
    sta_y = 0;
    end_x = sizex;
    end_y = sizey - 1;
    row_x = 1;
    mid_x = 0;
  }
  else if(angle == N_1_8) {
    sizeMidx = sizex;
    sizeMidy = sizey - 1;
    add_x = 1;
    add_y = 0;
    sta_x = 0;
    sta_y = 0; 
    end_x = sizex - 1;
    end_y = sizey - 1;
    row_x = 1;
    mid_x = 1;
  }
  else if(angle == N_7_8) {
    sizeMidx = sizex;
    sizeMidy = sizey - 1;
    add_x = -1;
    add_y = 0;
    sta_x = 1;
    sta_y = 0; 
    end_x = sizex;
    end_y = sizey - 1;
    row_x = 1;
    mid_x = 1;
  }
  else {
    printf("Wrong angle %d in Gradient_angled()\n", angle);
    exit(1);
  }

  ptrBlk = block + sta_x + sta_y*sizex;
  ptrMid = block_mid + (sta_x + add_x) + (sta_y + add_y)*sizeMidx;

  for(y=sta_y; y<end_y; y++, ptrBlk+=row_x, ptrMid+=mid_x) {
    for(x=sta_x; x<end_x; x++, ptrBlk++, ptrMid++) {
      gradsum += (*ptrMid > *ptrBlk ? *ptrMid - *ptrBlk : *ptrBlk - *ptrMid);
    }
  }

  amount = (int32)(end_y-sta_y)*(end_x-sta_x);

  if(gradsum < 1)
    gradsum = 1;

  /* Result has 10 fraction bits */
  return ((int32)gradsum*1024/amount);
}


/*
 * calcDirGradient:
 *
 * Parameters:
 *      source          I: Source image or block pointer
 *      sizex           I: Columns in source image
 *      sizey           I: Rows in source image
 *      dir             I: Direction where the gradient is calculated
 *
 * Function:
 *      Calculates the gradient image for given direction to the given image.
 *      Function returns the amount of correct gradient values in image.
 *
 * Returns:
 *      amount          Amount of gradient values.
 */

int32  calcDirGradient(int *source, int sizex, int sizey, int dir)
{
  int
    x, y,          /* x- and y-coordinates of pixel */
    amount = 0,    /* Amount of correct gradient values */
    *ptr,          /* Pointer to other pixels location	*/
    gradsum = 0,   /* Sum of absolute fradient values */
    asizex;         /* # of columns to be processed */

  /*
   * Find the direction dependent constants.
   */
  asizex = sizex;
  if(dir == N_0_8) {
    asizex--;
    ptr = source + 1;
  } else if(dir == N_2_8) {
    sizey--;
    asizex--;
    ptr = source + 1 + sizex;
  } else if(dir == N_4_8) {
    sizey--;
    ptr = source + sizex;
  } else if(dir == N_6_8) {
    sizey--;
    asizex--;
    ptr = source + sizex;
    source++;
  }

  amount = asizex * sizey;
  /*
   * Calculate the gradient value pixel by pixel.
   */
  switch (asizex) {
  case 3:
    for(y=0; y<sizey; y++, source += sizex, ptr += sizex) {
      gradsum += abs(ptr[0] - source[0]);
      gradsum += abs(ptr[1] - source[1]);
      gradsum += abs(ptr[2] - source[2]);
    }
    break;
  case 4:
    for(y=0; y<sizey; y++, source += sizex, ptr += sizex) {
      gradsum += abs(ptr[0] - source[0]);
      gradsum += abs(ptr[1] - source[1]);
      gradsum += abs(ptr[2] - source[2]);
      gradsum += abs(ptr[3] - source[3]);
    }
    break;
  case 7:
    for(y=0; y<sizey; y++, source += sizex, ptr += sizex) {
      gradsum += abs(ptr[0] - source[0]);
      gradsum += abs(ptr[1] - source[1]);
      gradsum += abs(ptr[2] - source[2]);
      gradsum += abs(ptr[3] - source[3]);
      gradsum += abs(ptr[4] - source[4]);
      gradsum += abs(ptr[5] - source[5]);
      gradsum += abs(ptr[6] - source[6]);
    }
    break;
  case 8:
    for(y=0; y<sizey; y++, source += sizex, ptr += sizex) {
      gradsum += abs(ptr[0] - source[0]);
      gradsum += abs(ptr[1] - source[1]);
      gradsum += abs(ptr[2] - source[2]);
      gradsum += abs(ptr[3] - source[3]);
      gradsum += abs(ptr[4] - source[4]);
      gradsum += abs(ptr[5] - source[5]);
      gradsum += abs(ptr[6] - source[6]);
      gradsum += abs(ptr[7] - source[7]);
    }
    break;
  default:
    for(y=0; y<sizey; y++, source += sizex, ptr += sizex) {
      for(x=0; x<asizex; x++) {
        gradsum += abs(ptr[x] - source[x]);
      }
    }
    break;
  }

  if(gradsum < 1)
    gradsum = 1;

  /* Result has 10 fraction bits */
  return((int32)gradsum*1024/(int32)amount);
}


/*
 *
 *  blockEdge:
 *
 * Parameters:
 *      block           I: Image block.
 *      sizex           I: width of the block.
 *      sizey           I: length of the block.
 *      edge            O: Edge Type
 *
 * Function:
 *      Finds the Edge of the block.
 *
 * Returns:
 *      void
 */
void blockEdge(int *block, int sizex, int sizey, u_char *edge)
{
  int
    i, minPosit,
    flag, index[4];

  int32
    gradient[8], gradRatio[4], ratio[4],
    minGrad, maxGrad, minGradRatio,
    textThrHor, textThrMid, textThrDia;

  const int32
    edgeThr        =  154,        /* 0.15f  */
    textThr        =  614,        /* 0.60f  */
    flatThr        =  5 * 512,    /* 2.50f  */
    textSmallThr   =  10 * 1024,  /*10.00f  */
    edgeThrDiaOnly =  1 * 512,    /* 0.50f  */
    edgeThrMidOnly =  410,        /* 0.40f  */
    edgeThrHorOnly =  307,        /* 0.30f  */
    edgeThrDiaSure =  614,        /* 0.60f  */ 
    edgeThrMidSure =  1 * 512,    /* 0.50f  */
    edgeThrHorSure =  410,        /* 0.40f  */
    veryFlatThr    =  2 * 1024;   /* 2.00f  */

  if(sizex <= 4 /* && sizey == 4 */ ) {         /* 4x4 Block */
    textThrHor =  9 * 1024;   /*  9.0 */
    textThrMid = 23 *  512;   /* 11.5 */
    textThrDia = 14 * 1024;   /* 14.0 */
  }
  else /* if(sizex == 8 && sizey == 8) */ {     /* 8x8 Block */
    textThrHor =  7 * 1024;   /*  7.0 */
    textThrMid =  9 * 1024;   /*  9.0 */
    textThrDia = 11 * 1024;   /* 11.0 */
  }

  /* This part is possible to generate much faster */

  gradient[N_0_8] = calcDirGradient(block, sizex, sizey, N_0_8);   /* - */
  gradient[N_4_8] = calcDirGradient(block, sizex, sizey, N_4_8);   /* | */

  if(gradient[N_0_8] <= veryFlatThr && gradient[N_4_8] <= veryFlatThr)
    *edge = FLAT0;
  else
  {
    gradient[N_2_8] = calcDirGradient(block, sizex, sizey, N_2_8);   /* \ */
    gradient[N_6_8] = calcDirGradient(block, sizex, sizey, N_6_8);   /* / */

    /* gradRatio has 10 fraction bits */
    gradRatio[0/2] = gradient[0] * 1024 / gradient[4];
    gradRatio[2/2] = gradient[2] * 1024 / gradient[6];
    gradRatio[4/2] = gradient[4] * 1024 / gradient[0];
    gradRatio[6/2] = gradient[6] * 1024 / gradient[2];

    minGrad = maxGrad = gradient[0];

    for(i=2; i<8; i+=2)
    {
      if (gradient[i] < minGrad)
        minGrad = gradient[i];

      if (gradient[i] > maxGrad)
        maxGrad = gradient[i];
    }

    flag = 0;
    if(maxGrad <= flatThr)
      *edge = FLAT0;
    else
    {
      minPosit = minInt32(gradRatio, &minGradRatio, 4);
      if(minGradRatio <= edgeThr)
        *edge = minPosit*2;
      else if(minGradRatio >= textThr)
      {
        *edge = UCHAR_MAX;
        flag  = 1;
      }
      else
      {
        for(i=0; i<4; i++)
        {
          index[i] = i * 2;
          ratio[i] = gradRatio[i];
        }
        sortInt32SmallToBigIndex(ratio, index, 4, 3);

        if(3 * (ratio[1]-ratio[0]) < (ratio[2]-ratio[1]) )
        {
          if((index[0] == N_2_8 || index[0] == N_6_8) &&
              ratio[0] < edgeThrDiaSure)
          {
            *edge = index[0];
            flag  = 2;
          }
          else if((index[1] == N_2_8 || index[1] == N_6_8) &&
                   ratio[0] < edgeThrDiaSure)
          {
            *edge = index[1];
            flag  = 2;
          }
        }

        if(flag == 0)
        {
          if((index[0] == N_2_8 || index[0] == N_6_8) &&
              ratio[0] < edgeThrDiaOnly)
          {
            *edge = index[0];
            flag  = 2;
          }
          else if((index[0] == N_0_8 || index[0] == N_4_8) &&
                   ratio[0] < edgeThrHorOnly)
          {
            *edge = index[0];
            flag  = 2;
          }
        }

        if(flag == 0)
        {
          if((index[0] == N_2_8 || index[0] == N_6_8) &&
              ratio[0] < edgeThrDiaSure)
            *edge = index[0];
          else if((index[0] == N_0_8 || index[0] == N_4_8) &&
                   ratio[0] < edgeThrHorSure)
            *edge = index[0];
          else
          {
            *edge = UCHAR_MAX;
            flag  = 1;
          }
        }
      }

      if(*edge == N_0_8 || *edge == N_4_8)
      {
        if(minGrad > textThrHor)
          flag = 1;
      }
      else if (*edge == N_2_8 || *edge == N_6_8)
      {
        if(minGrad > textThrDia)
          flag = 1;
      }

      if (flag == 1)
      {
        if(maxGrad < textSmallThr)
          *edge = FLAT0;
        else
          *edge = TEXT2;
      }
    }
  }

}







/*
 *
 *  predIntra:
 *
 * Parameters:
 *      RecoImag        I: Reconstructed Image.
 *      pred_blk        O: Predicted block.
 *      x               I: x Position.
 *      y               I: y Position.
 *      method          I: Prediction method Of the Block.
 *      blocksizex      I: width of the block.
 *      blocksizey      I: length of the block.
 *      x_size          I: width of the Image.
 *      y_size          I: length of the Image.
 *      dcValues        I: Values of previous blocks DCs.
 *
 * Function:
 *      Performs Prediction of the block based on the method.
 *
 * Returns:
 *      dcValueSum      Sum of prediction block values.
 */

int32 predIntra(u_char *RecoImag, int32 *pred_blk, int32 x, int32 y,
                int32 method, int32 blocksizex, int32 blocksizey,
                int32 x_size, int32 y_size, int32 sliceMinx, int32 sliceMaxx,
                int32 sliceMiny, int32 sliceMaxy, int32 leftDc, int32 upperDc,
                int32 upleftDc, int16 *slicePredMap, int32 sliceSx,
                int32 sliceSy, int32 currSliceID)
{
  int i, j, ii, jj, value[3], vect[8], a, b, sliceBufSizeX;
  int addX, addY, dcValueSum=0;
  int blkTaken, blkTakenx, blkTakeny;
  int sliceFlag, stepx, stepy, tx, ty;
  static u_char 
    tmpBuffer[180];
  u_char
    *tmpImage = &tmpBuffer[38];
  const u_char
     bufferSize = 18;

  if(slicePredMap != NULL) {
    sliceFlag = 1;
    stepx   = (sliceMaxx-sliceMinx) / sliceSx;  /* integer division */
    stepy   = (sliceMaxy-sliceMiny) / sliceSy;  /* integer division */
    sliceBufSizeX = x_size / stepx;				  /* integer division */
  }
  else {
    sliceFlag = 0;
    stepx   = 4;
    stepy   = 4;
  }
  if(stepx != stepy || stepx < 4 || stepy < 4) {
    printf("Wrong stepx %d or stepy %d in predIntra()\n", stepx, stepy);
    exit(1);
  }

  switch(method) {

  case  0:
    addX = 3;           /* no prediction */ 
    addY = 3;
    break;

  case  1:
    addX = 0;           /* DC prediction */ 
    addY = 0;
    break;

  case  2:
    addX =  2;          /* up   */
    addY = -1;
    break;
  case  3:
    addX =  1;          /* up   */
    addY = -1;
    break;
  case  4:
    addX =  1;          /* up   */
    addY = -2;
    break;

  case  5: case 14:
    addX =  0;          /* up   */
    addY = -1;
    break;

  case  6:
    addX = -1;          /* up & left */
    addY = -2;
    break;
  case  7:
    addX = -1;          /* up & left */
    addY = -1;
    break;
  case  8:
    addX = -2;          /* up & left */
    addY = -1;
    break;

  case  9: case 15:
    addX = -1;          /* left */
    addY =  0;
    break;

  case 10:
    addX = -2;          /* left */
    addY =  1;
    break;
  case 11:
    addX = -1;          /* left */
    addY =  1;
    break;
  case 12:
    addX = -1;          /* left */
    addY =  2;
    break;

  default:
    printf("Wrong method %d in predIntra()\n", method);
    exit(1);
    break;
  }

  if(addX > stepx || addY > stepy) {
    printf("Wrong addX %d or addY %d in predIntra()\n", addX, addY);
    exit(1);
  }

  if(addX >= 2 && addY >= 2) {                          /* no prediction */
    dcValueSum = -1;
  }
  
  else if(addX == 0 && addY == 0) {                     /* DC prediction */
    dcValueSum = predictDCValue(leftDc, upperDc, upleftDc, LUMA);
    for(i=0; i<blocksizex*blocksizey; i++) {
      pred_blk[i] = dcValueSum;
    }
    dcValueSum *= (blocksizex*blocksizey);
  }

  else if(addX == 0) {                                  /* from up | */
    if(y >= sliceMiny + stepy) {
      blkTakenx = blocksizex;

      if(sliceFlag == 1) {
        i = blkTakenx;
        while(i > 0) {
          ty = (y-1) / stepy;     /* integer division */
          tx = (x+i-1) / stepx;  /* integer division */
          i -= stepx;
          if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {   /* outside */
            blkTakenx = i;
          }
        }
      }

      if(blkTakenx >= blocksizex) {
        if(method == 5) {
          for(j=0; j<blocksizex; j++) {
            vect[j] = (int)RecoImag[(y-1)*x_size+(x+j)];
            dcValueSum += vect[j];
          }
        }
        else if(method == 14) {
          for(j=0; j<blocksizex; j++) {
            value[0] = (int)RecoImag[(y-1)*x_size+(x+j)];
            value[1] = (int)RecoImag[(y-2)*x_size+(x+j)];
            value[2] = (int)RecoImag[(y-3)*x_size+(x+j)];
            vect[j]  = median(value,3);
            dcValueSum += vect[j];
          }
        }
        else {
          printf("Wrong up method %d in predIntra()\n", method);
          exit(1);
        }
        dcValueSum *= blocksizey;
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            pred_blk[i*blocksizex+j] = vect[j];
          }
        }
      }
      else {
        dcValueSum = -2;
      }
    }      
    else {
      dcValueSum = -2;
    }
  }

  else if(addY == 0) {                                  /* from left - */
    if(x >= sliceMinx + stepx) {
      blkTakeny = blocksizey;

      if(sliceFlag == 1) {
        i = blkTakeny;
        while(i > 0) {
          ty = (y+i-1) / stepy;  /* integer division */
          tx = (x-1) / stepx;	  /* integer division */
          i -= stepy;
          if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {   /* outside */
            blkTakeny = i;
          }
        }
      }

      if(blkTakeny >= blocksizey) {
        if(method == 9) {
          for(i=0; i<blocksizey; i++) {
            vect[i] = (int)RecoImag[(y+i)*x_size+(x-1)];
            dcValueSum += vect[i];
          }
        }
        else if(method == 15) {
          for(i=0; i<blocksizey; i++) {
            value[0] = (int)RecoImag[(y+i)*x_size+(x-1)];
            value[1] = (int)RecoImag[(y+i)*x_size+(x-2)];
            value[2] = (int)RecoImag[(y+i)*x_size+(x-3)];
            vect[i]  = median(value, 3);
            dcValueSum += vect[i];
          }
        }
        else {
          printf("Wrong left method %d in predIntra()\n", method);
          exit(1);
        }
        dcValueSum *= blocksizex;
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            pred_blk[i*blocksizex+j] = vect[i];
          }
        }
      }
      else {
        dcValueSum = -2;
      }
    }
    else {
      dcValueSum = -2;
    }
  }
  
  else if(addX > 0 && addY < 0) {                       /* from up right / */
    if(y >= sliceMiny+abs(addY)) {
      blkTakenx = sliceMaxx - x;

      if(blkTakenx > (1+1)*blocksizex) {
        blkTakenx = (1+1)*blocksizex;
      }

      if(sliceFlag == 1) {
        i = blkTakenx;
        while(i > 0) {
          ty = (y-1) / stepy;    /* integer division */
          tx = (x+i-1) / stepx;	  /* integer division */
          i -= stepx;
          if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {/*outside*/
            blkTakenx = i;
          }
        }
      }

      if(blkTakenx >= blocksizex) {
        for(i = y + addY, ii = addY; i < y; i++, ii++) {
          for(j = x + addX, jj = addX; j < x + blkTakenx; j++, jj++) {
            tmpImage[ii * bufferSize + jj] = RecoImag[i * x_size + j];
          }
        }

        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blkTakenx; j++) {
            a = j + addX;
            b = i + addY;
            if(a >= blkTakenx) {
              a = blkTakenx-1;
              b = -1;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            if(j < blocksizex) {
              pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
              dcValueSum += pred_blk[i*blocksizex+j];
            }
          }
        }
      }
      else {
        dcValueSum = -2;
      }
    }
    else {
      dcValueSum = -2;
    }
  }

  else if(addX < 0 && addY < 0) {                       /* from up left \ */
    if(x >= sliceMinx+abs(addX) || y >= sliceMiny+abs(addY)) {
      blkTaken  = 0;
      blkTakeny = 0;
      blkTakenx = 0;

      if(x >= sliceMinx+abs(addX) && y >= sliceMiny+abs(addY)) {
        blkTaken = -stepx;

        if(sliceFlag == 1) {
          i = blkTaken;
          while(i < 0) {
            ty = (y+i+1) / stepy;  /* integer division */
            tx = (x+i+1) / stepx;  /* integer division */
            i += stepx;
            if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {   /* outside */
              blkTaken = i;
            }
          } 
        }
      }
      if(x >= sliceMinx+abs(addX)) {
        blkTakeny = blocksizey;

        if(sliceFlag == 1) {
          i = blkTakeny;
          while(i > 0) {
            ty = (y+i-1) / stepy;  /* integer division */
            tx = (x-1) / stepx;		/* integer division */
            i -= stepy;
            if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {   /* outside */
              blkTakeny = i;
            }
          }
        }
      }
      if(y >= sliceMiny+abs(addY)) {
        blkTakenx = blocksizex;

        if(sliceFlag == 1) {
          i = blkTakenx;
          while(i > 0) {
            ty = (y-1) / stepy;   /* integer division */
            tx = (x+i-1) / stepx; /* integer division */
            i -= stepx;
            if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {   /* outside */
              blkTakenx = i;
            }
          }
        }
      }

      if(blkTaken <= -stepx) {  /* x>=sliceMinx+abs(addX) && y>=sliceMiny+abs(addY) */
        for(i = y + addY, ii = addY; i < y; i++, ii++) {
          for(j = x + addX, jj = addX; j < x; j++, jj++) {
            tmpImage[ii * bufferSize + jj] = RecoImag[i * x_size + j];
          }
        }
      }
      if(blkTakeny >= blocksizey) {     /* x >= sliceMinx+abs(addX) */
        for(i = y, ii = 0; i < y + blocksizey + addY; i++, ii++) {
          for(j = x + addX, jj = addX; j<x; j++, jj++) {
            tmpImage[ii * bufferSize + jj] = RecoImag[i * x_size + j];
          }
        }
      }
      if(blkTakenx >= blocksizex) {     /* y >= sliceMiny+abs(addY) */
        for(i = y + addY, ii = addY; i < y; i++, ii++) {
          for(j = x, jj = 0; j < x + blocksizex + addX; j++, jj++) {
            tmpImage[ii * bufferSize + jj] = RecoImag[i * x_size + j];
          }
        }
      }

      if(blkTakenx >= blocksizex && blkTakeny >= blocksizey &&
         blkTaken  <= -stepx) {                 /* up + left + leftup */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex+j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTakenx >= blocksizex && blkTaken <= -stepx) {  /* up + leup */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(a < 0 && b >= 0) {
              a = addX;
              b = -1;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex+j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTakeny >= blocksizey && blkTaken <= -stepy) {  /* le + leup */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(b < 0 && a >= 0) {
              a = -1;
              b = addY;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTakenx >= blocksizex && blkTakeny >= blocksizey) { /* up+le */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(a < 0 && b < 0) {
              if(addY < addX) {
                if(b < a) {
                  a =  0;
                  b = -1;
                }
                else {
                  a = -1;
                  b =  0;
                }
                tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
              }
              else if(addX < addY) {
                if(a < b) {
                  a = -1;
                  b =  0;
                }
                else {
                  a =  0;
                  b = -1;
                }
                tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
              }
              else {
                tmpImage[i * bufferSize + j] = 1;
                a =  0;
                b = -1;
                tmpImage[i * bufferSize + j] += tmpImage[b * bufferSize + a];
                a = -1;
                b =  0;
                tmpImage[i * bufferSize + j] += tmpImage[b * bufferSize + a];
                tmpImage[i * bufferSize + j] >>= 1;
              }
            }
            else {
              tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            }
            pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTakenx >= blocksizex) {                /* only up */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(a < 0) {
              a =  0;
              b = -1;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTakeny >= blocksizey) {                /* only left */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(b < 0) {
              a = -1;
              b =  0;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else if(blkTaken <= -stepx) {                     /* only leftup */
        for(i=0; i<blocksizey; i++) {
          for(j=0; j<blocksizex; j++) {
            a = j + addX;
            b = i + addY;
            if(b < 0 && a >= 0) {
              a = -1;
              b = addY;
            }
            else if(a < 0 && b >= 0) {
              a = addX;
              b = -1;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
            dcValueSum += pred_blk[i*blocksizex+j];
          }
        }
      }
      else {
        dcValueSum = -2;
      }
    }
    else {
      dcValueSum = -2;
    }
  }

  else if(addY > 0 && addX < 0) {               /* from bottom left / */
    if(x >= sliceMinx+abs(addX)) {

      blkTakeny = blocksizey;

      if(sliceFlag == 1) {
        i = blkTakeny;
        while(i > 0) {
          ty = (y+i-1) / stepy;  /* integer division */
          tx = (x-1) / stepx;    /* integer division */
          i -= stepy;
          if (slicePredMap[ty * sliceBufSizeX + tx] != currSliceID) {/*outside*/
            blkTakeny = i;
          }
        }
      }

      if(blkTakeny >= blocksizey) {
        for(i = y + addY, ii = addY; i < y + blkTakeny; i++, ii++) {
          for(j = x + addX, jj = addX; j < x; j++, jj++) { 
            tmpImage[ii * bufferSize + jj] = RecoImag[i * x_size + j];
          }
        }

        for(j=0; j<blocksizex; j++) {
          for(i=0; i<blkTakeny; i++) {
            a = j + addX;
            b = i + addY;
            if(b >= blkTakeny) {
              a = -1;
              b = blkTakeny-1;
            }
            tmpImage[i * bufferSize + j] = tmpImage[b * bufferSize + a];
            if(i < blocksizey) {
              pred_blk[i * blocksizex + j] = (int)tmpImage[i * bufferSize + j];
              dcValueSum += pred_blk[i*blocksizex+j];
            }
          }
        }
      }
      else {
        dcValueSum = -2;
      }
    }
    else {
      dcValueSum = -2;
    }
  }
  else {
    printf("No pred for addX %d or addY %d in predIntra()\n", addX, addY);
    exit(1);
  }

  if(dcValueSum == -2) {        /* no directional prediction */
    for(i=0; i<blocksizex*blocksizey; i++) {
      pred_blk[i] = 128;
    }
    dcValueSum = 128*blocksizex*blocksizey;
  }

  if(dcValueSum == -1) {        /* no prediction */
    for(i=0; i<blocksizex*blocksizey; i++) {
      pred_blk[i] = 0;
    }
  }

  if(dcValueSum < -1 || dcValueSum > 255*blocksizex*blocksizey) {
    printf("Wrong dcValueSum %d in predIntra()\n", dcValueSum);
    exit(1);
  }

  return dcValueSum;
}


int32 findValidVectors(int32 *validX, int32 *validY, int32 *validI,
                       int32 maxValid, int32 maxVector, int32 x, int32 y,
                       int32 blocksizex, int32 blocksizey, int32 x_size,
                       int32 sliceMinx, int32 sliceMaxx, int32 sliceMiny,
                       int32 sliceMaxy, int16 *slicePredMap,
                       int32 sliceSx, int32 sliceSy, int32 currSliceID,
                       int32 mode)
{
  int 
    valids, vectNo, size, validFlag,
    sliceFlag, stepx, stepy, sliceBufSizeX,
    dX, dY, bX, bY, ty, tx, maxCheck, checkNo,
    checkBufX[9], checkBufY[9];

  if(slicePredMap != NULL) {
    sliceFlag = 1;
    stepx = (sliceMaxx-sliceMinx) / sliceSx;  /* integer division */
    stepy = (sliceMaxy-sliceMiny) / sliceSy; /* integer division */
    sliceBufSizeX = x_size / stepx;             /* integer division */
  }
  else {
    sliceFlag = 0;
    stepx = 4;
    stepy = 4;
  }
  if(stepx != stepy || stepx < 4 || stepy < 4) {
    printf("Wrong stepx %d or stepy %d in findValidVectors()\n", stepx, stepy);
    exit(1);
  }

  /*
   * This part has just done for 8x8 blocks and only 4x4 or 8x8 stepsizes.
   */
  if(sliceFlag == 1) {
    if(blocksizex == stepx) {
      maxCheck = 4;
      checkBufX[0] = 0;            checkBufY[0] = 0;
      checkBufX[1] = blocksizex-1; checkBufY[1] = 0;
      checkBufX[2] = 0;            checkBufY[2] = blocksizey-1;
      checkBufX[3] = blocksizex-1; checkBufY[3] = blocksizey-1;
    }
    else {
      if(mode == FIRST_BLOCK) {
        maxCheck = 4;
        checkBufX[0] = 0;            checkBufY[0] = 0;
        checkBufX[1] = blocksizex-1; checkBufY[1] = 0;
        checkBufX[2] = 0;            checkBufY[2] = blocksizey-1;
        checkBufX[3] = blocksizex-1; checkBufY[3] = blocksizey-1;
      }
      else {
        maxCheck = 9;
        checkBufX[0] = 0;            checkBufY[0] = 0;
        checkBufX[1] = stepx;        checkBufY[1] = 0;
        checkBufX[2] = blocksizex-1; checkBufY[2] = 0;
        checkBufX[3] = 0;            checkBufY[3] = stepy;
        checkBufX[4] = stepx;        checkBufY[4] = stepy;
        checkBufX[5] = blocksizex-1; checkBufY[5] = stepy;
        checkBufX[6] = 0;            checkBufY[6] = blocksizey-1;
        checkBufX[7] = stepx;        checkBufY[7] = blocksizey-1;
        checkBufX[8] = blocksizex-1; checkBufY[8] = blocksizey-1;
      }
    }
  }

  vectNo = 0; valids = 0;
  size = blocksizex*blocksizey;
  while(valids < maxValid && vectNo < maxVector) {
    dX = motionTableX[vectNo];
    dY = motionTableY[vectNo];
    validFlag = 1;

    /*
     * Check if the whole block after motion is inside the bounded box.
     */
    if((y+dY)<sliceMiny || (y+dY+blocksizey)>sliceMaxy ||
       (x+dX)<sliceMinx || (x+dX+blocksizex)>sliceMaxx) {   /* outside */
      validFlag = 0;
    }
    else {                                                  /* inside  */

      /*
       * Check if the checking of sliceMap is needed.
       */
      if(sliceFlag == 1) {
        checkNo = 0;
        while(validFlag == 1 && checkNo < maxCheck) {
          bX = checkBufX[checkNo];
          bY = checkBufY[checkNo];
          if (stepx == 8) {
            tx = (x+dX+bX) >> 3;
            ty = (y+dY+bY) >> 3;
          }
          else if (stepx == 4) {
            tx = (x+dX+bX) >> 2;
            ty = (y+dY+bY) >> 2;
          }
          else {
            tx = (x+dX+bX) / stepx;  /* integer division */
            ty = (y+dY+bY) / stepy; /* integer divsiion */
		  }

          /*
           * Check if the checked pixel is inside the slice with same id.
           */
          if(slicePredMap[ty*sliceBufSizeX + tx] != currSliceID) { /* outside */
            validFlag = 0;
          }
          checkNo++;
        }
      }
    }

    if(validFlag == 1) {
      validX[valids] = dX;
      validY[valids] = dY;
      validI[valids] = vectNo;
      valids++;
    }

    vectNo++;
  }

  return valids;
}


/*
 *
 *  predIntraBlockMatch:
 *
 * Parameters:
 *     recoImag        I: Reconstructed Image.
 *     predBlk         O: Predicted block.
 *     x               I: x Position.
 *     y               I: y Position.
 *     dY              I: Vertical intra BM component.
 *     dX              I: horizontal intra BM component.
 *     blocksizex      I: width of the block.
 *     blocksizey      I: length of the block.
 *     x_size          I: width of the Image.
 *     y_size          I: length of the Image.
 *     sliceMinx, ..., sliceMaxy I: slice bounding box coordinates
 *     slicePredMap     I: slice map buffer
 *     sliceSx, sliceSy I: size (in 8x8 luma blocks) of slice map
 *
 * Function:
 *      Performs Prediction by block matching for decoder use.
 *
 * Returns:
 *      dcValueSum      Sum of prediction block values.
 */

int32 predIntraBlockMatch(u_char *recoImag, int32 *predBlk, int32 x, int32 y,
                          int32 blocksizex, int32 blocksizey,
                          int32 x_size, int32 dY, int32 dX)
{
  int
    pelNo, size,
    tmp, dcValueSum = 0, pelY, pelX;

  size = blocksizex*blocksizey;

  pelNo = 0;
  pelY = 0;
  while(pelY < blocksizey) {
    pelX = 0;
    while(pelX < blocksizex) {
      predBlk[pelNo] = (tmp = recoImag[(y+dY+pelY)*x_size + (x+dX+pelX)]);
      dcValueSum+= tmp;
      pelNo++;
      pelX++;
    }
    pelY++;
  }

  return dcValueSum;
}


int find4x4Start(int codingMethod, int size)
{
  int
    start;

  switch(codingMethod) { 

    case METH_4x4_8:
      start = 3 * size;
      break;

    case METH_4x4_4:
      start = size << 1;
      break;
 
    case METH_4x4_2:
      start = size;
      break;
  
    default:
      start = 0;
      break;
  }

  return start;
}



int find8x8Start(int codingMethod, int size)
{
  int
    start;

  switch (codingMethod) {
    case METH_4x4_DCT_8:
      start = 3 * size;
      break;

    case METH_4x4_DCT_4: case METH_4x4_DCT_12:
      start = size << 1;
      break;

    case METH_4x4_DCT_2: case METH_4x4_DCT_6:
    case METH_4x4_DCT_10: case METH_4x4_DCT_14:
      start = size;
      break;

    case  METH_4x8_DCT_2: case  METH_8x4_DCT_2:
      start = size;
      break;

    default:
      start = 0;
      break;
  }

  return start;
}


int findNumMethods(int method, int blkSize)
{
  int
    codedBlks;

  switch(method) {
    case METH_4x4_DCT_15:
      codedBlks = 4;
      break;

    case METH_4x4_DCT_14: case METH_4x4_DCT_13:
    case METH_4x4_DCT_11: case METH_4x4_DCT_7:
      codedBlks = 3;
      break;

    case METH_4x4_DCT_12: case METH_4x4_DCT_10:
    case METH_4x4_DCT_9 : case METH_4x4_DCT_6:
    case METH_4x4_DCT_5 : case METH_4x4_DCT_3:
    case METH_4x8_DCT_3 : case METH_8x4_DCT_3:
      codedBlks = 2;
      break;

    case NOT_CODED:
    case METH_EXTRAPOL_12: case METH_EXTRAPOL_14:
    case METH_EXTRAPOL_28: case METH_EXTRAPOL_48:
      codedBlks = 0;
      break;

    default:
      codedBlks = 1;
      break;
  }

  return codedBlks;
}


void findTripletInfos(int method8x8, int method4x4, int qp, int chroma,
                      int mode, int *tableNo, int *runBits, int *levelBits)
{
  int
    cols, rows,
    kltNum;
 
  switch(method8x8) {
    case METH_8x8_DCT:
    case METH_DCT_8x8_N_0_8: case METH_DCT_8x8_N_4_8:
      cols = COLS_8x8;
      rows = ROWS_8x8;
      kltNum = -1;
      break;

    case METH_4x8_DCT_1: case METH_4x8_DCT_2:
    case METH_4x8_DCT_3: case METH_4x8_DCT_4:
    case METH_8x4_DCT_1: case METH_8x4_DCT_2:
    case METH_8x4_DCT_3: case METH_8x4_DCT_4:
      cols = COLS_4x8;
      rows = ROWS_4x8;
      kltNum = -1;
      break;

    case METH_KLT_8x8_N_1_8: case METH_KLT_8x8_N_3_8:
    case METH_KLT_8x8_N_5_8: case METH_KLT_8x8_N_7_8:
      cols = COLS_8x8;
      rows = ROWS_8x8;
      kltNum = 1;
      break;

    case METH_KLT_8x8_N_2_8: case METH_KLT_8x8_N_6_8:
      cols = COLS_8x8;
      rows = ROWS_8x8;
      kltNum = 2;
      break;

    case METH_4x4_1: case METH_4x4_15:
    case METH_4x4_DCT_1:  case METH_4x4_DCT_2:
    case METH_4x4_DCT_3:  case METH_4x4_DCT_4:
    case METH_4x4_DCT_5:  case METH_4x4_DCT_6:
    case METH_4x4_DCT_7:  case METH_4x4_DCT_8:
    case METH_4x4_DCT_9:  case METH_4x4_DCT_10:
    case METH_4x4_DCT_11: case METH_4x4_DCT_12:
    case METH_4x4_DCT_13: case METH_4x4_DCT_14:
    case METH_4x4_DCT_15:
      cols = COLS_4x4;
      rows = ROWS_4x4;

      switch (method4x4) {

        case KLT_4x4_N_1_8: case KLT_4x4_N_3_8:
        case KLT_4x4_N_5_8: case KLT_4x4_N_7_8:
          kltNum = 1;
          break;

        case KLT_4x4_N_2_8: case KLT_4x4_N_6_8:
          kltNum = 2;
          break;

        case METH_DCT_4x4: case METH_DCT_4x4_N_0_8:
        case METH_DCT_4x4_N_4_8:
          kltNum = -1;
          break;
      }
      break;
  }

  if (kltNum < 0)
    vdxGetDctTable(cols, rows, qp, chroma, mode,
                   tableNo, runBits, levelBits);
  else
    vdxGetKltTable(cols, rows, qp, chroma, mode, kltNum,
                   tableNo, runBits, levelBits);

}


