/*
 * toolSplineInterp.c
 *
 *    Copyright (c) 2000 Nokia Research Center, Tampere, FINLAND
 *
 * Project:
 *    Nokia Very Low Bitrate Codec
 *
 * Contains:
 *    Cubic spline interpolation functions.
 *
 * Author:
 *    Nokia Research Center
 */

#include "mainFlags.h"


/*
 * The "spline interpolation" in this code refers to cubic convolution 
 * interpolation. The samples of convolution kernel are stored in a global
 * array. The derivative of the convolution kernel is used for derivative
 * calculation. Hence, the samples of derivative of spline convolution kernel
 * are stored in another global array. These arrays must be initialized
 * in the beginning of the program.
 *
 * The constants below specify the resolution of sampling of the convolution
 * kernels. 
 */


#define STEP_SPLINE  0.00390625
#define INTERV       256
#define SIZE_SPLINE  1024

#define flor(X) (((X) < 0) ? ((int)(X) - 1) : ((int)(X)))
#define double2byte(X) (X < 0.5 ? 0 : (X>=254.5 ? 255 : (Byte)(X+0.5)))
#define toShort(X) ((X) >= 0 ? (short)((X) + 0.5) : (short)((X) - 0.5))

/* Array for samples of cubic convolution kernel. It is stored in a way
 * that coefficients corresponding to consecutive integer neighbors are
 * consecutive in the array.
 */

double splineArray[SIZE_SPLINE];
short int splineArrayInt[SIZE_SPLINE];   /* Fixed-point kernel */

/* Array for samples of derivative of cubic convolution kernel.It is
 * stored in a way that coefficients corresponding to consecutive
 * integer neighbors are consecutive in the array.
 */

double derArray[SIZE_SPLINE];
short int derArrayInt[SIZE_SPLINE];      /* Fixed-point derivative kernel */

/* Frame size in number of pels */

int picSizeXglob, picSizeYglob;


/*
 * splSetPicSize:
 *
 * Parameters:
 *     sx           I: x-size
 *     sy           I: y-size
 *
 * Function:
 *     Sets global variables for picture size.
 *
 * Returns:
 *     Nothing
 *
 */

void splSetPicSize(int sx, int sy)
{
  picSizeXglob = sx;
  picSizeYglob = sy;
}


/*
 * g_inter0:
 *
 * Parameters:
 *     x          I: A non-integer value in the range [-2, 2]
 *
 * Function:
 *     Given a value in the range [-2, 2], returns the value of
 *     cubic convolution kernel at that location in continuous
 *     domain. It is used in the initialization of sample array.
 *
 * Returns:
 *     sample value.
 *
 */

double g_inter0(double x)
{
  static double 
    p =  1.5,
    q = -2.5,
    s =  1.0,
    t = -0.5,
    u =  2.5,
    v = -4.0,
    w =  2.0,
    y;
  
  if ((y=fabs(x)) < 1) 
    y = (p*y + q)*y*y + s;
  else if (y<2) 
    y = ((t*y + u)*y +v)*y + w;
  else 
    y = 0.0;

  return(y);
}


/*
 * g_der0:
 *
 * Parameters:
 *     x          I: A non-integer value in the range [-2, 2]
 *
 * Function:
 *     Given a value in the range [-2, 2], returns the value of
 *     derivative of cubic convolution kernel at that location in 
 *     continuous domain. It is used in the initialization of 
 *     sample array.
 *
 * Returns:
 *     sample value.
 *
 */

double g_der0(double x)
{
  static double 
    p =  4.5,
    q = -5.0,
    t = -1.5,
    u =  5.0,
    v = -4.0,
    y;

  if (x >= 0) {
    if (x < 1)
      y =(p*x + q)*x;
    else {
      if (x < 2)
        y =(t*x + u)*x + v;
      else
        y = 0.0;
    }
  }
  else {
    if (x > -1)
      y = (-p*x + q)*x;
    else {
      if (x > -2)
        y = (-t*x + u)*x - v;
      else
        y = 0.0;
    }
  }

  return(y);
}


/*
 * splInitInterpArray:
 *
 * Parameters:
 *     None.
 *
 * Function:
 *     Initializes the sample array (splineArrayInt) for cubic
 *     convolution kernel. Pixel interval is divided into 256
 *     (INTERV) reconstruction points. splineArrayInt contains
 *     four filter coefficients for each of the reconstruction
 *     point.
 *
 * Returns:
 *     Nothing.
 *
 */

void splInitInterpArray(void)
{
  int i,tot;

  for (i=0;i<INTERV;i++){
    splineArray[4*i]   = g_inter0( (double)i * STEP_SPLINE+1.0);
    splineArray[4*i+1] = g_inter0( (double)i * STEP_SPLINE);
    splineArray[4*i+2] = g_inter0( (double)i * STEP_SPLINE-1.0);
    splineArray[4*i+3] = g_inter0( (double)i * STEP_SPLINE-2.0);
  }

   for (i=0;i<INTERV;i++){
    tot = (splineArrayInt[4*i]=
      toShort(g_inter0( (double)i * STEP_SPLINE+1.0) * 16384.0));
    tot += (splineArrayInt[4*i+1] =
      toShort(g_inter0( (double)i * STEP_SPLINE) * 16384.0));
    tot += (splineArrayInt[4*i+2] =
      toShort(g_inter0( (double)i * STEP_SPLINE-1.0) * 16384.0));
    splineArrayInt[4*i+3] = (short int)(16384-tot); 
  }

}


/*
 * splInitDerivativeArray:
 *
 * Parameters:
 *     None.
 *
 * Function:
 *     Initializes the sample array (derArrayInt) for the derivative
 *     of cubic convolution kernel. Array contains four filter
 *     coefficients for each of the 256 (INTERV) reconstruction points.
 *
 * Returns:
 *     Nothing.
 *
 */

void splInitDerivativeArray(void)
{
  int i;

  for (i=0;i<INTERV;i++){
    derArray[4*i] = g_der0( (double)i * STEP_SPLINE+1.0);
    derArrayInt[4*i] = toShort(derArray[4*i] * 8192.0);
    derArray[4*i+1] = g_der0( (double)i * STEP_SPLINE);
    derArrayInt[4*i+1] = toShort(derArray[4*i+1] * 8192.0);
    derArray[4*i+2] = g_der0( (double)i * STEP_SPLINE-1.0);
    derArrayInt[4*i+2] = toShort(derArray[4*i+2] * 8192.0);
    derArray[4*i+3] = g_der0( (double)i * STEP_SPLINE-2.0);
    derArrayInt[4*i+3] = toShort(derArray[4*i+3] * 8192.0);
  }

}


/*
 * splInterpolation:
 *
 * Parameters:
 *     yin              I: Y component of the frame
 *     i, j             I: Coordinates of the pel to be interpolated
 *     m                I: Frame width
 *     n                I: Frame height
 *
 * Function:
 *     Cubic convolution interpolation at an arbitrary non-integer location,
 *     using the values at integer locations.
 *     First finds the nearest integer neighbors in the array, then
 *     accomplishes the cubic convolution interpolation using the integer
 *     neighbor values and samples of the cubic convolution kernel.
 *
 * Returns:
 *     Interpolated value at the given non-integer location.
 *
 */

double splInterpolation(Byte *yin, double i, double j,
                        int picSizeX, int picSizeY)
{
  double
    *spx, *spy,       /* spline array pointers for i and j values */
    ib, jb,           /* Bounded i and j values */
    sum1, sum;
  Byte
    *img;
  int
    wrap,
    offsetx, offsety,
    p, q,
    cy, cx, bx, by;


  /* Boundary Checks */
  if((i<1) || (i>=picSizeY-2.004) || (j<1) || (j>=picSizeX-2.004)) {
    
    /* Exceptions for failing boundary checks */
    
    bx = (int)flor(j)+3;
    by = (int)flor(i)+3;

    /* Let's quantize the coordinate to 1/INTERV (1/256) pixel
     * accuracy. Interpolation filter coefficients (four) for each
     * reconstruction point are stored consecutively in the
     * splineArray. Thus, offsets for this array are calculated as
     * 4*INTERV*fractional pixel value.
     */
    if((offsetx = (int)(0.5+INTERV*(j - (bx-3)))<<2)>=SIZE_SPLINE){
      offsetx = 0;
      bx++;
    }

    if((offsety = (int)(0.5+INTERV*(i - (by-3)))<<2)>=SIZE_SPLINE){
      offsety = 0;
      by++;
    }

    /* 4x4 interpolation filtering for one subpixel value */
    for(spx = splineArray+offsetx, spy = splineArray+offsety, sum=0.0, p=by-4; p<by; spx-=4, p++){
      
      cy = p<0 ? 0 : ( p>=picSizeY ? picSizeY-1 : p); 
      
      /* Horizontal filtering */
      for(sum1=0.0, q=bx-4; q<bx; q++){
        cx    = q<0 ? 0 : ( q>=picSizeX ? picSizeX-1 : q);
        sum1 += (*spx++) * yin[cy*picSizeX+cx];
      }
      
      sum += (*spy++) * sum1;
    }
  }
  else {

    /* 4x4 neighborhood is inside the frame -> no boundary checks needed */
    ib = (double)i;
    jb = (double)j;
    
    /* Pointer to the upper left pixel of the 4x4 neighborhood */
    img = yin + ((int)jb - 1) + (((int)ib-1)*picSizeX);
    
    /* Let's quantize the coordinate to 1/INTERV (1/256) pixel
     * accuracy. Interpolation filter coefficients (four) for each
     * reconstruction point are stored consecutively in the
     * splineArray. Thus, offsets for this array are calculated as
     * 4*INTERV*fractional pixel value.
     */
    if((offsetx = (int)(0.5+INTERV*(jb - (int)jb))<<2)>=SIZE_SPLINE){
      offsetx = 0;
      img++;
    }
    
    if((offsety = (int)(0.5+INTERV*(ib - (int)ib))<<2)>=SIZE_SPLINE){
      offsety = 0;
      img += picSizeX;
    }
    
    /* Pointers to the filter coefficient quartets */
    spx = splineArray + offsetx;  
    spy = splineArray + offsety;

    wrap = picSizeX - 3;
    
    /* Spline interpolation starts */
    sum1  = (*spx) * (*img);          /* Horiz. interp. for first line */
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    
    sum   = sum1 * (*spy);            /* Contribution from the first line */
    
    sum1  = *(spx-=3) * *(img+=wrap); /* Horiz. interp. for second line */
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    
    sum  += sum1 * *(++spy);          /* Contribution from the second line */
    
    sum1  = *(spx-=3) * *(img+=wrap); /* Horiz. interp. for third line */
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    
    sum  += sum1 * *(++spy);          /* Contribution from the third line */
    
    sum1  = *(spx-=3) * *(img+=wrap); /* Horiz. interp. for fourth line */
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    sum1 += *(++spx) * *(++img);
    
    sum  += sum1 * *(++spy);          /* Contribution from the fourth line */
  }

  /* Now sum contains the desired output */
  return(sum);
}


/*
 * splDerivativeXxX:
 *
 * Parameters:
 *     xx, xy, yx, yy   I: Horizontal & vertical step values in fixed-point
 *                         format
 *     pels             O: Interpolated pixels in 20-bit fraction fixed
 *                         point representation
 *     drx              O: x-derivatives in 19-bit fraction fixed point
 *                         representation
 *     dry              O: y-derivatives in 19-bit fraction fixed point
 *                         representation
 *     yin              I: Y component of the frame
 *     i32, j32         I: Block coordinates in fixed-point format
 *     picSizeX         I: Frame width
 *     picSizeY         I: Frame height
 *     blkSideLen       I: Block side length (4 or 8)
 *
 * Function:
 *     Cubic convolution interpolation and calculation of derivatives for
 *     a square block.
 *     First finds the nearest integer neighbors in the array, then
 *     accomplishes the cubic convolution interpolation and derivative
 *     calculation, using the integer neighbor values and samples of the
 *     cubic convolution kernel.
 *
 * Returns:
 *     Nothing
 *
 * History:
 *
 */

void splDerivativeXxX(int32 xx, int32 xy, int32 yx, int32 yy, int32 *pels,
                      int32 *drx, int32 *dry, Byte *yin, int32 i32, int32 j32,
                      int picSizeX, int picSizeY, int blkSideLen)
{
  int16
    *spx,*spy,*sdy,*sdx;          /* spline array pointers for i & j values */
  
  Byte 
    *img, block[4][4];            /* Temp. storage for 4x4 pixels */
    
  int32
    vCount, hCount,               /* Horiz./vert. looping variables */
    sum, sum1,                    /* Sums used in interpolation */
    gx, gy, p, q, bx, by, cx, cy; /* Temp vars. used when boundary checks fail */

  int
    offset, index, safe=1;


  /*
   * Boundary checks are performed for whole block by checking each of the four
   * corners of the block.
   */

  /* Check for big step values */
  if (xx>100000000 || xx<-100000000 || xy>100000000 || xy<-100000000 ||
      yx>100000000 || yx<-100000000 || yy>100000000 || yy<-100000000) safe = 0;

  /* Upper left corner */
  if( (i32 < 65536) || (i32 >= ((picSizeY-2)<<16)) ||
      (j32 < 65536) || (j32 >= ((picSizeX-2)<<16)) ) safe = 0;

  /* Upper right corner */
  i32 += (blkSideLen-1)*yx;
  j32 += (blkSideLen-1)*xx;
  if( (i32 < 65536) || (i32 >= ((picSizeY-2)<<16)) ||
      (j32 < 65536) || (j32 >= ((picSizeX-2)<<16)) ) safe = 0;

  /* Lower left corner */
  i32 += (blkSideLen-1)*yy;
  j32 += (blkSideLen-1)*xy;
  if( (i32 < 65536) || (i32 >= ((picSizeY-2)<<16)) ||
      (j32 < 65536) || (j32 >= ((picSizeX-2)<<16)) ) safe = 0;

  /* Lower right corner */
  i32 -= (blkSideLen-1)*yx;
  j32 -= (blkSideLen-1)*xx;
  if( (i32 < 65536) || (i32 >= ((picSizeY-2)<<16)) ||
      (j32 < 65536) || (j32 >= ((picSizeX-2)<<16)) ) safe = 0;

  i32 -= (blkSideLen-1)*yy;
  j32 -= (blkSideLen-1)*xy;

  xy -= blkSideLen*xx;
  yy -= blkSideLen*yx;


  if (safe) {

    /*
     * If safe is true, no boundary are checks needed for any of the pixels in
     * the block.
     */

    /* Interpolate blkSideLen*blkSideLen pixels */
    for(index=0, vCount=0; vCount<blkSideLen; vCount++) {
      for(hCount=0; hCount<blkSideLen; hCount++, index++) {

        img = yin + ((j32>>16) - 1) + (((i32>>16) - 1) * picSizeX);

        /* Offset where filter coefficients can be found for this
         * 1/256th pixel position (horizontal)
         */
        offset = (j32 & 0xff00)>>6;
        spx = splineArrayInt+offset;
        sdx = derArrayInt+offset;

        /* Offset where filter coefficients can be found for this
         * 1/256th pixel position (vertical)
         */
        offset = (i32 & 0xff00)>>6;
        spy = splineArrayInt+offset;
        sdy = derArrayInt+offset;

        /* Now filter coefficients can be accessed as:
         *   spx[i] pixel interpolation for horizontal direction
         *   spy[i] pixel interpolation for vertical direction
         *   sdx[i] derivative calculation for horizontal direction
         *   sdy[i] derivative calculation for vertical direction
         * i = 0, 1, 2, 3;
         */

        /* First line */
        sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
        sum = spy[0]*sum1;
        gy = sdy[0]*sum1;
        gx = spy[0]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
        
        img += picSizeX;
        
        /* Second line */
        sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
        sum += spy[1]*sum1;
        gy += sdy[1]*sum1;
        gx += spy[1]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
        
        img += picSizeX;
        
        /* Third line */
        sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
        sum += spy[2]*sum1;
        gy += sdy[2]*sum1;
        gx += spy[2]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
        
        img += picSizeX;
        
        /* Fourth (last) line */
        sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
        sum += spy[3]*sum1;
        gy += sdy[3]*sum1;
        gx += spy[3]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
        
        /* Interpolated pixel value (20-bit fraction) and derivatives
         * (19-bit fraction)
         */
        pels[index] = sum;
        dry[index]  = gy;
        drx[index]  = gx;

        i32 += yx;
        j32 += xx;
      }
      i32 += yy;
      j32 += xy;
    }


  } /* if */

  else {

    /*
     * safe is false, boundary checks are needed
     */

    for(index=0, vCount=0; vCount<blkSideLen; vCount++) {
      for(hCount=0; hCount<blkSideLen; hCount++, index++) {

        if( (i32 < 65536) || (i32 >= ((picSizeY-2)<<16)) ||
            (j32 < 65536) || (j32 >= ((picSizeX-2)<<16)) ) {

          /* Boundary check fails */
          bx = (j32>>16) + 3;
          by = (i32>>16) + 3;
          
          /* Offset where filter coefficients can be found for this
           * 1/256th pixel position (horizontal)
           */
          offset = (j32 & 0xff00)>>6;
          spx    = splineArrayInt + offset; 
          sdx    = derArrayInt + offset; 
          
          /* Offset where filter coefficients can be found for this
           * 1/256th pixel position (vertical)
           */
          offset = (i32 & 0xff00)>>6;
          spy    = splineArrayInt + offset;
          sdy    = derArrayInt + offset;

          /* Now filter coefficients can be accessed as:
           *   spx[i] pixel interpolation for horizontal direction
           *   spy[i] pixel interpolation for vertical direction
           *   sdx[i] derivative calculation for horizontal direction
           *   sdy[i] derivative calculation for vertical direction
           * i = 0, 1, 2, 3;
           */

          /* block is a 4x4 matrix where neighboring pixel values
           * are collected (copying boundary pixels if coordinates
           * point outside the frame)
           */
          img = (Byte *)block;
          for(p=by-4; p<by; p++) {
            cy = p<0 ? 0 : ( p>=picSizeY ? picSizeY-1 : p) ;
            cy *= picSizeX;
            for(q=bx-4; q<bx; q++) {
              cx = q<0 ? 0 : ( q>=picSizeX ? picSizeX-1 : q) ;
              *img++ = yin[cy+cx];
            }
          }

          /* Pixel interpolation and vertical derivative */
          for(gy=sum=0, p=0; p<4; p++) {
            sum1  = spx[0] * block[p][0];
            sum1 += spx[1] * block[p][1];
            sum1 += spx[2] * block[p][2];
            sum1 += spx[3] * block[p][3];
            sum1 >>= 8;  /* sum1 shifted right to fit into 16-bit */
            sum += spy[p] * sum1;
            gy  += sdy[p] * sum1;
          }

          /* Horizontal derivative */
          for(gx=0, q=0; q<4; q++) {
            sum1  = spy[0] * block[0][q];
            sum1 += spy[1] * block[1][q];
            sum1 += spy[2] * block[2][q];
            sum1 += spy[3] * block[3][q];
            gx += sdx[q] * (sum1>>8); /* sum1 shifted right to fit into 16-bit */
          }

          /* Interpolated pixel value (20-bit fraction) and derivatives
           * (19-bit fraction)
           */
          pels[index] = sum;
          dry[index] = gy;
          drx[index] = gx;
        }
  
        else {  /* When boundary checks are O.K. */


          img = yin + ((j32>>16) - 1) + (((i32>>16) - 1) * picSizeX);

          /* Offset where filter coefficients can be found for this
           * 1/256th pixel position (horizontal)
           */
          offset = (j32 & 0xff00)>>6;
          spx=splineArrayInt+offset;
          sdx=derArrayInt+offset;
          
          /* Offset where filter coefficients can be found for this
           * 1/256th pixel position (vertical)
           */
          offset = (i32 & 0xff00)>>6;
          spy=splineArrayInt+offset;
          sdy=derArrayInt+offset;
          
          /* First line */
          sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
          sum = spy[0]*sum1;
          gy = sdy[0]*sum1;
          gx = spy[0]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
          
          img += picSizeX;
          
          /* Second line */
          sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
          sum += spy[1]*sum1;
          gy += sdy[1]*sum1;
          gx += spy[1]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
          
          img += picSizeX;
          
          /* Third line */
          sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
          sum += spy[2]*sum1;
          gy += sdy[2]*sum1;
          gx += spy[2]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
          
          img += picSizeX;
          
          /* Fourth (last) line */
          sum1 = (spx[0]*img[0]+spx[1]*img[1]+spx[2]*img[2]+spx[3]*img[3])>>8;
          sum += spy[3]*sum1;
          gy += sdy[3]*sum1;
          gx += spy[3]*((sdx[0]*img[0]+sdx[1]*img[1]+sdx[2]*img[2]+sdx[3]*img[3])>>8);
          
          pels[index] = sum;
          dry[index] = gy;
          drx[index] = gx;


        }   /* else (boundary check)*/
        i32 += yx;
        j32 += xx;
      }   /* for (horizontal) */
      i32 += yy;
      j32 += xy;
    }   /* for (vertical) */

  }   /* else (safe) */

}

/*
 * splDerivativeNoAffineXxX:
 *
 * Parameters:
 *     pels             O: "Interpolated" pixels (20 fraction bits)
 *     drx              O: x-derivatives (19 fraction bits)
 *     dry              O: y-derivatives (19 fraction bits)
 *     yin              I: Y component of the frame
 *     i, j             I: Block coordinates
 *     picSizeX         I: Frame width
 *     picSizeY         I: Frame height
 *     blkSideLen       I: Block side length (4 or 8)
 *
 * Function:
 *     Cubic convolution interpolation and calculation of  derivatives for
 *     a square block without affine movement. This is a special case of
 *     cubic interpolation. Interpolation is not needed in this case
 *     since pixels are always taken from integer positions.
 *
 * Returns:
 *     Nothing
 *
 */

void splDerivativeNoAffineXxX(int32 *pels, int32 *dx, int32 *dy, Byte *yin,
                              int32 i, int32 j, int picSizeX, int picSizeY,
                              int blkSideLen)
{
  int k, l;
  Byte *yPtr, *yPtr2, *yPtr3;

  i = i - 1;
  j = j - 1;

  /*
   * Check if per pixel bound checks are needed
   */
  if (j >= 0 && i >= 0 && j < (picSizeX-blkSideLen-1) && i < (picSizeY-blkSideLen-1)) {

    /* Bound check are not needed */

    yPtr = yin + i*picSizeX + j;
    for (k = 0; k < blkSideLen; k++, yPtr+=(picSizeX-blkSideLen)) {
      for (l = 0; l < blkSideLen; l++, yPtr++) {

        /* Pixel value in 20-bit fraction fixed point representation */
        *pels++ = (int)yPtr[picSizeX + 1] << 20;

        /* Derivatives in 19-bit fraction fixed point representation */
        *dx++ = ((int)yPtr[picSizeX + 2] - (int)yPtr[picSizeX]) << 19;
        *dy++ = ((int)yPtr[2*picSizeX + 1] - (int)yPtr[1]) << 19;
      }
    }
  }

  else {

    /* Bound checks are needed */

    for (k = 0; k < blkSideLen; k++, i++, j-=blkSideLen) {

      /* Boundary checks in vertical direction */
      if (i < 0) {
        yPtr = yPtr2 = yin;
        if (i == -1)
          yPtr3 = yin + picSizeX;
        else
          yPtr3 = yin;
      }
      else if (i >= picSizeY-2) {
        yPtr2 = yPtr3 = yin + (picSizeY-1)*picSizeX;
        if (i == picSizeY-2)
          yPtr = yin + (picSizeY-2)*picSizeX;
        else
          yPtr = yin + (picSizeY-1)*picSizeX;
      }
      else {
        yPtr = yin + i*picSizeX;
        yPtr2 = yPtr + picSizeX;
        yPtr3 = yPtr + 2*picSizeX;
      }

      for (l = 0; l < blkSideLen; l++, j++) {

        /* Boundary checks in horizontal direction and evaluation of
         * pixel value in 20-bit and derivatives in 19-bit fraction
         * fixed point representation
         */
        if (j <= -1) {
          *pels++ = (int)yPtr2[0] << 20;
          *dy++ = ((int)yPtr3[0] - (int)yPtr[0]) << 19;
          if (j == -1)
            *dx++ = ((int)yPtr2[1] - (int)yPtr2[0]) << 19;
          else
            *dx++ = 0;
        }
        else if (j >= picSizeX-2) {
          *pels++ = (int)yPtr2[picSizeX-1] << 20;
          *dy++ = ((int)yPtr3[picSizeX-1] - (int)yPtr[picSizeX-1]) << 19;
          if (j == picSizeX-2)
            *dx++ = ((int)yPtr2[picSizeX-1] - (int)yPtr2[picSizeX-2]) << 19;
          else
            *dx++ = 0;
        }
        else {
          *pels++ = (int)yPtr2[j+1] << 20;
          *dx++ = ((int)yPtr2[j+2] - (int)yPtr2[j]) << 19;
          *dy++ = ((int)yPtr3[j+1] - (int)yPtr[j+1]) << 19;
        }

      }

    }
  }

}

/*
 * Se_8x8:
 *
 * Parameters:
 *     xstepx, xstepy   I: Horizontal & vertical step values
 *     ystepx, ystepy   I: when generating motion vector field
 *     curr             I: Y component of current frame
 *     i, j             I: Block coordinates
 *     iprev1, jprev1   I: Block coordinates in previous frame
 *     prev             I: Y component of previous reconstructed frame
 *
 * Function:
 *     Computes square prediction error of an 8x8 block. It uses
 *     spline_interpolation() for spline interpolation
 *
 *     Assumes that picSizeXglob and picSizeYglob contain correct frame
 *     dimensions
 *
 * Returns:
 *     Square prediction error
 *
 */

double Se_8x8(double xstepx, double xstepy, double ystepx, double ystepy,
              Byte *curr, int i, int j, double iprev1, double jprev1,
              Byte *prev)
{
   int
     m,n;

   double 
     sum=0.0, tmp, jPrev, iPrev;

   Byte 
     *curptr;

   curptr = curr + i*picSizeXglob + j;
   
   for(m=0; m<8; m++, curptr += picSizeXglob){
     iPrev = iprev1 + m*ystepy;
     jPrev = jprev1 + m*xstepy;

     for(n=0; n<8; n++){
       tmp = curptr[n] - splInterpolation(prev, iPrev, jPrev,
                                          picSizeXglob, picSizeYglob);
       sum   += tmp*tmp;
       iPrev += ystepx;
       jPrev += xstepx;
     }
   }

   return(sum); 
}


/*
 * mseRegionLum:
 *
 * Parameters:
 *     A                I: Motion coefficients of the region in local
 *                         coordinate system
 *     region           I: Region for which the error is calculated
 *     curr             I: Y component of current frame
 *     prev             I: Y component of previous reconstructed frame
 *
 * Function:
 *     Computes the mean square error of a region.
 *
 * Returns:
 *     Mean square error
 *
 */

double mseRegionLum(double A[], object *region, Byte *curr, Byte *prev)
{
  double 
    sum = 0.0, 
    xstepx, xstepy, ystepx, ystepy, iprev, jprev, Fx, Fy,
    x, y, centx, centy, invSizex, invSizey;
  int 
    i,j, xmin, xmax, ymin, ymax, sizex, sizey, safe=1;

  xmin  = region->xmin;
  xmax  = region->xmax;
  ymin  = region->ymin;
  ymax  = region->ymax;

  sizex = xmax - xmin;
  sizey = ymax - ymin;

  centx = (double) (xmin+xmax)/2.0;
  centy = (double) (ymin+ymax)/2.0;

  invSizex = 1.0/sizex;
  invSizey = 1.0/sizey;

  /* Differences between neighboring motion vectors when
   * generating the motion vector field
   */
  ystepx = (2.0*A[5])*invSizex;
  xstepx = 1.0 + (2.0*A[2])*invSizex;
  ystepy = 1.0 + (2.0*A[4])*invSizey;
  xstepy = (2.0*A[1])*invSizey;

  for(i=ymin; i <=ymax; i+=8){

    y  = 2.0*((double) i - centy)*invSizey;
    
    Fx = A[0] + A[1]*y;
    Fy = A[3] + A[4]*y;
    
    for(j=xmin; j <=xmax; j+=8){

      x     = (2.0*((double) j - centx))*invSizex;

      /* Location where motion vector points to in the upper left
       * corner of this 8x8 block
       */
      jprev = (double)j + (Fx + A[2]*x);
      iprev = (double)i + (Fy + A[5]*x);

      /* Evaluate squared error for this 8x8 block */
      sum  += Se_8x8(xstepx,xstepy,ystepx,ystepy,curr,i,j, iprev, jprev,prev);
    }
  }

  /* Return MSE */
  return(sum/(double)region->no_of_pels);
}

